/*
		Official ShipWars Reality Constants


	Notation legend:

	ms = milliseconds
	us = microseconds

	rl = real life
	ru = XSW real units

	su = screen units


	Unit conversion legend:

	1 XSW second = 1 real life second

	1 screen unit = 1 pixel (at 1:1 zoom)
	1 screen unit = 0.001 XSW real units
	1 screen unit = 8 meters

	1 XSW real unit = 1000 screen units
	1 XSW real unit = 8000 meters

	1 meter = 0.125 screen units
	1 meter = 0.000125 XSW real units


 */


#ifndef REALITY_H
#define REALITY_H


/*
 *	Cycle Lapse (in ms and us):
 *
 *	Specifies the sleep time for the main loop.
 *
 *	All velocity, speed, and other timmed calculations are based
 *	on this value.
 *
 *	Catchup time (time compensation) for speed should be multiplied
 *	by the differance of this value, for example:
 *
 *	time_compensation_coeff = cycle_delta_ms / SW_CYCLE_LAPSE_MS.
 */
#define SW_CYCLE_LAPSE_MS		16
#define SW_CYCLE_LAPSE_US		16000


/*
 *	Sector Sizes (in real units):
 */
#define SW_SECTOR_SIZE_X_MIN		-1024
#define SW_SECTOR_SIZE_X_MAX		1024
#define SW_SECTOR_SIZE_Y_MIN		-1024
#define SW_SECTOR_SIZE_Y_MAX		1024
#define SW_SECTOR_SIZE_Z_MIN		-1024
#define SW_SECTOR_SIZE_Z_MAX		1024


/*
 *	Drag Coefficents (in real units / cycle):
 */
#define SUBLIGHT_DRAG_COEF		0.0
#define WARP_DRAG_COEF			0.009
#define TRANSWARP_DRAG_COEF		0.03


/*
 *	Velocity Thresholds (in real units / cycle):
 */
#define WARP_THRESHOLD			0.04
#define TRANSWARP_THRESHOLD		0.3


/*
 *	Frequency ranges (in kHz):
 *
 *	(Note: Frequency compare resolution is 2 decimal places).
 */
#define SW_FREQ_MIN			1.00
#define SW_FREQ_MAX			900.00


/*
 *	Short-Range Communications Range (in real units):
 *
 *	For hails and com messages.
 */
#define SW_COM_SHORT_RANGE_MAX		100.0


/*
 *	Shield Phaze Visibility Interval (in ms):
 */
#define SHIELD_VISIBILITY_INTERVAL	3000


/*
 *	Visibility in a nebula:
 */
#define VISIBILITY_NEBULA		0.15


/*
 *	Wormhole Enter Delay (in ms):
 *
 *	The delay before you can enter another wormhole again.
 */
#define SW_WORMHOLE_ENTER_DELAY		5000

/*
 *	Maximum switchable throttle power:
 *
 *	Cannot switch throttle to reverse if throttle setting is
 *	greater than this value.
 */
#define THROTTLE_MAX_SWITCH		0.3

   
/*
 *	PI Constants:
 */
#ifndef PI
# define PI				3.141592654
#endif

#ifndef TWOPI
# define TWOPI				6.283185307
#endif

#ifndef HALFPI
# define HALFPI				1.570796327
#endif



/*
 *	Default real units to astronomical units coeff:
 *
 *	Must be greater than 0!
 *
 *	1 AU = 149,600,000 km
 */
#define DEF_UNITCONV_RU_TO_AU		0.1


/*
 *	Unit conversions:
 */
typedef struct {                           

	double ru_to_au;        /* SW Real units to Astronomical units. */

} sw_units_struct;



/* In unitconv.c */
extern double ConvertRUToAU(sw_units_struct *units, double x);
extern double ConvertRUToSU(sw_units_struct *units, double x);

extern double ConvertVelocityRUPCToRUPS(sw_units_struct *units, double x);
extern double ConvertVelocityRUPCToAUPS(sw_units_struct *units, double x);



#endif /* REALITY_H */
