#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/netio.h"

#include "siteban.h"
#include "sysparm.h"
#include "swserv.h"
#include "../include/auxstatcodes.h"


int AUXConIsAllocated(int n);
int AUXConInitialize(int socket, int format);

int AUXConSend(sws_aux_connection_struct *aux_con, char *sndbuf);
int AUXConSendStats(int n);

int AUXHandleLogin(sws_aux_connection_struct *aux_con, char *arg);
int AUXHandleSegment(sws_aux_connection_struct *aux_con, char *buf);
int AUXConHandleRecv(int n);

int AUXConManageAll(void);

void AUXConClose(int n);
void AUXConDelete(int n);
void AUXConDeleteAll(void);

int AUXConManageNewConnections(int socket);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the AUX connection is allocated.
 */
int AUXConIsAllocated(int n)
{
	if((aux_connection == NULL) ||
	   (n < 0) ||
	   (n >= total_aux_connections)
	)
	    return(0);
	else if(aux_connection[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Creates a new AUX connection from the specified socket.
 *
 *	Returns the new AUX connection's index or -1 on error.
 */
int AUXConInitialize(int socket, int format)
{
	int i, n;
	sws_aux_connection_struct *aux_con;

	/* Sanitize total */
	if(total_aux_connections < 0)
	    total_aux_connections = 0;

	/* Search for an available pointer */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)
		break;

	    if(aux_con->socket < 0)
		break;
	}
	if(i < total_aux_connections)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_aux_connections;
	    total_aux_connections = n + 1;
	    aux_connection = (sws_aux_connection_struct **)realloc(
		aux_connection,
		total_aux_connections * sizeof(sws_aux_connection_struct *)
	    );
	    if(aux_connection == NULL)
	    {
		total_aux_connections = 0;
		return(-3);
	    }

	    /* Reset the new pointer to NULL so that it gets allocated
	     * below
	     */
	    aux_connection[n] = NULL;
	}

	/* Allocate as needed */
	if(aux_connection[n] == NULL)
	{
	    aux_connection[n] = (sws_aux_connection_struct *)calloc(
		1, sizeof(sws_aux_connection_struct)
	    );
	    if(aux_connection[n] == NULL)
		return(-1);
	}


	aux_con = aux_connection[n];

	/* Set new values */
	aux_con->socket = socket;
	aux_con->format = format;
	aux_con->logged_in = 0;
	aux_con->next = cur_millitime;

	aux_con->conhost[0] = '\0';
	aux_con->contime = cur_systime;

	aux_con->bytes_received = 0;
	aux_con->bytes_sent = 0;
	aux_con->errors_received = 0;
	aux_con->errors_sent = 0;


	return(n);
}

/*
 *	Sends the data to the AUX connection.
 */
int AUXConSend(sws_aux_connection_struct *aux_con, char *sndbuf)
{
	int socket, bytes_written, error_level;

	if((aux_con == NULL) || (sndbuf == NULL))
	    return(-1);

	socket = aux_con->socket;

	/* Check if socket is valid and writeable */
	if(NetIsSocketWritable(socket, &error_level))
	{
	    const int len = strlen(sndbuf);
	    bytes_written = send(
	        socket,
	        sndbuf,
	        len,
	        0
	    );
	    if(bytes_written == 0)
	    {
	        return(bytes_written);
	    }
	    else if(bytes_written < 0)
	    {
		aux_con->errors_sent++;

		return(bytes_written);
	    }

	    aux_con->bytes_sent += bytes_written;
	}
	else
	{
	    aux_con->errors_sent++;
	    switch(error_level)
	    {
	      case 3:
		return(-1);
		break;

	      default:
		return(-1);
		break;
	    }
	}

	return(bytes_written);
}

/*
 *	Procedure to send all server stats to all AUX pipes.
 */
int AUXConSendStats(sws_aux_connection_struct *aux_con)
{
	int i, total_con, total_guest_con;
	char buf[AUXSTAT_MAX_LEN];
	sws_memory_stats_struct m;
	time_t uptime;
	sws_connection_struct *con;

	if(aux_con == NULL) 
	    return(-1);

	/* Calculate connections */
	for(i = 0, total_con = 0, total_guest_con = 0;
	    i < total_connections;
	    i++
	)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;
	    if(con->socket < 0)
		continue;

	    if(con->flags & CON_FLAG_IS_GUEST)
	        total_guest_con++;

	    total_con++;
	}

	/* Get memory stats */
	SWServGetMemoryStats(&m);

	/* Up time (use first incoming socket) */
	if(IncomingSocketIsAllocated(0))
	    uptime = cur_systime - incoming_socket[0]->start_time;
	else
	    uptime = 0l;


	/* Format buffer */
	sprintf(
	    buf,
"%s: %i %i\n\
%s: %ld %ld %ld\n\
%s: %ld %ld\n\
%s: %s\n\
%s: %i\n\
%s: %i\n\
%s: %ld\n",
	    STAT_PFX_CONNECTIONS, total_con, total_guest_con,
	    STAT_PFX_MEMORY, m.total, m.con, m.obj,
	    STAT_PFX_NEXT, (next.unv_save - cur_systime),
		(next.stats_export - cur_systime),
	    STAT_PFX_TITLE, unv_head.title,
	    STAT_PFX_OBJECTS, total_objects,
	    STAT_PFX_PID, getpid(),
	    STAT_PFX_UPTIME, uptime
	);

	return(AUXConSend(aux_con, buf));
}


/*
 *	Handles a login.
 */
int AUXHandleLogin(sws_aux_connection_struct *aux_con, char *arg)
{
	char *s;
	int obj_num;
	xsw_object_struct *obj;

	char name[AUXSTAT_SEGMENT_MAX];
	char password[AUXSTAT_SEGMENT_MAX];
	char sndbuf[AUXSTAT_SEGMENT_MAX + AUXSTAT_SEGMENT_MAX + 256];
	char welcome[AUXSTAT_SEGMENT_MAX + 256];

	if((aux_con == NULL) || (arg == NULL))
	    return(-1);

	/*
	 *	sname;spassword
	 */

	/* Search for delimiter */
	s = strchr(arg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	    return(-1);

	strncpy(password, s + 1, AUXSTAT_SEGMENT_MAX);
	password[AUXSTAT_SEGMENT_MAX - 1] = '\0';

	strncpy(name, arg, AUXSTAT_SEGMENT_MAX);
	name[AUXSTAT_SEGMENT_MAX - 1] = '\0';
	s = strchr(name, CS_STRING_DELIMINATOR_CHAR);
	if(s != NULL)
	    *s = '\0';

	strstrip(name);
	strstrip(password);


	/* Does name specify a guest login */
	if(sysparm.allow_guest &&
	   !strcasecmp(name, sysparm.guest_login_name)
	)
	{
	    /* Do nothing for Guest logins */
	}
	else
	{
	    /* Find player object by the given name */
	    obj_num = MatchObjectByName(
		xsw_object, total_objects, name, SW_OBJ_TYPE_PLAYER
	    );
	    if(DBIsObjectGarbage(obj_num))
	    {
		/* No such player object */
		sprintf(sndbuf, "%s: %s\n",
		    STAT_PFX_MESSAGE,
		    sysparm.mesg_wrong_login
	        );
		AUXConSend(aux_con, sndbuf);
		return(-1);
	    }
	    else
	    {
		/* Found player object */

		obj = xsw_object[obj_num];

		/* Check password */
		if(CryptHandleVerify(password, obj->password) < 1)
		{
		    /* Wrong password */
		    sprintf(sndbuf, "%s: %s\n",
			STAT_PFX_MESSAGE,
			sysparm.mesg_wrong_login
		    );
		    AUXConSend(aux_con, sndbuf);
		    return(-1);
		}
		else
		{
		    /* Password matched */

		    /* Log in connection */
		    aux_con->logged_in = 1;

		    /* Print welcome message */
		    strncpy(welcome, sysparm.mesg_welcome, AUXSTAT_SEGMENT_MAX);
		    welcome[AUXSTAT_SEGMENT_MAX - 1] = '\0';
		    substr(welcome, "%title", unv_head.title);
		    substr(welcome, "%name", obj->name);

		    sprintf(sndbuf, "%s: %s\n",
			STAT_PFX_MESSAGE,
			welcome
		    );
		    AUXConSend(aux_con, sndbuf);
		}
	    }
	}

	return(0);
}


/*
 *	Manages a segment of data, called by AUXConHandleRecv().
 */
int AUXHandleSegment(sws_aux_connection_struct *aux_con, char *buf)
{
	char *s;
	char parm[AUXSTAT_SEGMENT_MAX];
	char val[AUXSTAT_SEGMENT_MAX];

	if((aux_con == NULL) || (buf == NULL))
	    return(-1);

	/* Search for parmameter and value delimiter */  
	s = strchr(buf, ':');
	if(s == NULL)
	{
	    parm[0] = '\0';
	    
	    strncpy(val, buf, AUXSTAT_SEGMENT_MAX);
	    val[AUXSTAT_SEGMENT_MAX - 1] = '\0';
	}
	else
	{
	    strncpy(parm, buf, AUXSTAT_SEGMENT_MAX);
	    parm[AUXSTAT_SEGMENT_MAX - 1] = '\0';
	
	    s = strchr(parm, ':');
	    if(s != NULL)
		*s = '\0';

	    strncpy(val, s + 1, AUXSTAT_SEGMENT_MAX);
	    val[AUXSTAT_SEGMENT_MAX - 1] = '\0';
	}
	strstrip(val);
	strstrip(parm);


	/* Login */
	if(!strcasecmp(parm, STAT_PFX_LOGIN))
	{
	    AUXHandleLogin(aux_con, val);
	}


	return(0);
}

/*
 *	Handles incoming data on AUX connection aux_con.
 *
 *	Returns 0 or positive on success, -1 on error,
 *	and -3 on disconnected socket.
 */
int AUXConHandleRecv(sws_aux_connection_struct *aux_con)
{
	int socket;
	int bytes_read;
	char buf[AUXSTAT_MAX_LEN];
	char *buf_ptr;
	int buf_cnt;  

	char workbuf[AUXSTAT_SEGMENT_MAX];
	char *workbuf_ptr;
	int workbuf_cnt;  


	if(aux_con == NULL)
	    return(-1);

	socket = aux_con->socket;


	/* Is socket valid and contains data to be read? */
	if(!NetIsSocketReadable(socket, NULL))
	    return(0);

	/* Recieve incoming data */
	bytes_read = recv(socket, buf, AUXSTAT_MAX_LEN, 0);
	if(bytes_read == 0)
	{
	    /*   When polling of socket says there are bytes to be
	     *   read and recv() returns 0, it implies that the
	     *   socket has died.
	     */
	    return(-3);
	}
	/* Recieve error? */
	if(bytes_read < 0)
	{
	    /* Handle error */
	    switch(errno) 
	    {
	      /* Invalid descriptor */
	      case EBADF:
		aux_con->errors_received++;
		return(-3);
		break;

	      /* Not a socket */
	      case ENOTSOCK:
		aux_con->errors_received++;
		return(-3);
		break;

	      /* Operation would block a non-blocking socket */
	      case EWOULDBLOCK:
		/* Not an error */
		return(0); 
		break;
	
	      /* Segmentation fault */
	      case EFAULT:
		aux_con->errors_received++;
		return(-3);
		break;

	     default:
		aux_con->errors_received++;
		return(-3);
		break;
	    }
	}

	/* Increment bytes received */
	aux_con->bytes_received += bytes_read;


	/* Null terminate */
	if(bytes_read > AUXSTAT_MAX_LEN)
	    bytes_read = AUXSTAT_MAX_LEN;

	buf[bytes_read - 1] = '\0';


	/* Begin parsing and handling */
	buf_cnt = 0;
	buf_ptr = buf;

	while(buf_cnt < bytes_read)
	{
	    /* Reset work buffer count and pointer */
	    workbuf_cnt = 0;
	    workbuf_ptr = workbuf;

	    /* Begin fetching data from recvbuf to workbuf */
	    while(1)
	    {
		/* Work buffer overflowed? */
		if(workbuf_cnt >= AUXSTAT_SEGMENT_MAX)
		{
		    /* Increment buf_cnt to next delimiter char */
		    while(buf_cnt < bytes_read)
		    {      
			if((*buf_ptr == '\n') ||
			   (*buf_ptr == '\r') ||
			   (*buf_ptr == '\0')
			)
			{
			    buf_cnt++;
			    buf_ptr++;
			    break;
			}
			buf_cnt++;
			buf_ptr++;
		    }
		    /*   Null terminating character for workbuf will
		     *   be added farther below.
		     */
		    break;
		}

		/* End of buffer reached? */
		else if(buf_cnt >= bytes_read)
		{
		    *workbuf_ptr = '\0';
	    
		    /*   Copy workbuf_cnt to net_parms.co_data for
		     *   processing in the next call to this function.
		     */
/*
		    net_parms.co_data_len = MIN(
			workbuf_cnt,
			AUXSTAT_SEGMENT_MAX
		    );
		    
		    if(net_parms.co_data_len > 0)
			memcpy(
			    net_parms.co_data,  
			    workbuf,
			    net_parms.co_data_len
			);
 */
		     /*   Return, don't continue. This is so that the
		     *   the carry over buffer is not reset and its
		     *   values just set above get carryed over to
		     *   the next call to this function.
		     */
		    return(0);

		    break;
		}

		/* Deliminator in recieve buffer encountered? */
		else if((*buf_ptr == '\n') ||
			(*buf_ptr == '\r') ||
			(*buf_ptr == '\0')
		)
		{
		    *workbuf_ptr = '\0';

		    buf_cnt++;
		    buf_ptr++;

		    break;
		}

		/* Copy data from buffer to work buffer normally */
		else
		{
		    *workbuf_ptr = *buf_ptr;

		    buf_cnt++;
		    buf_ptr++;

		    workbuf_cnt++;
		    workbuf_ptr++;
		}
	    }

	    /* Skip if workbuf is empty */
	    if(*workbuf == '\0')
		continue;

	    /* Set null terminating character for workbuf */
	    workbuf[AUXSTAT_SEGMENT_MAX - 1] = '\0';


	    /* Handle this segment */
	    AUXHandleSegment(aux_con, workbuf);
	}


	return(0);
}


int AUXConManageAll()
{
	int i, status;
	sws_aux_connection_struct **ptr, *aux_con;


	for(i = 0, ptr = aux_connection;
	    i < total_aux_connections;
	    i++, ptr++
	)
	{
	    aux_con = *ptr;
	    if(aux_con == NULL)
		continue;

	    if(aux_con->socket < 0)
		continue;

	    /* Handle by format */
	    switch(aux_con->format)
	    {
	      default:	/* AUX_CON_FORMAT_STANDARD */

		/* Handle incoming data */
		status = AUXConHandleRecv(aux_con);
		if(status == -1)
		{
		    /* General error */
		    AUXConClose(i);
		    continue;
		}
		else if(status == -3)
		{
		    /* Socket has died, need to close connection */
		    AUXConClose(i);
		    continue;
		}

		/* Next send? */
		if(aux_con->next <= cur_millitime)
		{
		    AUXConSendStats(aux_con);

		    aux_con->next = cur_millitime +
			sysparm.int_aux_con_stats;
		}
		break;
	    }
	}


	return(0);
}


/*
 *	Closes the AUX connection.
 */
void AUXConClose(int n)
{
	sws_aux_connection_struct *aux_con;

	if(!AUXConIsAllocated(n))
	    return;

	aux_con = aux_connection[n];

	/* Close socket if connected */
	if(aux_con->socket > -1)
	{
	    while(close(aux_con->socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{   
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"AUXConClose(): connection %i: close socket %i error: %s.",
			    n,
			    aux_con->socket,
			    s
			);
			free(s);
		    }
		}    
		break;
	    }
	    aux_con->socket = -1;
	}

	aux_con->format = AUX_CON_FORMAT_STANDARD;
	aux_con->logged_in = 0;


	/* Notify close of AUX connection */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
		"AUX Connection %i: Disconnected.\n",
		n
	    );
}

/*
 *	Closes and deletes the AUX connection.
 */
void AUXConDelete(int n)
{
	if(AUXConIsAllocated(n))
	{
	    AUXConClose(n);
	    free(aux_connection[n]);
	    aux_connection[n] = NULL;
	}
}

/*
 *	Closes and deletes all AUX connections.
 */
void AUXConDeleteAll(void)
{
	int i;

	for(i = 0; i < total_aux_connections; i++)
	    AUXConDelete(i);

	free(aux_connection);
	aux_connection = NULL;
	total_aux_connections = 0;
}


/*
 *	Manages the new AUX connection from the specified socket.
 */
int AUXConManageNewConnections(int socket)
{
	int i, n, new_con;
	char sndbuf[AUXSTAT_SEGMENT_MAX];
	sws_aux_connection_struct *aux_con;

	int sin_size;
	int new_socket;
	struct sockaddr_in foreign_addr;

	siteban_ip_union ip;


	/* `Answer' new connection */
	sin_size = sizeof(struct sockaddr_in);
	new_socket = accept(
	    socket,
	    (struct sockaddr *)&foreign_addr,
	    reinterpret_cast<socklen_t *>(&sin_size)
	);
	if(new_socket == -1)
	    return(new_socket);

	/* Set socket nonblocking */
	fcntl(new_socket, F_SETFL, O_NONBLOCK);


	/* Get IP (in nbo) */
	ip.whole = foreign_addr.sin_addr.s_addr;

	/* Is this IP banned? */
	if(SiteBanIsBanned(&ip, 0))
	{
	    sprintf(
		sndbuf,
		"%s: Your address %i.%i.%i.%i has been banned.\n",
		STAT_PFX_MESSAGE,
		ip.part_u8[0],
		ip.part_u8[1],
		ip.part_u8[2],
		ip.part_u8[3]
	    );
	    send(new_socket, sndbuf, strlen(sndbuf), 0);

	    /* Close socket */
	    while(close(new_socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"Banned site AUX connection close socket %i error: %s.",
			    new_socket,
			    s
			);
			free(s);
		    }
		}
		break;
	    }

	    /* Pretend like we did not do anything */
	    return(0);
	} 


	/* Too many connections? */
	for(i = 0, n = 0; i < total_aux_connections; i++)
	{
	    if(aux_connection[i] == NULL)
		continue;

	    if(aux_connection[i]->socket < 0)
		continue;

	    n++;
	}
	if(n >= sysparm.max_aux_connections)
	{
	    sprintf(
		sndbuf,
"%s: Maximum of %i AUX connections reached.\n",
		STAT_PFX_MESSAGE,
		sysparm.max_aux_connections
	    );
	    send(new_socket, sndbuf, strlen(sndbuf), 0);

	    /* Close socket */
	    while(close(new_socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"Excessive AUX connection (max %i) close socket %i error: %s.",
			    sysparm.max_aux_connections,
			    new_socket,
			    s
			);   
			free(s);
		    }  
		}
		break;
	    }

	    return(-3);
	}


	/* Get a new connection descriptor for new connection */
	new_con = AUXConInitialize(
	    new_socket,
	    AUX_CON_FORMAT_STANDARD
	);
	if(AUXConIsAllocated(new_con))
	{
	    aux_con = aux_connection[new_con];
	}
	else
	{
	    if(sysparm.log_net)
		LogAppend(
		    fname.log,
"AUXConInitialize(): Unable to create a new AUX connection."
		);

	    /* Close socket */
	    while(close(new_socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"Failed AUX connection creation close socket %i error: %s.",
			    new_socket,
			    s
			);
			free(s);
		    }
		}
		break;
	    }

	    return(-3);
	}


	/* Consider it logged in for now */
/*
	aux_con->logged_in = 1;
 */

	strncpy(
	    aux_con->conhost,
	    inet_ntoa(foreign_addr.sin_addr),
	    sizeof(aux_con->conhost)
	);
	aux_con->conhost[sizeof(aux_con->conhost) - 1] = '\0';


	/* Notify of new AUX connection */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
		"AUX Connection %i: Got connection from `%s'.\n",
		new_con,
		aux_con->conhost
	    );


	/* Send request for login to connection (just once on initial
	 * connect).
	 */
	sprintf(sndbuf, "%s:\n", STAT_PFX_REQUESTLOGIN);
	AUXConSend(aux_con, sndbuf);


	return(0);
}
