#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/prochandle.h"
#include "../include/cfgfmt.h"
#include "../include/fio.h"
#include "../include/disk.h"

#ifdef PLUGIN_SUPPORT
# include "plugins.h"
#endif
#include "siteban.h"
#include "sysparm.h"
#include "cfgfile.h"
#include "swserv.h"


int CfgFileOpen(const char *path);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Opens the configuration from the specified file.
 */
int CfgFileOpen(const char *path)
{
	int i, n, lines_read = 0;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	off_t filesize;
	struct stat stat_buf;

	int *lsocket_port = NULL;
	int total_lsocket_ports = 0;

	if(path == NULL)
	    return(-1);

	/* Get the file's stats */
	if(stat(path, &stat_buf))
	{
	    char *s = STRDUP(strerror(errno));
	    if(s == NULL)
		s = strdup("no such file");
	    *s = toupper(*s);
	    fprintf(
		stderr,
		"%s: %s.\n",
		path, s
	    );
	    free(s);
	    return(-1);
	}

	/* Get the size of the file */
	filesize = stat_buf.st_size;

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    fprintf(
		stderr,
		"%s: Unable to open file for reading.\n",
		path
	    );
	    return(-1);
	}

	/* Delete resources that would be recreated when the
	 * configuration is loaded
	 */
#ifdef PLUGIN_SUPPORT
	/* Plugins */
	PluginUnloadAll();
#endif

	/* Site ban list */
	SiteBanDeleteAll();


	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Free previous line and allocate/read next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Fetch parameter */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Fetch value */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';


	    /* ListeningPort */
	    if(!strcasecmp(parm, "ListeningPort"))
	    {
		const int port = atoi(val);

		/* Create a new incoming socket */
		IncomingSocketNew(
		    port,
		    INCOMING_SOCKET_TYPE_STANDARD
		);

		/* Record this listening socket's port number */
		n = total_lsocket_ports;
		total_lsocket_ports = n + 1;
		lsocket_port = (int *)realloc(
		    lsocket_port,
		    total_lsocket_ports * sizeof(int)
		);
		if(lsocket_port != NULL)
		{
		    lsocket_port[n] = port;
		}
		else
		{
		    total_lsocket_ports = 0;
		}
	    }
	    /* AUXStatsListeningPort */
	    else if(!strcasecmp(parm, "AUXStatsListeningPort"))
	    {
		const int port = atoi(val);

		/* Create a new incoming socket */
		IncomingSocketNew(
		    port,
		    INCOMING_SOCKET_TYPE_AUXSTATS
		);

		/* Record this listening socket's port number */
		n = total_lsocket_ports;
		total_lsocket_ports = n + 1;
		lsocket_port = (int *)realloc(   
		    lsocket_port,
		    total_lsocket_ports * sizeof(int)
		);
		if(lsocket_port != NULL)
		{
		    lsocket_port[n] = port;
		}
		else
		{
		    total_lsocket_ports = 0;
		}
	    }
	    /* ToplevelDir */
	    else if(!strcasecmp(parm, "ToplevelDir") ||
		    !strcasecmp(parm, "ServerToplevelDir")
	    )
	    {
#if !defined(__WIN32__)		/* Disable this for windows */
		strncpy(
		    dname.toplevel, val,
		    sizeof(dname.toplevel)
		);
		dname.toplevel[
		    sizeof(dname.toplevel) - 1] = '\0';

		/* Warn if not an absolute path */
		if(!ISPATHABSOLUTE(dname.toplevel) &&
		   (*dname.toplevel != '\0')
		)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value must be an absolute path.\n",
			path,  
			lines_read,
			parm
		    );
		}

		/* Path does not exist? */
		if(stat(dname.toplevel, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.toplevel, s
		    );
		    free(s);
		}
#endif
	    }
	    /* EtcDir */
	    else if(!strcasecmp(parm, "EtcDir") ||
		    !strcasecmp(parm, "ServerEtcDir")
	    )
	    {
		strncpy(
		    dname.etc, val,
		    sizeof(dname.etc)
		);
		dname.etc[
		    sizeof(dname.etc) - 1] = '\0';

		/* Complete path as needed */			
		if(!ISPATHABSOLUTE(dname.etc) &&
		   (*dname.etc != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.etc
		    );
		    if(s != NULL)
			strncpy(dname.etc, s, sizeof(dname.etc));
		    dname.etc[sizeof(dname.etc) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.etc, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.etc, s
		    );
		    free(s);
		}
	    }
	    /* LogsDir */
	    else if(!strcasecmp(parm, "LogsDir") ||
		    !strcasecmp(parm, "ServerLogsDir")
	    )
	    {
		strncpy(
		    dname.logs, val,
		    sizeof(dname.logs)
		);
		dname.logs[
		    sizeof(dname.logs) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(dname.logs) &&
		   (*dname.logs != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.logs
		    );
		    if(s != NULL)
			strncpy(dname.logs, s, sizeof(dname.logs));
		    dname.logs[sizeof(dname.logs) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.logs, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.logs, s
		    );
		    free(s);
		}
	    }
	    /* BinDir */
	    else if(!strcasecmp(parm, "BinDir") ||
		    !strcasecmp(parm, "ServerBinDir")
	    )
	    {
#if !defined(__WIN32__)		/* Disable item for windows */
		strncpy(
		    dname.bin, val,
		    sizeof(dname.bin)
		);
		dname.bin[
		    sizeof(dname.bin) - 1] = '\0';

		/* Complete path as needed */ 
		if(!ISPATHABSOLUTE(dname.bin) &&
		   (*dname.bin != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.bin
		    );
		    if(s != NULL)
			strncpy(dname.bin, s, sizeof(dname.bin));
		    dname.bin[sizeof(dname.bin) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.bin, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.bin, s
		    );
		    free(s);
		}
#endif
	    }
	    /* DBDir */
	    else if(!strcasecmp(parm, "DBDir") ||
		    !strcasecmp(parm, "ServerDBDir")
	    )
	    {
		strncpy(
		    dname.db, val,
		    sizeof(dname.db)
		);
		dname.db[
		    sizeof(dname.db) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(dname.db) &&
		   (*dname.db != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.db
		    );
		    if(s != NULL)
			strncpy(dname.db, s, sizeof(dname.db));
		    dname.db[sizeof(dname.db) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.db, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.db, s
		    );
		    free(s);
		}
	    }
	    /* TmpDir */
	    else if(!strcasecmp(parm, "TmpDir") ||
		    !strcasecmp(parm, "ServerTmpDir")
	    )
	    {
		strncpy(
		    dname.tmp, val,
		    sizeof(dname.tmp)
		);
		dname.tmp[sizeof(dname.tmp) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(dname.tmp) &&
		   (*dname.tmp != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.tmp
		    );
		    if(s != NULL)
			strncpy(dname.tmp, s, sizeof(dname.tmp));
		    dname.tmp[sizeof(dname.tmp) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.tmp, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.tmp, s
		    );
		    free(s);
		}
	    }
	    /* PluginsDir */
	    else if(!strcasecmp(parm, "PluginsDir") ||
		    !strcasecmp(parm, "ServerPluginsDir")
	    )
	    {
		strncpy(
		    dname.plugins, val,
		    sizeof(dname.plugins)
		);
		dname.plugins[
		    sizeof(dname.plugins) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(dname.plugins) &&
		   (*dname.plugins != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.plugins
		    );
		    if(s != NULL)
			strncpy(dname.plugins, s, sizeof(dname.plugins));
		    dname.plugins[sizeof(dname.plugins) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.plugins, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.plugins, s
		    );
		    free(s);
		}
	    }
	    /* PublicHTMLDir */
	    else if(!strcasecmp(parm, "PublicHTMLDir"))
	    {
		strncpy(
		    dname.public_html, val,
		    sizeof(dname.public_html)
		);
		dname.public_html[
		    sizeof(dname.public_html) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(dname.public_html) &&
		   (*dname.public_html != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.toplevel, dname.public_html
		    );
		    if(s != NULL)
			strncpy(dname.public_html, s, sizeof(dname.public_html));
		    dname.public_html[sizeof(dname.public_html) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(dname.public_html, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such directory");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, dname.public_html, s
		    );
		    free(s);
		}
	    }
	    /* Monitor */
	    else if(!strcasecmp(parm, "Monitor"))
	    {
		if(!fname.monitor_set)
		{
		    strncpy(
			fname.monitor, val,
			sizeof(fname.monitor)
		    );
		    fname.monitor[
			sizeof(fname.monitor) - 1] = '\0';
		    fname.monitor_set = 1;
		}

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.monitor) &&
		   (*fname.monitor != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.bin, fname.monitor
		    );
		    if(s != NULL)
			strncpy(fname.monitor, s, sizeof(fname.monitor));
		    fname.monitor[sizeof(fname.monitor) - 1] = '\0';
		}

		/* Path does not exist? */
		if(stat(fname.monitor, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such file");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, fname.monitor, s
		    );
		    free(s);
		}
	    }
	    /* UnvIn */
	    else if(!strcasecmp(parm, "UnvIn") ||
		    !strcasecmp(parm, "DBIn")
	    )
	    {
		if(!fname.unv_in_set)
		{
		    strncpy(
			fname.unv_in, val,
			sizeof(fname.unv_in)
		    );
		    fname.unv_in[sizeof(fname.unv_in) - 1] = '\0';
		    fname.unv_in_set = 1;
		}

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.unv_in) &&
		   (*fname.unv_in != '\0')
		)
		{
		    const char *s = PrefixPaths(dname.db, fname.unv_in);
		    if(s != NULL)
		        strncpy(fname.unv_in, s, sizeof(fname.unv_in));
		}
		fname.unv_in[sizeof(fname.unv_in) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.unv_in, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such file");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, fname.unv_in, s
		    );
		    free(s);
		}
	    }
	    /* UnvOut */ 
	    else if(!strcasecmp(parm, "UnvOut") ||
		    !strcasecmp(parm, "DBOut")
	    )
	    {
		if(!fname.unv_out_set)
		{
		    strncpy(
			fname.unv_out, val,
			sizeof(fname.unv_out)
		    );
		    fname.unv_out[sizeof(fname.unv_out) - 1] = '\0';
		    fname.unv_out_set = 1;
		}

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.unv_out) &&
		   (*fname.unv_out != '\0')
		)
		{
		    const char *s = PrefixPaths(dname.db, fname.unv_out);
		    if(s != NULL)
			strncpy(fname.unv_out, s, sizeof(fname.unv_out));
		}
		fname.unv_out[sizeof(fname.unv_out) - 1] = '\0';
	    }
	    /* OCSFile */
	    else if(!strcasecmp(parm, "OCSFile"))
	    {
		if(!fname.ocs_set)
		{
		    strncpy(
			fname.ocs, val,
			sizeof(fname.ocs)
		    );
		    fname.ocs[sizeof(fname.ocs) - 1] = '\0';
		    fname.ocs_set = 1;
		}

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.ocs) &&
		   (*fname.ocs != '\0')
		)
		{
		    const char *s = PrefixPaths(dname.db, fname.ocs);
		    if(s != NULL)
		        strncpy(fname.ocs, s, sizeof(fname.ocs));
		}
		fname.ocs[sizeof(fname.ocs) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.ocs, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such file");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, fname.ocs, s
		    );
		    free(s);
		}
	    }
	    /* OPMFile */
	    else if(!strcasecmp(parm, "OPMFile"))
	    {
		if(!fname.opm_set)
		{
		    strncpy(
			fname.opm,
			val,
			sizeof(fname.opm)
		    );
		    fname.opm[sizeof(fname.opm) - 1] = '\0';
		    fname.opm_set = 1;
		}

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.opm) &&
		   (*fname.opm != '\0')
		)
		{
		    const char *s = PrefixPaths(dname.db, fname.opm);
		    if(s != NULL)
			strncpy(fname.opm, s, sizeof(fname.opm));
		}
		fname.opm[sizeof(fname.opm) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.opm, &stat_buf))
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("no such file");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: %s.\n",
			path, lines_read, fname.opm, s
		    );
		    free(s);
		}
	    }
	    /* LogFile */
	    else if(!strcasecmp(parm, "LogFile") ||
		    !strcasecmp(parm, "PrimaryLogFile")
	    )
	    {
		strncpy(
		    fname.log, val,
		    sizeof(fname.log)
		);
		fname.log[
		    sizeof(fname.log) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.log) &&
		   (*fname.log != '\0')
		)
		{
		    const char *s = PrefixPaths(dname.logs, fname.log);
		    if(s != NULL)
			strncpy(
			    fname.log, s,
			    sizeof(fname.log)
			);
		    fname.log[
			sizeof(fname.log) - 1] = '\0';
		}
	    }
	    /* PIDFile */
	    else if(!strcasecmp(parm, "PIDFile"))
	    {
		strncpy(
		    fname.pid, val,
		    sizeof(fname.pid)
		);
		fname.pid[sizeof(fname.pid) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.pid) &&
		   (*fname.pid != '\0')
		)
		{
		    const char *s = PrefixPaths(dname.tmp, fname.pid);
		    if(s != NULL)
			strncpy(
			    fname.pid, s,   
			    sizeof(fname.pid)
			);
		    fname.pid[sizeof(fname.pid) - 1] = '\0';
		}
	    }
	    /* ConListExportFile */
	    else if(!strcasecmp(parm, "ConListExportFile") ||
		    !strcasecmp(parm, "ConListHTML")
	    )
	    {
		strncpy(
		    fname.conlist_export, val,
		    sizeof(fname.conlist_export)
		);
		fname.conlist_export[
		    sizeof(fname.conlist_export) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.conlist_export) &&
		   (*fname.conlist_export != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.public_html, fname.conlist_export
		    );
		    if(s != NULL)
			strncpy(
			    fname.conlist_export, s,
			    sizeof(fname.conlist_export)
			);
		    fname.conlist_export[
			sizeof(fname.conlist_export) - 1] = '\0';
		}
	    }
	    /* ScoresExportFile */
	    else if(!strcasecmp(parm, "ScoresExportFile") ||
		    !strcasecmp(parm, "ScoresHTML")
	    )
	    {
		strncpy(
		    fname.scores_export, val,
		    sizeof(fname.scores_export)
		);
		fname.scores_export[
		    sizeof(fname.scores_export) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.scores_export) &&
		   (*fname.scores_export != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.public_html, fname.scores_export
		    );
		    if(s != NULL)
			strncpy(
			    fname.scores_export, s,
			    sizeof(fname.scores_export)
			);
		    fname.scores_export[
			sizeof(fname.scores_export) - 1] = '\0';
		}
	    }
	    /* EventsExportFile */
	    else if(!strcasecmp(parm, "EventsExportFile") ||
		    !strcasecmp(parm, "EventsHTML")
	    )
	    {
		strncpy(
		    fname.events_export, val,
		    sizeof(fname.events_export)
		);
		fname.events_export[
		    sizeof(fname.events_export) - 1] = '\0';

		/* Complete path as needed */
		if(!ISPATHABSOLUTE(fname.events_export) &&
		   (*fname.events_export != '\0')
		)
		{
		    const char *s = PrefixPaths(
			dname.public_html, fname.events_export
		    );
		    if(s != NULL)
			strncpy(
			    fname.events_export, s,
			    sizeof(fname.events_export)
			);
		    fname.events_export[
			sizeof(fname.events_export) - 1] = '\0';
		}
	    }

	    /* Sysparms */
	    /* AllowGuestLogins */
	    else if(!strcasecmp(parm, "AllowGuestLogins"))
	    {
		sysparm.allow_guest = StringIsYes(val);
	    }
	    /* AllowGuestSignup */
	    else if(!strcasecmp(parm, "AllowGuestSignup"))
	    {
		sysparm.allow_guest_signup = StringIsYes(val);
	    }
	    /* CeaseFire */
	    else if(!strcasecmp(parm, "CeaseFire"))
	    {
		sysparm.cease_fire = StringIsYes(val);
	    }
	    /* CloakAMRate */  
	    else if(!strcasecmp(parm, "CloakAMRate"))
	    {
		/* In size_ru * antimatter / cycle */
		sysparm.cloak_am_rate = MAX(atof(val), 0.0);
	    }
	    /* ComRange */
	    else if(!strcasecmp(parm, "ComRange"))
	    {
		/* In real units */
		sysparm.com_range = MAX(atof(val), 0.0);
	    }
	    /* ConnectNotify */
	    else if(!strcasecmp(parm, "ConnectNotify"))
	    {
		sysparm.con_notify = StringIsYes(val);
	    }
	    /* ConsoleQuiet */
	    else if(!strcasecmp(parm, "ConsoleQuiet"))
	    {
/* Not supported from configuration file, command line only */
	    }
	    /* CPUPriority */
	    else if(!strcasecmp(parm, "CPUPriority"))
	    {
		if(isdigit(*val))
		{
		    sysparm.cpu_priority = atoi(val);
		}
		else
		{
		    if(!strcasecmp(val, "highest"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGHEST;
		    else if(strcasepfx(val, "h"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGH;
		    else if(strcasepfx(val, "m"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_MEDIUM;
		    else if(strcasepfx(val, "l"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_LOWEST;
		    else
			fprintf(
			    stderr,
"%s: Line %i: Warning: %s: Invalid value `%s' (valid values are;\
 highest, high, medium, or lowest).\n",
			    path,
			    lines_read,
			    parm,
			    val
			);
		}
	    }
	    /* DmgCtlRate */
	    else if(!strcasecmp(parm, "DmgCtlRate") ||
		    !strcasecmp(parm, "DamageControlRate")
	    )
	    {
		sysparm.dmg_ctl_rate = MAX(atof(val), 0.0);
	    }
	    /* GuestLoginName */
	    else if(!strcasecmp(parm, "GuestLoginName"))
	    {
		strncpy(
		    sysparm.guest_login_name, val,
		    sizeof(sysparm.guest_login_name)
		);
		sysparm.guest_login_name[
		    sizeof(sysparm.guest_login_name) - 1] = '\0';
	    }
	    /* HidePlayersHome */
	    else if(!strcasecmp(parm, "HidePlayersHome") ||
		    !strcasecmp(parm, "HidePlayers")
	    )
	    {
		sysparm.hide_players_home = StringIsYes(val);
	    }
	    /* HitPlayerBonus */
	    else if(!strcasecmp(parm, "HitPlayerBonus"))
	    {
		sysparm.hit_player_bonus = MAX(atof(val), 0.0);
	    }
	    /* HomesDestroyable */
	    else if(!strcasecmp(parm, "HomesDestroyable"))
	    {
		sysparm.homes_destroyable = StringIsYes(val);
	    }
	    /* IntervalSystemCheck */
	    else if(!strcasecmp(parm, "IntervalSystemCheck"))
	    {
/* Not settable from configuration file */
	    }
	    /* IntervalNewConnectionPoll */
	    else if(!strcasecmp(parm, "IntervalNewConnectionPoll"))
	    {
/* Not settable from configuration file */
	    }
	    /* IntervalObjectValues */
	    else if(!strcasecmp(parm, "IntervalObjectValues"))
	    {
		sysparm.int_object_values = atol(val);
		if(sysparm.int_object_values < SWS_MIN_OBJECT_UPDATE_INT)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value too small, setting to minimum %i.\n",
			path,
			lines_read,
			parm,
			SWS_MIN_OBJECT_UPDATE_INT
		    );
		    sysparm.int_object_values = SWS_MIN_OBJECT_UPDATE_INT;
		}
		next.object_values = cur_millitime + sysparm.int_object_values;
	    }
	    /* IntervalWeaponValues */
	    else if(!strcasecmp(parm, "IntervalWeaponValues"))
	    {
		sysparm.int_weapon_values = atol(val);
		if(sysparm.int_weapon_values < SWS_MIN_OBJECT_UPDATE_INT)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value too small, setting to minimum %i.\n",
			path,
			lines_read,
			parm,
			SWS_MIN_OBJECT_UPDATE_INT
		    );
		    sysparm.int_weapon_values = SWS_MIN_OBJECT_UPDATE_INT;
		}
		next.need_weapon_values = 1;
	    }
	    /* IntervalMemoryClean */
	    else if(!strcasecmp(parm, "IntervalMemoryClean"))
	    {
/* Not settable from configuration file */
	    }
	    /* IntervalAUXStats */
	    else if(!strcasecmp(parm, "IntervalAUXStats"))
	    {
		sysparm.int_aux_con_stats = atol(val);
		if(sysparm.int_aux_con_stats < 1000l)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value too small, setting to minimum %i.\n",
			path,
			lines_read,
			parm,
			1000
		    );
		    sysparm.int_aux_con_stats = 1000l;
		}
	    }
	    /* IntervalOSStats */
	    else if(!strcasecmp(parm, "IntervalOSStats"))
	    {
		sysparm.int_os_stats = atol(val);	/* In seconds */
		if(sysparm.int_os_stats < 10l)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value too small, setting to minimum %i.\n",
			path,
			lines_read,
			parm,
			10
		    );
		    sysparm.int_os_stats = 10l;
		}
		next.os_stats = cur_systime;	/* Now */
	    }
	    /* IntervalStatsExportFile */
	    else if(!strcasecmp(parm, "IntervalStatsExportFile") ||
		    !strcasecmp(parm, "IntervalStatisticsExport")
	    )
	    {
		sysparm.int_stats_export = atol(val);	/* In seconds */
		if(sysparm.int_stats_export < 10l)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value too small, setting to minimum %i.\n",
			path,
			lines_read,
			parm,
			10
		    );
		    sysparm.int_stats_export = 10l;
		}
		next.stats_export = cur_systime;	/* Now */
	    }
	    /* IntervalUnvSave */
	    else if(!strcasecmp(parm, "IntervalUnvSave") ||
		    !strcasecmp(parm, "IntervalDBSave")
	    )
	    {
		sysparm.int_unv_save = atol(val);	/* In seconds */
		if(sysparm.int_unv_save < 10l)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value too small, setting to minimum %i.\n",
			path,
			lines_read,
			parm,
			10
		    );
		    sysparm.int_unv_save = 10l;
		}
		next.unv_save = cur_systime + sysparm.int_unv_save;
	    }
	    /* KillerGetsCredits */
	    else if(!strcasecmp(parm, "KillerGetsCredits"))
	    {
		sysparm.killer_gets_credits = StringIsYes(val);
	    }
	    /* LogGeneral */
	    else if(!strcasecmp(parm, "LogGeneral"))
	    {
		sysparm.log_general = StringIsYes(val);
	    }
	    /* LogEvents */
	    else if(!strcasecmp(parm, "LogEvents"))
	    {
		sysparm.log_events = StringIsYes(val);
	    }
	    /* LogNet */
	    else if(!strcasecmp(parm, "LogNet"))
	    {
		sysparm.log_net = StringIsYes(val);
	    }
	    /* LogErrors */
	    else if(!strcasecmp(parm, "LogErrors"))
	    {
		sysparm.log_errors = StringIsYes(val);
	    }
	    /* LoginTimeout */
	    else if(!strcasecmp(parm, "LoginTimeout"))
	    {
		sysparm.login_timeout = atol(val);	/* Seconds */
		if(sysparm.login_timeout < 1)
		    sysparm.login_timeout = 1;
	    }
	    /* MaxAUXStatsConnections */
	    else if(!strcasecmp(parm, "MaxAUXStatsConnections"))
	    {
		sysparm.max_aux_connections = atol(val);
		if(sysparm.max_aux_connections < 0)
		    sysparm.max_aux_connections = 0;
	    }
	    /* MaxConnections */
	    else if(!strcasecmp(parm, "MaxConnections"))
	    {
		sysparm.max_connections = atol(val);
		if(sysparm.max_connections < 1)
		{
		    fprintf(
			stderr,
"%s: Line %i: Warning: %s: Value may not be less than %i, setting to %i.\n",
			path,
			lines_read,
			parm,
			1,
			1
		    );
		    sysparm.max_connections = 1;
		}
	    }
	    /* MaxFailedLogins */
	    else if(!strcasecmp(parm, "MaxFailedLogins"))
	    {
		sysparm.max_failed_logins = MAX(atoi(val), 1);
	    }
	    /* MaxGuests */
	    else if(!strcasecmp(parm, "MaxGuests"))
	    {
		sysparm.max_guests = MAX(atol(val), 0);
	    }
	    /* MessageWrongLogin */
	    else if(!strcasecmp(parm, "MessageWrongLogin"))
	    {
		strncpy(
		    sysparm.mesg_wrong_login, val,
		    sizeof(sysparm.mesg_wrong_login)
		);
		sysparm.mesg_wrong_login[
		    sizeof(sysparm.mesg_wrong_login) - 1] = '\0';
	    }
	    /* MessageNoGuests */
	    else if(!strcasecmp(parm, "MessageNoGuests"))
	    {
		strncpy(
		    sysparm.mesg_no_guests, val,
		    sizeof(sysparm.mesg_no_guests)
		);
		sysparm.mesg_no_guests[
		    sizeof(sysparm.mesg_no_guests) - 1] = '\0';
	    }
	    /* MessageWelcome */
	    else if(!strcasecmp(parm, "MessageWelcome"))
	    {
		strncpy(
		    sysparm.mesg_welcome, val,
		    sizeof(sysparm.mesg_welcome)
		);
		sysparm.mesg_welcome[
		    sizeof(sysparm.mesg_welcome) - 1] = '\0';
	    }
	    /* MessageLeave */
	    else if(!strcasecmp(parm, "MessageLeave"))
	    {
		strncpy(
		    sysparm.mesg_leave, val,
		    sizeof(sysparm.mesg_leave)
		);
		sysparm.mesg_leave[
		    sizeof(sysparm.mesg_leave) - 1] = '\0';
	    }
	    /* ReportDestroyedWeapons */
	    else if(!strcasecmp(parm, "ReportDestroyedWeapons"))
	    {
		sysparm.report_destroyed_weapons = StringIsYes(val);
	    }
	    /* RespawnPlayers */
	    else if(!strcasecmp(parm, "RespawnPlayers"))
	    {
		sysparm.respawn_players = StringIsYes(val);
	    }
	    /* SendStarChart */
	    else if(!strcasecmp(parm, "SendStarChart"))
	    {
		sysparm.send_starchart = StringIsYes(val);
	    }
	    /* SingleConnection */
	    else if(!strcasecmp(parm, "SingleConnection"))
	    {
		sysparm.single_connection = StringIsYes(val);
	    }
	    /* TractorBeamAMRate */
	    else if(!strcasecmp(parm, "TractorBeamAMRate"))
	    {  
		/* In size_ru * antimatter / cycle */
		sysparm.tractor_beam_am_rate = MAX(atof(val), 0.0);
	    }
	    /* TractorBeamRange */
	    else if(!strcasecmp(parm, "TractorBeamRange"))
	    {
		/* In real units */
		sysparm.tractor_beam_range = MAX(atof(val), 0.0);
	    }
	    /* TransporterRange */
	    else if(!strcasecmp(parm, "TransporterRange"))
	    {  
		/* In real units */
		sysparm.transporter_range = MAX(atof(val), 0.0);
	    }
	    /* UseServerAI */
	    else if(!strcasecmp(parm, "UseServerAI"))
	    {
		sysparm.use_server_ai = StringIsYes(val);
	    }
	    /* WeaponDisarmLifespan */
	    else if(!strcasecmp(parm, "WeaponDisarmLifespan"))
	    {
		if(isdigit(*val))
		{
		    sysparm.weapon_disarm_lifespan = (time_t)atol(val);
		}
		else
		{
		    if(!strcasecmp(val, "infinate") ||
		       strcasepfx(val, "i")
		    )
			sysparm.weapon_disarm_lifespan = 0l;
		    else if(!strcasecmp(val, "no_change") ||
			    !strcasecmp(val, "no change") ||
			    strcasepfx(val, "n")
		    )
			sysparm.weapon_disarm_lifespan = 1l;
		    else
			fprintf(
			    stderr,
"%s: Line %i: Warning: %s: Value must be a number in milliseconds,\
 \"infinate\", or \"no_change\".\n",
			    path,
			    lines_read,
			    parm
			);
		}
	    }

#ifdef PLUGIN_SUPPORT
	    /* BeginPlugin */
	    else if(!strcasecmp(parm, "BeginPlugin"))
	    {
		plugin_id_t id;
/*		u_int64_t flags = 0; */
		char *path = NULL;
		char **argv = NULL;
		int argc = 0;

		while(1)
		{
		    /* Free previous line and allocate/read next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Fetch parameter */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Fetch value */ 
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Path */
		    if(!strcasecmp(parm, "Path"))
		    {
			free(path);
			if(ISPATHABSOLUTE(val))
			    path = STRDUP(val);
			else
			    path = STRDUP(PrefixPaths(dname.plugins, val));
		    }
		    /* Flags */
		    else if(!strcasecmp(parm, "Flags"))
		    {
			int i, strc;
			const char *s;
			char **strv = strexp(val, &strc);

			for(i = 0; i < strc; i++)
			{
			    s = strv[i];
			    if(s == NULL)
				continue;


			}

			strlistfree(strv, strc);
		    }
		    /* Arguments */ 
		    else if(!strcasecmp(parm, "Arguments"))
		    {
			strlistfree(argv, argc);
			argv = ExecExplodeCommand(val, &argc);
		    }

		    /* EndPlugin */
		    else if(!strcasecmp(parm, "EndPlugin"))
		    {
			/* Check if we got all the information */
			if(path != NULL)
			{
			    id = PluginLoad(
				path,
				0,
				argc, argv,
				-1	/* No connection started this */
			    );
			    if(id < 0)
			        fprintf(
				    stderr,
"%s: Line %i: Unable to load plugin `%s'.\n",
				    path, lines_read, path
				);
			}
			else
			{
			    if(path == NULL)
				fprintf(
				    stderr,
"%s: Line %i: Insufficient information to load plugin.\n",
				    path, lines_read
				);
			    else
				fprintf(   
				    stderr,
"%s: Line %i: Insufficient information to load plugin `%s'.\n",
				    path, lines_read, path
				);
			}

			/* Delete plugin loading values */
			free(path);
			strlistfree(argv, argc);

			break;
		    }
		}
	    }
#endif	/* PLUGIN_SUPPORT */
	    /* ******************************************************** */
	    /* BeginSiteBan */
	    else if(!strcasecmp(parm, "BeginSiteBan"))
	    {
		siteban_struct *sb_ptr;

		/* Allocate new site ban entry */
		if(total_sitebans < 0)
		    total_sitebans = 0;

		n = total_sitebans;
		total_sitebans = n + 1;

		siteban = (siteban_struct **)realloc(
		    siteban,
		    total_sitebans * sizeof(siteban_struct *)
		);
		if(siteban == NULL)
		{
		    total_sitebans = 0;
		    continue;
		}

		siteban[n] = (siteban_struct *)calloc(
		    1, sizeof(siteban_struct)
		);
		if(siteban[n] == NULL)
		{
		    total_sitebans--;
		    continue;
		}

		/* Get pointer to structure */
		sb_ptr = siteban[n];


		while(1)
		{
		    /* Free previous line and allocate/read next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Fetch parameter */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Fetch value */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';


		    /* Address */
		    if(!strcasecmp(parm, "Address"))
		    {
			StringParseIP(
			    val,
			    &(sb_ptr->ip.part_u8[0]),
			    &(sb_ptr->ip.part_u8[1]),
			    &(sb_ptr->ip.part_u8[2]),
			    &(sb_ptr->ip.part_u8[3])
			);
		    }
		    /* Restrict */
		    else if(!strcasecmp(parm, "Restrict"))
		    {
			sb_ptr->restrict = atoi(val);
		    }
		    /* EndSiteBan */
		    else if(!strcasecmp(parm, "EndSiteBan"))
		    {
			break;
		    }
		}
	    }


 	    /* Unknown parameter */
	    else
	    {
		fprintf(
		    stderr,
"%s: Line %i: Unknown parameter: `%s'\n",
		    path,
		    lines_read,
		    parm
		);
	    }
	}

	/* Close the file */
	fclose(fp);


	/* Close incoming connections listening sockets that do not
	 * match any port number in the listening socket ports list
	 */
	if(total_incoming_sockets > 0)
	{
	    int port_num;
	    incoming_socket_struct *insock;

	    /* Iterate through existing incoming sockets list */
	    for(i = 0; i < total_incoming_sockets; i++)
	    {
		insock = incoming_socket[i];
	        if(insock == NULL)
		    continue;

	        /* Get port number */
	        port_num = insock->port_num;

	        /* Iterate through incoming socket port numbers */
		for(n = 0; n < total_lsocket_ports; n++)
	        {
	            if(lsocket_port[n] == port_num)
		        break;
	        }
	        if(n >= total_lsocket_ports)
	        {
		    /* Port number not in the list, so close and
		     * delete this incoming socket
		     */
		    IncomingSocketDelete(i);
	        }
	    }
	}


	/* Delete listening socket ports list */
	free(lsocket_port);
	lsocket_port = NULL;
	total_lsocket_ports = 0;


	return(0);
}
