#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/string.h"

#include "cmd.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


#ifdef PLUGIN_SUPPORT
#include "plugins.h"
#endif 


int CmdHandleInput(int condescriptor, const char *cmd);


/*
 *	Parses and handles a server command from a connection.
 */
int CmdHandleInput(int condescriptor, const char *cmd)
{
	int status, con_obj_num, command_handled;
	char	*s,
		arg[CS_DATA_MAX_LEN],
		command[CS_DATA_MAX_LEN],
		sndbuf[CS_DATA_MAX_LEN + 256],
		text[CS_DATA_MAX_LEN + 256];
	xsw_object_struct *con_obj;
	sws_connection_struct *con;

	if(cmd == NULL)
	    return(-1);

	while(ISBLANK(*cmd))
	    cmd++;

	if(*cmd == '\0')
	    return(-2);

	/* Connection must be valid, allocated, and connected */
	if(!ConIsConnected(condescriptor))
	    return(-1);

	con = connection[condescriptor];

	/* Connection must also be logged in (have a valid object) */
	con_obj_num = con->object_num;
	if(DBIsObjectGarbage(con_obj_num))
	{
	    NetSendLiveMessage(
		condescriptor,
		"You must be logged in to execute a server command."
	    );
	    return(-1); 
	}

	con_obj = xsw_object[con_obj_num];


	/* Parse command and argument */
	strncpy(command, cmd, sizeof(command));
	command[sizeof(command) - 1] = '\0';

	s = strchr(command, ' ');
	if(s == NULL)
	{
	    *arg = '\0';
	}
	else
	{
	    strncpy(arg, s + 1, sizeof(arg));
	    arg[sizeof(arg) - 1] = '\0';

	    *s = '\0';
	}

	/* Strip spaces from command */
	strstrip(command);

	/* Do not strip spaces from argument */


	/* Log this command being typed */
	sprintf(
	    text,
	    "%s: Executed: Command: `%s'  Argument: `%s'",
	    DBGetFormalNameStr(con_obj_num),
	    command,
	    arg
	);
	if(sysparm.log_general)
	    LogAppend(fname.log, text);


	/* The command and argument are now properly parsed and stored
	 * as strings command and arg (respectivly) with the cmd
	 * variable holding the original entire command.
	 *
	 * Now begin handling of the command, first check if the command
	 * matches any of the critical surface level commands, which
	 * are debug, help, plugin, and sysparm.
	 *
	 * Afterwards, if the command is not handled it will be passed
	 * to any plugins to allow them to handle it.  If no plugins
	 * handle the command then pass it to the other internal server
	 * commands.
	 */
	status = 0;
	command_handled = 0;

	/* Debug */
	if(!strcasecmp(command, "debug"))
	{
	    status = CmdDebug(condescriptor, arg);
	    command_handled = 1;
	}
	/* Help */
	else if(!strcasecmp(command, "help") ||
		!strcasecmp(command, "?")
	)
	{
	    status = CmdHelp(condescriptor, arg);
	    command_handled = 1;
	}
	/* Version */ 
	else if(!strcasecmp(command, "version"))
	{
	    status = CmdVersion(condescriptor, arg);
	    command_handled = 1;
	} 
	/* Sysparm */
	else if(!strcasecmp(command, "sysparam") ||
		!strcasecmp(command, "sysparm") ||
		!strcasecmp(command, "tune")
	)
	{
	    status = CmdSysparm(condescriptor, arg);
	    command_handled = 1;
	}
#ifdef PLUGIN_SUPPORT
	/* Plugin */
	else if(!strcasecmp(command, "plugin") ||
		!strcasecmp(command, "plugins")
	)
	{
	    status = CmdPlugin(condescriptor, arg);
	    command_handled = 1;
	}
#endif  /* PLUGIN_SUPPORT */
	if(command_handled)
	    return(status);

#ifdef PLUGIN_SUPPORT
	/* If the command has not been handled above then pass it
	 * to the plugins
	 */
	if(!command_handled)
	{
	    int i;
	    for(i = 0; i < total_plugins; i++)
	    {
		if(PluginCommand(
		    i,
		    condescriptor, con_obj_num,
		    command, arg,
		    con_obj->permission.uid,
		    con_obj->permission.gid
		))
		{
		    status = 0;
		    command_handled = 1;
		    break;
		}
	    }

	    /* If the command was handled by a plugin then return
	     * instead of passing it to the lower level server commands
	     * below
	     */
	    if(command_handled)
		return(status);
	}
#endif  /* PLUGIN_SUPPORT */

        /* Laston */
        if(!strcasecmp(command, "laston"))
        {
            status = CmdLaston(condescriptor, arg);
        }
	/* Link */
	else if(strcasepfx(command, "li"))
	{
	    status = CmdLink(condescriptor, arg);
	}
	/* Memory */
	else if(strcasepfx(command, "mem"))
	{
	    status = CmdMemory(condescriptor, arg);
	}
	/* Disk */
	else if(strcasepfx(command, "disk"))
	{
	    status = CmdDisk(condescriptor, arg);
	}
	/* Password */
	else if(strcasepfx(command, "passw"))
	{
	    status = CmdPassword(condescriptor, arg);
        }
	/* Set */
	else if(strcasepfx(command, "set"))
	{
	    status = CmdSet(condescriptor, arg);
	}
	/* Wall */
	else if(strcasepfx(command, "wal"))
	{
	    status = CmdWall(condescriptor, arg);
	}
	/* Who */
	else if(strcasepfx(command, "who"))
	{
	    status = CmdWho(condescriptor, arg);
	}
	/* Netstat */
	else if(strcasepfx(command, "nets"))
	{
	    status = CmdNetstat(condescriptor, arg);
	}
	/* Score */
	else if(strcasepfx(command, "score"))
	{
	    status = CmdScore(condescriptor, arg);
	}
	/* ETA */
	else if(strcasepfx(command, "eta"))
	{
	    status = CmdETA(condescriptor, arg);
	}
	/* Save */
	else if(strcasepfx(command, "save")) 
	{
	    status = CmdSaveUniverse(condescriptor, arg);
	}
	/* Shutdown */
	else if(strcasepfx(command, "shutdown"))
	{
	    status = CmdShutdown(condescriptor, arg);
	}
	/* Shutdown */
	else if(strcasepfx(command, "siteb"))
	{
	    status = CmdSiteBan(condescriptor, arg);
	}
	/* Boot */
	else if(strcasepfx(command, "boot"))
	{
	    status = CmdBoot(condescriptor, arg);
	}
	/* Sync time */
	else if(strcasepfx(command, "sync"))
	{
	    status = CmdSyncTime(condescriptor, arg);
	}
	/* ID */
	else if(strcasepfx(command, "id"))
	{
	    status = CmdID(condescriptor, arg);
	}
	/* Create player */
	else if(strcasepfx(command, "createp"))
	{
	    status = CmdCreatePlayer(condescriptor, arg);
	}
	/* Create object */
	else if(strcasepfx(command, "create"))
	{
	    status = CmdCreate(condescriptor, arg);
	}
	/* Chown object */
	else if(strcasepfx(command, "cho"))
	{
	    status = CmdChown(condescriptor, arg);
	}
	/* Recycle player */
	else if(strcasepfx(command, "recyclep"))
	{
	    status = CmdRecyclePlayer(condescriptor, arg);
	}
	/* Recycle object */
	else if(strcasepfx(command, "rec"))
	{
	    status = CmdRecycle(condescriptor, arg);
	}
	/* Unrecycle object */
	else if(strcasepfx(command, "unrec"))
	{
	   status = CmdUnrecycle(condescriptor, arg);
	}
	/* Examine object */
	else if(strcasepfx(command, "ex"))
	{
	    status = CmdExamine(condescriptor, arg);
	}
	/* Find */
	else if(strcasepfx(command, "find"))
	{
	    status = CmdFind(condescriptor, arg);
	}
	/* Eco product create */
	else if(strcasepfx(command, "ecoproductc") ||
		strcasepfx(command, "ecoprodc")
	)
	{
	    status = CmdEcoProductCreate(condescriptor, arg);
	}
	/* Eco product create */
	else if(strcasepfx(command, "ecoproducts") ||
		strcasepfx(command, "ecoprods")
	)
	{
	    status = CmdEcoProductSet(condescriptor, arg);
	}
	/* Eco product delete */
	else if(strcasepfx(command, "ecoproductd") ||
		strcasepfx(command, "ecoprodd")
	)
	{
	    status = CmdEcoProductDelete(condescriptor, arg);
	}
	/* Signup */
	else if(strcasepfx(command, "signup"))
	{
	    status = CmdSignup(condescriptor, arg);
	}


	/* Test */
	else if(!strcasecmp(command, "test")) 
	{
	    status = CmdTest(condescriptor, arg);
	}

	/* Unknown command */
	else
	{
	    status = -1;

	    sprintf(
		sndbuf,
		"%s: no such command or macro.",
		command
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}


	return(status);
}
