#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int CMDECO_PERMISSION_CHECK(int condescriptor, int permission);
int CmdEcoProductCreate(int condescriptor, const char *arg);
int CmdEcoProductSet(int condescriptor, const char *arg);
int CmdEcoProductDelete(int condescriptor, const char *arg);


/*
 *	Checks if the connection matches or surpasses the specified
 *	permission level.
 */
int CMDECO_PERMISSION_CHECK(int condescriptor, int permission)
{
	int obj_num;
	xsw_object_struct *obj;

	/* Connection must be logged in */
	if(!ConIsLoggedIn(condescriptor))
	    return(0);

	/* Get connection's object */
	obj_num = connection[condescriptor]->object_num;
	if(DBIsObjectGarbage(obj_num))
	    return(0);
	else
	    obj = xsw_object[obj_num];

	/* Check object's permission */
	if(obj->permission.uid <= permission)
	    return(1);
	else
	    return(0);
}


/*
 *	Creates an eco product on an object.
 *
 *	Syntax: "<object>=<product_name>"
 */
int CmdEcoProductCreate(int condescriptor, const char *arg)
{
	int i, obj_num, status;
	char		*s,
			sndbuf[CS_DATA_MAX_LEN],
			name[SW_OBJ_NAME_MAX],
			product_name[ECO_PRODUCT_NAME_MAX],
			tmp_name1[SW_OBJ_NAME_MAX],
			tmp_name2[SW_OBJ_NAME_MAX];
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[condescriptor];
	xsw_ecodata_struct *eco;

	/* Parse object name from argument */
	strncpy(name, arg, sizeof(name));
	name[sizeof(name) - 1] = '\0';
	s = strchr(name, '=');
	if(s != NULL)
	{
	    *s = '\0';
	}
	else
	{
	    sprintf(
		sndbuf,
"Usage: `ecoprodcreate <object>=<product_name>'"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}
	strstrip(name);

	/* Parse product name from argument */
	while((*arg != '=') && (*arg != '\0'))
	    arg++;
	if(*arg == '=')
	    arg++;
	strncpy(product_name, arg, sizeof(product_name));
	product_name[sizeof(product_name) - 1] = '\0';
	s = strchr(product_name, '=');
	if(s != NULL)
	    *s = '\0';
	strstrip(product_name);


	/* Product name must be specified */
	if(*product_name == '\0')
	    return(-1);


	/* Match object by name, first check for special cases */
	/* Locked on object */
	if(*name == '\0')
	{
	    obj_num = con->object_num;
	    if(!DBIsObjectGarbage(obj_num))
		obj_num = xsw_object[obj_num]->locked_on;
	}
	/* Connection's object */
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con->object_num;
	}
	/* All else match by name */
	else
	{
	    obj_num = MatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    sprintf(
		sndbuf,
		"%s: No such object.",
		name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}

	obj = xsw_object[obj_num];


	/* Permission check */
	if(obj->owner != con->object_num)
	{
	    if(!CMDECO_PERMISSION_CHECK(condescriptor, SWS_ACCESS_UID_SETO))
	    {
		sprintf(
		    sndbuf,
"ecoprodcreate: You do not own that object."
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(-1);
	    }
	}
	else
	{
	    if(!CMDECO_PERMISSION_CHECK(condescriptor, SWS_ACCESS_UID_SET))
	    {
		sprintf(
		    sndbuf,
"ecoprodcreate: Permission denied: Requires access level %i.",
		    SWS_ACCESS_UID_SET
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(-1);
	    }
	}


	/* Allocate economy data on the specified object as needed */
	if(UNVAllocEco(obj))
	{
	    sprintf(
		sndbuf,
"ecoprodcreate: %s: Unable to create economy data on this object.",
		name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);

	    return(-1);
	}

	/* Get pointer to the economy data on the object */
	eco = obj->eco;
	if(eco == NULL)
	    return(-1);


	/* Check if the product already exists */
	for(i = 0; i < eco->total_products; i++)
	{
	    if(eco->product[i] == NULL)
		continue;

	    if(!strcasecmp(eco->product[i]->name, product_name))
	    {
		sprintf(
		    sndbuf,
"ecoprodcreate: %s: Product `%s' already exists.",
		    name, product_name
		);
		NetSendLiveMessage(condescriptor, sndbuf);

		return(-2);
	    }
	}


	/* Create new product */
	status = DBCreateObjectEconomyProduct(
	    obj_num,
	    product_name,
	    1.0,		/* Sell price */
	    1.0,		/* Buy price */
	    1.0,		/* Amount */
	    1.0			/* Amount max */
	);
	if(status)
	{
	    sprintf(
		sndbuf,
"ecoprodcreate: %s: Unable to create product `%s'.",
		name, product_name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}
	else
	{
	    /* Print response and log */
	    strncpy(
		tmp_name1, DBGetFormalNameStr(obj_num),
		sizeof(tmp_name1)
	    );
	    tmp_name1[sizeof(tmp_name1) - 1] = '\0';
	    strncpy(
		tmp_name2,
		DBGetFormalNameStr(con->object_num),
		sizeof(tmp_name2)
	    );
	    tmp_name2[sizeof(tmp_name2) - 1] = '\0';

	    sprintf(
		sndbuf,
"ecoprodcreate: %s: Created product `%s'.",
		tmp_name1, product_name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);

	    sprintf(
		sndbuf,
"%s: Created economy product %s on %s.",
		tmp_name2,
		product_name,
		tmp_name1
	    );
	    if(sysparm.log_general)
		LogAppend(fname.log, sndbuf);
	}

	/* Sort products */
	DBSortEconomyProducts(obj_num);

	return(0);
}

/*
 *	Sets parameter of product on object.
 *
 *	Syntax: "<object>=<product_name>=<parm>:<value>"
 */
int CmdEcoProductSet(int condescriptor, const char *arg)
{
	int i, obj_num;
	char		*s,
			sndbuf[CS_DATA_MAX_LEN],
			name[SW_OBJ_NAME_MAX],
			product_name[ECO_PRODUCT_NAME_MAX],
			parm[CS_DATA_MAX_LEN],
			val[CS_DATA_MAX_LEN],
			tmp_name1[SW_OBJ_NAME_MAX],
			tmp_name2[SW_OBJ_NAME_MAX];
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[condescriptor];
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *product;

	/* Parse object name from argument */
	strncpy(name, arg, sizeof(name));
	name[sizeof(name) - 1] = '\0';
	s = strchr(name, '=');
	if(s != NULL)
	{
	    *s = '\0';
	}
	else
	{
	    sprintf(
		sndbuf,
"Usage: `ecoprodset <object>=<product_name>=<parm>:<value>'"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}
	strstrip(name);

	/* Parse product name from argument */
	while((*arg != '=') && (*arg != '\0'))
	    arg++;
	if(*arg == '=')
	    arg++;
	strncpy(product_name, arg, sizeof(product_name));
	product_name[sizeof(product_name) - 1] = '\0';
	s = strchr(product_name, '=');
	if(s != NULL)
	    *s = '\0';
	strstrip(product_name);

	/* Parse parameter from argument */
	while((*arg != '=') && (*arg != '\0'))
	    arg++;
	if(*arg == '=')
	    arg++;
	strncpy(parm, arg, sizeof(parm));
	parm[sizeof(parm) - 1] = '\0';
	s = strchr(parm, ':');
	if(s != NULL)
	    *s = '\0';
	strstrip(parm);

	/* Parse value from argument */
	while((*arg != ':') && (*arg != '\0'))
	    arg++;
	if(*arg == ':')
	    arg++;
	strncpy(val, arg, sizeof(val));
	val[sizeof(val) - 1] = '\0';
	strstrip(val);


	/* Product name must be specified */
	if(*product_name == '\0')
	    return(-1);


	/* Match object by name, first check for special cases */
	/* Locked on object */
	if(*name == '\0')
	{
	    obj_num = con->object_num;
	    if(!DBIsObjectGarbage(obj_num))
		obj_num = xsw_object[obj_num]->locked_on;
	}
	/* Connection's object */
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con->object_num;
	}
	/* All else match by name */
	else
	{
	    obj_num = MatchObjectByName( 
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    sprintf(
		sndbuf,
		"%s: No such object.",
		name   
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}

	obj = xsw_object[obj_num];


	/* Permission check */
	if(obj->owner != con->object_num)
	{
	    if(!CMDECO_PERMISSION_CHECK(condescriptor, SWS_ACCESS_UID_SETO))
	    {
		sprintf(
		    sndbuf,
"ecoprodset: You do not own that object."
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(-1);
	    }
	}
	else
	{
	    if(!CMDECO_PERMISSION_CHECK(condescriptor, SWS_ACCESS_UID_SET))
	    {
		sprintf(
		    sndbuf,
"ecoprodset: Permission denied: Requires access level %i.",
		    SWS_ACCESS_UID_SET
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(-1);
	    }
	}

	/* Get pointer to the economy data on the object */
	eco = obj->eco;
	if(eco == NULL)
	{
	    sprintf(
		sndbuf,
"ecoprodset: %s: No such product `%s'.",
		name,
		product_name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}

	/* Get pointer to product */
	product = NULL;
	if(isdigit(*product_name))
	{
	    /* Match product by number */
	    i = atoi(product_name);
	    if((i >= 0) && (i < eco->total_products))
		product = eco->product[i];
	}
	else
	{
	    /* Match product by name */
	    for(i = 0; i < eco->total_products; i++)
	    {
	        if(eco->product[i] == NULL)
		    continue;

	        if(!strcasecmp(eco->product[i]->name, product_name))
	        {
		    product = eco->product[i];
		    break;
	        }
	    }
	}
	if(product == NULL)
	{
	    sprintf(
		sndbuf,
"ecoprodset: %s: No such product `%s'.",
		name,
		product_name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}


	/* Name */
	if(!strcmp(parm, "name"))
	{
	    strncpy(product->name, val, sizeof(product->name));
	    product->name[sizeof(product->name) - 1] = '\0';
	}
	/* Buy price */
	else if(!strcmp(parm, "buy_price"))
	{
	    product->buy_price = atof(val);
	}
	/* Sell price */
	else if(!strcmp(parm, "sell_price"))
	{
	    product->sell_price = atof(val);
	}
	/* Amount */
	else if(!strcmp(parm, "amount"))
	{
	    product->amount = atof(val);
	}
	/* Amount max */
	else if(!strcmp(parm, "amount_max"))
	{
	    product->amount_max = atof(val);
	}
	else
	{
	    sprintf(
		sndbuf,
"ecoprodset: %s: No such economy product parameter `%s'.",
		name, parm
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}

 
	/* Print response and log */
	strncpy(
	    tmp_name1, DBGetFormalNameStr(obj_num),
	    sizeof(tmp_name1)
	);
	tmp_name1[sizeof(tmp_name1) - 1] = '\0';
	strncpy(
	    tmp_name2,
	    DBGetFormalNameStr(con->object_num),
	    sizeof(tmp_name2)
	);
	tmp_name2[sizeof(tmp_name2) - 1] = '\0';

	sprintf(
	    sndbuf,
"ecoprodset: %s: Set product `%s' parameter `%s' to `%s'.",
	    tmp_name1, product_name, parm, val
	);
	NetSendLiveMessage(condescriptor, sndbuf);

	sprintf(
	    sndbuf,
"%s: Set economy product `%s' %s to `%s' on %s.",
	    tmp_name2,
	    product_name,
	    parm,
	    val,
	    tmp_name1
	);
	if(sysparm.log_general)
	    LogAppend(fname.log, sndbuf);

	/* Sort products */
	DBSortEconomyProducts(obj_num);

	return(0);
}


/*
 *	Deletes a product entry on an object.
 *
 *	Syntax: "<object>=<product_name>"
 */
int CmdEcoProductDelete(int condescriptor, const char *arg)
{
	int i, obj_num;
	char		*s,
			sndbuf[CS_DATA_MAX_LEN],
			name[SW_OBJ_NAME_MAX],
			product_name[ECO_PRODUCT_NAME_MAX],
			tmp_name1[SW_OBJ_NAME_MAX],
			tmp_name2[SW_OBJ_NAME_MAX];
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[condescriptor];
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *product;

	/* Parse object name from argument */
	strncpy(name, arg, sizeof(name));
	name[sizeof(name) - 1] = '\0';
	s = strchr(name, '=');
	if(s != NULL)
	{
	    *s = '\0';
	}
	else
	{   
	    sprintf(
		sndbuf,
"Usage: `ecoproddelete <object>=<product_name>'"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}
	strstrip(name);

	/* Parse product name from argument */
	while((*arg != '=') && (*arg != '\0'))
	    arg++;
	if(*arg == '=')
	    arg++;
	strncpy(product_name, arg, sizeof(product_name));
	product_name[sizeof(product_name) - 1] = '\0';
	s = strchr(product_name, '=');
	if(s != NULL)
	    *s = '\0';
	strstrip(product_name);


	/* Product name must be specified */
	if(*product_name == '\0')
	    return(-1);


	/* Match object by name, first check for special cases */
	/* Locked on object */
	if(*name == '\0')
	{
	    obj_num = con->object_num;
	    if(!DBIsObjectGarbage(obj_num))
		obj_num = xsw_object[obj_num]->locked_on;
	}
	/* Connection's object */
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con->object_num;
	}
	/* All else match by name */
	else
	{
	    obj_num = MatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    sprintf(
		sndbuf,
		"%s: No such object.",
		name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}

	obj = xsw_object[obj_num];


	/* Permission check */
	if(obj->owner != con->object_num)
	{
	    if(!CMDECO_PERMISSION_CHECK(condescriptor, SWS_ACCESS_UID_SETO))
	    {
		sprintf(
		    sndbuf,
"ecoproddelete: You do not own that object."
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(-1);
	    }
	}
	else
	{
	    if(!CMDECO_PERMISSION_CHECK(condescriptor, SWS_ACCESS_UID_SET))
	    {
		sprintf(
		    sndbuf,
"ecoproddelete: Permission denied: Requires access level %i.",
		    SWS_ACCESS_UID_SET
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(-1);
	    }
	}

	/* Get pointer to the economy data on the object */
	eco = obj->eco;
	if(eco == NULL)
	    return(-1);


	/* Get pointer to product */
	product = NULL;
	if(isdigit(*product_name))
	{
	    /* Match product by number */
	    i = atoi(product_name);
	    if((i >= 0) && (i < eco->total_products))
		product = eco->product[i];
	}
	else
	{
	    /* Match product by name */
	    for(i = 0; i < eco->total_products; i++)
	    {
		if(eco->product[i] == NULL)
		    continue;

		if(!strcasecmp(eco->product[i]->name, product_name))
		{
		    product = eco->product[i];
		    break;
		}
	    }
	}
	if(product == NULL)
	{
	    sprintf(
		sndbuf,
"ecoproddeletet: %s: No such product `%s'.",
		name,
		product_name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}

	/* Delete product */
	free(product);
	eco->product[i] = product = NULL;

	/* Print response and log */
	strncpy(
	    tmp_name1, DBGetFormalNameStr(obj_num),
	    sizeof(tmp_name1)
	);
	tmp_name1[sizeof(tmp_name1) - 1] = '\0';
	strncpy(
	    tmp_name2,
	    DBGetFormalNameStr(con->object_num),
	    sizeof(tmp_name2)
	);
	tmp_name2[sizeof(tmp_name2) - 1] = '\0';

	sprintf(
	    sndbuf,
"ecoproddelete: %s: Deleted product `%s'.",
	    tmp_name1, product_name
	);
	NetSendLiveMessage(condescriptor, sndbuf);

	sprintf(
	    sndbuf,
"%s: Deleted economy product %s on %s.",
	    tmp_name2,
	    product_name,
	    tmp_name1
	);
	if(sysparm.log_general)
	    LogAppend(fname.log, sndbuf);


	/* If all product pointers are NULL then delete products */
	for(i = 0; i < eco->total_products; i++)
	{
	    if(eco->product[i] != NULL)
		break;
	}
	if(i >= eco->total_products)
	{
	    free(eco->product);
	    eco->product = NULL;
	    eco->total_products = 0;
	}

	/* Sort products */
	DBSortEconomyProducts(obj_num);

	return(0);
}
