#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fnmatch.h>

#include "../include/string.h"
#include "../include/strexp.h"

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


#define THIS_CMD_NAME	"find"


static int CmdFindMatch(const char *pattern, const char *string);


/*
 *	Matches string with pattern, using fnmatch().
 */
static int CmdFindMatch(const char *pattern, const char *string)
{
	int status;
	char	*tmp_pattern = strdup(pattern),
		*tmp_string = strdup(string);

	strtoupper(tmp_pattern);
	strtoupper(tmp_string);

	status = fnmatch(tmp_pattern, tmp_string, 0);

	free(tmp_pattern);
	free(tmp_string);

	return(status);
}


int CmdFind(int condescriptor, const char *arg)
{
	int len, status, argc;
	char	larg[CS_DATA_MAX_LEN],
		pattern[CS_DATA_MAX_LEN],
		type_str[CS_DATA_MAX_LEN],
		name1[SW_OBJ_NAME_MAX + 80],
		name2[SW_OBJ_NAME_MAX + 80],
		**argv;

	const int delta_str_len = 256;
	char delta_str[delta_str_len];

	const int position_str_len = 256;
	char position_str[position_str_len];

	int obj_num, objects_found, con_obj_num;
	xsw_object_struct *con_obj, *obj;
	sws_connection_struct *con = connection[condescriptor];
	char sndbuf[CS_DATA_MAX_LEN + delta_str_len + position_str_len];

	/* Get connection's object (assumed valid) */
	con_obj_num = con->object_num;
	con_obj = xsw_object[con_obj_num];

	/* Print usage? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    NetSendLiveMessage(
		condescriptor,
		"Usage: `find <name_pattern>[=<type>]'"
	    );
	    return(-1);
	}


	/* Copy argument */
	strncpy(larg, arg, sizeof(larg));
	larg[sizeof(larg) - 1] = '\0';   
	strstrip(larg);


	/* Check if connection object is allowed to find? */
	if(con_obj->permission.uid > SWS_ACCESS_UID_FIND)
	{
	    sprintf(
		sndbuf,
		"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_FIND
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}


	/* Parse and get pattern and type_str */
	argv = strchrexp(larg, '=', &argc);

	strncpy(
	    pattern,
	    (argc >= 1) ? argv[0] : "*",
	    sizeof(pattern)
	);
	pattern[sizeof(pattern) - 1] = '\0';
	strstrip(pattern);

	strncpy(
	    type_str,
	    (argc >= 2) ? argv[1] : "*",
	    sizeof(type_str)
	);
	type_str[sizeof(type_str) - 1] = '\0';
	strstrip(type_str);


	strlistfree(argv, argc);


	/* Iterate through the objects list */
	objects_found = 0;
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Allowed to match others? */
	    if((obj->owner != con_obj_num) &&
	       (con_obj->permission.uid > SWS_ACCESS_UID_FINDO)
	    )
		continue;


	    /* Type match? */
	    switch(obj->type)
	    {
	      case SW_OBJ_TYPE_STATIC:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_STATIC))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_DYNAMIC:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_DYNAMIC))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_CONTROLLED:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_CONTROLLED))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_PLAYER:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_PLAYER))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_WEAPON:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_WEAPON))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_STREAMWEAPON:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_STREAMWEAPON))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_SPHEREWEAPON:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_SPHEREWEAPON))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_HOME:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_HOME))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_AREA:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_AREA))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_ANIMATED:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_ANIMATED))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_WORMHOLE:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_WORMHOLE))
		    status = -1;
		else
		    status = 0;
		break;

	      case SW_OBJ_TYPE_ELINK:
		if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_ELINK))
		    status = -1;
		else
		    status = 0;
		break;

	      default:
		status = 0;
		break;
	    }
	    if(status)
		continue;


	    /* Name match? */
	    if(CmdFindMatch(pattern, obj->name))
		continue;

	    /* Consider this object found */
	    objects_found++;

	    /* First object found? */
	    if(objects_found == 1)
	    {
		/* Print heading for the first time */
		NetSendLiveMessage(
		    condescriptor,
"Name                     Owner                    Location"
		);
	    }

	    /* Format delta string */
	    if(Mu3DInSectorPtr(
		obj,
		con_obj->sect_x,
		con_obj->sect_y,
		con_obj->sect_z
	    ))
	    {
		double dx = obj->x - con_obj->x;
		double dy = obj->y - con_obj->y;
		const char *units_str = "RU";

		sprintf(
		    delta_str,
		    "Brg: %.2f' D: %.2f %s",
		    RADTODEG(MuCoordinateDeltaVector(dx, dy)),
		    Mu3DDistance(dx, dy, 0),
		    units_str
		);
	    }
	    else
	    {
		double dx = obj->sect_x - con_obj->sect_x;
		double dy = obj->sect_y - con_obj->sect_y;
		const char *units_str = "S";

		sprintf(
		    delta_str,
		    "Brg: %.2f' D: %.2f %s",
		    RADTODEG(MuCoordinateDeltaVector(dx, dy)),
		    Mu3DDistance(dx, dy, 0),
		    units_str
		);
	    }

	    /* Format position string */
	    UNVLocationFormatString(
		position_str,
		&obj->sect_x, &obj->sect_y, &obj->sect_z,
		&obj->x, &obj->y, &obj->z,
		position_str_len
	    );

	    /* Format names */
	    strncpy(
		name1, DBGetFormalNameStr(obj_num),
		sizeof(name1)
	    );
	    name1[sizeof(name1) - 1] = '\0';
	    strncpy(
		name2, DBGetFormalNameStr(obj->owner),
		sizeof(name2)
	    );
	    name2[sizeof(name2) - 1] = '\0';

	    strcpy(sndbuf, name1);
	    len = strlen(sndbuf);
	    strpad(&sndbuf[len], 25 - len);

	    strcpy(&sndbuf[25], name2);
	    len = strlen(sndbuf);
	    strpad(&sndbuf[len], 50 - len);

	    strcpy(&sndbuf[50], delta_str);
	    len = strlen(sndbuf);
	    strpad(&sndbuf[len], 78 - len);

	    strcpy(&sndbuf[78], position_str);

	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	/* Print number of objects found */
	if(objects_found > 0)
	    sprintf(
		sndbuf, "*** %i Object%s Found ***",
		objects_found,
		(objects_found == 1) ? "" : "s"
	    );
	else
	    sprintf(sndbuf, "*** No Objects Found ***");
	NetSendLiveMessage(condescriptor, sndbuf);


	return(0);
}
