#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fnmatch.h>
#include <sys/types.h>

#include "../include/string.h"
#include "../include/unvutil.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


#define THIS_CMD_NAME	"sysparm"


static int CmdSysparmMatchParm(const char *str, const char *parm);
static char *CmdSysparmFmtDeltaTime(time_t dt);
int CmdSysparm(int condescriptor, const char *arg);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns true if str matches parm, using fnmatch() style
 *	wildcard matching.
 */
static int CmdSysparmMatchParm(const char *str, const char *parm)
{
	return(!fnmatch(str, parm, 0));
}


/*
 *	Returns a statically allocated string containing a verbose
 *	explaination of the delta time in seconds.
 */
static char *CmdSysparmFmtDeltaTime(time_t dt)
{
	static char s[256];

	if(dt < 0l)
	    dt = 0l;

	if(dt < 60)
	    sprintf(s, "%ld s", dt);
	else if(dt < 3600)
	    sprintf(s, "%ld m", dt / 60);
	else if(dt < 86400)
	    sprintf(s, "%ld h", dt / 3600);
	else
	    sprintf(s, "%ld d", dt / 86400);
	s[sizeof(s) - 1] = '\0';

	return(s);
}


/*
 *      Set or view system parameters.
 */
int CmdSysparm(int condescriptor, const char *arg)
{
	int con_obj_num, matches = 0;
	char		*s, *parm_name,
			parm[CS_DATA_MAX_LEN],
			val[CS_DATA_MAX_LEN],
			sndbuf[CS_DATA_MAX_LEN + SW_UNV_TITLE_MAX + 512];
	xsw_object_struct *con_obj;

	/* Get connection's object number (assumed valid) */
	sws_connection_struct *con = connection[condescriptor];
	con_obj_num = con->object_num;
	con_obj = xsw_object[con_obj_num];

	/* Reset send buffer */
	*sndbuf = '\0';

	/* Check if allowed to view or set system parameters */
	if(con_obj->permission.uid > SWS_ACCESS_UID_SYSPARM)
	{
	    sprintf(
		sndbuf,
		"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME,
		SWS_ACCESS_UID_SYSPARM
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(-1);
	}


	/* Parse argument */
	s = strchr(arg, '=');
	if(s == NULL)
	{
	    strncpy(parm, arg, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';
	    strstrip(parm);

	    *val = '\0';

	    if(*parm == '\0')
	    {
		strncpy(parm, "*", sizeof(parm));
		parm[sizeof(parm) - 1] = '\0';
	    }
	}
	else
	{
	    strncpy(val, s + 1, sizeof(val));
	    val[sizeof(val) - 1] = '\0';
	    strstrip(val);

	    *s = '\0';
	    strncpy(parm, arg, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';
	    strstrip(parm);
	}


	/* Check input parameter through each parameter, set the
	 * new value if new value is specified
	 */
	parm_name = "allow_guest";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.allow_guest = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.allow_guest ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "allow_guest_signup";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.allow_guest_signup = StringIsYes(val);
	    }
	    sprintf(
		sndbuf, 
		"%s = %s",
		parm_name,
		sysparm.allow_guest_signup ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "cease_fire";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.cease_fire = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.cease_fire) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "cloak_am_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.cloak_am_rate = MAX(atof(val), 0.0);
	    }
	    sprintf(
		sndbuf,
		"%s = %.4f",
		parm_name,   
		sysparm.cloak_am_rate
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}     

	parm_name = "com_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.com_range = MAX(atof(val), 0.0);
	    }
	    sprintf(
		sndbuf,
		"%s = %.2f ru",
		parm_name,
		sysparm.com_range
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "con_notify"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.con_notify = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.con_notify ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "convert_ru_to_au";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sw_units.ru_to_au = atof(val);
		if(sw_units.ru_to_au <= 0.0)
		    sw_units.ru_to_au = 1.0;

		/* Update units to all connections */
		NetSendUnits(-1);
	    }
	    sprintf(
		sndbuf,
		"%s = %.4f",
		parm_name,
		sw_units.ru_to_au
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "cpu_priority";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];

	    matches++;
	    if(*val != '\0')
	    {
		if(isdigit(*val))
		{
		    sysparm.cpu_priority = atoi(val);
		}
		else
		{
		    if(!strcasecmp(val, "highest"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGHEST;
		    else if(strcasepfx(val, "h"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGH;
		    else if(strcasepfx(val, "m"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_MEDIUM;
		    else if(strcasepfx(val, "l"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_LOWEST;
		}
	    }
	    switch(sysparm.cpu_priority)
	    {
	      case SWS_CPU_PRIORITY_HIGHEST:
		strncpy(text, "highest", sizeof(text));
		break;
	      case SWS_CPU_PRIORITY_HIGH:
		strncpy(text, "high", sizeof(text));
		break;
	      case SWS_CPU_PRIORITY_MEDIUM:
		strncpy(text, "medium", sizeof(text));
		break;
	      default:	/* SWS_CPU_PRIORITY_LOWEST */
		strncpy(text, "lowest", sizeof(text));
		break;
	    }
	    text[sizeof(text) - 1] = '\0';
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name, text
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "dmg_ctl_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.dmg_ctl_rate = MAX(atof(val), 0.0);
	    }
	    sprintf(
		sndbuf,
		"%s = %f",
		parm_name,
		sysparm.dmg_ctl_rate
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "file_imgrefs";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    unv_head.imgrefs, val,
		    sizeof(unv_head.imgrefs)
		);
		unv_head.imgrefs[sizeof(unv_head.imgrefs) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		unv_head.imgrefs
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "file_ocsn"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {       
		strncpy(
		    unv_head.ocsn, val, sizeof(unv_head.ocsn)
		);
		unv_head.ocsn[sizeof(unv_head.ocsn) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		unv_head.ocsn   
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "file_sndrefs";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    unv_head.sndrefs, val, sizeof(unv_head.sndrefs)
		);
		unv_head.sndrefs[sizeof(unv_head.sndrefs) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		unv_head.sndrefs
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "guest_login_name";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    sysparm.guest_login_name, val,
		    sizeof(sysparm.guest_login_name)
		);
		sysparm.guest_login_name[
		    sizeof(sysparm.guest_login_name) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.guest_login_name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "guest_start_direction";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];

	    matches++;
	    if(*val != '\0')
	    {
		UNVParseDirection(
		    val,
		    &unv_head.guest_start_heading,
		    &unv_head.guest_start_pitch,
		    &unv_head.guest_start_bank
		);
	    }
	    UNVDirectionFormatString(
		text,
		&unv_head.guest_start_heading,
		&unv_head.guest_start_pitch,
		&unv_head.guest_start_bank,
		sizeof(text)
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "guest_start_location";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];

	    matches++;
	    if(*val != '\0')
	    { 
		UNVParseLocation(
		    val,
		    &unv_head.guest_start_sect_x,
		    &unv_head.guest_start_sect_y,
		    &unv_head.guest_start_sect_z,
		    &unv_head.guest_start_x,
		    &unv_head.guest_start_y,
		    &unv_head.guest_start_z
		);
	    }
	    UNVLocationFormatString(
		text,
		&unv_head.guest_start_sect_x,
		&unv_head.guest_start_sect_y,
		&unv_head.guest_start_sect_z,
		&unv_head.guest_start_x,
		&unv_head.guest_start_y,
		&unv_head.guest_start_z,
		sizeof(text)
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "hide_players_home";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.hide_players_home = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.hide_players_home) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "hit_player_bonus";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.hit_player_bonus = atof(val);
		if(sysparm.hit_player_bonus < 0.0)
		    sysparm.hit_player_bonus = 0.0;
	    }

	    if(sysparm.hit_player_bonus == 0.0)
		sprintf(
		    sndbuf,
		    "%s = %s",
		    parm_name,
		    "none"
		);
	    else
		sprintf(
		    sndbuf,
		    "%s = %f",
		    parm_name,
		    sysparm.hit_player_bonus
		);
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "homes_destroyable";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.homes_destroyable = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.homes_destroyable ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_aux_con_stats";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_aux_con_stats = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_aux_con_stats
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_memory_clean";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_memory_clean = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_memory_clean
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_new_connection_poll"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_new_connection_poll = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_new_connection_poll
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_object_values";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_object_values = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_object_values
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_os_stats";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_os_stats = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld s  (next: %s)",
		parm_name,
		sysparm.int_os_stats,
		CmdSysparmFmtDeltaTime(next.os_stats - cur_systime)
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_stats_export";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_stats_export = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld s  (next: %s)",
		parm_name,
		sysparm.int_stats_export,
		CmdSysparmFmtDeltaTime(next.stats_export - cur_systime)
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_system_check";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_system_check = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_system_check
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_unv_save";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_unv_save = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld s  (next: %s)",
		parm_name,
		sysparm.int_unv_save,
		CmdSysparmFmtDeltaTime(next.unv_save - cur_systime)
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "int_weapon_values";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_weapon_values = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_weapon_values
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "killer_gets_credits";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.killer_gets_credits = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.killer_gets_credits) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "log_errors"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.log_errors = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.log_errors) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "log_events";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.log_events = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.log_events) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "log_general";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.log_general = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.log_general) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "log_net";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.log_net = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.log_net) ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "login_timeout";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.login_timeout = dt;
	    }
	    sprintf(
		sndbuf,
		"%s = %ld s",
		parm_name,
		sysparm.login_timeout
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "lost_found_owner";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		const int obj_num = MatchObjectByName(
		    xsw_object, total_objects,
		    val, SW_OBJ_TYPE_PLAYER
		);
		if(DBIsObjectGarbage(obj_num))
		{
		    sprintf(
			sndbuf,
			"%s: No such player object.",
			val
		    );
		}
		else
		{
		    unv_head.lost_found_owner = obj_num;
		    sprintf(
			sndbuf,
			"%s = %s",
			parm_name,
			DBGetFormalNameStr(unv_head.lost_found_owner)
		    );
		}
	    }
	    else
	    {
		sprintf(
		    sndbuf,
		    "%s = %s",
		    parm_name,
		    DBGetFormalNameStr(unv_head.lost_found_owner)
		);
	    }
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "max_aux_connections";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		int i = atoi(val);
		if(i < 0)
		    i = 0;
		sysparm.max_aux_connections = i;
	    }
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_aux_connections
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "max_connections";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		int i = atoi(val);
		if(i < 1)
		    i = 1;
		sysparm.max_connections = i;
	    }
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_connections
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "max_failed_logins";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		int i = atoi(val);
		if(i < 1)
		    i = 1;
		sysparm.max_failed_logins = i;
	    }
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_failed_logins
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "max_guests";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		int i = atoi(val);
		if(i < 0)
		    i = 0;
		sysparm.max_guests = i;
	    }
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_guests
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "mesg_leave";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    sysparm.mesg_leave, val, sizeof(sysparm.mesg_leave)
		);
		sysparm.mesg_leave[sizeof(sysparm.mesg_leave) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_leave
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "mesg_no_guests";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    sysparm.mesg_no_guests, val,
		    sizeof(sysparm.mesg_no_guests)
		);
		sysparm.mesg_no_guests[
		    sizeof(sysparm.mesg_no_guests) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_no_guests
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "mesg_welcome";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    sysparm.mesg_welcome, val,
		    sizeof(sysparm.mesg_welcome)
		);
		sysparm.mesg_welcome[
		    sizeof(sysparm.mesg_welcome) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_welcome
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "mesg_wrong_login";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    sysparm.mesg_wrong_login, val,
		    sizeof(sysparm.mesg_wrong_login)
		);
		sysparm.mesg_wrong_login[
		    sizeof(sysparm.mesg_wrong_login) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_wrong_login
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "player_start_direction";  
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];
	 
	    matches++;
	    if(*val != '\0')
	    {
		UNVParseDirection(
		    val,
		    &unv_head.player_start_heading,
		    &unv_head.player_start_pitch,
		    &unv_head.player_start_bank
		);
	    }
	    UNVDirectionFormatString(
		text,
		&unv_head.player_start_heading,
		&unv_head.player_start_pitch,
		&unv_head.player_start_bank,
		sizeof(text)
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "player_start_location";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];
		
	    matches++;
	    if(*val != '\0')
	    {
		UNVParseLocation(
		    val,
		    &unv_head.player_start_sect_x,
		    &unv_head.player_start_sect_y,
		    &unv_head.player_start_sect_z,
		    &unv_head.player_start_x,
		    &unv_head.player_start_y,
		    &unv_head.player_start_z
		);
	    }
	    UNVLocationFormatString(
		text,
		&unv_head.player_start_sect_x,
		&unv_head.player_start_sect_y,
		&unv_head.player_start_sect_z,
		&unv_head.player_start_x,
		&unv_head.player_start_y,
		&unv_head.player_start_z,
		sizeof(text)
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "report_destroyed_weapons";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.report_destroyed_weapons = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.report_destroyed_weapons ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "respawn_players";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.respawn_players = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.respawn_players ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "send_starchart";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.send_starchart = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.send_starchart ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "unv_title";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		strncpy(
		    unv_head.title, val, sizeof(unv_head.title)
		);
		unv_head.title[
		    sizeof(unv_head.title) - 1] = '\0';
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		unv_head.title
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "tractor_beam_am_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.tractor_beam_am_rate = MAX(atof(val), 0.0);
	    }
	    sprintf(
		sndbuf,
		"%s = %.4f",
		parm_name,
		sysparm.tractor_beam_am_rate
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "tractor_beam_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.tractor_beam_range = MAX(atof(val), 0.0);
	    }
	    sprintf(
		sndbuf,
		"%s = %.2f ru",
		parm_name,
		sysparm.tractor_beam_range
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "transporter_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.transporter_range = MAX(atof(val), 0.0);
	    }
	    sprintf(
		sndbuf,
		"%s = %.2f ru",
		parm_name,
		sysparm.transporter_range
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "use_server_ai";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		sysparm.use_server_ai = StringIsYes(val);
	    }
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.use_server_ai ? "yes" : "no"
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}

	parm_name = "weapon_disarm_lifespan";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(*val != '\0')
	    {
		time_t nval;

		if(isdigit(*val))
		{
		    nval = (time_t)atol(val);
		}
		else
		{
		    if(!strcasecmp(val, "infinate") ||
		       strcasepfx(val, "i")
		    )
			nval = 0l;
		    else if(!strcasecmp(val, "no_change") ||
			    !strcasecmp(val, "no change") || 
			    strcasepfx(val, "n")
		    )
			nval = 1l;
		    else
			nval = 1l;
		}
		sysparm.weapon_disarm_lifespan = nval;
	    }
	    if(sysparm.weapon_disarm_lifespan == 0l)
		sprintf(
		    sndbuf,
		    "%s = infinate",
		    parm_name
		);
	    else if(sysparm.weapon_disarm_lifespan == 1l)
		sprintf(
		    sndbuf,
		    "%s = no_change",
		    parm_name
		);
	    else
		sprintf(     
		    sndbuf,
		    "%s = %ld ms",
		    parm_name,
		    sysparm.weapon_disarm_lifespan
		);
	    NetSendLiveMessage(condescriptor, sndbuf);
	}


	/* No matches? */
	if(matches == 0)
	{
	    sprintf(
		sndbuf,
		"%s: Unsupported sysparm parameter.",
		parm
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}


	return(0);
}
