#include <stdlib.h>
#include <string.h>

#include "net.h"
#include "swserv.h"


int ConIsAllocated(int condescriptor);
int ConIsConnected(int condescriptor);
int ConIsLoggedIn(int condescriptor);

int ConGetByObject(int obj_num);
int ConGetTop(void);
void ConReclaim(void);
void ConReset(int condescriptor);
int ConCreateNew(void);
void ConRecycle(int condescriptor);
void ConDeleteAll(void);


/*
 *	Checks if condescriptor is allocated.
 */
int ConIsAllocated(int condescriptor)
{
	if((condescriptor < 0) ||
	   (condescriptor >= total_connections) ||
	   (connection == NULL)
	)
	    return(0);
	else if(connection[condescriptor] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Checks if condescriptor is connected.
 */
int ConIsConnected(int condescriptor)
{
	if(!ConIsAllocated(condescriptor))
	    return(0);
	else if(connection[condescriptor]->socket < 0)
	    return(0);
	else
	    return(1);
}

/*
 *	Checks if condescriptor is connected and logged in,
 *	by checking if the connection is allocated and its
 *	object is not -1.
 *
 *	Does not check if connection's object is valid.
 */
int ConIsLoggedIn(int condescriptor)
{
	/* Make sure condescriptor is valid */
	if(!ConIsConnected(condescriptor))
	    return(0);
	else if(connection[condescriptor]->object_num > -1)
	    return(1);
	else
	    return(0);
}

/*
 *	Returns the first connection number who's object matches the
 *	specified object number.
 */
int ConGetByObject(int obj_num)
{
	int i;
	sws_connection_struct *con;

	if(obj_num < 0)
	    return(-1);

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    if(con->object_num == obj_num)
		break;
	}
	if(i < total_connections)
	    return(i);
	else
	    return(-1);
}

/*
 *	Reclaims connections memory.
 */
void ConReclaim(void)
{
	int i, h;
	sws_connection_struct *con;

	if(total_connections < 0)
	    total_connections = 0;

	for(i = 0, h = -1; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    h = i;
	}


	for(i = total_connections - 1; i > h; i--)
	{
	    free(connection[i]);
	    connection[i] = NULL;
	}

	total_connections = h + 1;


	/* Reallocate or free pointer array */
	if(total_connections > 0)
	{
	    connection = (sws_connection_struct **)realloc(
	        connection,
	        total_connections * sizeof(sws_connection_struct *)
	    );
	    if(connection == NULL)
	    {
	        total_connections = 0;
		return;
	    }
	}
	else
	{
	    free(connection);
	    connection = NULL;
	    total_connections = 0;
	}
}

/*
 *	Returns the highest connection number that is connected or
 *	-1 if there are no connections.
 */
int ConGetTop(void)
{
	int i, h;
	sws_connection_struct *con;

	for(i = 0, h = -1; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    h = i;
	}

	return(h);
}

/*
 *	Reset connection values and delete any allocated resources but
 *	not the connection itself.
 *
 *	Connection's socket will NOT be disconnected.
 */
void ConReset(int condescriptor)
{
	sws_connection_struct *con;

	if(ConIsAllocated(condescriptor))
	    con = connection[condescriptor];
	else
	    return;

	con->socket = -1;
	con->client_type = 0;
	con->flags = 0;

	con->conhost[0] = '\0';
	con->object_num = -1;
	con->contime = 0;
	con->bytes_received = 0;
	con->bytes_sent = 0;
	con->errors_received = 0;
	con->errors_sent = 0;

	con->badlogins = 0;

	con->obj_ud_interval = SWS_DEF_OBJECT_UPDATE_INT;
	con->obj_ud_next = cur_millitime;

	con->conbuf_items = 0;
}


/*
 *	Deletes all connections, closing any connected sockets.
 */
void ConDeleteAll()
{
	int i;
	sws_connection_struct *con;

	/* Close all connections first */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    /* Close connection standardly if connected */
	    if(con->socket > -1)
		NetCloseConnection(i);

	    /* Reset connection values */
	    ConReset(i);
	}

	/* Delete all connections */
	for(i = 0; i < total_connections; i++)
	    free(connection[i]);
	free(connection);
	connection = NULL;
	total_connections = 0;
}

/*
 *	Creates a new connection in the connections list.
 *
 *	Returns the connection number of -1 on error.
 *
 *	Global highest_connection will be updated as needed.
 */
int ConCreateNew(void)
{
	int i, n;
	sws_connection_struct *con;

	/* Sanitize total */
	if(total_connections < 0)
	    total_connections = 0;

	/* Search for available structure already allocated */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		break;

	    if(con->socket < 0)
	        break;
	}
	if(i < total_connections)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_connections;
	    total_connections = n + 1;
	    connection = (sws_connection_struct **)realloc(
	        connection,
	        total_connections * sizeof(sws_connection_struct *)
	    );
	    if(connection == NULL)
	    {
	        total_connections = 0;
		return(-1);
	    }

	    /* Reset new pointer to NULL so it gets allocated below */
	    connection[n] = NULL;
	}

	/* Allocate connection as needed */
	con = connection[n];
	if(con == NULL)
	{
	    connection[n] = con = (sws_connection_struct *)calloc(
		1, sizeof(sws_connection_struct)
	    );
	    if(con == NULL)
		return(-1);
	}

	/* Reset new connection */
	ConReset(n);

	/* Increment global highest connection as needed */
	if(n >= highest_connection)
	    highest_connection = n + 1;

	return(n);
}

/*
 *	Recycles the connection.
 *
 *	Values will be reset, but the socket will not be disconnected.
 */
void ConRecycle(int condescriptor)
{
	/* Make sure condescriptor is valid */
	if(!ConIsAllocated(condescriptor))
	    return;

	/* Reset connection (closing any connected sockets) */
	ConReset(condescriptor);
}
