#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <crypt.h>
#include "swserv.h"


/*	Table for valid seed characters is more portable than ASCII
 *	values.
 *
 *	This contains 64 characters.
 *
 *	A valid DES seed is 2 characters long, made up of any of these
 *	characters.
 */
static char seed_lib[] = { \
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', \
        'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', \
        'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', \
        '.', '/', \
        'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', \
        'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', \
};



/*
 *	Encrypted Word:
 *
 *	The encrypted version of some word.  It is 13 characters, I add
 *	another to print it as a null-terminated string.  The crypt()
 *	function always returns this length of word (at least on linux,
 *	probably on all unix).
 */
#define ENCRYPTED_WORD_LEN	13
#if 0
char encrypted[ENCRYPTED_WORD_LEN + 1];  
#endif


const char *CryptCreateSeed(void);
char *CryptGetEncrypted(const char *word, const char *seed);
const char *CryptExtractSeed(const char *encryption);

int CryptHandleVerify(const char *password, const char *crypted);
const char *CryptHandleEncrypt(const char *password);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/* Assumes that time_t is unsigned int...true for linux, but all UNIX?
 * This returns a 3-character string, the 3rd character being NULL, and
 * the first 2 characters being the seed.
 *              
 * Note that this is the weakest point of the encryptions.  To use the C
 * random value category of commands, a seed must be supplied. Thus, to
 * get a random number, which we intend to create a more useful seed with,
 * we must supply a random number to start with. This initial random number
 * will simply be seconds since 00:00:00 GMT, January 1, 1970.  Minor effort
 * is made to not use the same number of seconds twice in a row, via
 * old_time_seed.  Once this initial seed is available, random number functions 
 * provide a pseudo-random number as an index to seed_lib[]; this is done twice,
 * to create a 2-character, NULL-terminated string.  The NULL is ignored except 
 * for printf().
 */
const char *CryptCreateSeed(void)
{
	int i;

        /* This is where the actual crypt seed will be stored.
         */
        static char seed[3] = {0, 0, 0};

        /* Use time as an initial seed.
         */
        time_t time_seed;

        /* Remember prior time so we won't use a time twice; this will NOT
         * work if the program exits between uses, but this WILL work if this
         * function is used in a program that continues to exist between uses of
         * the function. old_time_seed simply remembers the last time, and let's
         * us avoid using the same time twice in a row. Just in case 2 passwords
         * are created within the same second.
         */                      
        static time_t old_time_seed = 0;

        /* Create a random value through rand(), and possibly
         * other operations if time hasn't changed a full second.
         */
        unsigned int rand_val = 0;

        time_seed = time(NULL);
        if(time_seed == old_time_seed)
	{
            time_seed += 1234567;	/* Just an odd number */
            time_seed = abs(time_seed);
        }

        srand(time_seed);
        rand_val = rand();

        /* Indexes into seed_lib[], 0 to 63 as range */
	i = (int)(64.0 * rand_val / (RAND_MAX + 1.0));
	if(i < 0)
	    i = 0;
	if(i > 63)
	    i = 63;
        seed[0] = seed_lib[i];

        rand_val = rand(); 

        rand_val += 1234567891;	/* Just an odd number */
        rand_val = abs(rand_val);

        /* Indexes into seed_lib[], 0 to 63 as range */
        i = (int)(64.0 * rand_val / (RAND_MAX + 1.0)); 
        if(i < 0)
            i = 0;
        if(i > 63)
            i = 63;
        seed[1] = seed_lib[i];

        /* remember the time we used as initial seed */
        old_time_seed = time_seed;

	return(seed);
}


/*
 *	Pass any word you wish to create an encryption for, and the
 *	seed to use in the creation.
 *
 *	Returns a dynamically allocated string describing the encryped
 *	word.
 */
char *CryptGetEncrypted(const char *word, const char *seed)
{
	if((word == NULL) || (seed == NULL))
	    return(NULL);

	return(STRDUP(crypt(word, seed)));
}

const char *CryptExtractSeed(const char *encryption)
{
        static char extracted[3] = { 0, 0, 0 };

        if(STRLEN(encryption) > 2)
	{
            strncpy(extracted, encryption, sizeof(extracted));
	    extracted[sizeof(extracted) - 1] = '\0';
        }

        return(extracted);
}

int CryptHandleVerify(const char *password, const char *crypted)
{
	int status;
	char	seed[3],
		*encrypted;

	/* If the crypted password is not set then this always passes */
	if(crypted == NULL)
	    return(1);

	/* No password specified, then always fail? */
	if(password == NULL)
	    return(0);

	/* Fetch seed */
	strncpy(seed, CryptExtractSeed(crypted), sizeof(seed));
	seed[sizeof(seed) - 1] = '\0';

	/* Get encrypted */
	encrypted = CryptGetEncrypted(password, seed);
	if(encrypted == NULL)
	    return(0);

	/* Compare */
	status = (strcmp(crypted, encrypted) == 0) ? 1 : 0;

	free(encrypted);

	return(status);
}

/*
 *	Returns a statically allocated string describing the encrypted
 *	password.
 */
const char *CryptHandleEncrypt(const char *password)
{
	const char *s;
        char	*encrypted,
		seed[3];
        static char encrypted_rtn[ENCRYPTED_WORD_LEN + 1];

	if(password == NULL)
	    return(NULL);

	/* Create seed */
	s = CryptCreateSeed();
	strncpy(seed, s, sizeof(seed));
	seed[sizeof(seed) - 1] = '\0';

	/* Get encrypted word */
	encrypted = CryptGetEncrypted(password, seed);
	if(encrypted == NULL)
	    return(NULL);

	strncpy(encrypted_rtn, encrypted, sizeof(encrypted_rtn));
        encrypted_rtn[sizeof(encrypted_rtn) - 1] = '\0';

	free(encrypted);

	return(encrypted_rtn);
}
