#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/imgrefs.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int DBIsObjectGarbage(int object_num);

xsw_object_struct **DBAllocObjectPointers(
	xsw_object_struct **cur_obj_ptrs, int num_objects
);
xsw_object_struct *DBAllocObject(void);
int DBCreateObject(
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
);
int DBCreateExplicitObject(
	int object_num,
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
);
int DBCreateObjectByOPM(
	const char *opm_name,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
);
int DBCreateObjectByOPMPtr(
	xsw_object_struct *opm_ptr,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
);

void DBDeleteObject(xsw_object_struct *obj_ptr);
void DBDeleteAllObjects(void);

void DBRecycleObject(int obj_num);

void DBReclaim(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if object number is valid, allocated and garbage.
 *
 *	Returns false if object is valid, allocated and non-garbage.
 */
int DBIsObjectGarbage(int object_num)
{
	return(
	    UNVIsObjectGarbage(
		xsw_object,
		total_objects,
		object_num
	    )
	);
}

/*
 *	Reallocates given object pointer array.
 *
 *	If num_objects <= 0 then cur_obj_ptrs is free'ed if it is not
 *	NULL.
 */
xsw_object_struct **DBAllocObjectPointers(
	xsw_object_struct **cur_obj_ptrs,
	int num_objects   
)
{
	xsw_object_struct **obj_ptrs_rtn = NULL;


	if(num_objects > 0)
	{
	    obj_ptrs_rtn = (xsw_object_struct **)realloc(
		cur_obj_ptrs,
		num_objects * sizeof(xsw_object_struct *)
	    );
	}
	else
	{
	    /* Free current pointer array since new total is 0 */
	    free(cur_obj_ptrs);
	}

	return(obj_ptrs_rtn);
}

/*
 *	Allocates an object structure, returns NULL on error.
 *
 *	If successful, the allocated object structure will be
 *	reset to default values by a call to UNVResetObject().
 */
xsw_object_struct *DBAllocObject()
{
	xsw_object_struct *obj_ptr;


	obj_ptr = (xsw_object_struct *)calloc(
	    1,
	    sizeof(xsw_object_struct)
	);
	if(obj_ptr != NULL)
	{
	    /* Reset values on new object */
	    UNVResetObject(obj_ptr);
	}

	return(obj_ptr);
}

/*
 *	Creates a new object in the database.
 *
 *	Returns the new object's index number or -1 on error.
 */
int DBCreateObject(
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
)
{       
	int i, object_rtn, prev_total_objects;
	xsw_object_struct *obj;

	/* Type may not be garbage */
	if(type <= SW_OBJ_TYPE_GARBAGE)   
	{
	    fprintf(
		stderr,
"DBCreateObject: Error: Useless request to create garbage.\n"
	    ); 
	    return(-1);
	}       

	/* Is the imgref an invalid value? */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	{
	    fprintf(
		stderr,
"DBCreateObject: Error: ImgRef value %i out of range.\n",
		imgref_num
	    );
	    return(-1);
	}

	/* Look for an available object already in the list */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type == SW_OBJ_TYPE_GARBAGE)
		break;
	}
	if(i < total_objects)
	{
	    obj = xsw_object[i];

	    UNVResetObject(obj);
	    obj->type = type;
	    obj->owner = owner_obj_num;
	    obj->imageset = imgref_num;
	    obj->sect_x = sect_x;
	    obj->sect_y = sect_y;
	    obj->sect_z = sect_z;
	    obj->x = x;
	    obj->y = y;
	    obj->z = z;
	    obj->heading = heading;
	    obj->pitch = pitch;
	    obj->bank = bank;
	    MuSetUnitVector2D(
		&obj->attitude_vector_compoent,
		obj->heading
	    );
	    obj->birth_time = cur_millitime;
	    obj->last_updated = cur_millitime;

	    return(i);
	}

	/* No garbage objects, can we allocate more? */
	if((total_objects + SWS_OBJECT_ALLOCATE_AHEAD) > SW_OBJECTS_MAX)
	    return(-1);

	/* Allocate more objects in memory */
	prev_total_objects = total_objects;
	total_objects += SWS_OBJECT_ALLOCATE_AHEAD;

	/* Allocate pointer array */
	xsw_object = DBAllocObjectPointers(xsw_object, total_objects);
	if(xsw_object == NULL)
	{
	    total_objects = 0;
	    return(-3);
	}

	/* Allocate each object */
	for(i = prev_total_objects; i < total_objects; i++)
	{
	    xsw_object[i] = DBAllocObject();
	    if(xsw_object[i] == NULL)
	    {
		total_objects = i;
		return(-3);
	    }
	}

	/* New object will be the first object allocated above */
	object_rtn = prev_total_objects;   

	/* Set values on the new object */
	obj = xsw_object[object_rtn];
	obj->type = type;
	obj->owner = owner_obj_num;
	obj->imageset = imgref_num;
	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;
	obj->x = x;
	obj->y = y;
	obj->z = z;
	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	MuSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
	obj->birth_time = cur_millitime;
	obj->last_updated = cur_millitime;

	return(object_rtn);
}

/*
 *	Creates a new object in the database at the index specified by
 *	object_num.
 *
 *	If the object already exists then it will be recycled and
 *	recreated.
 *
 *	Returns the new object's index number or -1 on error.
 */
int DBCreateExplicitObject(
	int object_num,
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
)
{
	int i, obj_num_rtn;
	xsw_object_struct *obj;

	/* Type may not be garbage or error */
	if(type <= SW_OBJ_TYPE_GARBAGE)
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Useless request to create garbage.\n"
	    );
	    return(-1);
	}

	/* Is the object number an invalid value? */
	if((object_num < 0) || (object_num >= SW_OBJECTS_MAX))
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Object number %i out of range.\n",
		object_num
	    );
	    return(-1);
	}

	/* Is the imgref an invalid value? */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Error: ImgRef value %i out of range.\n",
		imgref_num
	    );
	    return(-1);
	}

	/* If object_num is not garbage then recycle it first */
	if(!DBIsObjectGarbage(object_num))
	    DBRecycleObject(object_num);

	/* Allocate memory if object_num is greater than total_objects */
	if(object_num >= total_objects)
	{
	    /* Can we allocate more? */
	    if((object_num + SWS_OBJECT_ALLOCATE_AHEAD) >= SW_OBJECTS_MAX)
	    {
		fprintf(
		    stderr,
"DBCreateExplicitObject(): Maximum objects %i reached.\n",
		    SW_OBJECTS_MAX
		);
		return(-3);
	    }

	    /* Reallocate objects pointer array */ 
	    xsw_object = DBAllocObjectPointers(
		xsw_object,
		object_num + SWS_OBJECT_ALLOCATE_AHEAD + 1
	    );
	    if(xsw_object == NULL)
	    {
		total_objects = 0;
		return(-3);
	    }

	    /* Allocate each new object */
	    for(i = total_objects;
		i <= (object_num + SWS_OBJECT_ALLOCATE_AHEAD);
		i++
	    )
	    {
		xsw_object[i] = (xsw_object_struct *)DBAllocObject();
		if(xsw_object[i] == NULL)
		{
		    total_objects = i;
		    return(-3);
		}
	    }

	    /* Adjust global total */
	    i = total_objects;
	    total_objects = object_num + SWS_OBJECT_ALLOCATE_AHEAD + 1;

	    /* Reset newly created objects */
	    while(i < total_objects)
	    {
		UNVResetObject(xsw_object[i]);
		i++;
	    } 
	}

	/* Create explicit object */
	obj_num_rtn = object_num;

	/* Reset values on the new object */
	obj = xsw_object[obj_num_rtn];
	UNVResetObject(obj);
	obj->type = type; 
	obj->owner = owner_obj_num;
	obj->imageset = imgref_num;
	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;
	obj->x = x;
	obj->y = y;
	obj->z = z;
	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	MuSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
	obj->birth_time = cur_millitime;
	obj->last_updated = cur_millitime;

	return(obj_num_rtn);
}


/*
 *	Creates an object and models it to the specified OPM.
 */
int DBCreateObjectByOPM(
	const char *opm_name,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
)
{
	int opm_num;

	if((opm_name == NULL) || (type <= SW_OBJ_TYPE_GARBAGE))
	    return(-1);

	/* Get the OPM */
	opm_num = OPMGetByName(opm_name, type);
	if(OPMIsGarbage(opm_num))
	{
	    fprintf(
		stderr,
		"DBCreateObjectByOPM(): %s: No such OPM.\n",
		opm_name
	    );
	    return(-2);
	}

	return(DBCreateObjectByOPMPtr(
	    opm[opm_num], name, type,
	    owner_obj_num,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    heading, pitch, bank
	));
}

/*
 *	Creates an object and models it to the specified OPM.
 */
int DBCreateObjectByOPMPtr(
	xsw_object_struct *opm_ptr,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	double x, double y, double z,
	double heading, double pitch, double bank
)
{
	int obj_num;
	xsw_object_struct *obj;

	if((opm_ptr == NULL) || (type <= SW_OBJ_TYPE_GARBAGE))
	    return(-1);

	/* Create the new object */
	obj_num = DBCreateObject(
	    IMGREF_DEFAULT_IMGREF,
	    type,
	    owner_obj_num,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    heading, pitch, bank
	);
	if(DBIsObjectGarbage(obj_num))
	    return(-3);

	obj = xsw_object[obj_num];

	/* Set model the new object */
	OPMModelObjectPtr(obj, opm_ptr);

	/* Set values */
	strncpy(
	    obj->name, (name != NULL) ? name : "",
	    sizeof(obj->name)
	);
	obj->name[sizeof(obj->name) - 1] = '\0';

	obj->owner = owner_obj_num;

	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;

	obj->x = x; 
	obj->y = y;
	obj->z = z;

	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;

	MuSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);

	obj->birth_time = cur_millitime;
	obj->last_updated = cur_millitime;
	obj->animation.last_interval = cur_millitime;

	return(obj_num);
}

/*
 *	Deletes the object.
 */
void DBDeleteObject(xsw_object_struct *obj_ptr)
{
	if(obj_ptr == NULL)
	    return;

	UNVDeleteObject(obj_ptr);
}

/*
 *	Deletes all objects.
 */     
void DBDeleteAllObjects(void)
{
	UNVDeleteAllObjects(xsw_object, total_objects);
	xsw_object = NULL;
	total_objects = 0;
}

/*
 *	Recycles the object.
 *
 *	If the object is a player then all of its posessions will be
 *	be chowned to the global lost and found owner.
 *
 *	If the object is a player and is still connected then the
 *	connection will be disconnected first.
 *
 *	All members on the object will be deleted and reset.
 *
 *	The objects type will be set to garbage.
 */
void DBRecycleObject(int obj_num)   
{
	int condescriptor;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* If this is a player then perform additional checks */
	if(obj->type == SW_OBJ_TYPE_PLAYER)
	{
	    /* Change ownership of all objects owned by this player
	     * object to the global lost_found_owner object
	     */
	    if(!DBIsObjectGarbage(unv_head.lost_found_owner))
	    {
		int i;
		char	name1[SW_OBJ_NAME_MAX + 80],
			name2[SW_OBJ_NAME_MAX + 80];
		xsw_object_struct *tobj;

		strncpy(
		    name2, DBGetFormalNameStr(unv_head.lost_found_owner),
		    sizeof(name2)
		);
		name2[sizeof(name2) - 1] = '\0';

		for(i = 0; i < total_objects; i++)
		{
		    tobj = xsw_object[i];
		    if(tobj == NULL)
			continue;

		    if(tobj->type <= SW_OBJ_TYPE_GARBAGE)
			continue;

		    if(tobj->owner != obj_num)
			continue;

		    /* Skip object itself */
		    if(i == obj_num)
			continue;

		    tobj->owner = unv_head.lost_found_owner;

		    strncpy(
			name1, DBGetFormalNameStr(i), sizeof(name1)
		    );
		    name1[sizeof(name1) - 1] = '\0';

		    if(sysparm.log_general)
		    {
			char *s = (char *)malloc(
			    STRLEN(name1) + STRLEN(name2) + 80
			);
			sprintf(
			    s,
"Chowned %s to %s due to recycling of previous owner.",
			    name1, name2
			);
			LogAppend(fname.log, s);
			free(s);
		    }
		}
	    }
	    /* If no valid lost and found owner then set to owner 0 */
	    else
	    {
		int i;
		char name1[SW_OBJ_NAME_MAX + 80];
		xsw_object_struct *tobj;

		for(i = 0; i < total_objects; i++)
		{
		    tobj = xsw_object[i];
		    if(tobj == NULL)
			continue;
		    if(tobj->type <= SW_OBJ_TYPE_GARBAGE)
			continue;
		    if(tobj->owner != obj_num)
			continue;

		    /* Skip object itself */
		    if(i == obj_num)
			continue;

		    tobj->owner = 0;

		    strncpy(
			name1, DBGetFormalNameStr(i), sizeof(name1)
		    );

		    if(sysparm.log_general)
		    {
			char *s = (char *)malloc(
			    STRLEN(name1) + 80
			);
			sprintf(
			    s,
"Chowned %s to #0 due to recycling of previous owner.",
			    name1
			);
			LogAppend(fname.log, s);
			free(s);
		    }
		}
	    }


	    /* Close all connections logged in as obj_num */
	    while(1)
	    {
		condescriptor = ConGetByObject(obj_num);
		if(condescriptor < 0)
		    break;

		NetCloseConnection(condescriptor);
	    }
	}

	/* Reset and delete all allocated substructures on this
	 * object
	 */
	UNVResetObject(obj);

	/* Set this object to type garbage to mark it as recycled */  
	obj->type = SW_OBJ_TYPE_GARBAGE;
}



/*
 *	Performs maintainance on the universe database.
 *
 *	Reallocates unused object pointers.
 *
 *	Updates player object connected and hidden flags.
 *
 *	Sanitizes hp and power values on all objects.
 */        
void DBReclaim(void)
{
	int i, h;
	xsw_object_struct *obj;

	if((total_objects <= 0) || (xsw_object == NULL))
	    return;

	/* Get highest valid and non-garbage object */
	h = DBGetTopObjectNumber();
	if((h >= 0) && (h < total_objects))
	{
	    /* Free all allocated objects from total_objects - 1
	     * to h + 1
	     */
	    for(i = total_objects - 1; i > h; i--)
	    {
		DBDeleteObject(xsw_object[i]);
		xsw_object[i] = NULL;
	    }

	    /* Reallocate object pointers */
	    total_objects = h + 1;
	    if(total_objects > 0)
	    {
	        xsw_object = DBAllocObjectPointers(
		    xsw_object, total_objects
	        );
		if(xsw_object == NULL)
		{
		    total_objects = 0;
		    return;
	        }
	    }
	    else
	    {
		free(xsw_object);
		xsw_object = NULL;
		total_objects = 0;  
	    }
	}


	/* Perform maintainance on objects */

	/* Iterate through the objects list */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Player? */
	    if(obj->type == SW_OBJ_TYPE_PLAYER)
	    {
		/* Hide players at home objects? */
		if(sysparm.hide_players_home)
		{
		    /* Check if the player is not connected and is
		     * at a home object
		     */
		    if(ConGetByObject(i) < 0)
		    {
			/* Not connected, now check if it has not
			 * yet been marked hidden
			 */
			if(!(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON))
			{
			    int n;
			    xsw_object_struct *tar_obj;

			    /* Search for a home object that this
			     * player is at
			     */
			    for(n = 0; n < total_objects; n++)
			    {
			        tar_obj = xsw_object[n];
			        if(tar_obj == NULL)
				    continue;

			        if(tar_obj->type != SW_OBJ_TYPE_HOME)
				    continue;

			        /* Player at this home object? */
			        if(Mu3DInContactPtr(obj, tar_obj))
			        {
				    /* Mark this player hidden */
				    obj->ext_flags |=
				        SWS_OBJ_FLAG_HIDEFROMCON;
				    break;
				}
			    }
			}
		    }
		    else
		    {
			/* Object is connected */

			/* Marked hidden? */
			if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			{
			    /* Mark as not hidden */
			    obj->ext_flags &= ~SWS_OBJ_FLAG_HIDEFROMCON;
			}
			/* Not marked as connected? */
			if(!(obj->ext_flags & SWS_OBJ_FLAG_CONNECTED))
			{
			    /* Mark as connected */
			    obj->ext_flags |= SWS_OBJ_FLAG_CONNECTED;
			}
		    }
		}
	    }
	    /* Error object */
	    else if(obj->type < SW_OBJ_TYPE_GARBAGE)
	    {
/* TODO */

	    }

	    /* General checks for all object types */

	    /* Sanitize power */
	    if(obj->power_max < 0.0)
		obj->power_max = 0.0;
	    if(obj->power > obj->power_max)
		obj->power = obj->power_max;
	    if(obj->power < 0.0)
		obj->power = 0.0;

	    /* Sanitize hp */
	    if(obj->hp > obj->hp_max)
		obj->hp = obj->hp_max;


	}
}
