#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../include/disk.h"
#include "../include/unvfile.h"

#include "sysparm.h"
#include "swserv.h"


int DBOpenFromFile(const char *path);
int DBSaveToFile(const char *path);
int DBEmergencySave(xsw_object_struct **list, int total);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int DBOpenFromFile(const char *path)
{
	int total = 0;

	/* Delete all objects currently in memory */
	DBDeleteAllObjects();

	/* Load objects from the universe file into the universe
	 * database
	 */
	xsw_object = UNVFileOpen(
	    path,
	    &total,
	    &unv_head,
	    NULL,	/* Client data */
	    NULL	/* Notify callback function */
	);
	total_objects = total;

	/* Get header information */
	sw_units.ru_to_au = unv_head.ru_to_au;

	/* Delete unused object pointers, update object hidden states
	 * and sanitize values
	 */
	DBReclaim();

	return(0);
}


int DBSaveToFile(const char *path)
{
	int status;

	/* Set header information */
	strcpy(unv_head.version, PROG_VERSION);
	unv_head.ru_to_au = sw_units.ru_to_au;

	/* Save to file */
	status = UNVFileSave(
	    path,
	    xsw_object,
	    total_objects,
	    &unv_head,
	    NULL,	/* Client data */
	    NULL	/* Notify callback function */
	);

	return(status);
}


int DBEmergencySave(xsw_object_struct **list, int total)
{
	int status;
	char	*s, *s2,
		text[PATH_MAX + NAME_MAX + 80],
		tmp_path[PATH_MAX + NAME_MAX];
	const pid_t pid = getpid();

	/* Print and log incident */
	fprintf(
	    stderr,
	    "*** Server: Emergency database save in progress...\n"
	);
	if(sysparm.log_errors)
	    LogAppend(
		fname.log,
	        "*** Server: Emergency database save in progress..."
	    );


	/* ********************************************************** */
	/* Plan A */

	/* Format new filename for emergency save */
	sprintf(tmp_path, "%s.EMERGENCY.%i", fname.unv_out, pid);

	/* Attempt to save */
	status = DBSaveToFile(tmp_path);
	if(status > -1)
	{
	    /* Emergency save was successful */

	    fprintf(
		stderr,
		"%s: Plan A: Database sucessfully saved!\n",
		tmp_path
	    );
	    sprintf(
		text,
		"%s: Plan A: Database sucessfully saved!",
		tmp_path
	    );
 	    if(sysparm.log_errors)
		LogAppend(fname.log, text);

	    /* Return, no need for additional saving */
	    return(0);
	}


	/* ********************************************************** */
	/* Plan B */

	/* Format new filename for emergency save */
	s = getenv("TMPDIR");
	if(s == NULL)
#if defined(P_tmpdir)
	    s = P_tmpdir;
#elif defined(WIN32)
	    s = "C:\\TEMP";
#else
	    s = "/tmp";
#endif
	s2 = PrefixPaths(s, "sw-unv-XXXXXX");
	strncpy(
	    tmp_path, (s2 != NULL) ? s2 : "",
	    sizeof(tmp_path)
	);
	tmp_path[sizeof(tmp_path) - 1] = '\0';  
	mkstemp(tmp_path);

	/* Attempt to save */
	status = DBSaveToFile(tmp_path);
	if(status > -1)
	{
	    /* Emergency save was successful */

	    fprintf(
		stderr,
		"%s: Plan B: Database sucessfully saved!\n",
		tmp_path
	    );
	    sprintf(
		text,
		"%s: Plan B: Database sucessfully saved!",
		tmp_path
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, text);

	    return(0);
	}


	/* ********************************************************* */
	/* Out of ideas for emergency save, we're screwed */

	fprintf(
	    stderr,
	    "*** Emergency save failed! ***\n"
	);
	sprintf(
	    text,
	    "*** Emergency save failed! ***"
	);

	if(sysparm.log_errors)
	    LogAppend(fname.log, text);


	return(0);
}
