#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "swserv.h"


int DBRecycleBufferInit(int entries);
void DBRecycleBufferDeleteAll(void);

int DBSaveRecycledObject(int obj_num);
int DBRecoverRecycledObject(const char *name, int owner);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*      
 *	Initializes the recycled objects list.
 *
 *	The global recycled_xsw_object list must already be empty
 *	before calling this function.
 *
 *	Returns the number of recycled objects slots allocated.
 */
int DBRecycleBufferInit(int entries)
{
	int i, allocated;

	/* Set new total */
	total_recycled_objects = entries;

	/* Allocate pointer array */
	recycled_xsw_object = (xsw_object_struct **)realloc(
	    recycled_xsw_object,
	    total_recycled_objects * sizeof(xsw_object_struct *)
	);
	if(recycled_xsw_object == NULL)
	{
	    total_recycled_objects = 0;
	    return(-3);
	}

	/* Allocate each structure */
	for(i = 0, allocated = 0; i < total_recycled_objects; i++)
	{
	    recycled_xsw_object[i] = (xsw_object_struct *)calloc(
		1,
		sizeof(xsw_object_struct)
	    );
	    if(recycled_xsw_object[i] == NULL)
	    {
		total_recycled_objects = i;
		break;
	    }
	    else
	    {
		/* Increment allocated count */
		allocated++;

		/* Set type to garbage */
		recycled_xsw_object[i]->type = SW_OBJ_TYPE_GARBAGE;
	    }
	}

	return(allocated);
}

/*
 *	Deletes the recycled objects list.
 */
void DBRecycleBufferDeleteAll(void)
{
	int i;

	for(i = 0; i < total_recycled_objects; i++)
	    DBDeleteObject(recycled_xsw_object[i]);
	free(recycled_xsw_object);
	recycled_xsw_object = NULL;
	total_recycled_objects = 0;
}


/*
 *	Coppies the specified object into the recycled objects list.
 *
 *	Returns 0 on success or -1 on error.
 */
int DBSaveRecycledObject(int obj_num)
{
	int i, n;
	xsw_object_struct *obj, *rec_obj, *rec_obj2;

	if(total_recycled_objects < 1)
	    return(-3);

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];


	/* Look for an available pointer in the recycled objects list */
	for(i = 0; i < total_recycled_objects; i++)
	{
 	    if(recycled_xsw_object[i] == NULL)
		continue;

	    if(recycled_xsw_object[i]->type <= SW_OBJ_TYPE_GARBAGE)
		break;
	}
	if(i < total_recycled_objects)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Delete the highest (oldest) recycled object and shift
	     * pointers
	     */
	    UNVResetObject(recycled_xsw_object[total_recycled_objects - 1]);
	    for(i = total_recycled_objects - 1; i > 0; i--)
	    {
		rec_obj = recycled_xsw_object[i];
		rec_obj2 = recycled_xsw_object[i - 1];

		if((rec_obj == NULL) || (rec_obj2 == NULL))
		    continue;

		/* Don't worry about allocated substructures,
		 * the pointer values to them will be coppied accordingly
		 * when we copy over each object's value. Recycled object 0
		 * will still contain duplicate pointers to substructures,
		 * but will be set with the new object's values farther below.
		 */
		memcpy(
		    rec_obj,		/* Destination */
		    rec_obj2,		/* Source */
		    sizeof(xsw_object_struct)
		);
	    }

	    /* Set new available structure to number 0 (first one) */
	    n = 0;
   	}

	/* Get pointer to recycled object n */
	rec_obj = recycled_xsw_object[n];
	if(rec_obj == NULL)
	    return(-3);


	/* Begin copying the object to the available recycled object
	 * structure n
	 */

	/* Copy core object values */
	memcpy(
	    rec_obj,		/* Destination */
	    obj,		/* Source */
	    sizeof(xsw_object_struct)
	);

	rec_obj->password = STRDUP(obj->password);
	rec_obj->description = STRDUP(obj->description);
	rec_obj->elink = STRDUP(obj->elink);

	/* Do not copy tractored objects list, delete them */
	free(obj->tractored_object);
	obj->tractored_object = NULL;
	obj->total_tractored_objects = 0;

	rec_obj->tractored_object = NULL;
	rec_obj->total_tractored_objects = 0;


	/* Copy weapons as needed */ 
	if(rec_obj->total_weapons > 0)
	{
	    /* Allocate weapons pointer array */
	    rec_obj->weapons = (xsw_weapon_struct **)malloc(
		rec_obj->total_weapons * sizeof(xsw_weapon_struct *)
	    );
	    if(rec_obj->weapons == NULL)
		rec_obj->total_weapons = 0;

	    /* Allocate and copy each weapon */
	    for(i = 0; i < rec_obj->total_weapons; i++)
	    {
		rec_obj->weapons[i] = (xsw_weapon_struct *)malloc(
		    sizeof(xsw_weapon_struct)
		);
		memcpy(
		    rec_obj->weapons[i],
		    obj->weapons[i],
		    sizeof(xsw_weapon_struct)
		);
	    }
	}

	/* Copy scores as needed */
	if(obj->score != NULL)
	{
	    /* Allocate and copy score */
	    rec_obj->score = (xsw_score_struct *)malloc(
		sizeof(xsw_score_struct)
	    );
	    memcpy(
		rec_obj->score,
		obj->score,
		sizeof(xsw_score_struct)
	    );
 	}

	/* Copy economy as needed */
	if(obj->eco != NULL)
	{
	    /* Allocate and copy economy data structure */
	    rec_obj->eco = (xsw_ecodata_struct *)malloc(
		sizeof(xsw_ecodata_struct)
	    );
	    memcpy(
		rec_obj->eco,
		obj->eco,
		sizeof(xsw_ecodata_struct)
	    );
	    if(rec_obj->eco != NULL)
	    {
		/* Copy seconomy products as needed */
		if(rec_obj->eco->total_products > 0)
		{
		    rec_obj->eco->product = (xsw_ecoproduct_struct **)malloc(
			rec_obj->eco->total_products *
			sizeof(xsw_ecoproduct_struct *)
		    );
		    if(rec_obj->eco->product == NULL)
			rec_obj->eco->total_products = 0;

		    for(i = 0; i < rec_obj->eco->total_products; i++)
		    {
		        rec_obj->eco->product[i] = (xsw_ecoproduct_struct *)malloc(
			    sizeof(xsw_ecoproduct_struct)
			);
			memcpy(
			    rec_obj->eco->product[i],
			    obj->eco->product[i],
			    sizeof(xsw_ecoproduct_struct)
			);
		    }
		}
	    }
	}


	return(0);
}

/*
 *	Recovers an object in the recycled objects list, returns
 *	the object number recovered as or -1 on failure.
 */
int DBRecoverRecycledObject(const char *name, int owner)
{
	int i, owner_uid, rec_obj_num, obj_num;
	xsw_object_struct *obj, *owner_obj;

	if(DBIsObjectGarbage(owner))
	    return(-1);

	owner_obj = xsw_object[owner];

	/* Get permission level of owner object */
	owner_uid = owner_obj->permission.uid;

	/* Is owner object allowed to unrecycle? */
	if(owner_uid > SWS_ACCESS_UID_UNRECYCLE)
	    return(-2);


	/* Search for object in recycled objects buffer starting from
	 * the oldest
	 */
	for(i = total_recycled_objects - 1; i >= 0; i--)
	{
	    obj = recycled_xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Match name case sensitive */
	    if(strstr(obj->name, name) == NULL)
		continue;

	    /* Got matched recycled object, now check if it is a
	     * player object
	     */
	    if(obj->type != SW_OBJ_TYPE_PLAYER)
	    {
		/* Recycled object is not a player */

		/* Check if owner owns this recycled object */
		if(obj->owner != owner)
		{
		    /* Owner does not own this object, so check if owner
		     * is allowed to unrecycle other objects
		     */
		    if(owner_uid > SWS_ACCESS_UID_UNRECYCLEO)
			continue;
		}
	    }
	    else
	    {
		/* Recycled object is a player, so check if owner has
		 * create player permission and allowed to unrecycle
		 * others?
		 */
		if((owner_uid > SWS_ACCESS_UID_CREATEPLAYER) ||
		   (owner_uid > SWS_ACCESS_UID_UNRECYCLEO)
		)
		    continue;
	    }

	    /* Found recycled object that matches name and owner is
	     * allowed to unrecycle it
	     */
	    break;
	}
	rec_obj_num = i;


	/* Did we find anything? */
	if((rec_obj_num < 0) ||
	   (rec_obj_num >= total_recycled_objects)
	)
	    return(-1);

	obj = recycled_xsw_object[rec_obj_num];
	if(obj == NULL)
	    return(-1);


	/* Begin recovering object */

	/* Create a new object */
	obj_num = DBCreateObject(
	    obj->imageset,
	    obj->type,
	    obj->owner,
	    obj->sect_x, obj->sect_y, obj->sect_z,
	    obj->x, obj->y, obj->z,
	    obj->heading, obj->pitch, obj->bank
	);
	if(DBIsObjectGarbage(obj_num))
	    return(-3);


	/* Copy the recycled object data to the recovered object */
	memcpy(
	    xsw_object[obj_num],	/* Destination */
	    obj,			/* Source */
	    sizeof(xsw_object_struct)
	);

	/* NOTE: We copy the pointers to allocated substructures
	 * of the object in the recycled buffers list to the newly
	 * created recovered object
	 */

	/* Reset the object in the recycle list so that its
	 * substructure pointers are not deleted
	 */
	memset(
	    obj, 0x00,
	    sizeof(xsw_object_struct)
	);
	obj->type = SW_OBJ_TYPE_GARBAGE;


	/* Player objects need to own themselves */
	if(xsw_object[obj_num]->type == SW_OBJ_TYPE_PLAYER)
	    xsw_object[obj_num]->owner = obj_num;


	return(obj_num);
}
