#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "sysparm.h"
#include "swserv.h"


int DBCreateObjectEconomyProduct(
	int obj_num,
	const char *product_name,
	sw_credits_t sell_price,
	sw_credits_t buy_price,
	double amount, double amount_max
);

int DBObjectTractor(int obj_num, int tar_obj_num);
int DBIsObjectTractorablePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj
);
void DBObjectUntractor(int obj_num, int tar_obj_num);

int DBObjectDoSetSector(
	int obj_num,
	long sect_x, long sect_y, long sect_z,
	char allow_wrapping
);

void DBSortEconomyProducts(int obj_num);

char *DBGetTypeName(int type);
char *DBGetFormalNameStr(int obj_num);
char *DBGetOCSOPMName(int ocs_num);

int DBValidateObjectName(const char *name);
int DBValidateObjectPassword(const char *password);

double DBGetObjectVisibility(int obj_num);
double DBGetObjectVisibilityPtr(xsw_object_struct *obj);

int DBGetPlayerOwner(int obj_num);

int DBGetTopObjectNumber(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *      Allocates economy product structure on obj_num as needed.
 *      Returns -1 on error.
 */
int DBCreateObjectEconomyProduct(
	int obj_num,
	const char *product_name,
	sw_credits_t sell_price,
	sw_credits_t buy_price,
	double amount,
	double amount_max
)
{
	int i, n;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *product;
	xsw_object_struct *obj;

	if(product_name == NULL)
	    return(-2);

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	/* Allocate economy on the object as needed */
	if(obj->eco == NULL)
	{
	    if(UNVAllocEco(obj))
		return(-3);
	}

	eco = obj->eco;

	/* Sanitize total */
	if(eco->total_products < 0)
	    eco->total_products = 0;

	/* Is product_name in list? */
	for(i = 0, n = -1; i < eco->total_products; i++)
	{
	    product = eco->product[i];
	    if(product == NULL)
	    {
		n = i;
		continue;
	    }

	    if(!strcasecmp(product->name, product_name))
		break;
	}
	if(i >= eco->total_products)
	{
	    /* No such product, allocate a new one */

	    if(n > -1)
	    {
		/* Have allocated pointer n */

		i = n;
	    }
	    else
	    {
		/* No unallocated pointers, allocate more */
		i = eco->total_products; 
		eco->total_products = i + 1;
		eco->product = (xsw_ecoproduct_struct **)realloc(
		    eco->product,
		    eco->total_products * sizeof(xsw_ecoproduct_struct *)
		);
		if(eco->product == NULL)
		{
		    eco->total_products = 0;
		    return(-3);
		}
	    }

	    /* Allocate a new product */
	    eco->product[i] = (xsw_ecoproduct_struct *)calloc(
		1, sizeof(xsw_ecoproduct_struct)
	    );
	    if(eco->product[i] == NULL)
		return(-3);
	}

	product = eco->product[i];
	strncpy(
	    product->name, product_name, sizeof(product->name)
	);
	product->name[sizeof(product->name) - 1] = '\0';

	product->sell_price = sell_price;
	product->buy_price = buy_price;
	product->amount = amount;
	product->amount_max = amount_max;

	return(0);
}


/*
 *      Sets the object to lock tractor beam on the target object.
 *
 *	If tar_obj_num is -1 then the object's tractor beam will be
 *	unlocked.
 */
int DBObjectTractor(int obj_num, int tar_obj_num)
{
	int n;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	/* Unlock tractor beam? */
	if(tar_obj_num < 0)
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	    return(0);
	}


	/* tar_obj is assumed valid */

	/* Is the object already in the list? */
	for(n = 0; n < obj->total_tractored_objects; n++)
	{
	    if(obj->tractored_object[n] == tar_obj_num)
		return(0);
	}

	/* Sanitize total */
	if(obj->total_tractored_objects < 0)
	    obj->total_tractored_objects = 0;

	/* Maximum tractorable objects reached? */
	if(obj->total_tractored_objects >= SW_TRACTOR_BEAM_MAX_OBJECTS)
	    return(-3);

	/* Add to the tractored objects list */
	n = obj->total_tractored_objects;
	obj->total_tractored_objects = n + 1;
	obj->tractored_object = (int *)realloc(
	    obj->tractored_object,
	    obj->total_tractored_objects * sizeof(int)
	);
	if(obj->tractored_object == NULL)
	{
	    obj->total_tractored_objects = 0;
	    return(-3);
	}

	/* Set new tractored object */
	obj->tractored_object[n] = tar_obj_num;

	return(0);
}


/*
 *	Checks if the object can tractor the target object.
 */
int DBIsObjectTractorablePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj
)
{
	if((obj == NULL) || (tar_obj == NULL))
	    return(0);

	/* These object types may not be tractored */
	if((tar_obj->type == SW_OBJ_TYPE_HOME) ||
	   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
	   (tar_obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (tar_obj->type == SW_OBJ_TYPE_ELINK)
	)
	    return(0);

	/* Check if objects are in range */
	if(!Mu3DInRangePtr(obj, tar_obj, sysparm.tractor_beam_range))
	    return(0);

	/* Check if target object has shields up and if so, check
	 * if target object's power is greater than the source
	 * object's
	 */
	if((tar_obj->shield_state == SHIELD_STATE_UP) &&
	   ((tar_obj->power * tar_obj->power_purity) >=
	    (obj->power * obj->power_purity)
	   )
	)
	    return(0);


	return(1);
} 

/*
 *	Untractors the target object.
 */
void DBObjectUntractor(int obj_num, int tar_obj_num)
{
	int i, n;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Sanitize total */
	if(obj->total_tractored_objects < 0)   
	    obj->total_tractored_objects = 0;

	/* If tar_obj_num is -1 then unlock all tractor beam objects
	 * for now
	 */
	if(tar_obj_num < 0)
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	    return;
	}

	/* tar_obj_num is assumed valid */

	/* Untractor object */
	for(i = 0; i < obj->total_tractored_objects; i++)
	{
	    if(obj->tractored_object[i] == tar_obj_num)
		obj->tractored_object[i] = -1;
	}

	/* Reclaim tractored objects memory */
	n = -1;
	for(i = 0; i < obj->total_tractored_objects; i++)
	{
	    if(DBIsObjectGarbage(obj->tractored_object[i]))
		continue;
	    else
		n = i;
	}

	/* Reclaim memory as needed */
	obj->total_tractored_objects = n + 1;
	if(obj->total_tractored_objects > 0)
	{
	    obj->tractored_object = (int *)realloc(
		obj->tractored_object,
		obj->total_tractored_objects * sizeof(int)
	    );
	    if(obj->tractored_object == NULL)
	    {
		obj->total_tractored_objects = 0;
		return;
	    }
	}
	else
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	}
}


/*
 *	Procedure to change sector of an object in accordance with
 *	sector changing rules. Returns 0 on success, -1 on error,
 *	or -3 if the sector change would violate sector changing
 *	rules.
 *
 *	Tractored objects (if any) will have their sectors changed as
 *	well.
 *
 *	If allow_wrapping is set to 1, then the position of
 *	the object (and its tractored objects if any) will have
 *	their positions wrapped.
 */
int DBObjectDoSetSector(
	int obj_num,
	long sect_x, long sect_y, long sect_z,
	char allow_wrapping
)
{
	int i, tract_obj_num;
	long sect_dx, sect_dy, sect_dz;
	xsw_object_struct *obj, *tract_obj;

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	/* Is there a change in sectors? */
	if((obj->sect_x != sect_x) ||
	   (obj->sect_y != sect_y) ||
	   (obj->sect_z != sect_z)
	)
	{
	    /* Get sector change deltas */
	    sect_dx = sect_x - obj->sect_x;
	    sect_dy = sect_y - obj->sect_y; 
	    sect_dz = sect_z - obj->sect_z;

	    if(allow_wrapping)
	    {
		/* Wrap object position within sector */
		if(sect_dx < 0)
		    obj->x += sector_legend.x_len;
		else if(sect_dx > 0)
		    obj->x -= sector_legend.x_len;

		if(sect_dy < 0)
		    obj->y += sector_legend.y_len;
		else if(sect_dy > 0)
		    obj->y -= sector_legend.y_len;
/*
		if(sect_dz < 0)
		    obj->z += sector_legend.z_len;
		else if(sect_dz > 0)
		    obj->z -= sector_legend.z_len;
*/
	    }

	    /* Sanitize coordinates */
	    if(obj->x < sector_legend.x_min)
		obj->x = sector_legend.x_min + 1;
	    if(obj->x > sector_legend.x_max)
		obj->x = sector_legend.x_max - 1;

	    if(obj->y < sector_legend.y_min)
		obj->y = sector_legend.y_min + 1;
	    if(obj->y > sector_legend.y_max)
		obj->y = sector_legend.y_max - 1;

/*
	    if(obj->z < sector_legend.z_min)
		obj->z = sector_legend.z_min + 1;
	    if(obj->z > sector_legend.z_max)
		obj->z = sector_legend.z_max - 1;
 */

	    /* Set new sector of object */
	    obj->sect_x = sect_x;
	    obj->sect_y = sect_y;
	    obj->sect_z = sect_z;

	    /* Go through list of tractored objects */
	    for(i = 0; i < obj->total_tractored_objects; i++)
	    {
		tract_obj_num = obj->tractored_object[i];
		if(DBIsObjectGarbage(tract_obj_num))
		    continue;   
		else
		    tract_obj = xsw_object[tract_obj_num];

		/* Set tractored object's sector to match object's */
		tract_obj->sect_x = sect_x; 
		tract_obj->sect_y = sect_y;
		tract_obj->sect_z = sect_z; 

		/* Move tractored object's coordinate position
		 * within the sector to the object's position
		 */
		tract_obj->x = obj->x;
		tract_obj->y = obj->y;
/*
		tract_obj->z = obj->z;
 */
	    }
	}

	return(0);
}


/*
 *	Alphabitizes the products on obj_num by name of
 *	the product.  If object has no economy or has economy
 *	but no products, then no operation is performed.
 */
void DBSortEconomyProducts(int obj_num)
{
	int i, n, k;
	xsw_ecodata_struct *eco;
	xsw_object_struct *obj;
	xsw_ecoproduct_struct **product_ptr;	/* To hold tmp ptrs */
	char **strv;
	int strc;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	eco = obj->eco;
	if(eco == NULL)
	    return;

	if(eco->total_products <= 0)
	    return;

	/* Build products name list in strv */
	for(i = 0, strv = NULL, strc = 0;
	    i < eco->total_products;
	    i++
	)
	{
	    if(eco->product[i] == NULL)
		continue;

	    /* Copy name to string list */
	    strc++;
	    strv = (char **)realloc(strv, strc * sizeof(char *));
	    if(strv == NULL)
		return;

	    strv[strc - 1] = STRDUP(
		eco->product[i]->name
	    );
	    if(strv[strc - 1] == NULL)
	    {
		strc--;
		continue;
	    }
	}

	/* Got no strings? */
	if(strc == 0)
	    return;

	/* Sort strings */
	strv = StringQSort(strv, strc);
	if(strv == NULL)
	    return;


	/* Copy products pointers to tempory pointer array */
	product_ptr = (xsw_ecoproduct_struct **)calloc(
	    1,
	    eco->total_products * sizeof(xsw_ecoproduct_struct *)
	);
	memcpy(
	    product_ptr,	/* Destination */
	    eco->product,	/* Source */
	    eco->total_products * sizeof(xsw_ecoproduct_struct *)
	);


	/* Sort products with respect to the strv list */
	for(i = 0, k = 0; i < strc; i++)
	{
	    if(k >= eco->total_products)
		break;

	    if(strv[i] == NULL)
		continue;

	    /* Get pointer matching string in tempory products list */
	    for(n = 0; n < eco->total_products; n++)
	    {
		if(product_ptr[n] == NULL)
		    continue;

		if(!strcasecmp(product_ptr[n]->name, strv[i]))
		    break;
	    }
	    if(n < eco->total_products)
	    {
		/* Matched product name, now put it in order */

		eco->product[k] = product_ptr[n];
	        k++;
	    }
	    else
	    {
		fprintf(stderr,
 "DBSortEconomyProducts(): Warning: Missing name from tmp products array (#%i %i %i)\n",
		    obj_num, strc, eco->total_products
		);
	    }
	}
	/* Reallocate the rest of the product pointers */
	if(k > 0)
	{
	    eco->total_products = k;
	    eco->product = (xsw_ecoproduct_struct **)realloc(
		eco->product,
		eco->total_products * sizeof(xsw_ecoproduct_struct *)
	    );
	    if(eco->product == NULL)
	    {
		eco->total_products = 0;
		/* Don't return, need to free other stuff */
	    }
	}
	else
	{
	    free(eco->product);
	    eco->product = NULL;

	    eco->total_products = 0;
	}


	/* Free tempory products pointers */
	free(product_ptr);
	product_ptr = NULL;

	/* Free string array */
	strlistfree(strv, strc);
}


/*
 *      Return a statically allocated string containing the official
 *      name for the given object type code.
 */     
char *DBGetTypeName(int type)
{
	switch(type)
	{
	  case SW_OBJ_TYPE_ERROR:
	    return(SW_OBJ_TYPE_NAME_ERROR);
	    break;

	  case SW_OBJ_TYPE_GARBAGE:
	    return(SW_OBJ_TYPE_NAME_GARBAGE);
	    break;

	  case SW_OBJ_TYPE_STATIC:
	    return(SW_OBJ_TYPE_NAME_STATIC);
	    break; 

	  case SW_OBJ_TYPE_DYNAMIC:
	    return(SW_OBJ_TYPE_NAME_DYNAMIC);
	    break;

	  case SW_OBJ_TYPE_CONTROLLED:
	    return(SW_OBJ_TYPE_NAME_CONTROLLED);
	    break;

	  case SW_OBJ_TYPE_PLAYER:
	    return(SW_OBJ_TYPE_NAME_PLAYER);
	    break;
  
	  case SW_OBJ_TYPE_WEAPON:
	    return(SW_OBJ_TYPE_NAME_WEAPON);   
	    break;

	  case SW_OBJ_TYPE_STREAMWEAPON:
	    return(SW_OBJ_TYPE_NAME_STREAMWEAPON);
	    break;

	  case SW_OBJ_TYPE_SPHEREWEAPON:
	    return(SW_OBJ_TYPE_NAME_SPHEREWEAPON);
	    break;

	  case SW_OBJ_TYPE_HOME:
	    return(SW_OBJ_TYPE_NAME_HOME);
	    break;

	  case SW_OBJ_TYPE_AREA:
	    return(SW_OBJ_TYPE_NAME_AREA);
	    break;

	  case SW_OBJ_TYPE_ANIMATED:
	    return(SW_OBJ_TYPE_NAME_ANIMATED);
	    break;

	  case SW_OBJ_TYPE_WORMHOLE:
	    return(SW_OBJ_TYPE_NAME_WORMHOLE);
	    break;

	  case SW_OBJ_TYPE_ELINK:
	    return(SW_OBJ_TYPE_NAME_ELINK);
	    break;

	  default:
	    return("*Unknown*");
	    break;
	}

	return("*Unknown*");
}


/*
 *      Returns a statically allocated format name string in
 *      the format of Name(#12345)
 */
char *DBGetFormalNameStr(int obj_num)
{
	if(DBIsObjectGarbage(obj_num))
	    return(SW_OBJ_NAME_GARBAGE);
	else
	    return(UNVGetObjectFormalName(
		xsw_object[obj_num],
		obj_num
	    ));
}

/*
 *	Returns a pointer to the name of the OPM referanced
 *	by the given OCS code. Can return NULL on no match or
 *	error.
 *
 *	Do not free the returned pointer.
 */
char *DBGetOCSOPMName(int ocs_code)
{
	const int ocs_num = OCSGetByCode(ocs_code);
	if(ocs_num < 0)
	    return(NULL);

	return(ocs[ocs_num]->opm_name);
}

/*      
 *	Checks if the specified name is valid for an object to have.
 *      
 *	-1	General error
 *	0	Success
 *	1	Name too long
 *	2	Name too short.
 */        
int DBValidateObjectName(const char *name)
{          
	int i, n;  
	char *s = SW_INVALID_NAME_CHARACTERS;

	/* Null pointer? */
	if(name == NULL)
	    return(-1);

	/* Empty string? */
	if(*name == '\0')
	    return(2);

	/* Invalid characters? */
	for(i = 0; name[i] != '\0'; i++)   
	{
	    if(i >= SW_OBJ_NAME_MAX)
		return(1);

	    for(n = 0; s[n] != '\0'; n++)
	    {
		if(name[i] == s[n])
		    return(3);
	    }
	}

	return(0);
}


/*         
 *      Checks if this is a valid password.
 */        
int DBValidateObjectPassword(const char *password)
{
	return(0);
}


/*
 *      Returns object n's visibility, can be 0 if the object is
 *      invalid or garbage.
 */
double DBGetObjectVisibility(int obj_num)
{
	return(UNVGetObjectVisibility(
	    xsw_object, total_objects, obj_num
	));
}

double DBGetObjectVisibilityPtr(xsw_object_struct *obj)
{
	return(UNVGetObjectVisibilityPtr(obj));
}

/*
 *	Returns the player object that owns the specified object (not
 *      always just the immediate owner).
 */
int DBGetPlayerOwner(int obj_num)
{
	xsw_object_struct *obj;

	while(!DBIsObjectGarbage(obj_num))
	{
	    obj = xsw_object[obj_num];
	    if(obj->type == SW_OBJ_TYPE_PLAYER)
		return(obj_num);

	    obj_num = obj->owner;
	}

	return(-1);
}

/*
 *      Returns object number of highest non-garbage object.
 *
 *      Returns -1 on error or no objects found.
 */
int DBGetTopObjectNumber(void)
{
	int i, h;
	xsw_object_struct **ptr, *obj_ptr;

	for(i = 0, h = -1, ptr = xsw_object;
	    i < total_objects;
	    i++, ptr++
	)
	{
	    obj_ptr = *ptr;
	    if(obj_ptr == NULL)
		continue;

	    if(obj_ptr->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    h = i;
	}

	return(h);
}
