#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int EcoDoBuy(
	int condescriptor,
	int customer_num, int proprietor_num,
	xsw_ecoproduct_struct product
);
int EcoDoSell(
	int condescriptor,
	int customer_num, int proprietor_num,
	xsw_ecoproduct_struct product
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Procedure to buy a product specified by product from the
 *	proprietor to the customer.
 *
 *	The product specifies the product's name and amount to buy,
 *	it is not a pointer to an actual product on either the
 *	customer or proprietor.
 */
int EcoDoBuy(
	int condescriptor,	/* Connection of customer object */
	int customer_num, int proprietor_num,
	xsw_ecoproduct_struct product
)
{
	int iff_code, p_product_num;
	double tax, buy_price_taxed, qty;
	const char *product_name_clean;
	char sndbuf[CS_DATA_MAX_LEN];
	sw_credits_t charged = 0.0;
	xsw_object_struct *customer, *proprietor;
	xsw_ecodata_struct *c_eco, *p_eco;
	xsw_ecoproduct_struct *p_product;


	/* Priorietor object must be valid */
	if(DBIsObjectGarbage(proprietor_num))
	    return(0);

	proprietor = xsw_object[proprietor_num];

	/* Customer object must be valid */
	if(DBIsObjectGarbage(customer_num))
	    return(0);

	customer = xsw_object[customer_num];


	/* Are customer and proprietor the same? */
	if(customer == proprietor)
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
"Customer and proprietor are the same."
	    );
	    return(-1);
	}

	/* Customer's shields up? */
	if(customer->shield_state == SHIELD_STATE_UP)
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
"Unable to buy or obtain, your shields are up."
	    );
	    return(-1);
	}

	/* Proprietor's shields up? */
	if(proprietor->shield_state == SHIELD_STATE_UP)   
	{
	    char *mesg = (char *)malloc(STRLEN(proprietor->name) + 80);
	    sprintf(
		mesg,
"Unable to buy or obtain, %s's shields are up.",
		proprietor->name
	    );
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		mesg
	    );
	    free(mesg);
	    return(-1);
	}

	/* Proprietor must have an economy and score structure */
	p_eco = proprietor->eco;
	if(p_eco == NULL)
	    return(0);
	if(UNVAllocScores(proprietor))
	    return(0);

	/* Get customer economy structure (may be NULL) */
	c_eco = customer->eco;

	/* Customer must have score structure */
	if(UNVAllocScores(customer))
	    return(0); 


	/* Check if the customer and proprietor are in range */
	if(!Mu3DInRange(
	    xsw_object, total_objects,
	    customer_num, proprietor_num, sysparm.transporter_range
	))
	{
	    char *mesg = (char *)malloc(
		STRLEN(proprietor->name) + 80
	    );
	    sprintf(
		mesg,
"You are too far away to buy from %s.",
		proprietor->name
	    );
	    NetSendSysMessage(
		condescriptor,      
		CS_SYSMESG_CODE_ERROR,
		mesg
	    );
	    free(mesg);
	    return(-1);
	}

	/* Note, the customer object is assumed already checked to be
	 * owned by the connetion
	 */

	/* Check if the proprietor is closed */
	if(!(p_eco->flags & ECO_FLAG_OPEN))
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		"Proprietor object is closed."
	    );
	    return(-1);
	}
	/* Check if the proprietor is willing to buy */
	if(!(p_eco->flags & ECO_FLAG_BUY_OK))
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		"Proprietor object does not permit purchasing."
	    );
	    return(-1);
	}


	/* Product name may not be empty */
	if(product.name[0] == '\0')
	    return(0);

	/* Get clean product name */
	product_name_clean = strchr(product.name, ':');
	if(product_name_clean != NULL)
	    product_name_clean++;
	else
	    product_name_clean = product.name;

	/* Request product amount must be positive */
	if(product.amount <= 0.0)
	    return(0);

	/* Get the real product from the proprietor */
	p_product_num = EcoGetProductNumByName(proprietor_num, product.name);
	if(p_product_num < 0)
	    return(-1);

	p_product = p_eco->product[p_product_num];


	/* Check IFF for customer and proprietor to determine tax level */
	iff_code = MatchIFFPtr(proprietor, customer);
	if(iff_code == SW_IFF_FRIENDLY)
	    tax = p_eco->tax_friend;
	else if(iff_code == SW_IFF_HOSTILE)
	    tax = p_eco->tax_hostile;
	else
	    tax = p_eco->tax_general;

	/* Calculate the unit buy price with tax applied */
	buy_price_taxed = p_product->buy_price * tax;

	/* Get initial quantity, this value may be reduced due to
	 * the limit that the customer can carry, the proprietor has,
	 * and how many credits the customer has
	 */
	qty = product.amount;


	/* Handle buy by product name */
	/* Antimatter */
	if(!strcasecmp(product.name, ECO_PROD_NAME_ANTIMATTER))
	{
	    const double	cn = customer->antimatter,
				cm = customer->antimatter_max,
				pn = p_product->amount,
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the customer can carry */
	    if((qty + cn) > cm)
		qty = MAX(cm - cn, 0.0);

	    /* Reduce quantity by how much the proprietor has left */
	    if(pm > 0.0)
	    {
		if(qty > pn)
		    qty = pn;
	    }

	    /* Being charged for this product? */
	    if(buy_price_taxed > 0.0)
	    {
		/* Charge customer */
		charged = EcoTransCredits(
		    customer_num, -(qty * buy_price_taxed)
		) * -1.0;

		/* Give credits to proprietor */
		EcoTransCredits(proprietor_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / buy_price_taxed;
	    }

	    /* Perform transaction */
	    customer->antimatter += qty;
	    if(pm > 0.0)
		p_product->amount -= qty;

	    NetSendObjectValues(-1, customer_num);
	}
	/* Hull repair */
	else if(!strcasecmp(product.name, ECO_PROD_NAME_REPAIRHULL))
	{
	    const double	cn = customer->hp,
				cm = customer->hp_max,
				pn = p_product->amount,
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the customer can carry */
	    if((qty + cn) > cm)
		qty = MAX(cm - cn, 0.0);

	    /* Reduce quantity by how much the proprietor has left */
	    if(pm > 0.0)
	    {
		if(qty > pn)
		    qty = pn;
	    }

	    /* Being charged for this product? */
	    if(buy_price_taxed > 0.0)
	    {   
		/* Charge customer */
		charged = EcoTransCredits(
		    customer_num, -(qty * buy_price_taxed)
		) * -1.0;

		/* Give credits to proprietor */
		EcoTransCredits(proprietor_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / buy_price_taxed;
	    }

	    /* Perform transaction */
	    if(cm > 0.0)
		customer->hp += qty;
	    if(pm > 0.0)
		p_product->amount = MAX(
		    p_product->amount - qty, 0.0
		);

	    NetSendObjectValues(-1, customer_num);
	}
	/* Raw Material */
	else if(!strcasecmp(product.name, ECO_PROD_NAME_RMU))
	{
	    const double	cn = customer->score->rmu,
				cm = customer->score->rmu_max,
				pn = p_product->amount,
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the customer can carry */
	    if((qty + cn) > cm)
		qty = MAX(cm - cn, 0.0);

	    /* Reduce quantity by how much the proprietor has left */
	    if(pm > 0.0)
	    {
		if(qty > pn)
		    qty = pn;
	    }

	    /* Being charged for this product? */
	    if(buy_price_taxed > 0.0)
	    {
		/* Charge customer */
		charged = EcoTransCredits(
		    customer_num, -(qty * buy_price_taxed)
		) * -1.0;

		/* Give credits to proprietor */
		EcoTransCredits(proprietor_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / buy_price_taxed;
	    }

	    /* Perform transaction */
	    if(cm > 0.0)
		customer->score->rmu += qty;
	    if(pm > 0.0)
		p_product->amount = MAX(
		    p_product->amount - qty, 0.0
		);

	    NetSendObjectValues(-1, customer_num);
	}
	/* OPM player ships */
	else if(strcasepfx(product.name, ECO_PROD_NAME_OPMPFX) &&
		!OPMIsGarbage(OPMGetByName(product.name, 4))
	)
	{
#if 0
	    int wep_num, opm_num;
	    xsw_object_struct *saveplayer_ptr;

	    /* Sanitize amount for customer */
	    d_amount = product.amount;

	    /* Any ship at all? */
	    if(d_amount <= 0)
		return(0);

	    /* A player can never buy more than one ship */
	    if((d_amount > 1.0) || (d_amount < 1.0))
		d_amount = 1.0;

	    /* Set action item code to 0 so it's just valid */
	    schedual_act_item_code = 0;

	    /* Calculate total price */
	    d_credits = d_amount * p_product->buy_price
		* proprietor->eco->tax_general;

	    /* Check if player has enough credits for this ship */
	    if(customer->score->credits < d_credits)
	    {
		NetSendSysMessage(
		    condescriptor,
		    CS_SYSMESG_CODE_ERROR,
 "Sorry but you do not have enough credits to buy this ship."
		);
		return(-1);
	    }

	    /* Set action item code to 0 so it's just valid */
	    schedual_act_item_code = 0;
	    
	    /* Charge customer and get sanitize credits */
 	    EcoTransCredits(customer_num,-d_credits);

	    /* Give credits to proprietor */
	    EcoTransCredits(proprietor_num, d_credits);

	    /* Calculate increment limit */
	    schedual_act_inc_limit = d_amount;

	    /* Copy customer object for setting back values
	     * later
	     */
	    saveplayer_ptr = UNVDupObject(customer);
	    if(saveplayer_ptr == NULL)
	    {
		NetSendSysMessage(
		    condescriptor,
		    CS_SYSMESG_CODE_ERROR,
		    "Memory allocation error."
		);
		return(-1);
	    }

	    /* Get OPM number for modelling */
	    opm_num = OPMGetByName(product.name, SW_OBJ_TYPE_PLAYER);
	    if(opm_num < 0)
	    {
		sprintf(
		    sndbuf,
		    "%s: No such OPM.",
		    product.name
		);
		NetSendLiveMessage(condescriptor, sndbuf);

		/* Delete saved player object, no longer needed */
		UNVDeleteObject(saveplayer_ptr);

 		return(-1);
	    }

	    /* Model customer object with OPM values, any current
	     * allocated substructures on the customer object will
	     * be deallocated and replaced by (if any) ones specified
	     * in the OPM
	     */
	    if(OPMModelObject(customer_num, opm_num))
	    {
		/* Error occured when attempting to remodel object */
		sprintf(
		    sndbuf,
		    "%s: Unable to remodel.",
		    DBGetFormalNameStr(customer_num)
		);
	    }
	    else
	    {
		sprintf(
		    sndbuf,
		    "%s changed vessel to %s. (OPM #%i)",
		    saveplayer_ptr->name,
		    opm[opm_num]->name,
		    opm_num
		);

		/* Restore player values that would be changed by the
		 * modeling to the OPM's values
		 */
		customer->owner = saveplayer_ptr->owner;
		customer->permission.uid = saveplayer_ptr->permission.uid;
		customer->permission.gid = saveplayer_ptr->permission.gid;
		customer->intercepting_object = saveplayer_ptr->intercepting_object;
		customer->locked_on = saveplayer_ptr->locked_on;
		customer->lighting = saveplayer_ptr->lighting;
		customer->shield_frequency = saveplayer_ptr->shield_frequency;
		customer->com_channel = saveplayer_ptr->com_channel;
		customer->sect_x = saveplayer_ptr->sect_x;
		customer->sect_y = saveplayer_ptr->sect_y;
		customer->sect_z = saveplayer_ptr->sect_z;
		customer->x = saveplayer_ptr->x;
		customer->y = saveplayer_ptr->y;
		customer->z = saveplayer_ptr->z;
		strncpy(
		    customer->empire,
		    saveplayer_ptr->empire,
		    sizeof(customer->empire)
		);
		customer->empire[sizeof(customer->empire) - 1] = '\0';
		free(customer->password);
		customer->password = STRDUP(saveplayer_ptr->password);
		strncpy(
		    customer->name,
		    saveplayer_ptr->name,
		    sizeof(customer->name)
		);
		customer->name[sizeof(customer->name) - 1] = '\0';

		/* Score structure if exists on saved object */
		if(saveplayer_ptr->score != NULL)
		{
		    if(!UNVAllocScores(customer))
		    {
			memcpy(
			    customer->score,
			    saveplayer_ptr->score,  
			    sizeof(xsw_score_struct)
			);
		    }
		}
	    }
	    NetSendLiveMessage(condescriptor, sndbuf);

	    /* Delete the saved object */
	    UNVDeleteObject(saveplayer_ptr);

	    /* Notify customer */
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_WARNING,
 "Congratulations, your new starship is waiting in the dock!"
	    );

	    /* Update weapon values on customer */
	    for(wep_num = 0; wep_num <= customer->total_weapons; wep_num++)
	    {
		NetSendWeaponValues(
		    -1,			/* All connections */
		    customer_num,
		    wep_num		/* Weapon number on object */
		);
	    }
#endif
	}
	/* OPM Weapons */
	else if(strcasepfx(product.name, ECO_PROD_NAME_OPMPFX) &&
	        !OPMIsGarbage(OPMGetByName(product.name, -1))
	)
	{
	    xsw_weapon_struct *wep;

	    /* Match the weapon number on the customer */
	    const int wep_num = EcoGetWeaponNumByName(
		customer, product.name
	    );
	    if(wep_num < 0)
	    {
		sprintf(
		    sndbuf,
		    "Your vessel is unable to carry `%s'.",
		    product_name_clean
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(0);
	    }

	    wep = customer->weapons[wep_num];
	    if(wep != NULL)
	    {
		const int	cn = wep->amount,
				cm = wep->max,
				pn = (int)p_product->amount,
				pm = (int)p_product->amount_max;

		qty = floor(qty);

		/* Reduce quantity by how much the customer can carry */
		if((qty + cn) > cm)
		    qty = MAX(cm - cn, 0);

		/* Reduce quantity by how much the proprietor has left */
		if(pm > 0.0)
		{
		    if(qty > pn)
			qty = pn;
		}

		/* Being charged for this product? */
		if(buy_price_taxed > 0.0)
		{
		    /* Enough to buy at least one unit? */
		    if(customer->score->credits >= buy_price_taxed)
		    {
			/* Reduce quantity if the customer does not
			 * have enough credits to buy a whole numbered
			 * amount of the specified quantity
			 */
			if((qty * buy_price_taxed) > customer->score->credits)
			    qty = floor(
				customer->score->credits / buy_price_taxed
			    );

			/* Charge customer */
			charged = EcoTransCredits(
			    customer_num, -(qty * buy_price_taxed)
			) * -1.0;

			/* Give credits to proprietor */
			EcoTransCredits(proprietor_num, charged);

			/* Update quantity to the actual price paied */
			qty = floor(charged / buy_price_taxed);
		    }
		    else
		    {
			qty = 0.0;
		    }
		}

		/* Perform transaction */
		if(cm > 0)
		    wep->amount += (int)qty;
		if(pm > 0)
		    p_product->amount = MAX(
			p_product->amount - qty, 0.0
		    );

		/* Update weapon value on customer */
		NetSendWeaponValues(
		    -1,			/* All connections */
		    customer_num,
		    wep_num		/* Weapon number */
		);
	    }
	}
	/* Other products */
	else if((c_eco != NULL) &&
	        (EcoGetProductNumByName(customer_num, product.name) > -1)
	)
	{
	    const int c_product_num = EcoGetProductNumByName(
		customer_num, product.name
	    );
	    xsw_ecoproduct_struct *c_product = c_eco->product[c_product_num];
	    const double	cn = c_product->amount,
				cm = c_product->amount_max,
				pn = p_product->amount,
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the customer can carry */
	    if((qty + cn) > cm)
		qty = MAX(cm - cn, 0.0);

	    /* Reduce quantity by how much the proprietor has left */
	    if(pm > 0.0)
	    {
		if(qty > pn)
		    qty = pn;
	    }
	     
	    /* Being charged for this product? */
	    if(buy_price_taxed > 0.0)
	    {
		/* Charge customer */
		charged = EcoTransCredits(
		    customer_num, -(qty * buy_price_taxed)
		) * -1.0;

		/* Give credits to proprietor */
		EcoTransCredits(proprietor_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / buy_price_taxed;
	    }

	    /* Perform transaction */
	    if(cm > 0.0)
		c_product->amount += qty;
	    if(pm > 0.0)
		p_product->amount = MAX(
		    p_product->amount - qty, 0.0
		);
	}
	else
	{
	    /* Unsupported product */
	    char s[ECO_PRODUCT_NAME_MAX + 128];
	    sprintf(
		s,
		"Your vessel is unable to carry `%s'",
		product_name_clean
	    );
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		s
	    );
	}

#if 0
/* Some universes do not work well with this, make this an option? */
	/* Adjust buy and sell prices */
	EcoAdjustPriceAuto(p_product, -qty);
#endif

	/* Update product values to connection */
	NetSendEcoSetProductValues(
	    condescriptor,	/* Connection */
	    proprietor_num,	/* Proprietor object */
	    p_product_num	/* Product number on proprietor */
	);

	/* Update scores */
	NetSendScore(condescriptor, proprietor_num);
	NetSendScore(condescriptor, customer_num);

	/* Notify connection of buy results */
	EcoNotifyBuy(
	    condescriptor,
	    product_name_clean,
	    proprietor->name,
	    qty,
	    charged
	);

	return(0);
}

/*
 *	Procedure to sell a product specified by product to the
 *	proprietor from the customer.
 *
 *	The product specifies the product's name and amount to sell,
 *	it is not a pointer to an actual product on either the
 *	customer or proprietor.
 */
int EcoDoSell(
	int condescriptor,	/* Connection of customer object */
	int customer_num, int proprietor_num,
	xsw_ecoproduct_struct product
)
{
	int iff_code, p_product_num;
	double tax, sell_price_taxed, qty;
	const char *product_name_clean;
	char sndbuf[CS_DATA_MAX_LEN];
	sw_credits_t charged = 0.0;    
	xsw_object_struct *customer, *proprietor;
	xsw_ecodata_struct *c_eco, *p_eco;       
	xsw_ecoproduct_struct *p_product; 

	/* Priorietor object must be valid */
	if(DBIsObjectGarbage(proprietor_num))
	    return(0);

	proprietor = xsw_object[proprietor_num];

	/* Customer object must be valid */
	if(DBIsObjectGarbage(customer_num))
	    return(0);

	customer = xsw_object[customer_num];

	/* Are customer and proprietor the same? */
	if(customer == proprietor)
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		"Customer and proprietor are the same."
	    );
	    return(-1);
	}

	/* Customer's shields up? */
	if(customer->shield_state == SHIELD_STATE_UP)
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
"Unable to sell or distribute, your shields are up."
	    );
	    return(-1); 
	}

	/* Proprietor's shields up? */
	if(proprietor->shield_state == SHIELD_STATE_UP)
	{
	    char *mesg = (char *)malloc(STRLEN(proprietor->name) + 80);
	    sprintf(
		mesg,
"Unable to sell or distribute, %s's shields are up.",
		proprietor->name
	    );
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		mesg
	    );
	    free(mesg);
	    return(-1);
	}

	/* Proprietor must have an economy and score structure */
	p_eco = proprietor->eco;
	if(p_eco == NULL)
	    return(0);
	if(UNVAllocScores(proprietor))
	    return(0);

	/* Get customer economy structure (may be NULL) */
	c_eco = customer->eco;

	/* Customer must have score structure */
	if(UNVAllocScores(customer))
	    return(0);

	/* Check if the customer and proprietor are in range */
	if(!Mu3DInRange(
	    xsw_object, total_objects,
	    customer_num, proprietor_num, ECO_MAX_TRANSACTION_RANGE
	))
	{
	    char *mesg = (char *)malloc(
		STRLEN(proprietor->name) + 80
	    );
	    sprintf(
		mesg,
"You are too far away to sell to %s.",
		proprietor->name
	    );
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		mesg
	    );
	    free(mesg);
	    return(-1);
	}

	/* Note, the customer object is assumed already checked to be
	 * owned by the connetion
	 */

	/* Check if the proprietor is closed */
	if(!(p_eco->flags & ECO_FLAG_OPEN))
	{       
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		"Proprietor object is not opened for business."
	    );
	    return(-1);
	}
	/* Check if the proprietor is willing to sell */
	if(!(p_eco->flags & ECO_FLAG_SELL_OK))
	{
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		"Proprietor object does not permit selling."
	    );
	    return(-1);
	}


	/* Product name may not be empty */
	if(product.name[0] == '\0')  
	    return(0);

	/* Get clean product name */
	product_name_clean = strchr(product.name, ':');
	if(product_name_clean != NULL)
	    product_name_clean++;
	else
	    product_name_clean = product.name;

	/* Request product amount must be positive */
	if(product.amount <= 0.0)
	    return(0);

	/* Get the real product from the proprietor */
	p_product_num = EcoGetProductNumByName(proprietor_num, product.name);
	if(p_product_num < 0)
	    return(-1);

	p_product = p_eco->product[p_product_num];


	/* Check IFF for customer and proprietor to determine tax level */
	iff_code = MatchIFFPtr(proprietor, customer);
	if(iff_code == SW_IFF_FRIENDLY)
	    tax = p_eco->tax_friend;
	else if(iff_code == SW_IFF_HOSTILE)
	    tax = p_eco->tax_hostile;
	else
	    tax = p_eco->tax_general;

	/* Calculate the unit sell price with tax applied */
	sell_price_taxed = p_product->sell_price * tax;

	/* Get initial quantity, this value may be reduced due to
	 * the limit that the proprietor can carry, the quantity the
	 * customer has left, and how many credits the proprietor has
	 * left
	 */
	qty = product.amount;


	/* Handle sell by product name */
	/* Antimatter */
	if(!strcasecmp(product.name, ECO_PROD_NAME_ANTIMATTER))
	{
	    const double        cn = customer->antimatter,
				cm = customer->antimatter_max,
				pn = p_product->amount,       
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the proprietor can carry */
	    if(pm > 0.0)
	    {
		if((qty + pn) > pm)
		    qty = MAX(pm - pn, 0.0);
	    }

	    /* Reduce quantity by how much the customer has left */
	    if(cm > 0.0)
	    {
	        if(qty > cn)
		    qty = cn;
	    }

	    /* Being charged for this product? */
	    if(sell_price_taxed > 0.0)
	    {
		/* Charge proprietor */
		charged = EcoTransCredits(
		    proprietor_num, -(qty * sell_price_taxed)
		) * -1.0;

		/* Give credits to customer */
		EcoTransCredits(customer_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / sell_price_taxed;
	    }

	    /* Perform transaction */
	    if(pm > 0.0)
		p_product->amount += qty;
	    if(cm > 0.0)
		customer->antimatter = MAX(
		    customer->antimatter - qty, 0.0
		);

	    NetSendObjectValues(-1, customer_num);
	}
	/* Hull repair */
	else if(!strcasecmp(product.name, ECO_PROD_NAME_REPAIRHULL))
	{                                                         
	    /* Not supported in selling */
	    sprintf(
		sndbuf,
"%s is not a product that can be sold.",
		product.name
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	    return(0);
	}
	/* Raw Material */
	else if(!strcasecmp(product.name, ECO_PROD_NAME_RMU))
	{
	    const double	cn = customer->score->rmu,
				cm = customer->score->rmu_max,
				pn = p_product->amount,
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the proprietor can carry */
	    if(pm > 0.0)
	    {
		if((qty + pn) > pm)
		    qty = MAX(pm - pn, 0.0);
	    }

	    /* Reduce quantity by how much the customer has left */
	    if(cm > 0.0)
	    {
		if(qty > cn)
		    qty = cn;
	    }

	    /* Being charged for this product? */
	    if(sell_price_taxed > 0.0)
	    {
		/* Charge proprietor */
		charged = EcoTransCredits(
		    proprietor_num, -(qty * sell_price_taxed)
		) * -1.0;

		/* Give credits to customer */
		EcoTransCredits(customer_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / sell_price_taxed;
	    }

	    /* Perform transaction */
	    if(pm > 0.0)
		p_product->amount += qty;
	    if(cm > 0.0)
		customer->score->rmu = MAX(
		     customer->score->rmu - qty, 0.0
		);

	    NetSendObjectValues(-1, customer_num);
	}
	/* OPM Weapons */
	else if(strcasepfx(product.name, ECO_PROD_NAME_OPMPFX) &&
		!OPMIsGarbage(OPMGetByName(product.name, -1))
	)
	{
	    xsw_weapon_struct *wep;

	    /* Match the weapon number on the customer */
	    const int wep_num = EcoGetWeaponNumByName(
		customer, product.name
	    );
	    if(wep_num < 0)
	    {
		sprintf(
		    sndbuf,
		    "Your vessel does not carry `%s'.",
		    product_name_clean
		);
		NetSendLiveMessage(condescriptor, sndbuf);
		return(0);
	    }

	    wep = customer->weapons[wep_num];
	    if(wep != NULL)
	    {
		const int	cn = wep->amount,
				cm = wep->max,
				pn = (int)p_product->amount,
				pm = (int)p_product->amount_max;

		qty = floor(qty);

		/* Reduce quantity by how much the proprietor can carry */
		if(pm > 0)
		{
		    if((qty + pn) > pm)
			qty = MAX(pm - pn, 0);
		}

		/* Reduce quantity by how much the customer has left */
		if(cm > 0)
		{
		    if(qty > cn)
		        qty = cn;
		}

		/* Being charged for this product? */
		if(sell_price_taxed > 0.0)
		{
		    /* Enough credits to sell at least one? */
		    if(proprietor->score->credits >= sell_price_taxed)
		    {
			/* Charge proprietor */
			charged = EcoTransCredits(
			    proprietor_num, -(qty * sell_price_taxed)
			) * -1.0;

			/* Give credits to customer */
			EcoTransCredits(customer_num, charged);

			/* Update quantity to the actual price paied */
			qty = floor(charged / sell_price_taxed);
		    }
		    else
		    {
			qty = 0.0;
		    }
		}

		/* Perform transaction */
		if(pm > 0)
		    p_product->amount += qty;
		if(cm > 0)
		    wep->amount = MAX(wep->amount - (int)qty, 0);

		/* Update weapon value on customer */
		NetSendWeaponValues(
		    -1,			/* All connections */
		    customer_num,
		    wep_num		/* Weapon number */
		);
	    }
	}
	/* Other products */
	else if((customer->eco != NULL) &&
	        (EcoGetProductNumByName(customer_num, product.name) >= 0)
	)
	{
	    const int c_product_num = EcoGetProductNumByName(
		customer_num, product.name
	    );
	    xsw_ecoproduct_struct *c_product = c_eco->product[c_product_num];
	    const double	cn = c_product->amount,
				cm = c_product->amount_max,   
				pn = p_product->amount,
				pm = p_product->amount_max;

	    /* Reduce quantity by how much the proprietor can carry */
	    if(pm > 0.0)
	    {
		if((qty + pn) > pm)
		    qty = MAX(pm - pn, 0.0);
	    }

	    /* Reduce quantity by how much the customer has left */
	    if(cm > 0.0)
	    {
		if(qty > cn)
		    qty = cn;
	    }

	    /* Being charged for this product? */
	    if(sell_price_taxed > 0.0)
	    {
		/* Charge proprietor */
		charged = EcoTransCredits(
		    proprietor_num, -(qty * sell_price_taxed)
		) * -1.0;

		/* Give credits to customer */
		EcoTransCredits(customer_num, charged);

		/* Update quantity to the actual price paied */
		qty = charged / sell_price_taxed;
	    }

	    /* Perform transaction */
	    if(pm > 0.0)
		p_product->amount += qty;
	    if(cm > 0.0)
		c_product->amount = MAX(
		    c_product->amount - qty, 0.0
		);
	}
	else    
	{
	    /* Unsupported product */
	    char s[ECO_PRODUCT_NAME_MAX + 128];
	    sprintf(
		s,
		"Your vessel does not carry `%s'",
		product_name_clean
	    );
	    NetSendSysMessage(
		condescriptor,
		CS_SYSMESG_CODE_ERROR,
		s
	    );
	}

#if 0
/* Some universes do not work well with this, make this an option? */
	/* Adjust buy and sell prices */
	EcoAdjustPriceAuto(p_product, d_amount);
#endif

	/* Update product values to connection */
	NetSendEcoSetProductValues(
	    condescriptor,	/* Connection */
	    proprietor_num,	/* Proprietor object */
	    p_product_num	/* Product number on proprietor */
	);

	/* Update scores */
	NetSendScore(condescriptor, proprietor_num);
	NetSendScore(condescriptor, customer_num);

	/* Notify connection of buy results */
	EcoNotifySell(
	    condescriptor,
	    product_name_clean,
	    proprietor->name,
	    qty,
	    -charged		/* Negative to indicate sell */
	);

	return(0);
}
