#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"
#include "../include/xsw_ctype.h"

#include "net.h"
#include "swserv.h"


int EcoGetOCSFromName(const char *s, int start_ocs_num);
int EcoGetWeaponNumByName(
	xsw_object_struct *obj, const char *name
);

int EcoGetProductNumByName(int object_num, char *name);

sw_credits_t EcoTransCredits(
	int object_num, sw_credits_t d_credits
);

void EcoNotifyBuy(
	int condescriptor,
	const char *product_name, const char *proprietor_name,
	double total_amount, double total_price
);
void EcoNotifySell(
	int condescriptor,
	const char *product_name, const char *proprietor_name,
	double total_amount, double total_price
);

int EcoAdjustPriceAuto(
	xsw_ecoproduct_struct *product, double diff_amount
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
	 

/*
 *	Returns the next object create script index from
 *	start_ocs_num + 1 by parsing the object parameter model name
 *	from the string specified by s.
 *
 *	start_ocs_num should be one less than the ocs number that you
 *	want to start searching for.
 *
 *	The string s must be of the format "$OPM:<OPM_name>"
 *	where <opm_name> is the object parameter model name.
 *
 *	The next OCS referancing that object parameter macro name
 *	will be returned or -1 when no more OCS can be found.
 *
 *	Can return -1 on error or failed match.
 */
int EcoGetOCSFromName(const char *s, int start_ocs_num)
{
	int i;
	const char *opm_name;
	sws_ocs_struct *ocs_ptr;

	if(s == NULL)
	    return(-1);

	while(ISBLANK(*s))
	    s++;

	/* If the first character is not a '$' then there is no
	 * OPM name
	 */
	if(*s != '$')
	    return(-1);

/* Should we check if what is between '$' and ':' is "OPM"? */

	/* Search for the ':' delimiter character */
	opm_name = strchr(s, ':');
	if(opm_name == NULL)
	    return(-1);

	opm_name++;
	while(ISBLANK(*opm_name))
	    opm_name++;

	if(*opm_name == '\0')
	    return(-1);

	/* Increment start_ocs_num */
	start_ocs_num++;
	if(start_ocs_num < 0)
	    start_ocs_num = 0;

	/* Search for next OCS that references an OPM who's name
	 * matches the one parsed from string s
	 */
	for(i = start_ocs_num; i < total_ocss; i++)
	{
	    ocs_ptr = ocs[i];
	    if(ocs_ptr == NULL)
		continue;

	    if(ocs_ptr->opm_name == NULL)
		continue;

	    if(!strcasecmp(opm_name, ocs_ptr->opm_name))
		return(i);
	}

	return(-1);
}


/*
 *	Returns the weapon index on the specified object that matches
 *	the object parameter model name string specified by name.
 *
 *	Returns the weapon index or -1 on no match.
 */
int EcoGetWeaponNumByName(
	xsw_object_struct *obj, const char *name
)
{
	int i, n, wep_num, ocs_code;
	xsw_weapon_struct *wep;

	if(obj == NULL)
	    return(-1);

	for(i = -1; i < total_ocss; i = n)
	{
	    /* Get next matched OCS */
	    n = EcoGetOCSFromName(
		name,
		i
	    );
	    if(n < 0)
		break;

	    if(!OCSIsAllocated(n))
		continue;

	    ocs_code = ocs[n]->code;

	    /* Check if object has a weapon that uses the OCS */
	    for(wep_num = 0; wep_num < obj->total_weapons; wep_num++)
	    {
		wep = obj->weapons[wep_num];
		if(wep == NULL)
		    continue;

		if(wep->ocs_code == ocs_code)
		    return(wep_num);
	    }
	}

	return(-1);
}

	
/*
 *	Returns eco product index number of -1 on error of failed
 *	match.
 *
 *	If a match is returned, it is comfermed to be allocated and
 *	valid.
 */
int EcoGetProductNumByName(int object_num, char *name)
{
	int i;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *product;

	if(name == NULL)
	    return(-1);

	if(DBIsObjectGarbage(object_num))
	    return(-1);
	else
	    obj = xsw_object[object_num];

	eco = obj->eco;
	if(eco == NULL)
	    return(-1);

	for(i = 0; i < eco->total_products; i++)
	{
	    product = eco->product[i];
	    if(product == NULL)
		continue;

	    if(!strcasecmp(product->name, name))
		return(i);
	}

	return(-1);
}


/*
 *	Gives or takes away credits from the specified object.
 *
 *	If d_credits is positive then the credits will be given to the
 *	object, otherwise if the credits is negative then credits will
 *	be taken away from the object.
 *
 *	Returns actual number of credits given or taken away.
 */
sw_credits_t EcoTransCredits(
	int object_num, sw_credits_t d_credits
)
{
	xsw_score_struct *score;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(object_num))
	    return(0.0);

	obj = xsw_object[object_num];

	/* Allocate scores as needed */
	if(UNVAllocScores(obj))
	    return(0.0);

	score = obj->score;

	if((score->credits + d_credits) < 0.0)
	    d_credits = score->credits * -1.0;
	if((score->credits + d_credits) > SW_CREDITS_MAX)
	    d_credits = SW_CREDITS_MAX - score->credits; 

	score->credits += d_credits;

	return(d_credits);
} 


/*
 *	Notifies the results of a buy to the connection.
 */
void EcoNotifyBuy(
	int condescriptor,
	const char *product_name, const char *proprietor_name,
	double total_amount, double total_price
)
{
	double item_price;
	char *sndbuf;

	/* Calculate item price */
	if(total_amount != 0.0)
	    item_price = total_price / total_amount;
	else
	    item_price = 0.0;

	if(total_price > 0.0)
	{
	    sndbuf = (char *)malloc(
		STRLEN(product_name) + STRLEN(proprietor_name) +
		40 + 40 + 40 + 256
	    );
	    if(sndbuf == NULL)
		return;
	    if(total_amount > 0.0)
		sprintf(
		    sndbuf,
"Bought %.2f of %s from %s for %.2f credits (%.2f/unit).",
		    total_amount,
		    product_name,
		    proprietor_name,
		    total_price,
		    item_price
	        );
	    else
		sprintf(
		    sndbuf,
"Unable to buy %s from %s (capacity full, out of stock, or\
 insufficient credits).",
		    product_name,
		    proprietor_name
		);
	}
	else
	{
	    sndbuf = (char *)malloc(
		STRLEN(product_name) + STRLEN(proprietor_name) +
		40 + 256
	    );
	    if(sndbuf == NULL)
		return;
	    if(total_amount > 0.0)
		sprintf(
		    sndbuf,
"Obtained %.2f of %s from %s.",
		    total_amount,
		    product_name,
		    proprietor_name
		);
	    else
		sprintf(
		    sndbuf,
"Unable to obtain %s from %s (capacity full, out of stock, or\
 insufficient credits).",
		    product_name,
		    proprietor_name
		);
	}     
	NetSendLiveMessage(condescriptor, sndbuf);
	free(sndbuf);
}

/*
 *	Notifies the results of a sell to the connection.
 */
void EcoNotifySell(
	int condescriptor,
	const char *product_name, const char *proprietor_name,
	double total_amount, double total_price
)
{
	double item_price;
	char *sndbuf;

	/* Calculate item price */
	if(total_amount != 0.0)
	    item_price = total_price / total_amount;
	else
	    item_price = 0.0;

	if(total_price < 0.0)
	{
	    sndbuf = (char *)malloc(
		STRLEN(product_name) + STRLEN(proprietor_name) +
		40 + 40 + 40 + 256   
	    );
	    if(sndbuf == NULL)
		return;
	    if(total_amount > 0.0)
		sprintf(
		    sndbuf,
"Sold %.2f of %s to %s for %.2f credits (%.2f/unit).",
		    total_amount,
		    product_name,
		    proprietor_name,
		    -total_price,
		    -item_price
		);
	    else
		sprintf(
		    sndbuf,
"Unable to sell %s to %s (capacity full, out of stock, or\
 insufficient credits).",
		    product_name,
		    proprietor_name
	        );
	}
	else
	{
	    sndbuf = (char *)malloc(
		STRLEN(product_name) + STRLEN(proprietor_name) +
		STRLEN(proprietor_name) +
		40 +
		256   
	    );
	    if(sndbuf == NULL)
		return;
	    if(total_amount > 0.0)
		sprintf(
		    sndbuf,
"Distributed %.2f of %s to %s.",
		    total_amount,
		    product_name,
		    proprietor_name
		);
	    else
		sprintf(
		    sndbuf,
"Unable to distribute %s to %s (capacity full, out of stock, or\
 insufficient credits).",
		    product_name,
		    proprietor_name
		);
	}
	NetSendLiveMessage(condescriptor, sndbuf);
	free(sndbuf);
}


/*
 *	This function will set prices based on current amount related to
 *	max_amount. It will first calculate a "maxprice" for this
 *	product by looking at th in relation to the  to the
 *	amount/maxamount ratio.
 *
 *	Then it will calculate a new price using this maxprice and the
 *	new amount.
 *
 *	This way the maxprice (which is the price when the amount is zero)
 *	(wil to reuse in the next transaction.
 *
 *	The prices is set specially when the amount is max or zero, to
 *	bypass without loosing the maxprice.
 *
 *	The sell price is used for calculations, the buy_price is
 *	calculated u
 *
 *	TODO: Add support for products which the properitar will not buy
 *	(sell
 */
int EcoAdjustPriceAuto(
	xsw_ecoproduct_struct *product,
	double diff_amount	/* Difference between old and new amount */
)
{
	double		old_sell_price, old_buy_price,
			sell_price, max_price,
			old_amount, new_amount,
			max_amount, buy_price;

	/* Is product valid? */
	if(product == NULL)
	    return(0);

	/* Is diff_amount valid? */
	if(diff_amount == 0)
	    return(0);

	/* Get product values */
	old_sell_price = product->sell_price;
	old_buy_price = product->buy_price;
	old_amount = product->amount;
	max_amount = product->amount_max;

	/* No amount possible? */
	if(max_amount <= 0)
	    return(0);

	/* Calculate new amount */
	new_amount = old_amount + diff_amount;

	/* Check if stock was empty or full */
	if(old_amount >= max_amount)
	{
	    /* Full stock - restore maxprice */
	    max_price = old_sell_price * 100.0;
	}
	else if(old_amount == 0)
	{
	    /* Empty stock - restore maxprice */
	    max_price = old_buy_price;
	}
	else
	{
	    if(old_sell_price == 0)
	    {
		old_sell_price = old_buy_price;
	    }
	    /* Calculate max price based on old price and amount ratio */
	    max_price = (max_amount * old_sell_price) /
		(max_amount - old_amount);
	}       
       
	if(fabs(new_amount) >= max_amount)	// Dan S: used floating point version rather than cast.
	{
	    /* Full stock - save maxprice */
	    sell_price = max_price / 100;
	    buy_price = 0;
	}
	else if(new_amount == 0)
	{
	    /* Empty stock - save maxprice */
	    sell_price = 0;
	    buy_price = max_price;
	}
	else
	{
	    /* Calculate new price based on max price and new amount */
	    sell_price = (max_amount - new_amount) *
		(max_price / max_amount);
	    buy_price = sell_price * 1.2;

	    if(old_sell_price == old_buy_price)
	    {
		buy_price = sell_price;
		sell_price = 0;
	    }
	}

	/* Update product sell and buy prices */
	product->sell_price = sell_price;
	product->buy_price = buy_price;

	return(0);
}
