#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>

#include "swserv.h"


int IncomingSocketIsAllocated(int n);

int IncomingSocketNew(int port, int type);
void IncomingSocketDelete(int n);
void IncomingSocketDeleteAll(void);

void IncomingSocketReclaim(void);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the incoming socket is allocated.
 */
int IncomingSocketIsAllocated(int n)
{
	if((n < 0) || (n >= total_incoming_sockets))
	    return(0);
	else if(incoming_socket[n] == NULL)
	    return(0);
	else
	    return(1);
}


/*                  
 *	Creates a new incoming socket to listen for incoming
 *	connections on the specified port number.
 *
 *	Returns the new incoming socket number or -1 on error.
 */
int IncomingSocketNew(int port, int type)
{
	int i, n, status;
	struct sockaddr_in addr;
	incoming_socket_struct *insock;

	if(port < 0)
	    return(-2);

	if(total_incoming_sockets < 0)
	    total_incoming_sockets = 0;

	/* Check if there is already an incoming socket listening to
	 * the specified port number
	 */
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    insock = incoming_socket[i];
	    if(insock == NULL)
		continue;

	    if(insock->port_num == port)
		return(i);
	}

	/* Look for available pointer */
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    if(incoming_socket[i] == NULL)
		break;
	}
	if(i < total_incoming_sockets)
	{
	    /* Found available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_incoming_sockets;
	    total_incoming_sockets = n + 1;
	    incoming_socket = (incoming_socket_struct **)realloc(
		incoming_socket,
		total_incoming_sockets * sizeof(incoming_socket_struct *)
	    );
	    if(incoming_socket == NULL)
	    {
		total_incoming_sockets = 0;
		return(-3);
	    }
	}

	/* Allocate a new incoming socket */
	incoming_socket[n] = insock = (incoming_socket_struct *)calloc(
	    1, sizeof(incoming_socket_struct)
	);
	if(insock == NULL)   
	    return(-3);

	/* Set up values */
	insock->type = type;
	insock->port_num = port;
	insock->start_time = cur_systime;

	/* Bind the listening port */
	insock->socket = -1;

	for(i = 0; i < BIND_RETRIES; i++)
	{
	    /* Create a new socket as needed */
	    if(insock->socket < 0)
	    {
		insock->socket = socket(AF_INET, SOCK_STREAM, 0);
		if(insock->socket < 0)
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("unknown error");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"Unable to create incoming socket: %s%s\n",
			s,
			(i < (BIND_RETRIES - 1)) ? ", retrying..." : ""
		    );
		    free(s);
		    sleep(BIND_RETRY_INTERVAL);
		    continue;
		}
	    } 
 
	    /* Reset address structure */
	    addr.sin_family = AF_INET;
	    addr.sin_port = htons(insock->port_num);
	    addr.sin_addr.s_addr = INADDR_ANY;
	    memset(&addr.sin_zero, 0x00, 8);

	    /* Bind the socket */
	    status = bind(  
		insock->socket,
		(struct sockaddr *)&addr,
		sizeof(struct sockaddr) 
	    );
	    if(status == -1)
	    {
		char *s = STRDUP(strerror(errno));
		if(s == NULL)
		    s = strdup("unknown error");
		*s = toupper(*s);
		fprintf(
		    stderr,
"Unable to bind socket to port %i: %s%s\n",
		    insock->port_num,
		    s,
		    (i < (BIND_RETRIES - 1)) ? ", retrying..." : ""
		);
		free(s);
		sleep(BIND_RETRY_INTERVAL);
		continue;
	    }
		
	    /* Set up listening and backlogged connection limit */
	    if(listen(insock->socket, SWS_LISTEN_SOCKET_BACKLOG) == -1)
	    {
		char *s = STRDUP(strerror(errno));
		if(s == NULL)
		    s = strdup("unable to listen() socket");
		*s = toupper(*s);
		fprintf(
		    stderr,
"Unable to listen to incoming socket on port %i: %s%s\n",
		    insock->port_num,
		    s,
		    (i < (BIND_RETRIES - 1)) ? ", retrying..." : ""
		);
		free(s);

		shutdown(insock->socket, SHUT_RDWR);
		while(close(insock->socket) != 0)
		{
		    if(errno == EAGAIN)
			continue;
		    else
			break;
		}
		insock->socket = -1;

		sleep(BIND_RETRY_INTERVAL);
		continue;
	    }

	    /* Set listening port socket to nonblocking */
	    fcntl(insock->socket, F_SETFL, O_NONBLOCK);

	    break;
	}
	/* Too many retries? */
	if(i >= BIND_RETRIES) 
	{
	    fprintf(
		stderr,
"Failed to establish an incoming socket on port %i.\n",
 		port
	    );

	    /* Delete this incoming socket */
	    IncomingSocketDelete(n);
	    IncomingSocketReclaim();

	    return(-1);
	}

	return(n);
}

/*
 *	Closes and deletes the incoming socket.
 */
void IncomingSocketDelete(int n)
{ 
	incoming_socket_struct *insock;

	if(!IncomingSocketIsAllocated(n))
	    return;

	insock = incoming_socket[n];

	/* Close the socket as needed */
	if(insock->socket > -1)
	{
	    shutdown(insock->socket, SHUT_RDWR);
            while(close(insock->socket) != 0)
            {
                if(errno == EAGAIN)
		{
                    continue;
		}
                else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s == NULL)
			s = strdup("unknown error");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"Unable to close incoming socket %i (socket = %i): %s.\n",
			n, insock->socket, s
		    );
		    free(s);
                    break;
		}
            }
	    insock->socket = -1;
	}

	/* Delete this incoming socket */
	free(insock);
	incoming_socket[n] = insock = NULL;
}

/*
 *	Closes and deletes all incomming sockets.
 */
void IncomingSocketDeleteAll(void)
{
	int i;

	for(i = 0; i < total_incoming_sockets; i++)
	    IncomingSocketDelete(i);
	free(incoming_socket);
	incoming_socket = NULL;
	total_incoming_sockets = 0;
}


/*
 *	Reclaims incoming socket memory.
 */
void IncomingSocketReclaim(void)
{
	int i, n;
	incoming_socket_struct *insock;

	n = -1;
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    insock = incoming_socket[i];
	    if(insock == NULL)
		continue;

	    n = i;
	}

	total_incoming_sockets = n + 1;
	if(total_incoming_sockets > 0)
	{
	    incoming_socket = (incoming_socket_struct **)realloc(
		incoming_socket,
		total_incoming_sockets * sizeof(incoming_socket_struct *)
	    );
	    if(incoming_socket == NULL)
	    {
		total_incoming_sockets = 0;
		return;
	    }
	}
	else
	{
	    free(incoming_socket);
	    incoming_socket = NULL;
	    total_incoming_sockets = 0;
	}
}
