#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#if defined(__WIN32__)
# include <windows.h>
#else
# include <unistd.h>
#endif

#ifndef _USE_BSD
# define _USE_BSD
#endif
#include <sys/resource.h>
#include <sys/wait.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/prochandle.h"
#include "../include/df.h"   
#include "../include/netio.h"
#include "../include/unvmain.h"
#include "../include/unvmatch.h"

#ifdef PLUGIN_SUPPORT
# include "plugins.h"
#endif
#include "siteban.h"
#include "cfgfile.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static void SWServDoDebug(void);
static void SWServDoHelp(void);
static void SWServDoVersion(void);
void SWServGetOSStats(void);
void SWServGetMemoryStats(sws_memory_stats_struct *buf);
static int SWServCheckBadSockets(void);
static void SWServSignalCB(int s);
static void SWServReset(const int level);

void SWServWritePIDFile(void);
int SWServSaveUnv(void);
static void SWServExportStats(void);
void SWServResetTimmers(const time_t cur_ms, const time_t cur_sec);

void SWServShutdownStart(const time_t t);

int SWServInit(int argc, char *argv[]);
void SWServMainIteration(void);
void SWServMain(void);
void SWServMainPop(void);
void SWServMainPopAll(void);
void SWServShutdown(const char save_unv);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/* Signal makeup definations for some platforms */
#ifndef SIGPOLL
# define SIGPOLL	SIGUSR1
#endif
#ifndef SIGPWR
# define SIGPWR		SIGUSR2
#endif


int		runlevel;
int		master_reset;

time_t		cur_millitime;
time_t		cur_systime;
time_t		lapsed_millitime;
double		time_compensation;

swserv_debug_struct	debug;
swserv_next_struct	next;
swserv_sysparm_struct	sysparm;
swserv_dname_struct	dname;
swserv_fname_struct	fname;
swserv_os_stat_struct	os_stat;

sws_connection_struct	**connection;
int			total_connections;
int			highest_connection;

sws_aux_connection_struct	**aux_connection;
int			total_aux_connections;

incoming_socket_struct	**incoming_socket;
int			total_incoming_sockets;

unv_head_struct	unv_head;
sw_units_struct	sw_units;
sector_legend_struct	sector_legend;

xsw_object_struct	**xsw_object;
int			total_objects;

xsw_object_struct	**recycled_xsw_object;
int			total_recycled_objects;

sws_ocs_struct	**ocs;
int		total_ocss;

xsw_object_struct	**opm;
int			total_opms;

sws_std_opm_struct	sws_std_opm;


/*
 *	Segfault counter.
 */
static int	segfault_count;


/*
 *	Procedure to print debug information.
 */
static void SWServDoDebug(void)
{
	/* Heading */
	fprintf(
	    stderr,
"%s %s debug report:\n",
	    PROG_NAME, PROG_VERSION
	);

	fprintf(
	    stderr,
"runlevel: %i  total_connections (allocated): %i  highest_connection: %i\n",
	    runlevel, total_connections, highest_connection
	);

	fprintf(
	    stderr,
"total_objects (allocated): %i\n",
	    total_objects
	);

	/* Footer */
	fprintf(stderr, "%s", SWS_DEBUG_FOOTER);
}


/*
 *	Prints the help message to stdout.
 */
static void SWServDoHelp(void)
{
	fprintf(
	    stdout,
	    "%s",
	    PROG_USAGE_MESSAGE
	);
}


/*
 *	Prints the version information to stdout.
 */
static void SWServDoVersion(void)
{
	fprintf(
	    stdout,
	    "%s",
	    PROG_NAME " Version " PROG_VERSION "\n"
	    PROG_COPYRIGHT "\n"
	);
}


/*
 *	Procedure to update OS stats.
 */
void SWServGetOSStats(void)
{
#ifdef __WIN32__
	char CurrDir[257];
	int SectorsPerCluster, BytesPerSector, FreeClusters, TotalClusters;

	if(GetCurrentDirectory(sizeof(CurrDir),CurrDir))
	{
	    CurrDir[3] = '\0'; /* only need the drive */
	    if(GetDiskFreeSpace(CurrDir,
				(PDWORD)&SectorsPerCluster,
				(PDWORD)&BytesPerSector,
				(PDWORD)&FreeClusters,
				(PDWORD)&TotalClusters
	       )
	    )
	    {
		os_stat.disk_used = (BytesPerSector *
		    SectorsPerCluster) * FreeClusters;

		os_stat.disk_total = (BytesPerSector *
		    SectorsPerCluster) * TotalClusters;

		return;
	    }
	}

	os_stat.disk_used = 0;
	os_stat.disk_total = 0;

#else
	int i;
	int total_df_stats;
	df_stat_struct **df_stat;


	/* Get disk space stats */
	df_stat = DiskFreeGetListing(&total_df_stats);
	if((df_stat != NULL) && (total_df_stats > 0))
	{
	    /*   Look for device stat index for the home
	     *   directory of this program.  Must do search in
	     *   reverse order to catch higher mounted paths first.
	     */
	    for(i = total_df_stats - 1; i >= 0; i--)
	    {
		if(df_stat[i] == NULL)
		    continue;

		if(strpfx(
		    dname.toplevel,
		    df_stat[i]->mounted_on
		))
		    break;
	    }
	    if(i < total_df_stats)
	    {
		os_stat.disk_used = df_stat[i]->used;
		os_stat.disk_total = df_stat[i]->total;
	    }
	    else
	    {
		os_stat.disk_used = 0;
		os_stat.disk_total = 0;
	    }
	}

	/* Free listing */
	DiskFreeDeleteListing(df_stat, total_df_stats);


	return;
#endif /* __WIN32__ */
}


/*
 *	Procedure to get statistics of all memory used by the
 *	program.
 */
void SWServGetMemoryStats(sws_memory_stats_struct *buf)
{
	int i, n;
	xsw_object_struct *obj;
	xsw_weapon_struct *wep;
	xsw_ecodata_struct *eco;
#ifdef PLUGIN_SUPPORT
	plugin_struct *pi;
#endif	/* PLUGIN_SUPPORT */

	if(buf == NULL)
	    return;

	/* Reset total */
	buf->total = 0;

	/* AUX pipes */
	buf->aux = (total_aux_connections * sizeof(sws_aux_connection_struct *)) +
	    (total_aux_connections * sizeof(sws_aux_connection_struct));
	buf->total += buf->aux;

	/* Connections */
	buf->con = (total_connections * sizeof(sws_connection_struct *)) +
	    (total_connections * sizeof(sws_connection_struct));
	buf->total += buf->con;

	/* Objects */
	buf->obj = (total_objects * sizeof(xsw_object_struct *)) +
	    (total_objects * sizeof(xsw_object_struct));

	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Password */
	    if(obj->password != NULL)
		buf->obj += (strlen(obj->password) + 1) * sizeof(char);

	    /* Description */
	    if(obj->description != NULL)
		buf->obj += (strlen(obj->description) + 1) * sizeof(char);

	    /* Elink */
	    if(obj->elink != NULL)
		buf->obj += (strlen(obj->elink) + 1) * sizeof(char);

	    /* Tractored objects */
	    buf->obj += obj->total_tractored_objects * sizeof(int);

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		buf->obj += sizeof(xsw_weapon_struct);
	    }
	    buf->obj += obj->total_weapons *
		sizeof(xsw_weapon_struct *);

	    /* Score */
	    if(obj->score != NULL)
		buf->obj += sizeof(xsw_score_struct);

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		buf->obj += sizeof(xsw_ecodata_struct);

		for(n = 0; n < eco->total_products; n++)
		{
		    if(eco->product[n] == NULL)
			continue;

		    buf->obj += sizeof(xsw_ecoproduct_struct);
		}
		buf->obj += eco->total_products *
		    sizeof(xsw_ecoproduct_struct *);
	    }
	}

	buf->total += buf->obj;


	/* Recycle buffers (struct same as xsw_object_struct) */
	buf->rec_obj = (total_recycled_objects * sizeof(xsw_object_struct *))
	    + (total_recycled_objects * sizeof(xsw_object_struct));

	for(i = 0; i < total_recycled_objects; i++)
	{
	    obj = recycled_xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Password */
	    if(obj->password != NULL)
		buf->rec_obj += (strlen(obj->password) + 1) * sizeof(char);

	    /* Description */
	    if(obj->description != NULL)
		buf->rec_obj += (strlen(obj->description) + 1) * sizeof(char);

	    /* Elink */
	    if(obj->elink != NULL)
		buf->rec_obj += (strlen(obj->elink) + 1) * sizeof(char);

	    /* Tractored objects */
	    buf->rec_obj += obj->total_tractored_objects * sizeof(int);

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;
	
		buf->rec_obj += sizeof(xsw_weapon_struct);
	    }
	    buf->rec_obj += obj->total_weapons *
		sizeof(xsw_weapon_struct *);

	    /* Score */
	    if(obj->score != NULL)
		buf->rec_obj += sizeof(xsw_score_struct);

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		buf->rec_obj += sizeof(xsw_ecodata_struct);

		for(n = 0; n < eco->total_products; n++) 
		{
		    if(eco->product[n] == NULL)
			continue;

		    buf->rec_obj += sizeof(xsw_ecoproduct_struct);
		}
		buf->rec_obj += eco->total_products *
		    sizeof(xsw_ecoproduct_struct *);
	    }
	}
	buf->total += buf->rec_obj;


	/* Object parameter models */
	buf->opm = (total_opms * sizeof(xsw_object_struct *)) +
	    (total_opms * sizeof(xsw_object_struct));

	for(i = 0; i < total_opms; i++)
	{
	    obj = opm[i];

	    /* Password */
	    if(obj->password != NULL)
		buf->opm += (strlen(obj->password) + 1) * sizeof(char);

	    /* Description */
	    if(obj->description != NULL)
		buf->opm += (strlen(obj->description) + 1) * sizeof(char);

	    /* Elink */
	    if(obj->elink != NULL)
		buf->opm += (strlen(obj->elink) + 1) * sizeof(char);

	    /* Tractored objects */
	    buf->opm += obj->total_tractored_objects * sizeof(int);

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		buf->opm += sizeof(xsw_weapon_struct);
	    }
	    buf->opm += obj->total_weapons *
		sizeof(xsw_weapon_struct *);

	    /* Score */
	    if(obj->score != NULL)
		buf->opm += sizeof(xsw_score_struct);

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		buf->opm += sizeof(xsw_ecodata_struct);

		for(n = 0; n < eco->total_products; n++)
		{
		    if(eco->product[n] == NULL)
			continue;

		    buf->opm += sizeof(xsw_ecoproduct_struct);
		}
		buf->opm += eco->total_products *
		    sizeof(xsw_ecoproduct_struct *);
	    }
	}
	buf->total += buf->opm;


	/* Object create scripts */
	buf->ocs = (total_ocss * sizeof(sws_ocs_struct *)) +
	    (total_ocss * sizeof(sws_ocs_struct));

	buf->total += buf->ocs;


	/* Plugins */
#ifdef PLUGIN_SUPPORT
	buf->plugins = (total_plugins * sizeof(SWServContext *)) +
	    (total_plugins * sizeof(SWServContext));

	for(i = 0; i < total_plugins; i++)
	{
	    pi = plugin[i];
	    if(pi == NULL)
		continue;

	    if(pi->name != NULL)
		buf->plugins += (strlen(pi->name) + 1) * sizeof(char);

	    for(n = 0; n < pi->argc; n++)
	    {
		if(pi->argv[n] != NULL)
		    buf->plugins += (strlen(pi->argv[n]) + 1) * sizeof(char);
	    }
	    buf->plugins += pi->argc * sizeof(char *);
	}
#else	/* PLUGIN_SUPPORT */
	buf->plugins = 0;
#endif	/* PLUGIN_SUPPORT */

	buf->total += buf->plugins;
}


/*
 *	Checks all connections for any bad or dead sockets and closes
 *	them as needed.
 *
 *	Returns the number of connections closed.
 */
static int SWServCheckBadSockets(void)
{
	int i, bad_sockets = 0;
	sws_connection_struct *con;

	/* Iterate through all connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

#if 0
/* This this is not the right way to check for bad connections,
 * skip it for now
 */
	    if(!NetIsSocketWritable(con->socket))
	    {
		close(con->socket);
		con->socket = -1;

		NetCloseConnection(i);
	    }
#endif
	}



	return(bad_sockets);
}


/*
 *	Signal callback.
 */
static void SWServSignalCB(int s)
{
	int status;

	switch(s)
	{
	  /* Reload configuration */
	  case SIGHUP:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"Received signal: SIGHUP: Reloading configuration...\n"
		);
	    if(sysparm.log_general)
		LogAppend(
		    fname.log,
"*** Received signal: SIGHUP: Reloading configuration..."
		);
	    /* Set the master reset to 2 so that the configuration will
	     * be reloaded (the universe will not be reloaded)
	     */
	    master_reset = 2;
	    break;

	  case SIGINT:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"Received signal: SIGINT: Server shutting down...\n"
		);
	    if(sysparm.log_general || sysparm.log_net || sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGINT: Server shutting down..."
		);
	    SWServMainPopAll();
	    break;

	  case SIGTERM:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"Received signal: SIGTERM: Server shutting down...\n"
		);
	    if(sysparm.log_general || sysparm.log_net || sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGTERM: Server shutting down..."
		);
	    SWServMainPopAll();
	    break;

	  /* Broken pipe */
	  case SIGPIPE:
	    /* Watch for SIGPIPE again */
	    signal(SIGPIPE, SWServSignalCB);
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"Received signal: SIGPIPE: Broken pipe.\n"
		);
	    if(sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGPIPE: Broken pipe."
		);
	    /* Check if any sockets died */
	    SWServCheckBadSockets();
	    break;

	  case SIGSEGV:
	    /* Increment segfault counter */
	    segfault_count++;
	    if(segfault_count >= 2)
		exit(-1);

	    /* Print debug */
	    SWServDoDebug();

	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"Received signal: SIGSEGV: Server performing emergency save and shutting down...\n"
		);
	    if(sysparm.log_general || sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGSEGV: Server performing emergency save and shutting down..."
		);
	    status = DBEmergencySave(xsw_object, total_objects);
	    if(!sysparm.console_quiet)
	    {
		if(status)
		    printf(
"Emergency universe save failed, error code %i.\n",
			status
		    );
		else
		    printf(
"Emergency universe save complete and successful.\n"
		    );
	    }
	    if(sysparm.log_general || sysparm.log_errors)
	    {
		if(status)
		{
		    char *buf = (char *)malloc(40 + 128);
		    sprintf(
			buf,
"Server: Emergency universe save failed: Error code %i.",
			status
		    );
		    LogAppend(fname.log, buf);
		    free(buf);
		}
		else
		{
		    LogAppend(
			fname.log,
"Server: Emergency universe save complete and successful."
		    );
		}
	    }
	    SWServMainPopAll();
	    break;

	  case SIGCHLD:
	    /* Catch child and watch for signal again */
	    while(wait3(&status, WNOHANG, (struct rusage *)0) > 0);
	    signal(SIGCHLD, SWServSignalCB);
	    break;

	  default:
	    /* Log the receiving of an unsupported signal */
	    if(!sysparm.console_quiet)
		printf(
"Received signal: %i: Unsupported signal.\n",
		    s
		);
	    if(sysparm.log_errors)
	    {
		char *mesg = (char *)malloc(
		    40 + 80
		);
		sprintf(
		    mesg,
"*** Received unsupported signal %i.",
		    s
		);
		LogAppend(fname.log, mesg);
		free(mesg);
	    }
	    /* Perform an emergency save just in case */
	    DBEmergencySave(xsw_object, total_objects);
	    break;
	}
}


/*
 *	Performs a global reset based on the specified reset level.
 */
static void SWServReset(const int level)
{
	char log_general;
	char *log_file;

	/* Nothing to reset? */
	if(level <= 0)
	    return;

	log_general = sysparm.log_general;
	log_file = strdup(fname.log);

	/* Log reset */
	if(log_general)
	{
	    char text[256];
	    sprintf(
		text,
"Server: Level %i reset in progress...",
		level
	    );
	    LogAppend(log_file, text);
	}

	/* Level 3 or greater */
	if(level >= 3)
	{
	    /* Reload universe database from universe on file */
	    int i;

	    /* Notify connections that they need to relogin */
	    NetSendLiveMessage(
		-1,
"*** Universe reloading from original, you will need to login again..."
	    );

	    /* Disconnect all connections */
	    for(i = 0; i < total_connections; i++)
		NetCloseConnection(i);

	    /* Delete old universe and load new universe */
	    DBOpenFromFile(fname.unv_in);

	    /* Delete all connections */
	    ConDeleteAll();
	}

	/* Level 2 or greater */
	if(level >= 2)
	{
	    /* Reload configurations (except universe database) */

	    NetSendLiveMessage(
		-1,
"Server: Memory reload in progress..."
	    );

	    /* Configuration */
	    CfgFileOpen(fname.cfg);

	    /* Object Parameter Models */
	    OPMFileOpen(fname.opm);

	    /* Object Create Scripts */
	    OCSFileOpen(fname.ocs);

	    NetSendLiveMessage(
		-1,
"Server: Memory reload complete."
	    );
	}

	/* Exactly level 1 */
	if(level == 1)
	{
	    /* Refresh memory, note that there is no reason to refresh
	     * memory if the level was higher than 1 since the memory
	     * would have been just loaded.
	     */

	    NetSendLiveMessage(
		-1,
"Server: Memory refresh in progress..."
	    );

	    /* Universe database */
	    DBReclaim();

	    /* Object parameter models */
	    OPMReclaim();

	    /* Object create scripts */
	    OCSReclaim();

	    /* Connections */
	    ConReclaim();

	    NetSendLiveMessage(
		-1,
"Server: Memory refresh complete."
	    );
	}

	/* Log reset */
	if(log_general)
	{
	    char text[80];
	    sprintf(
		text,
"Server: Reset complete."
	    );
	    LogAppend(log_file, text);
	}

	free(log_file);
}

/*
 *	Writes the PID file.
 *
 *	If the PID file exists then it will be overwritten.
 */
void SWServWritePIDFile(void)
{
	const int p = getpid();
	const char *path = fname.pid;
	FILE *fp;

	if(STRISEMPTY(path))
	    return;

	fp = fopen(path, "wb");
	if(fp == NULL)
	    return;

	fprintf(fp, "%i\n", p);
	fclose(fp);
}


/*
 *	Saves the universe to file.
 *
 *	Returns non-zero on error.
 */
int SWServSaveUnv(void)
{
	int status;
	char sndbuf[CS_DATA_MAX_LEN];

	if(!sysparm.console_quiet)
	    printf(
"Starting universe save...\n"
	    );

	/* Notify all connections about save */
	sprintf(
	    sndbuf,
"Server: Universe save in progress..."
	);
	NetSendLiveMessage(-1, sndbuf);

	/* Log save */
	if(sysparm.log_general)
	{
	    char text[80];
	    sprintf(
		text,
"Server: Starting universe save..."
	    );
	    LogAppend(fname.log, text);
	}

	/* Save universe */
	status = DBSaveToFile(fname.unv_out);
	/* Check for errors and print finish message */
	if(status)
	    sprintf(
		sndbuf,
"Server: Universe save failed: Error code %i: Attempting emergency universe save...",
		status
	    );
	else
	    sprintf(
		sndbuf,
"Server: Universe save complete and successful."
	    );
	NetSendLiveMessage(-1, sndbuf);

	if(sysparm.log_general)
	    LogAppend(fname.log, sndbuf);

	if(!sysparm.console_quiet)
	{
	    if(status)
		printf(
"Universe save failed, attempting emergency universe save...\n"
		);
	    else
		printf(
"Universe save complete and successful.\n"
		);
	}

	/* Attempt emergancy save incase of error (ignore emergency
	 * save status result)
	 */
	if(status)
	{
	    int status2 = DBEmergencySave(xsw_object, total_objects);
	    if(status2)
	    {
		if(sysparm.log_general || sysparm.log_errors)
		{
		    sprintf(
			sndbuf,
"Server: Emergency universe save failed: Error code %i.",
			status2
		    );
		    LogAppend(fname.log, sndbuf);
		}
		printf(
"Emergency universe save failed, error code %i.\n",
		    status2
		);
	    }
	    else
	    {
		if(sysparm.log_general || sysparm.log_errors) 
		    LogAppend(
			fname.log,
"Server: Emergency universe save complete and successful."
		    );
		printf(
"Emergency universe save complete and successful.\n"
		);
	    }
	}

	return(status);
}

/*
 *	Exports statistics.
 */
static void SWServExportStats(void)
{
	ExportScores(fname.scores_export);
	ExportConList(fname.conlist_export);
}

/*
 *	Resets all timmers.
 */
void SWServResetTimmers(const time_t cur_ms, const time_t cur_sec)
{
	const char shutdown_schedualed = (next.shutdown > 0l) ? 1 : 0;
	int i, n;
	sws_aux_connection_struct *aux_con;
	sws_connection_struct *con;
	xsw_object_struct *obj;

	/* Next instances */
	memset(&next, 0x00, sizeof(swserv_next_struct));
	next.system_check = cur_ms;
	next.object_values = cur_ms;
	next.memory_clean = cur_ms;
	next.new_connection_poll = cur_ms;
	next.os_stats = cur_sec;
	next.unv_save = cur_sec;
	next.stats_export = cur_sec;
	next.shutdown = shutdown_schedualed ? cur_sec : 0l;
	next.need_weapon_values = 0;

	/* Connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    con->obj_ud_next = cur_ms;
	}

	/* AUX Connections */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)
		continue;

	    aux_con->next = cur_ms;
	}

	/* Objects */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    obj->last_updated = cur_ms;
	    obj->birth_time = cur_ms;
	    obj->animation.last_interval = cur_ms;

	    for(n = 0; n < obj->total_weapons; n++)
		obj->weapons[n]->last_used = cur_ms;
	}
}


/*
 *	Starts the SWServ shutdown process.
 *
 *	The t specifies the delay (in seconds) before SWServ shuts down,
 *	to shutdown as soon as possible set t to 0.
 *
 *	All connections will be notified about the shutdown process.
 *
 *	The global next.shutdown will be set to when the shutdown
 *	should occure and SWServMainIteration() will check for that.
 */
void SWServShutdownStart(const time_t t)
{
	char text[128];

	if(!sysparm.console_quiet)
	{
	    if(t > 0l)
	        fprintf(
		    stdout,
"%s shutting down in %ld second%s...\n",
		    PROG_NAME,
		    t,
		    (t == 1) ? "" : "s"
		);
	    else
		fprintf(
		    stdout,
"%s shutting down...\n",
		    PROG_NAME
		);
	}


	/* Notify all connections that the server is shutting down */
	if(t > 0l)
	    sprintf(
		text,
"*** Server shutting down in %ld second%s! ***",
		t,
		(t == 1) ? "" : "s"
	    );
	else
	    sprintf(
		text,
"*** Server shutting down! ***"
	    );
	NetSendLiveMessage(-1, text);

	/* Schedual shutdown */
	next.shutdown = cur_systime + t;
}


/*
 *	Initializes SWServ.
 */
int SWServInit(int argc, char *argv[])
{
	char run_in_foreground = 0;
	int i, n, status;
	const char *s, *arg;
	char cwd[PATH_MAX];
	struct stat stat_buf;


	/* Check if command line arguments specify to run in
	 * SWServ in the foreground
	 */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    if(!strcasecmp(arg, "--foreground") ||
	       !strcasecmp(arg, "--fg") ||
	       !strcasecmp(arg, "-foreground") ||
	       !strcasecmp(arg, "-fg")
	    )
	    {
		run_in_foreground = 1;
		break;
	    }
	}
	/* Do not run in foreground? */
	if(!run_in_foreground)
	{
	    /* Run in the background */
	    switch(fork())
	    {
	      case -1:
		perror("fork");
		return(-3);
		break;
	      case 0:	/* Child */
		break;
	      default:	/* Parent */
		exit(0);
		break;
	    }
	}


	/* Initialize time zone */
	tzset();

	/* Reset globals */
	master_reset = 0;

	cur_millitime = MilliTime();
	cur_systime = time(NULL);

	debug.level = DEBUG_LEVEL_NONE;
	debug.val = 0;

	lapsed_millitime = 0l;
	time_compensation = 1.0;

	/* Reset sysparm */
	sysparm.allow_guest = 1;
	sysparm.allow_guest_signup = 1;
	sysparm.cease_fire = 0;
	sysparm.cloak_am_rate = SWS_DEF_CLOAK_AM_RATE;	/* In size_ru * antimatter / cycle */
	sysparm.com_range = SW_COM_SHORT_RANGE_MAX;	/* In real units */
	sysparm.con_notify = 1;
	sysparm.console_quiet = 0;
	sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGHEST;
	sysparm.dmg_ctl_rate = SWS_DEF_DMG_CTL_RATE;
	strncpy(
	    sysparm.guest_login_name, SWS_DEF_GUEST_NAME,
	    sizeof(sysparm.guest_login_name)
	);
	sysparm.guest_login_name[
	    sizeof(sysparm.guest_login_name) - 1] = '\0';
	sysparm.hide_players_home = 1;
	sysparm.hit_player_bonus = SWS_DEF_HIT_PLAYER_BONUS;
	sysparm.homes_destroyable = 0;
	sysparm.int_system_check = DEF_INT_SYSTEM_CHECK;
	sysparm.int_new_connection_poll = DEF_INT_NEW_CONNECTION_POLL;
	sysparm.int_object_values = DEF_INT_OBJECT_VALUES;
	sysparm.int_weapon_values = DEF_INT_WEAPON_VALUES;
	sysparm.int_memory_clean = DEF_INT_MEMORY_CLEAN;
	sysparm.int_aux_con_stats = DEF_INT_AUX_CON_STATS;
	sysparm.int_os_stats = DEF_INT_OS_STATS;	/* In seconds */
	sysparm.int_stats_export = DEF_INT_STATS_EXPORT;	/* In seconds */
	sysparm.int_unv_save = DEF_INT_UNV_SAVE;	/* In seconds */
	sysparm.killer_gets_credits = 1;
	sysparm.log_general = 1;
	sysparm.log_events = 1;
	sysparm.log_net = 1;
	sysparm.log_errors = 1;
	sysparm.login_timeout = SWS_DEF_LOGIN_TIMEOUT;
	sysparm.max_aux_connections = SWS_DEF_MAX_AUX_STATS_CONNECTIONS;
	sysparm.max_connections = SWS_DEF_MAX_CONNECTIONS;
	sysparm.max_failed_logins = SWS_DEF_MAX_FAILED_LOGINS;
	sysparm.max_guests = SWS_DEF_MAX_GUESTS;
	strncpy(
	    sysparm.mesg_wrong_login,
	    SWS_DEF_WRONG_LOGIN_MESG,
	    sizeof(sysparm.mesg_wrong_login)
	);
	sysparm.mesg_wrong_login[
	    sizeof(sysparm.mesg_wrong_login) - 1] = '\0';
	strncpy(
	    sysparm.mesg_no_guests,
	    SWS_DEF_NO_GUESTS_MESG,
	    sizeof(sysparm.mesg_no_guests)
	);
	sysparm.mesg_no_guests[
	    sizeof(sysparm.mesg_no_guests) - 1] = '\0';
	strncpy(
	    sysparm.mesg_welcome,
	    SWS_DEF_WELCOME_MESG,  
	    sizeof(sysparm.mesg_welcome)
	);
	sysparm.mesg_welcome[
	    sizeof(sysparm.mesg_welcome) - 1] = '\0';
	strncpy(
	    sysparm.mesg_leave,
	    SWS_DEF_LEAVE_MESG,
	    sizeof(sysparm.mesg_leave)
	);
	sysparm.mesg_leave[
	    sizeof(sysparm.mesg_leave) - 1] = '\0';
	sysparm.report_destroyed_weapons = 0;
	sysparm.respawn_players = 1;
	sysparm.send_starchart = 1;
	sysparm.single_connection = 1;
	sysparm.tractor_beam_range = SW_TRACTOR_BEAM_MAX_RANGE;	/* In real units */
	sysparm.tractor_beam_am_rate = SWS_DEF_TRACTOR_BEAM_AM_RATE;	/* In size_ru * antimatter / cycle */
	sysparm.transporter_range = SWS_DEF_TRANSPORTER_RANGE;
	sysparm.use_server_ai = 1;


	/* Reset units */
	sw_units.ru_to_au = DEF_UNITCONV_RU_TO_AU;

	/* Reset sector legend */
	sector_legend.x_max = SW_SECTOR_SIZE_X_MAX;
	sector_legend.x_min = SW_SECTOR_SIZE_X_MIN;
	sector_legend.y_max = SW_SECTOR_SIZE_Y_MAX;
	sector_legend.y_min = SW_SECTOR_SIZE_Y_MIN;
	sector_legend.z_max = SW_SECTOR_SIZE_Z_MAX;
	sector_legend.z_min = SW_SECTOR_SIZE_Z_MIN;
	sector_legend.x_len = sector_legend.x_max - sector_legend.x_min;
	sector_legend.y_len = sector_legend.y_max - sector_legend.y_min;
	sector_legend.z_len = sector_legend.z_max - sector_legend.z_min;

	/* Reset next scheduals */
	next.system_check = cur_millitime;
	next.object_values = cur_millitime;
	next.memory_clean = cur_millitime;
	next.new_connection_poll = cur_millitime;
	next.os_stats = cur_systime;
	next.unv_save = cur_systime + sysparm.int_unv_save;
	next.stats_export = cur_systime;
	next.shutdown = 0l;			/* Do not shutdown */
	next.need_weapon_values = 0;

	/* Reset objects database */
	xsw_object = NULL;
	total_objects = 0;

	/* Reset recycled objects */
	recycled_xsw_object = NULL;
	total_recycled_objects = 0;

	/* Reset object create scripts */
	ocs = NULL;
	total_ocss = 0;

	/* Reset object parameter models */
	opm = NULL;
	total_opms = 0;

	/* Reset standard opms indices */
	sws_std_opm.explosion_small_num = -1;
	sws_std_opm.explosion_large_num = -1;
	sws_std_opm.debris_num = -1;
	sws_std_opm.guest_num = -1;
	sws_std_opm.new_player_num = -1;

	/* Reset connections */
	connection = NULL;
	total_connections = 0;
	highest_connection = 0;

	/* Reset incoming sockets */
	incoming_socket = NULL;
	total_incoming_sockets = 0;

	/* Reset sitebans */
	siteban = NULL;
	total_sitebans = 0;

	/* Reset directories */
	strncpy(
	    dname.toplevel, SWS_TOPLEVEL_DIR,
	    sizeof(dname.toplevel)
	);
	strncpy(
	    dname.bin, SWS_BIN_DIR,
	    sizeof(dname.bin)
	);
	strncpy(
	    dname.etc, SWS_ETC_DIR,
	    sizeof(dname.etc)
	);
	strncpy(
	    dname.db, SWS_DB_DIR,
	    sizeof(dname.db)
	);
	strncpy(
	    dname.logs, SWS_LOGS_DIR,
	    sizeof(dname.logs)
	);
	strncpy(
	    dname.plugins, SWS_PLUGINS_DIR,
	    sizeof(dname.plugins)
	);
	strncpy(
	    dname.public_html, SWS_PUBLIC_HTML_DIR,
	    sizeof(dname.public_html)
	);
	strncpy(
	    dname.tmp, SWS_TMP_DIR,
	    sizeof(dname.tmp)
	);

	strncpy(
	    fname.monitor, "monitor",
	    sizeof(fname.monitor)
	);
	fname.monitor_set = 0;
	strncpy(
	    fname.unv_in, SWS_DEF_UNV_IN_FILE,
	    sizeof(fname.unv_in)
	);
	fname.unv_in_set = 0;
	strncpy(
	    fname.unv_out, SWS_DEF_UNV_OUT_FILE,
	    sizeof(fname.unv_out)
	);
	fname.unv_out_set = 0;
	strncpy(
	    fname.opm, SWS_DEF_OPM_FILE,
	    sizeof(fname.opm)
	);
	fname.opm_set = 0;
	strncpy(
	    fname.ocs, SWS_DEF_OCS_FILE,
	    sizeof(fname.ocs)
	);
	fname.ocs_set = 0;

	s = PrefixPaths(dname.logs, SWS_DEF_LOG_FILE);
	strncpy(
	    fname.log,
	    (s != NULL) ? s : SWS_DEF_LOG_FILE,
	    sizeof(fname.log)
	);

	s = PrefixPaths(dname.tmp, SWS_DEF_PID_FILE);
	strncpy(
	    fname.pid,
	    (s != NULL) ? s : SWS_DEF_PID_FILE,
	    sizeof(fname.pid)
	);

	*fname.conlist_export = '\0';
	*fname.scores_export = '\0';
	*fname.events_export = '\0';

#ifdef __WIN32__
	strcpy(cwd, ".");
#else
	if(getcwd(cwd, sizeof(cwd)) == NULL)
	    *cwd = '\0';
#endif /* __WIN32_ */
	s = PrefixPaths(cwd, SWS_CFG_FILE);
	strncpy(
	    fname.cfg,
	    (s != NULL) ? s : SWS_CFG_FILE,
	    sizeof(fname.cfg)
	);


	total_aux_connections = 0;
	aux_connection = NULL;

	strncpy(unv_head.title, "Untitled", sizeof(unv_head.title));


	/* Initialize universe object management resources */
	UNVInit(argc, argv);


	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    if(strcasepfx(arg, "--mon") ||
	       strcasepfx(arg, "-mon")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.monitor, argv[i], sizeof(fname.monitor));
		    fname.monitor[sizeof(fname.monitor) - 1] = '\0';
		    fname.monitor_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    else if(strcasepfx(arg, "--unvin") ||
		    strcasepfx(arg, "-unvin") ||
		    strcasepfx(arg, "--dbin") ||
		    strcasepfx(arg, "-dbin")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.unv_in, argv[i], sizeof(fname.unv_in));
		    fname.unv_in[sizeof(fname.unv_in) - 1] = '\0';
		    fname.unv_in_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    else if(strcasepfx(arg, "--unvout") ||
		    strcasepfx(arg, "-unvout") ||
		    strcasepfx(arg, "--dbout") ||
		    strcasepfx(arg, "-dbout")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.unv_out, argv[i], sizeof(fname.unv_out));
		    fname.unv_out[sizeof(fname.unv_out) - 1] = '\0';
		    fname.unv_out_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Object parameter values */
	    else if(strpfx(arg, "--opm") ||
		    strpfx(arg, "-opm")
	    )
	    {
		i++;
		if(i < argc)
		{   
		    strncpy(fname.opm, argv[i], sizeof(fname.opm));
		    fname.opm[sizeof(fname.opm) - 1] = '\0';
		    fname.opm_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Object create scripts */
	    else if(strpfx(arg, "--ocs") ||
		    strpfx(arg, "-ocs")  
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.ocs, argv[i], sizeof(fname.ocs));
		    fname.ocs[sizeof(fname.ocs) - 1] = '\0';
		    fname.ocs_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Port */
	    else if(strpfx(arg, "--p") ||
		    strpfx(arg, "-p")
	    )
	    {
		i++;
		if(i < argc)
		{
		    const int port = atoi(argv[i]);
		    const int insock_num = IncomingSocketNew(
			port,
			INCOMING_SOCKET_TYPE_STANDARD
		    );
		    if(insock_num < 0)
			return(-1);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Quiet */
	    else if(strpfx(arg, "--q") ||
		    strpfx(arg, "-q")
	    )
	    {
		sysparm.console_quiet = 1;
	    }
	    /* Foreground (just checked, don't handle since it was
	     * handled above)
	     */
	    else if(!strcasecmp(arg, "--foreground") ||
		    !strcasecmp(arg, "--fg") ||
		    !strcasecmp(arg, "-foreground") ||
		    !strcasecmp(arg, "-fg")
	    )
	    {

	    }
	    /* Check if argument is prefixed with '-' or '+' */
	    else if((*arg == '-') ||
		    (*arg == '+')
	    )
	    {
		fprintf(
		    stderr,
		    "%s: Unsupported argument.\n",
		    arg
		);
	    }
	    /* Sole argument is configuration file to use */
	    else
	    {
		strncpy(fname.cfg, arg, sizeof(fname.cfg));
		fname.cfg[sizeof(fname.cfg) - 1] = '\0';
	    }
	}

/* Check for any errors in arguments */


	/* Print start up messages */
	if(!sysparm.console_quiet)
	{
	    printf(
		"%s\n",
		PROG_NAME " " PROG_VERSION
	    );
	}


	/* Open the configuration from file */
	if(!stat(fname.cfg, &stat_buf))
	{
	    status = CfgFileOpen(fname.cfg);
	}
	else
	{
#if defined(__WIN32__)
	    status = -1;
#else
	    /* Is in the etc dir? */
	    s = PrefixPaths(dname.etc, SWS_CFG_FILE);
	    strncpy(
		fname.cfg, (s != NULL) ? s : SWS_CFG_FILE,
		sizeof(fname.cfg)
	    );
	    fname.cfg[sizeof(fname.cfg) - 1] = '\0';

	    if(!stat(fname.cfg, &stat_buf))
	    {
		status = CfgFileOpen(fname.cfg);
	    }
	    else
	    {
		/* Is in the global etc directory? */
		s = PrefixPaths(ETC_DIR, SWS_CFG_FILE);
		strncpy(
		    fname.cfg,
		    (s != NULL) ? s : SWS_CFG_FILE,
		    sizeof(fname.cfg)
		);
		fname.cfg[sizeof(fname.cfg) - 1] = '\0';
		status = CfgFileOpen(fname.cfg);
	    }
#endif
	}
	/* Error loading the configuration file? */
	if(status < 0)
	{
	    fprintf(
		stderr,
"%s: An error occured while loading the configuration file.\n",
		fname.cfg
	    );
	    return(-1);
	}

	/* Log startup, this is the earliest that we can log this
	 * since we just obtained the log option and the log file from
	 * the configuration file
	 */
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"*** " PROG_NAME " " PROG_VERSION " starting..."
	    );

	/* Write the PID file after getting the configuration */
	SWServWritePIDFile();

	/* Open a standard incoming connections listening socket to
	 * the default port as needed
	 */
	for(i = 0, n = 0; i < total_incoming_sockets; i++)
	{
	    if(incoming_socket[i] == NULL)
		continue;

	    /* Is this a standard incoming socket? */
	    if(incoming_socket[i]->type == INCOMING_SOCKET_TYPE_STANDARD)
		n++;
	}
	if(n <= 0)
	{
	    /* No standard incoming listening sockets exist, so create
	     * a new one
	     */
	    const int port = SWS_DEF_PORT;
	    const int insock_num = IncomingSocketNew(
		port,
		INCOMING_SOCKET_TYPE_STANDARD
	    );
	    if(insock_num < 0)
	    {
		fprintf(
		    stderr,
"Error: Unable to establish listening socket on port %i.\n",
		    port
		);
	        return(-1);
	    }
	}


	/* ********************************************************* */
	/* Check if directories are valid */

	/* Server toplevel/home directory */
	if(!ISPATHDIR(dname.toplevel))
	{
	    fprintf(
		stderr,
		"%s: Error: Invalid directory.\n",
		dname.toplevel
	    );
	    return(-1);
	}

	/* Server etc directory */
	if(!ISPATHDIR(dname.etc))
	{
	    fprintf(
		stderr,
		"%s: Warning: Invalid directory.\n",
		dname.etc
	    );
	    /* Non-fatal */
	}

	/* Server logs directory */
	if(!ISPATHDIR(dname.logs))
	{
	    fprintf(
		stderr,
		"%s: Warning: Invalid directory.\n",
		dname.logs
	    );
	    /* Non-fatal */
	}

	/* Server bin directory */
	if(!ISPATHDIR(dname.bin))
	{
	    fprintf(
		stderr,
		"%s: Warning: Invalid directory.\n",
		dname.bin
	    );
	    /* Non-fatal */
	}

	/* Server database directory */
	if(!ISPATHDIR(dname.db))
	{
	    fprintf(
		stderr,
		"%s: Error: Invalid directory.\n",
		dname.db
	    );
	    return(-1);
	}


	/* Run the monitor as needed, this needs to be done after
	 * the rcfile has been parsed so that the AUX Stat listening
	 * socket has been created
	 */
	if(fname.monitor_set &&
	   ISPATHEXECUTABLE(fname.monitor)
	)
	{
	    /* Use first AUX Stats listening socket port number */
	    for(i = 0, n = 0; i < total_incoming_sockets; i++)
	    {
		if(incoming_socket[i] == NULL)
		    continue;

		if(incoming_socket[i]->type == INCOMING_SOCKET_TYPE_AUXSTATS)
		    n = incoming_socket[i]->port_num;
	    }

	    if(n > 0)
	    {
		char cmd[1024 + PATH_MAX + NAME_MAX];
	        sprintf(
		    cmd,
		    "%s %s %i",
		    fname.monitor,
		    "localhost",
		    n
	        );
	        Exec(cmd);
	    }
	    else
	    {
		fprintf(
		    stderr,
"Warning: Requested use of monitor but no AUX Stats port available.\n"
		);
		fprintf(
		    stderr,
"Warning: Monitor not executed.\n"
		);
	    }
	}


	/* Allocate recycled objects buffers */
	status = DBRecycleBufferInit(MAX_RECYCLED_OBJECTS);
	if(status < 0)
	{
	    fprintf(
		stderr,
"Error: Unable to allocate recycled objects buffer.\n"
	    );
	    return(-3);
	}


	/* Open the input universe */
	status = DBOpenFromFile(fname.unv_in);
	if(status)
	{
	    fprintf(
		stderr,
"%s: Error: An error occured while loading the universe file.\n",
		fname.unv_in
	    );
	    return(-1);
	}


	/* Load the Object Parameters Models */
	status = OPMFileOpen(fname.opm);
	if(status < 0)
	{
	    fprintf(
		stderr,
"%s: Warning: An error occured while loading the object parameter models file.\n",
		fname.opm
	    );
	    /* Non-fatal error */
	}

	/* Load Object Create Scripts */
	status = OCSFileOpen(fname.ocs);
	if(status < 0)
	{
	    fprintf(
		stderr,
"%s: Warning: An error occured while loading the object create scripts file.\n",
		fname.ocs
	    );
	    /* Non-fatal error */
	}


	/* Print startup messages? */
	if(!sysparm.console_quiet)
	{
	    /* Universe title */
	    fprintf(
		stdout,
		"Universe: `%s'\n",
		unv_head.title
	    );

	    /* Configuration file */
	    fprintf(
		stdout,
		"Configuration file: %s\n",
		fname.cfg   
	    );

	    /* Listening port(s), both standard and aux */
	    for(i = 0; i < total_incoming_sockets; i++)
	    {
		if(incoming_socket[i] == NULL)
		    continue;

		if(incoming_socket[i]->type == INCOMING_SOCKET_TYPE_STANDARD)
		    fprintf(
			stdout,
			"Standard connection port: %i\n",
			incoming_socket[i]->port_num
		    );
		if(incoming_socket[i]->type == INCOMING_SOCKET_TYPE_AUXSTATS)
		    fprintf(
			stdout,
			"AUX monitoring port: %i\n",
			incoming_socket[i]->port_num
		    );

	    }

	    /* Processs ID */
	    fprintf(stdout, "Process ID: %i\n", getpid());
	}

	/* Initialize the reality engine */
	REngInit();

	/* Log the end of startup */
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"*** " PROG_NAME " " PROG_VERSION " started."
	    );

	return(0);
}

/*
 *	Manages one iteration of SWServ.
 */
void SWServMainIteration(void)
{
	time_t	new_systime = time(NULL),
		new_millitime = MilliTime();

	/* Check if timmers have cycled */
	if((new_systime < cur_systime) ||
	   (new_millitime < cur_millitime)
	)
	{
	    /* Reset timmers and time lapses */
	    SWServResetTimmers(new_millitime, new_systime);
	    lapsed_millitime = 0;
	    time_compensation = 1.0;
	}
	else
	{
	    /* Calculate time lapse ms and compensation coeff */
	    lapsed_millitime = new_millitime - cur_millitime;
	    time_compensation = (double)lapsed_millitime /
		(double)SW_CYCLE_LAPSE_MS;
#if 0
	    if(time_compensation < 1.0)
	        time_compensation = 1.0;
#endif
	}

	/* Set current time */
	cur_systime = new_systime;
	cur_millitime = new_millitime;


	/* Check if a global reset has been requested */
	if(master_reset)
	{
	    /* Reset SWServ */
	    SWServReset(master_reset);
	    master_reset = 0;
	}


	/* Time for system updates? */
	if(next.system_check <= cur_millitime)
	{
	    /* Shutdown schedualed? */
	    if(next.shutdown > 0l)
	    {
		/* Time to shutdown? */
		if(next.shutdown <= cur_systime)
		{
		    SWServMainPopAll();
		    return;
		}
	    }

	    /* Poll incoming sockets? */
	    if(next.new_connection_poll <= cur_millitime)
	    {
		int i, status;
		incoming_socket_struct *in_sock;
		sws_connection_struct *con;

		/* Iterate through each incoming socket */
		for(i = 0; i < total_incoming_sockets; i++)
		{
		    in_sock = incoming_socket[i];
		    if(in_sock == NULL)
			continue;

		    /* Handle by socket type */
		    switch(in_sock->type)
		    {
		      case INCOMING_SOCKET_TYPE_AUXSTATS:
			status = AUXConManageNewConnections(
			    in_sock->socket
			);
			break;
		      default:	/* INCOMING_SOCKET_TYPE_STANDARD */
			status = NetManageNewConnections(
			    in_sock->socket
			);
		        switch(status)
		        {
		          /* No error */
		          case 0:
			    break;
		          /* General error */
		          case -1:
			    break;
			  /* Connection would exceed sysparm.max_connections */
			  case -2:
			    break;
			  /* Connection was guest connection and exceed
			   * sysparm.max_guests
		           */
			  case -3:
			    break;
			  /* Other error */
			  default:
			    break;
			}
			break;
		    }
		}

		/* Resend login request to connections that are
		 * connected but haven't given us a name and
		 * password (not logged in)
		 */
		for(i = 0; i < total_connections; i++)
		{
		    con = connection[i];
		    if(con == NULL)
			continue;

		    /* Skip if connection is not connected */
		    if(con->socket < 0)
			continue;

		    /* Skip if connection's object *is* valid */
		    if(!DBIsObjectGarbage(con->object_num))
			continue;

		    /* Close socket if timmed out */
		    if((con->contime + sysparm.login_timeout) <
			cur_systime
		    )
		    {
			char s[80];
			sprintf(
			    s,
			    "Connection %i: Timmed out.",
			    i
			);
			if(sysparm.log_net)
			    LogAppend(fname.log, s);

			/* Close connection */
			NetCloseConnection(i);

			continue;
		    }

		    /* Resend login request */
		    NetSendLogin(i);
		}

		/* Schedual next poll for new connections */
		next.new_connection_poll = cur_millitime +
		    sysparm.int_new_connection_poll;
	    }

	    /* Time to fetch OS stats? */
	    if(next.os_stats <= cur_systime)
	    {
		SWServGetOSStats();

		/* Schedual next fetch OS stats */
		next.os_stats = cur_systime + sysparm.int_os_stats;
	    }

	    /* Time to clean memory? */
	    if(next.memory_clean <= cur_millitime)
	    {
		/* Universe objects */
		DBReclaim();

		/* Object parameter macros */
		OPMReclaim();

		/* Object create scripts */
		OCSReclaim();

		/* Connections */
		ConReclaim();

		/* Schedual next memory clean */
		next.memory_clean = cur_millitime + sysparm.int_memory_clean;
	    }

	    /* Time to save the universe? */
	    if(next.unv_save <= cur_systime)
	    {
		SWServSaveUnv();

		/* Schedual next universe save */
		next.unv_save = cur_systime + sysparm.int_unv_save;
	    }

	    /* Time to export stats? */
	    if(next.stats_export <= cur_systime)
	    {
		SWServExportStats();

		/* Schedual next time to export stats */
		next.stats_export = cur_systime + sysparm.int_stats_export;
	    }


	    /* Schedual next time for system check */
	    next.system_check = cur_millitime + sysparm.int_system_check;
	}


	/* Manage objects */
	REngManage();


	/* Network */

	/* AUX connections */
	AUXConManageAll();

	/* Send out pending outgoing network data */
	NetManageSend();

	/* Recieve and process incoming network data */
	NetManageRecv();


#ifdef PLUGIN_SUPPORT
	/* Plugins management */
	PluginManage();
#endif	/* PLUGIN_SUPPORT */
}

/*
 *	Manages SWServ.
 *
 *	Pushes an SWServ main loop and manages and blocks until
 *	SWServMainPop() is called.
 */
void SWServMain(void)
{
	const int entering_runlevel = runlevel;
	runlevel++;
	while(runlevel > entering_runlevel)
	{
	    switch(sysparm.cpu_priority)
	    {
	      case SWS_CPU_PRIORITY_HIGHEST:
		usleep(0);
		break;
	      case SWS_CPU_PRIORITY_HIGH:
		usleep(1000);
		break;
	      case SWS_CPU_PRIORITY_MEDIUM:
		usleep(SW_CYCLE_LAPSE_US / 2);
		break;
	      default:	/* SWS_CPU_PRIORITY_LOWEST */
		usleep(SW_CYCLE_LAPSE_US);
		break;
	    }

	    SWServMainIteration();
	}
}

/*
 *	Pops off the highest SWServ main loop.
 */
void SWServMainPop(void)
{
	if(runlevel > 0)
	    runlevel--;
}

/*
 *	Pops off all SWServ main loops.
 */
void SWServMainPopAll(void)
{
	runlevel = 0;
}

/*
 *	Shuts down SWServ.
 *
 *	If save_unv is true then the universe will be saved.
 */
void SWServShutdown(const char save_unv)
{
#ifdef PLUGIN_SUPPORT
	/* Unload plugins first so that they may modify things before
	 * the final universe save and shutdown
	 */
	PluginUnloadAll();
#endif

	/* Close and delete all connections so that Guest objects
	 * are not saved
	 */
	ConDeleteAll();

	/* Save the universe? */
	if(save_unv)
	    SWServSaveUnv();

	/* Begin shutdown */

	/* Reality engine */
	REngShutdown();

	/* Recycled objects buffers */
	DBRecycleBufferDeleteAll();

	/* Close and delete all connections */
	ConDeleteAll();

	/* XSW Objects */
	DBDeleteAllObjects();

	/* Object Create Scripts */
	OCSDeleteAll();

	/* Object Parameter Models */
	OPMDeleteAll();

	/* Universe management resources */
	UNVShutdown();

	/* AUX Monitoring Connections */
	AUXConDeleteAll();

	/* Site Bans */
	SiteBanDeleteAll();

	/* Incoming Sockets */
	IncomingSocketDeleteAll();

	/* Remove the PID file */
	if(!STRISEMPTY(fname.pid))
	    unlink(fname.pid);
}


int main(int argc, char *argv[])
{
	int i, status;
	const char *arg;

	segfault_count = 0;
	runlevel = 0;

	/* Select signals to watch for */
	signal(SIGHUP, SWServSignalCB);
	signal(SIGINT, SWServSignalCB);
	signal(SIGQUIT, SWServSignalCB);
	signal(SIGILL, SWServSignalCB);
	signal(SIGTRAP, SWServSignalCB);
	signal(SIGABRT, SWServSignalCB);
	signal(SIGBUS, SWServSignalCB);
	signal(SIGFPE, SWServSignalCB);
	signal(SIGKILL, SWServSignalCB);
	signal(SIGUSR1, SWServSignalCB);
	signal(SIGSEGV, SWServSignalCB);
	signal(SIGUSR2, SWServSignalCB);
	signal(SIGPIPE, SWServSignalCB);
	signal(SIGALRM, SWServSignalCB);
	signal(SIGTERM, SWServSignalCB);
	signal(SIGCHLD, SWServSignalCB);
	signal(SIGCONT, SWServSignalCB);
	signal(SIGSTOP, SWServSignalCB);
	signal(SIGTSTP, SWServSignalCB);
	signal(SIGTTIN, SWServSignalCB);
	signal(SIGTTOU, SWServSignalCB);
	signal(SIGURG, SWServSignalCB);
	signal(SIGXCPU, SWServSignalCB);
	signal(SIGXFSZ, SWServSignalCB);
	signal(SIGVTALRM, SWServSignalCB);
	signal(SIGPROF, SWServSignalCB);
	signal(SIGWINCH, SWServSignalCB);
	signal(SIGIO, SWServSignalCB);
	signal(SIGPOLL, SWServSignalCB);
#ifndef __CYGWIN32__
	signal(SIGPWR, SWServSignalCB);
#endif /* __CYGWIN32__ */

	/* Handle help and version arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!strcasecmp(arg, "--help") ||
	       !strcasecmp(arg, "-help") ||
	       !strcasecmp(arg, "--h") ||
	       !strcasecmp(arg, "-h") ||
	       !strcasecmp(arg, "-?") ||
	       !strcasecmp(arg, "/?")
	    )
	    {
		SWServDoHelp();
		return(0);
	    }
	    /* Version */
	    else if(!strcasecmp(arg, "--version") ||
		    !strcasecmp(arg, "-version")
	    )
	    {
		SWServDoVersion();
		return(0);
	    }
	}

	/* Initialize */
	status = SWServInit(argc, argv);
	if(status)
	{
	    /* Shutdown and do not save the universe */
	    SWServShutdown(0);
	    fprintf(
		stderr,
		"%s\n",
PROG_NAME " " PROG_VERSION " shutting down due to error during initialization."
	    );
	    return(-status);
	}

	/* Manage */
	SWServMain();

	/* Save the universe and shutdown */
	SWServShutdown(1);

	/* Notify and log shutdown */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
		"%s shut down.\n",
		PROG_NAME
	    );
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"*** " PROG_NAME " " PROG_VERSION " shut down."
	    );

	return(0);
}
