#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/netio.h"
#include "../include/swnetcodes.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "siteban.h"
#include "cmd.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int NetConGuests(void);

int NetManageNewConnections(int socket);

void NetCloseConnection(int condescriptor);

int NetSendDataToConnection(
	int condescriptor,
	const char *data,
	int priority			/* 1 or 2 */
);
void NetDoSend(int condescriptor, const char *sndbuf);

int NetManageSend(void);

static int NetHandleExtCmd(int condescriptor, char *arg);
static int NetManageRecvConnection(int condescriptor);
int NetManageRecv(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the current number of guest connections.
 */
int NetConGuests(void)
{
	int i, n = 0;
	sws_connection_struct *con;

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    if(con->flags & CON_FLAG_IS_GUEST)
		n++;
	}

	return(n);
}


/*
 *	Checks listening socket for incoming connections
 *	and processes them.  Returns non-zero on error.
 */
int NetManageNewConnections(int socket)
{
	int i, n, con_num;
	char sndbuf[CS_DATA_MAX_LEN];
	sws_connection_struct *con;

	int sin_size;
	int new_socket;
	struct sockaddr_in foreign_addr;
	siteban_ip_union ip;

	/* Check if the listening socket has any queued incoming
	 * connection(s)
	 */
	if(!NetIsSocketReadable(socket, NULL))
	    return(0);

	/* Answer new connection */
	sin_size = sizeof(struct sockaddr_in);
	new_socket = accept(
	    socket,
	    (struct sockaddr *)&foreign_addr,
	    reinterpret_cast<socklen_t *>(&sin_size)
	);
	if(new_socket == -1)
	    return(-1);

	/* Set socket nonblocking */
	fcntl(new_socket, F_SETFL, O_NONBLOCK);


	/* Get IP (in nbo) */
	ip.whole = foreign_addr.sin_addr.s_addr;

	/* Is this IP banned? */
	if(SiteBanIsBanned(&ip, 0))
	{
	    sprintf(
		sndbuf,
"%i Your address %i.%i.%i.%i has been banned.\n",
		CS_CODE_LIVEMESSAGE,
		ip.part_u8[0],
		ip.part_u8[1],
		ip.part_u8[2],
		ip.part_u8[3]
	    );
	    send(new_socket, sndbuf, strlen(sndbuf), MSG_NOSIGNAL);

	    sprintf(
		sndbuf,
"%i\n",
		CS_CODE_LOGOUT  
	    );
	    send(new_socket, sndbuf, strlen(sndbuf), MSG_NOSIGNAL);
 
	    /* Close socket */
	    while(close(new_socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"Banned site connection close error: %s.",
			    s
			);
			free(s);		    
		    }
		}
		break;
	    }

	    /* Pretend like we did not do anything */
	    return(0);
	}



	/* Do we have room for new connection? */
	for(i = 0, n = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    n++;
	}
	if(n >= sysparm.max_connections)
	{
	    sprintf(
		sndbuf,
"%i Maximum of %i connections reached.\n",
		CS_CODE_LIVEMESSAGE,
		sysparm.max_connections
	    );
	    send(new_socket, sndbuf, strlen(sndbuf), MSG_NOSIGNAL);

	    sprintf(
		sndbuf,
"%i\n",
		CS_CODE_LOGOUT
	    );
	    send(new_socket, sndbuf, strlen(sndbuf), MSG_NOSIGNAL);

	    /* Close socket */
	    while(close(new_socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"Excessive connection (max %i) close error: %s.",
			    sysparm.max_connections,
			    s
			);
			free(s);
		    }
		}    
		break;
	    }

	    return(-3);
	}


	/* Get a new connection descriptor for new connection */
	con_num = ConCreateNew();
	if(ConIsAllocated(con_num))
	{
	    con = connection[con_num];
	}
	else
	{
	    if(sysparm.log_net)
	    {
		char s[80 + 40];
		sprintf(
		    s,
"ConCreateNew(): Unable to create a new connection (total = %i)",
		    total_connections
		);
	        LogAppend(fname.log, s);
	    }

	    /* Close socket */
	    while(close(new_socket) != 0)
	    {
		if(errno == EAGAIN)
		{
		    continue;
		}
		else
		{
		    char *s = STRDUP(strerror(errno));
		    if(s != NULL)
		    {
			fprintf(
			    stderr,
"Failed connection creation close socket %i error: %s.",
			    new_socket,
			    s
			);
			free(s);
		    }
		}
		break;
	    }

	    return(-3);
	}


	/* Reset some statistics for the new connection */
	con->contime = time(NULL);

	con->obj_ud_interval = SWS_DEF_OBJECT_UPDATE_INT;
	con->obj_ud_next = cur_millitime;

	con->socket = new_socket;

	strncpy(
	    con->conhost,
	    inet_ntoa(foreign_addr.sin_addr),
	    sizeof(con->conhost)
	);
	con->conhost[sizeof(con->conhost) - 1] = '\0';


	/* We don't give connection a new object untill they send
	 * us a valid name and password
	 */
	con->object_num = -1;


	/* Log newly allocated descriptor for connection */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
		"Connection %i: Got connection from `%s'.\n",
		con_num,
		con->conhost
	    );

	if(sysparm.log_net)
	{
	    char *s = (char *)malloc(
		80 + 40 + STRLEN(con->conhost)
	    );
	    sprintf(
		s,
"Connection %i: Got connection from `%s'.",
		con_num,
		con->conhost
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	/* Request login name and password from the new connection */
	NetSendLogin(con_num);

	return(0);
}


/*
 *	Closes and recycles the connection.
 *
 *	If the connection's socket is connected then a
 *	CS_CODE_LOGOUT is sent to it before closing its socket.
 *
 *      If the connection is a guest connection then the guest object
 *      will be recycled.
 */
void NetCloseConnection(int condescriptor)
{
	int i;
	sws_connection_struct *con;

	if(!ConIsAllocated(condescriptor))
	    return;

	con = connection[condescriptor];

	/* Connected? */
	if(con->socket > -1)
	{
	    /* Send a logout to the connection
	     *
	     * If there was an error sending to the connection then
	     * this function NetCloseConnection() is called again
	     * but the socket will already be closed and set to -1
	     */
	    NetSendLogout(condescriptor);

	    /* If NetSendLogout() delted this connection then
	     * just return
	     */
	    if(!ConIsAllocated(condescriptor))
		return;

	    con = connection[condescriptor];

	    /* NetSendLogout() may have already closed this socket,
	     * but double check and close it as needed
	     */
	    if(con->socket > -1)
	    {
		while(close(con->socket) != 0)
		{
		    if(errno == EAGAIN)
		    {
			continue;
		    }
		    else
		    {   
			char *s = STRDUP(strerror(errno));
			if(s != NULL)
			{
			    fprintf(
				stderr,
"NetCloseConnection(): connection %i: close socket %i error: %s.",
				condescriptor,
				con->socket,
				s
			    );
			    free(s);
			}
		    }
		    break;
		}
		con->socket = -1;
	    }

	    /* Notify console of disconnect as needed */
	    if(!sysparm.console_quiet)
	    {
		int obj_num = con->object_num;
	        if(DBIsObjectGarbage(obj_num))
	            fprintf(
		        stdout,
		        "Connection %i: *Unknown* disconnected.\n",
		        condescriptor
		    );
	        else
		    fprintf(   
			stdout,
			"Connection %i: %s disconnected.\n",
		        condescriptor,
			DBGetFormalNameStr(obj_num)
		    );
	    }

	    /* Log disconnect */
	    if(sysparm.log_net)
	    {
		int obj_num = con->object_num;
		char text[SW_OBJ_NAME_MAX + 80 + 80];
		if(DBIsObjectGarbage(obj_num))
		    sprintf(
		        text,
	                "Connection %i: Disconnected.",
		        condescriptor
	            );
	        else
		    sprintf(
			text,
			"Connection %i: %s disconnected.",
			condescriptor,
			DBGetFormalNameStr(obj_num)
		    );
		LogAppend(fname.log, text);
	    }
	}


	/* Begin recycling connection */

	/* Recycle the connection's object if it was a guest connection
	 *
	 * WARNING: DBRecycleObject() should be called after the
	 * connection is closed and reset or else DBRecycleObject()
	 * will call *this* function to close if the object is a player
	 * (and the object IS usually a player)
	 */
	if(DBIsObjectGarbage(con->object_num))
	{
	    /* Object is invalid, so just recycle the connection */ 
	    ConRecycle(condescriptor);
	}
	else
	{
	    /* Connection's object is valid */
	    int obj_num = con->object_num;
	    xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;

	    /* Notify all connections of disconnect? */
	    if(sysparm.con_notify)
	    {
		char	*s = strdup("%name has disconnected."),
			*mesg = strsub(s, "%name", obj->name);
		free(s);

	        for(i = 0; i < total_connections; i++)
	        {
		    if(connection[i] == NULL)
		        continue;

		    if(connection[i]->socket < 0)
			continue;

		    if(connection[i]->object_num < 0)
		        continue;

		    if(i == condescriptor)
		        continue;

		    NetSendLiveMessage(i, mesg);
	        }

		free(mesg);
	    }

	    /* Mark the object as not connected */
	    obj->ext_flags &= ~SWS_OBJ_FLAG_CONNECTED;

	    /* Set the last connected time */
	    obj->last_connected = cur_systime;

	    /* Mark disconneted players hidden at home objects? */
	    if(sysparm.hide_players_home)
	    {
		/* Iterate through the objects list */
		for(i = 0; i < total_objects; i++)
		{
		    tar_obj = xsw_object[i];
		    if(tar_obj == NULL)
			continue;

		    /* Skip non-HOME objects */
		    if(tar_obj->type != SW_OBJ_TYPE_HOME)
			continue;

		    /* In contact with this home object? */
		    if(Mu3DInContactPtr(tar_obj, obj))
		    {
			/* Mark it hidden from connections */
			obj->ext_flags |= SWS_OBJ_FLAG_HIDEFROMCON;
			break;
		    }
		}
	    }

	    /* Set the object's throttle to 0 so that it does not
	     * burn antimatter while disconnected
	     */
	    obj->throttle = 0.0;

	    /* Check if connection is a guest connection */
	    if(con->flags & CON_FLAG_IS_GUEST)
	    {
		/* Is a guest connection, so recycle the connection
		 * and the object
		 */
		ConRecycle(condescriptor);
		NetSendRecycleObject(-1, obj_num);
		DBRecycleObject(obj_num);
	    }
	    else
	    {
		/* Not a guest connection, so only recycle the
		 * connection
		 */
		ConRecycle(condescriptor);
	    }
	}

	/* Schedual stats export to now in order to export the
	 * connections list now
	 */
	next.stats_export = cur_systime;
}


/*
 *      Sends sndbuf to condescriptor or to all connections if
 *      condescriptor is -1.
 * 
 *      The priority will be 1 if the condescriptor is not -1
 *      and the priority will be 2 if the condescriptor is -1.
 */
void NetDoSend(int condescriptor, const char *sndbuf)
{
	/* Send to all connections? */
	if(condescriptor == -1)
	{
	    int i;
	    sws_connection_struct *con;

	    for(i = 0; i < total_connections; i++)
	    {
		con = connection[i];
		if(con == NULL)
		    continue;

		if(con->socket < 0)
		    continue;

		NetSendDataToConnection(
		    i,
		    sndbuf,
		    2		/* Send priority bulk */
		);
	    }
	}  
	else
	{
	    /* Send to specific connection */
	    NetSendDataToConnection(
		condescriptor,
		sndbuf,   
		1		/* Send priority queue as needed */
	    );
	}
}

/*
 *      Sends data to the connection, queing it as needed.
 *      
 *      Before the data is sent however, if any queued data in the
 *      connection is buffered, the queued data will be sent first.
 * 
 *      Priority can be 1 or 2.
 * 
 *	Priority 1 means the data will be sent and if it could not
 *	be sent then it will be queued.
 *
 *	Priority 2 means the data will be sent on the first try and
 *	discarded even if it was not sent.
 */
int NetSendDataToConnection(
	int condescriptor,
	const char *data,
	int priority			/* 1 or 2 */
)
{
	int	s, status, buf_len, /* error_level, */
		bytes_sent = 0,
		cur_qbufs;
	sws_connection_struct *con;

	if(data == NULL)
	    return(-1);

	if(*data == '\0')
	    return(0);

	if(!ConIsConnected(condescriptor))
	    return(-1);

	con = connection[condescriptor];
	s = con->socket;

	/* Get length of data to be sent */
	buf_len = strlen(data);
	if(buf_len >= CS_DATA_MAX_LEN)
	    return(-2);

#if 0
	/* Make sure there is a newline character */
	if(strchr(data, '\n') == NULL)
	{
	    sprintf(text,
  "NetSendDataToConnection(): data missing newline character, not sent."
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, text);

	    sprintf(
		text,
		"NetSendDataToConnection(): `%s'",
		data
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, text);
	
	
	    return(-1);
	}
#endif

	/* Send any queued buffers first */
	cur_qbufs = con->conbuf_items;
	if(cur_qbufs > 0)
	{
	    int	i,
		total_qbufs = SWS_MAX_CON_BUF_ITEMS;
	    sws_conbuf_struct *qbuf = con->conbuf;

	    if(cur_qbufs > total_qbufs) 
		cur_qbufs = total_qbufs;
#if 1
	    /* Send each queued line */
	    for(i = 0; i < cur_qbufs; i++)
	    {
		status = send(
		    s,
		    qbuf[i].buffer, 
		    strlen(qbuf[i].buffer),
		    MSG_NOSIGNAL
		);
		if(status > 0)
		{
		    bytes_sent += status;
		}
		else
		{
		    switch(errno)
		    {
		      case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
		      case EWOULDBLOCK:
#endif
			break;
		      case EBADF:
		      case ENOTSOCK:
		      case EPIPE:
			con->errors_sent++;
			close(con->socket);
			con->socket = -1;
			NetCloseConnection(condescriptor);
			return(-1);
			break;
		      default:
			con->errors_sent++;
			break;
		    }
		    break;
		}
	    }
#else
	    /* Send each queued line */
	    for(i = 0; i < cur_qbufs; i++)
	    {
		/* Check if socket is writeable */
		if(NetIsSocketWritable(s, &error_level))
		{
		    /* Socket is writeable, send data */
		    bytes_sent += send(
			s,
			qbuf[i].buffer,
			strlen(qbuf[i].buffer),
			MSG_NOSIGNAL
		    );
	        }
	        else
	        {
	            /* Socket is not writable, unable to send */
		    con->errors_sent++;

		    /* Check the error level */
		    switch(error_level)
		    {
		      case 3:
			/* Must force closing of socket (and ignore
			 * error) before calling NetCloseConnection()
		         */
			close(con->socket);
			con->socket = -1;
			NetCloseConnection(condescriptor);
			return(-1);
		        break;
		    }
		    break;
	        }
	    }
#endif
	    if(i < cur_qbufs)
	    {
		/* Not able to send out all queued buffers, shift the
		 * remaining queued buffers
		 */
		if(i > 0)
		{
		    int n, k;

		    for(n = 0, k = i;
			k < total_qbufs;
			n++, k++
		    )
		    {
		        if(k >= cur_qbufs)
			    break;

		        memcpy(
			    qbuf[n].buffer,		/* Target */
			    qbuf[k].buffer,		/* Source */
			    CS_DATA_MAX_LEN
		        );
		    }

		    /* Set new number of queued buffers */
		    con->conbuf_items = n;
		}
	    }
	    else
	    {
		/* All queued buffers sent, reset number of queued
		 * buffers on this connection
		 */
		con->conbuf_items = 0;
	    }
	}


	/* Send out the specified data by priority
	 *
	 *      0	Highest, force send for up to 5 seconds (not
	 *		supported yet)
	 *      1	Moderate, try to send. If cannot then put into
	 *		connection's buffer
	 *	2	Lowest, try to send, If cannot then forget it
	 */
	switch(priority)
	{
	  case 0:
	    fprintf(
		stderr,
"NetSendDataToConnection(): Priority 0 not supported.\n"
	    );
	    break;

	  case 1:	/* High priority */
#if 1
	    status = send(s, data, buf_len, MSG_NOSIGNAL);
	    if(status > 0)
	    {
		bytes_sent += status;
	    }
	    else
	    {   
		switch(errno)
		{
		  case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
		  case EWOULDBLOCK:
#endif
		    /* Queue this data for later send */
		    cur_qbufs = con->conbuf_items;
		    if(cur_qbufs < SWS_MAX_CON_BUF_ITEMS)
		    {
			strcpy(
			    con->conbuf[cur_qbufs].buffer,
			    data
			);
			cur_qbufs++;
			con->conbuf_items = cur_qbufs;
		    }
		    else
		    {   
			/* Overflow, queued buffers at maximum,
			 * unable to queue this data segment
			 */
			if(sysparm.log_net || sysparm.log_errors)
			{
			    char s[128 + 40 + 40];
			    sprintf(
				s,
"Connection %i: Unable to queue data to send to connection,\
 queued buffers at maximum %i.\n",
				condescriptor,
				SWS_MAX_CON_BUF_ITEMS
			    );
			    LogAppend(fname.log, s);
			}
		    }
		    break;
		  case EBADF:
		  case ENOTSOCK:
		  case EPIPE:
		    con->errors_sent++;
		    close(con->socket);
		    con->socket = -1;
		    NetCloseConnection(condescriptor);
		    return(-1);
		    break;
		  default:
		    con->errors_sent++;
		    break;
		}
		break;
	    }
#else
	    if(NetIsSocketWritable(s, &error_level))
		bytes_sent += send(s, data, buf_len, MSG_NOSIGNAL);
	    if(error_level)
	    {
		/* Socket is not writeable or there was an error */
		switch(error_level)
		{
		  /* Sever error */
		  case 3:
		    /* Must force closing of socket (and ignore error)
		     * before calling NetCloseConnection()
		     */
		    close(con->socket);
		    con->socket = -1;
		    NetCloseConnection(condescriptor);
		    break;

		  /* Other error */
		  default:
		    /* Queue this data for later send */
		    cur_qbufs = con->conbuf_items;   
		    if(cur_qbufs < SWS_MAX_CON_BUF_ITEMS)
		    {
			strcpy(
			    con->conbuf[cur_qbufs].buffer,
			    data
			);
			cur_qbufs++;
			con->conbuf_items = cur_qbufs;
		    }
		    else
		    {
			/* Overflow, queued buffers at maximum,
			 * unable to queue this data segment
			 */
			if(sysparm.log_net || sysparm.log_errors)
			{
			    char s[128 + 40 + 40];
			    sprintf(
				s,
"Connection %i: Unable to queue data to send to connection,\
 queued buffers at maximum %i.\n",
				condescriptor,
				SWS_MAX_CON_BUF_ITEMS
			    );
			    LogAppend(fname.log, s);
			}
		    }
		    break;
		}
		con->errors_sent++;	/* Count send error */
	    }
#endif
	    break;

	  default:	/* Low priority */
#if 1
	    status = send(s, data, buf_len, MSG_NOSIGNAL);
	    if(status > 0)
	    {
		bytes_sent += status;
	    }
	    else
	    {
		switch(errno)   
		{
		  case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
		  case EWOULDBLOCK:  
#endif
		    /* Discard if not sendable */
		    break;
		  case EBADF:
		  case ENOTSOCK:
		  case EPIPE:  
		    con->errors_sent++;
		    close(con->socket);
		    con->socket = -1;   
		    NetCloseConnection(condescriptor);
		    return(-1);
		    break;
		  default:
		    con->errors_sent++;
		    break;
		}
		break;
	    }
#else
	    if(NetIsSocketWritable(s, &error_level))  
		bytes_sent += send(s, data, buf_len, MSG_NOSIGNAL);
	    if(error_level)
	    {
		/* Socket is not writeable or there was an error */
		switch(error_level)
		{
		  /* Sever error */
		  case 3:
		    /* Must force closing of socket (and ignore error)
		     * before calling NetCloseConnection()
		     */
		    close(con->socket);
		    con->socket = -1; 
		    NetCloseConnection(condescriptor);
		    return(-1);
		    break;
		}
		con->errors_sent++;	/* Count send error */
	    }
#endif
	    break;
	}

	/* Update number of bytes sent on that connection */
	con->bytes_sent += bytes_sent;

	return(0);
}


/*
 *	Sends out schedualed information such as object poses, values,
 *	maximums, weapons, etc to all connections.
 */
int NetManageSend(void)
{
	int con_num, obj_num, obj_count;
	xsw_object_struct *obj, *con_obj;
	sws_connection_struct *con;


	/* Begin sending object poses */

	/* Iterate through connections list */
	for(con_num = 0; con_num < total_connections; con_num++)
	{
	    con = connection[con_num];
	    if(con == NULL)
		continue;

	    obj_num = con->object_num;
	    if(DBIsObjectGarbage(obj_num))
		continue;

	    con_obj = xsw_object[obj_num];

	    /* Connection due to recieve XSW object updates? */
	    if(con->obj_ud_next <= cur_millitime)
	    {
		/* Iterate through objects list */
		for(obj_count = 0; obj_count < total_objects; obj_count++)
		{
		    obj = xsw_object[obj_count];
		    if(obj == NULL)
			continue;

		    /* Check if objects are valid and in range */
		    if(!Mu3DInRangePtr(
			con_obj, obj,
			con_obj->scanner_range
		    ))
			continue;

		    /* If object is not owned by the connection,
		     * then check the visibility of the object.
		     */
		    if(obj->owner != obj_num)
		    {
			if(DBGetObjectVisibilityPtr(obj) <= 0.0)
			    continue;
		    }

		    /* Check if marked hidden from connection */
		    if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;
       
		    /* Send object position (pose) */
		    NetSendObjectPose(con_num, obj_count);
		}

		/* Schedual next object update for connection */ 
		con->obj_ud_next = cur_millitime +
		    con->obj_ud_interval;
	    }
	}


	/* Time to send standard XSW object values? */
	if(next.object_values <= cur_millitime)
	{
	    /* Begin sending object values */

	    /* Iterate through connections list */
	    for(con_num = 0; con_num < total_connections; con_num++)
	    {
		con = connection[con_num];
		if(con == NULL)
		    continue;

		obj_num = con->object_num;
		if(DBIsObjectGarbage(obj_num))  
		    continue;

		con_obj = xsw_object[obj_num];

		/* Iterate through objects list */
		for(obj_count = 0; obj_count < total_objects; obj_count++)
		{
		    obj = xsw_object[obj_count];
		    if(obj == NULL)
			continue;

		    /* Check if objects are valid and in range */
		    if(!Mu3DInRangePtr(
			con_obj, obj,
			con_obj->scanner_range
		    ))
			continue;

		    /* Not owned objects, check for visibility */
		    if(obj->owner != obj_num)
		    {
			if(DBGetObjectVisibilityPtr(obj) <= 0.0)
			    continue;
		    }

		    /* Check if marked hidden from connection */
		    if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Send sector */
		    NetSendObjectSect(con_num, obj_count);

		    /* Send maximums and values */
		    NetSendObjectMaximums(con_num, obj_count);
		    NetSendObjectValues(con_num, obj_count);

		    /* Send tractor beam lock */
		    if(obj->total_tractored_objects > 0)
		    {
			int i, m = obj->total_tractored_objects;
			for(i = 0; i < m; i++)
			    NetSendTractorBeamLock(
				con_num, obj_count,
				obj->tractored_object[i]
			    );
		    }
		}
	    }

	    /* Schedual next object values send */
	    next.object_values = cur_millitime +
		sysparm.int_object_values;
	}


	/* Time to send weapon values? */
	if(next.need_weapon_values)
	{
	    /* Send weapon values */

	    /* Iterate through connections list */
	    for(con_num = 0; con_num < total_connections; con_num++)
	    {
		con = connection[con_num];
		if(con == NULL)
		    continue;

		obj_num = con->object_num;
		if(DBIsObjectGarbage(obj_num))
		    continue;

		con_obj = xsw_object[obj_num];

		/* Iterate through objects list */
		for(obj_count = 0; obj_count < total_objects; obj_count++)
		{
		    obj = xsw_object[obj_count];
		    if(obj == NULL)
			continue;

		    /* Check if objects are valid and in range */
		    if(!Mu3DInRangePtr(
			con_obj, obj,
			con_obj->scanner_range
		    ))
			continue;

		    /* Not owned objects, check for visibility */
		    if(obj->owner != obj_num)
		    {
			if(DBGetObjectVisibilityPtr(obj) <= 0.0)
			    continue;
		    }

		    /* Check if marked hidden from connection */
		    if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Send weapons values */
		    if(obj->total_weapons > 0)
		    {
			int i, m = obj->total_weapons;
			for(i = 0; i < m; i++)
			    NetSendWeaponValues(
				con_num, obj_count, i
			    );
		    }
		}
	    }

	    /* Reset need next weapon values sent */
	    next.need_weapon_values = 0;
	}


	return(0);
}



/*
 *      Handles an XSW extended command.
 *
 *      (This function called by NetHandleConnection() as needed.)
 */
static int NetHandleExtCmd(int condescriptor, char *arg)
{
	int ext_cmd;
	char	*s,
		ext_arg[CS_DATA_MAX_LEN];

	/* condescriptor and arg assumed valid */

	/* Get command and argument */          
	ext_cmd = StringGetNetCommand(arg);
	if(ext_cmd < 0)
	    return(-1);

	s = StringGetNetArgument(arg);
	if(s == NULL)
	    return(-1);

	strncpy(ext_arg, s, sizeof(ext_arg));
	ext_arg[sizeof(ext_arg) - 1] = '\0';


	/* ****************************************************** */
	/* Handle extended command */

	switch(ext_cmd)
	{
	  /* System and enviroment */
	  case SWEXTCMD_SETOCSN:
	    NetHandleSetOCSN(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETUNITS:
	    NetHandleSetUnits(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETTITLE:
	    NetHandleSetTitle(condescriptor, ext_arg);
	    break;  


	  /* Standards */
	  case SWEXTCMD_STDOBJVALS:
	    NetHandleObjectValues(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_STDOBJMAXS:
	    /* Server does not accept set object maximums */
	    break;
	
	  case SWEXTCMD_STDWEPVALS:
	    /* Server does not accept set weapon values */
	    break;
	    

	  /* Set */  
	  case SWEXTCMD_SETOBJNAME:
	    /* Server does not accept set names */
	    break;
   
	  case SWEXTCMD_SETOBJSECT:
	    NetHandleObjectSect(condescriptor, ext_arg);
	    break;

	 case SWEXTCMD_SETFOBJSECT:
	    /* Handle this as SWEXTCMD_SETOBJSECT */
	    NetHandleObjectSect(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETTHROTTLE:
	    NetHandleObjectThrottle(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETWEAPON:
	    NetHandleSelectWeapon(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETINTERCEPT:
	    NetHandleSetIntercept(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETWEPLOCK:
	    NetHandleSetWeaponsLock(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETSHIELDS:
	    NetHandleSetShields(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETDMGCTL:
	    NetHandleSetDmgCtl(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETCLOAK:
	    NetHandleSetCloak(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETSHIELDVIS:
	    /* Server does not handle this command */
	    break;

	  case SWEXTCMD_SETLIGHTING:
	    NetHandleLighting(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETCHANNEL:
	    NetHandleSetChannel(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETSCORE:
	    NetHandleScore(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETENGINE:
	    NetHandleSetEngine(condescriptor, ext_arg);
	    break;


	  /* Requests */
	  case SWEXTCMD_REQNAME:
	    NetHandleObjectName(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_REQSECT:
	    NetHandleReqObjectSect(condescriptor, ext_arg);
	    break;


	  /* Notifies */
	  case SWEXTCMD_NOTIFYHIT:
	    break;

	  case SWEXTCMD_NOTIFYDESTROY:
	    break;


	  /* Actions */
	  case SWEXTCMD_FIREWEAPON:
	    NetHandleFireWeapon(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_TRACTORBEAMLOCK:
	    NetHandleTractorBeamLock(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_HAIL:
	    NetHandleHail(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_COMMESSAGE:
	    NetHandleComMesg(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_WORMHOLEENTER:
	    NetHandleWormHoleEnter(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_ELINKENTER:
	    NetHandleELinkEnter(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_WEPDISARM:
	    NetHandleWeaponDisarm(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_AUTODESTRUCT:
	    NetHandleAutoDestruct(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_SETDESCRIPTION:
	    break;

	  case SWEXTCMD_REQDESCRIPTION:
	    NetHandleObjectDescription(condescriptor, ext_arg);
	    break;


	  /* Economy */
	  case SWEXTCMD_ECO_REQVALUES:
	    NetHandleEcoGetValues(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_ECO_SETVALUES:
	    break;

	  case SWEXTCMD_ECO_SETPRODUCTVALUES:
	    break;

	  case SWEXTCMD_ECO_BUY:
	    NetHandleEcoBuy(condescriptor, ext_arg);
	    break;

	  case SWEXTCMD_ECO_SELL:
	    NetHandleEcoSell(condescriptor, ext_arg);
	    break;

	}   

	return(0);
}


/*
 *      Checks and processes incoming network data on
 *      connection condescriptor.
 *
 *	Warning condescriptor is assumed to be valid.
 */
static int NetManageRecvConnection(int condescriptor)
{
	int bytes_read;

	char recvbuf[CS_DATA_MAX_BACKLOG];
	int recvbuf_cnt;
	char *recvbuf_ptr;

	char workbuf[CS_DATA_MAX_LEN];
	int workbuf_cnt;
	char *workbuf_ptr;

	char arg[CS_DATA_MAX_LEN];
	int command;

	sws_connection_struct *con = connection[condescriptor];
	const int socket = con->socket;
 
#if 0
	/* Is socket valid and contains data to be read? */
	if(!NetIsSocketReadable(socket, NULL))
	    return(0);
#endif

	/* Recieve incoming data */
	bytes_read = recv(socket, recvbuf, sizeof(recvbuf), 0);
	if(bytes_read == 0)
	{
	    /* When polling of socket says there are bytes to be
	     * read and recv() returns 0, it implies that the
	     * socket has died
	     */
	    if(sysparm.log_net || sysparm.log_errors)
	    {
		char *s = (char *)malloc(40 + 128);
		sprintf(
		    s,
"Connection %i: Received 0 bytes of data.",
		    condescriptor
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	    con->socket = -1;
	    NetCloseConnection(condescriptor);
	    return(-1);
	}

	/* Recieve error? */
	if(bytes_read < 0)
	{
	    const int error_code = errno;
	    char *error_msg;

	    /* Handle error */
	    switch(error_code)
	    {
	      case EAGAIN:	/* No data available */
#if (EWOULDBLOCK != EAGAIN)
	      case EWOULDBLOCK:
#endif
	      case EINTR:	/* Interrupted by signal */
		return(0);
		break;

	      case EBADF:	/* Invalid descriptor */
	      case ECONNRESET:	/* Connection terminated by peer */
	      case EFAULT:	/* Receive buffer out of address space */
	      case ENOTCONN:	/* Descriptor is not connected */
	      case ENOTSOCK:	/* Descriptor is not a socket */
	      case EPIPE:	/* Broken pipe */
		error_msg = STRDUP(strerror(error_code));
		if((error_msg != NULL) &&
		   (sysparm.log_net || sysparm.log_errors)
		)
		{
		    char *s = (char *)malloc(
			40 + STRLEN(error_msg) + 128
		    );
		    *error_msg = toupper(*error_msg);
		    sprintf(
			s,
"Connection %i: Receive error and broken connection: %s.",
			condescriptor,
			error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		NetCloseConnection(condescriptor);
		free(error_msg);
		return(-1);
		break;

	      default:
		error_msg = STRDUP(strerror(error_code));   
		if((error_msg != NULL) &&
		   (sysparm.log_net || sysparm.log_errors)
		)
		{
		    char *s = (char *)malloc(
			40 + STRLEN(error_msg) + 128
		    );
		    *error_msg = toupper(*error_msg);
		    sprintf(
			s,
"Connection %i: Receive error: %s.",
			condescriptor,
			error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		free(error_msg);
		return(-1);
		break;
	    }
	}


	/* bytes_read cannot be greater than CS_DATA_MAX_BACKLOG */
	if(bytes_read > (int)sizeof(recvbuf))
	    bytes_read = (int)sizeof(recvbuf);

	/* Increment bytes_received for this condescriptor */
	con->bytes_received += bytes_read;


	/* Set null character at the end of recvbuf just to be safe */
	recvbuf[sizeof(recvbuf) - 1] = '\0';
	    
	      
	/* *************************************************************** */
	/* Begin parsing recieve buffer */

	/* Reset recieve buffer count and pointer */
	recvbuf_cnt = 0;
	recvbuf_ptr = recvbuf;

	while(recvbuf_cnt < bytes_read)
	{
	    /* Reset work buffer count and pointer */
	    workbuf_cnt = 0;
	    workbuf_ptr = workbuf;

	    /* Fetch data for workbuf */
	    while(1)
	    {
		if(workbuf_cnt >= (int)sizeof(workbuf))
		{
		    /* Increment recvbuf count and ptr to next delimiter */
		    while(recvbuf_cnt < bytes_read)
		    {
			if((*recvbuf_ptr == '\n') ||
			   (*recvbuf_ptr == '\r') ||
			   (*recvbuf_ptr == '\0')
			)
			{
			    recvbuf_cnt++;
			    recvbuf_ptr++;
			    break;
			}
			recvbuf_cnt++;
			recvbuf_ptr++;
		    }
		    /* Null terminating character for workbuf will
		     * be added farther below
		     */
		    break;
		}
		/* End of recieve buffer data reached? */
		else if(recvbuf_cnt >= bytes_read)
		{
		    *workbuf_ptr = '\0';
		    break;
		}
		/* Deliminator in recieve buffer encountered? */
		else if((*recvbuf_ptr == '\n') ||
			(*recvbuf_ptr == '\r') ||
			(*recvbuf_ptr == '\0')
		)
		{   
		    *workbuf_ptr = '\0';

		    recvbuf_cnt++;
		    recvbuf_ptr++;

		    break;
		}

		/* Copy data from recieve buffer to work buffer normally */
		else
		{
		    *workbuf_ptr = *recvbuf_ptr;
		
		    recvbuf_cnt++;
		    recvbuf_ptr++;
			
		    workbuf_cnt++;
		    workbuf_ptr++;
		}
	    }

	    /* Skip if workbuf is empty */
	    if(*workbuf == '\0')
		continue;

	    /* Set null terminating character for workbuf */
	    workbuf[sizeof(workbuf) - 1] = '\0';


	    /* Get command command */
	    command = StringGetNetCommand(workbuf);
	    if(command < 0)
		continue;

	    /* Get argument arg */
	    strncpy(arg, StringGetNetArgument(workbuf), sizeof(arg));
	    arg[sizeof(arg) - 1] = '\0';
		 
	 
	    /* See which net command handling function to call */
	    switch(command)
	    {
	      case CS_CODE_LOGIN:
		NetHandleLogin(condescriptor, arg);
		break;

	      case CS_CODE_LOGOUT:
		NetHandleLogout(condescriptor);
		break;

	      case CS_CODE_WHOAMI:
		NetHandleWhoAmI(condescriptor, arg);
		break;

	      case CS_CODE_REFRESH:
		NetHandleRefresh(condescriptor, arg);
		break;

	      case CS_CODE_INTERVAL:
		NetHandleSetInterval(condescriptor, arg);
		break;

	      case CS_CODE_IMAGESET:
		NetHandleSetImageSet(condescriptor, arg);
		break;

	      case CS_CODE_SOUNDSET:
		NetHandleSetSoundSet(condescriptor, arg);
		break;

	      case CS_CODE_PLAYSOUND:
		/* server dosen't play sounds */
		break;


	      case CS_CODE_LIVEMESSAGE:
		NetHandleLiveMessage(condescriptor, arg);
		break;

	      case CS_CODE_SYSMESSAGE:
		/* Server does not handle this */
		break;

	      case CS_CODE_LITERALCMD:
		CmdHandleInput(condescriptor, arg);
		break;


	      case CS_CODE_CREATEOBJ:
		/* Server doesn't accept this command */
		break;

	      case CS_CODE_RECYCLEOBJ:
		/* Server dosen't accept this command */
		break;

	      case CS_CODE_POSEOBJ:
		NetHandleObjectPose(condescriptor, arg);
		break;

	      case CS_CODE_FORCEPOSEOBJ:
		/* Server does not accept this */
		break;

	      case CS_CODE_EXT:
		NetHandleExtCmd(condescriptor, arg);
		break;

	      /* Unknown command */
	      default:
#if 0
fprintf(
 stderr, "Recieved unsupported command %i.\n", command
);
fprintf(
 stderr, "Raw text: `%s'\n", workbuf
);
#endif
		con->errors_received++;
		break;
	    }


	    /*   The call to the net handling function may have reallocated
	     *   the connection pointers (espessially a call to
	     *   CmdHandleInput()) so we need to test if condescriptor is
	     *   still valid.
	     */
	    if(condescriptor >= total_connections)
		break;
	}

	return(0);
}


/*
 *      Procedure to check all connections for incoming network
 *      data and process the data.
 */
int NetManageRecv(void)
{
	int i;
	sws_connection_struct *con;

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    /* Note: The functions that NetManageRecvConnection() calls
	     * may cause the connections to be deallocated.
	     * If they are, global total_connections would have been reset
	     * to 0 and this loop would exit.
	     */
	    NetManageRecvConnection(i);
	}

	return(0);
}                   
