#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#ifdef PLUGIN_SUPPORT
# include "plugins.h"
#endif
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


int NetHandleComMesg(int condescriptor, const char *arg);
int NetSendComMesg(
        int condescriptor,
        int src_obj_num, int tar_obj_num,
        double bearing, int channel,
        const char *msg
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int NetHandleComMesg(int condescriptor, const char *arg)
{
	char *s, *darg, *msg;
	int obj_num, tar_obj_num;
	double bearing;
	int channel;
	xsw_object_struct *obj;

	/* Copy the argument for parsing and editing purposes */
	darg = STRDUP(arg);

	/* SWEXTCMD_COMMESSAGE format:
	 *
	 *	src_obj tar_obj bearing channel;message
	 */

	/* Get message */
	s = strchr(darg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	{
	    free(darg);
	    return(-1);
	}
	msg = STRDUP(s + 1);

	/* Remove the message argument from darg */
	*s = '\0';

	/* No message? */
	if(msg == NULL)
	{
	    free(darg);
	    return(-1);
	}

	/* Parse arguments */
	sscanf(
	    darg,
	    "%i %i %lf %i",
	    &obj_num,
	    &tar_obj_num,	/* Can be -1 */
	    &bearing,
	    &channel
	);

	free(darg);

	/* Connection must own the object */
	if(connection[condescriptor]->object_num != obj_num)
	{
	    free(msg);
	    return(-3);
	}

	if(DBIsObjectGarbage(obj_num))
	{
	    free(msg);
	    return(-1);
	}

	obj = xsw_object[obj_num];

	/* Send message to specific object? */
	if(tar_obj_num > -1)
	{
	    /* Objects in com range? */
	    if(Mu3DInRange(
		xsw_object, total_objects,
		obj_num, tar_obj_num, sysparm.com_range
	    ))
	    {
		const int tar_con_num = ConGetByObject(tar_obj_num);
		xsw_object_struct *tar_obj = xsw_object[tar_obj_num];
		if((tar_obj->com_channel == channel) && (tar_con_num > -1))
		{
		    /* Get bearing from target to source object */
		    bearing = MuCoordinateDeltaVector(
			obj->x - tar_obj->x,
			obj->y - tar_obj->y
		    );

		    /* Send message to that connection */
		    NetSendComMesg(
			tar_con_num,
			obj_num,	/* Sender */
			tar_obj_num,	/* Receiver */
			bearing,	/* Target to source */
			channel,
			msg
		    );
		}
	    }
	}
	else
	{
	    int i, tar_con_num;
	    xsw_object_struct *tar_obj;

	    /* Send message to all objects in com range */
	    for(i = 0; i < total_objects; i++)
	    {
		/* Object not valid or not in com range? */
		if(!Mu3DInRange(
		    xsw_object, total_objects,
		    obj_num, i, sysparm.com_range
		))
		    continue;

		tar_obj = xsw_object[i];

		/* Skip of object_count is not listening on channel */
		if(tar_obj->com_channel != channel)
		    continue;

		/* Get the connection of this object (if any) */
		tar_con_num = ConGetByObject(i);
		if(tar_con_num < 0)
		    continue;

		/* Get bearing from target to source object */
		bearing = MuCoordinateDeltaVector(
		    obj->x - tar_obj->x,
		    obj->y - tar_obj->y
		);

		/* Send message to that connection */
		NetSendComMesg(
		    tar_con_num,
		    obj_num,	/* Sender */
		    -1,		/* General message */
		    bearing,	/* Target to source */
		    channel,
		    msg
		);
	    }
	}

	/* Log message */
	if(sysparm.log_net)
	{
	    const char *name = DBGetFormalNameStr(obj_num);
	    s = (char *)malloc(
		STRLEN(name) + 40 + STRLEN(msg) + 80
	    );
	    sprintf(
		s,
"%s: Channel %.2f: Message: \"%s\"",
		name,
		(double)((double)channel / 100),
		msg
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

#ifdef PLUGIN_SUPPORT
	/* Forward this hail to plugins */
	if(total_plugins > 0)
	{
	    int i;
	    for(i = 0; i < total_plugins; i++)
		PluginHailNotify(
		    i,
		    condescriptor,
		    obj_num,		/* Sender */
		    tar_obj_num,	/* Target object or -1 */
		    channel,		/* Com channel */
		    msg			/* Message */
		);
	}
#endif

	free(msg);

	return(0);
}


int NetSendComMesg(
	int condescriptor,
	int src_obj_num,
	int tar_obj_num,
	double bearing,		/* Target to source in radians */
	int channel,
	const char *msg
)
{
	char	sndbuf[CS_DATA_MAX_LEN],
		lmsg[CS_MESG_MAX];

	if(msg == NULL)
	    return(-1);

	/* Copy and shorten message */
	strncpy(lmsg, msg, sizeof(lmsg));
	lmsg[sizeof(lmsg) - 1] = '\0';

	/* SWEXTCMD_COMMESSAGE format:
	 *
	 *	src_obj_num tar_obj_num
	 *	bearing channel;message
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %.4f %i;%s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_COMMESSAGE,
	    src_obj_num,
	    tar_obj_num,
	    bearing,
	    channel,
	    lmsg
	);
	NetDoSend(condescriptor, sndbuf);

	return(0);
}
