#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/cfgfmt.h"
#include "../include/fio.h"
#include "../include/unvmath.h"

#include "swserv.h"


int OCSFileOpen(const char *path);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


int OCSFileOpen(const char *path)
{
	int ocs_num = -1, lines_read = 0;
        char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
        FILE *fp;
        off_t filesize;
        struct stat stat_buf;
	sws_ocs_struct *ocs_ptr = NULL;

	if(path == NULL)
	    return(-1);

	/* Get the file's stats */
        if(stat(path, &stat_buf))
        {
            s = STRDUP(strerror(errno));
            if(s == NULL)
                s = strdup("no such file");
            *s = toupper(*s);
            fprintf(
                stderr,
                "%s: %s.\n",
                path, s
            );
            free(s);
            return(-1);
        }

	/* Get the size of the file */
        filesize = stat_buf.st_size;

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
            fprintf(
		stderr,
		"%s: Unable to open file for reading.\n",
                path
            );
            return(-1);
	}

	/* Delete all existing OCSs */
	OCSDeleteAll();

	/* Begin reading file */
	s = NULL;
	while(1)
	{
	    /* Did we load more than SWS_MAX_OCSS? */
	    if(ocs_num >= SWS_MAX_OCSS)
	    {
		fprintf(
		    stderr,
"Warning: Maximum of %i OCSs loaded.\n",
                    SWS_MAX_OCSS
                );
		fprintf(
		    stderr,
"Warning: Additional OCSs will not be loaded.\n"
                );
                break;
	    }

            /* Free previous line and allocate/read next line */
            free(s);
            s = FReadNextLineAllocCount(
                fp, UNIXCFG_COMMENT_CHAR, &lines_read
            );
            if(s == NULL)
		break;

            /* Fetch parameter */
            s2 = StringCfgParseParm(s);
            if(s2 == NULL)
		continue;
            strncpy(parm, s2, sizeof(parm));
            parm[sizeof(parm) - 1] = '\0';

            /* Fetch value */
            s2 = StringCfgParseValue(s);
            if(s2 == NULL)
		s2 = "0";
            strncpy(val, s2, sizeof(val));
            val[sizeof(val) - 1] = '\0';


	    /* BeginHeader */
	    if(!strcmp(parm, "BeginHeader"))
	    {
                while(1)
                {
                    /* Free previous line and allocate/read next line */
                    free(s);
                    s = FReadNextLineAllocCount(
                        fp, UNIXCFG_COMMENT_CHAR, &lines_read
                    );
                    if(s == NULL)
			break;

                    /* Fetch parameter */
                    s2 = StringCfgParseParm(s);
                    if(s2 == NULL)
			continue;
                    strncpy(parm, s2, sizeof(parm));
                    parm[sizeof(parm) - 1] = '\0';

                    /* Fetch value */
                    s2 = StringCfgParseValue(s);
                    if(s2 == NULL)
			s2 = "0";
                    strncpy(val, s2, sizeof(val));
                    val[sizeof(val) - 1] = '\0';

                    /* Title */
                    if(!strcmp(parm, "Title"))
                    {

                    }
                    /* EndHeader */
                    else if(!strcmp(parm, "EndHeader"))
                    {
                        break;
                    }
                }
		continue;
	    }
	    /* BeginOCSEntry */
	    else if(!strcmp(parm, "BeginOCSEntry"))
	    {
		/* Create a new OCS */
                ocs_num++;
		if(OCSCreateExplicit(ocs_num, 1))
		{
		    fprintf(
			stderr,
"%s: Line %i: Could not create OCS #%i.\n",
			path, lines_read, ocs_num
		    );
		    break;
		}

		ocs_ptr = ocs[ocs_num];

		while(1)
                {
                    /* Free previous line and allocate/read next line */
                    free(s);
                    s = FReadNextLineAllocCount(
                        fp, UNIXCFG_COMMENT_CHAR, &lines_read
                    );
                    if(s == NULL)
			break;

                    /* Fetch parameter */
                    s2 = StringCfgParseParm(s);
                    if(s2 == NULL)
			continue;
                    strncpy(parm, s2, sizeof(parm));
                    parm[sizeof(parm) - 1] = '\0';

                    /* Fetch value */
                    s2 = StringCfgParseValue(s);
                    if(s2 == NULL)
			s2 = "0";
                    strncpy(val, s2, sizeof(val));
                    val[sizeof(val) - 1] = '\0';

		    /* Code */
		    if(!strcmp(parm, "Code"))
		    {
                        ocs_ptr->code = atoi(val);

			/* Make sure the code is valid */
			if(ocs_ptr->code <= OCS_TYPE_GARBAGE)
			    ocs_ptr->code = OCS_TYPE_GARBAGE + 1;
		    }
                    /* OPMName */
                    else if(!strcmp(parm, "OPMName"))
                    {
                        strncpy(
			    ocs_ptr->opm_name, val,
			    sizeof(ocs_ptr->opm_name)
			);
                        ocs_ptr->opm_name[
			    sizeof(ocs_ptr->opm_name) - 1] = '\0';
                    }
                    /* Coppies */
                    else if(!strcmp(parm, "Coppies"))
                    {
                        ocs_ptr->coppies = atoi(val);
                        
                        /* Make sure coppies is something valid */
			if(ocs_ptr->coppies > SWS_MAX_OCS_COPPIES)
			    ocs_ptr->coppies = SWS_MAX_OCS_COPPIES;
                        if(ocs_ptr->coppies < 1)
                            ocs_ptr->coppies = 1;
                    }
                    /* Heading */
                    else if(!strcmp(parm, "Heading") ||
                            !strcmp(parm, "Angle")
		    )
                    {
                        ocs_ptr->heading = SANITIZERADIANS(atof(val));
                    }
                    /* Pitch */
                    else if(!strcmp(parm, "Pitch"))
                    {
                        ocs_ptr->pitch = SANITIZERADIANS(atof(val));
                    }
                    /* Bank */
                    else if(!strcmp(parm, "Bank"))
                    {
                        ocs_ptr->bank = SANITIZERADIANS(atof(val));
                    }
                    /* Radius */
                    else if(!strcmp(parm, "Radius"))
                    {
                        ocs_ptr->radius = MAX(atof(val), 0.0);
                    }
                    /* EndOCSEntry */
                    else if(!strcmp(parm, "EndOCSEntry"))
                    {
                        break;
                    }
		    /* Unsupported parameter */
		    else
		    {
			fprintf(
			    stderr,
"%s: Line %i: Warning: Unsupported parameter `%s'\n",
                            path, lines_read, parm
			);
		    }
		}
	    }
	}

	/* Close the file */
        fclose(fp);

	return(0);
}
