#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "swserv.h"


void OPMReclaim(void);
void OPMDeleteAll(void);
int OPMIsGarbage(int opm_num);
int OPMCreate(int type);
int OPMCreateExplicit(int opm_num, int type);
void OPMRecycle(int opm_num);
int OPMGetTop(void);
int OPMGetByName(const char *name, int type);
int OPMModelObjectPtr(
	xsw_object_struct *obj,
	xsw_object_struct *opm_ptr
);
int OPMModelObject(int obj_num, int opm_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


void OPMReclaim(void)
{
	int i;

	/* Get highest valid and allocated OPM index number */
	const int h = OPMGetTop();
	if(h < 0)
	    return;

	/* Do we need to reclaim memory? */
	if((total_opms - h - 1) <= SWS_OBJECT_ALLOCATE_AHEAD)
	    return;

	/* Free all allocation from total_opms - 1 to h + 1 */
	for(i = total_opms - 1; i > h; i--)
	{
#ifdef DEBUG_MEM_FREE
if(opm[opm_num] != NULL)
    printf("OPM %i: Free'ed (reclaim).\n", opm_num);
#endif
	    DBDeleteObject(opm[i]);
	    opm[i] = NULL;
	}

	/* Adjust total */
	total_opms = h + 1;


	/* Reallocate OPM pointers */
	if(total_opms > 0)
	{
	    opm = DBAllocObjectPointers(
	        opm, total_opms
	    );
	    if(opm == NULL)
	    {
	        total_opms = 0;
		return;
	    }
	}
	else
	{
#ifdef DEBUG_MEM_FREE
if(opm != NULL)
    printf("OPM pointers: Free'ed.\n");
#endif
	    free(opm);
	    opm = NULL;
	    total_opms = 0;
	}
}


/*
 *	Procedure to delete all OPMs.
 */
void OPMDeleteAll(void)
{
	int i;

	for(i = 0; i < total_opms; i++)
	    DBDeleteObject(opm[i]);
	free(opm);
	opm = NULL;
	total_opms = 0;
}


/*
 *	Checks if opm_num is allocated and non-garbage.
 */
int OPMIsGarbage(int opm_num)
{
	if((opm_num < 0) || (opm_num >= total_opms) || (opm == NULL))
	    return(-1);
	else if(opm[opm_num] == NULL)
	    return(-1);
	else if(opm[opm_num]->type <= SW_OBJ_TYPE_GARBAGE)
	    return(1);
	else
	    return(0);
}


/*
 *	Creates a new OPM.
 */
int OPMCreate(int type)
{
	int opm_num, opm_num_rtn;
	char name[SW_OBJ_NAME_MAX + 80];
	xsw_object_struct *opm_ptr;

	if(type <= SW_OBJ_TYPE_GARBAGE)
	    return(-2);

	/* Look for available OPM already allocated */
	opm_num_rtn = -1;
	for(opm_num = 0; opm_num < total_opms; opm_num++)
	{
	    if(opm[opm_num] == NULL)
		continue;

	    /* Look for garbage OPMs */
	    if(opm[opm_num]->type == SW_OBJ_TYPE_GARBAGE)
	    {
		opm_num_rtn = opm_num;
		break;
	    }
	}

	/* Did we find an OPM already allocated? */
	if((opm_num_rtn > -1) && (opm_num_rtn < total_opms))
	{
	    /* Reset OPM (free all allocated substructures) */
	    opm_ptr = opm[opm_num];
	    UNVResetObject(opm_ptr);

	    sprintf(name, "OPM #%i", opm_num_rtn);
	    strncpy(opm_ptr->name, name, sizeof(opm_ptr->name));
	    opm_ptr->name[sizeof(opm_ptr->name) - 1] = '\0';

	    opm_ptr->type = type;

	    return(opm_num_rtn);
	}

	/* Not enough memory, can we allocate more? */
	if((total_opms + SWS_OBJECT_ALLOCATE_AHEAD) > SWS_MAX_OPMS)
	{
	    fprintf(
		stderr,
"OPMCreate(): maximum %i OPMs reached.\n",
		SWS_MAX_OPMS
	    );
	    return(-3);
	}

	/* Allocate more memory */

	/* Allocate pointer array */
	opm = DBAllocObjectPointers(
	    opm,
	    total_opms + SWS_OBJECT_ALLOCATE_AHEAD
	);
	if(opm == NULL)
	{
	    total_opms = 0;
	    return(-3);
	}

	/* Allocate each OPM structure */
	for(opm_num = total_opms;
	    opm_num < (total_opms + SWS_OBJECT_ALLOCATE_AHEAD);
	    opm_num++
	)
	    opm[opm_num] = DBAllocObject();


	/* New object will be the first object allocated */
	opm_num_rtn = total_opms;


	/* Adjust global variable total_opms */
	opm_num = total_opms;
	total_opms += SWS_OBJECT_ALLOCATE_AHEAD;

	/* Reset values on newly allocated OPMs */
	while(opm_num < total_opms)
	{
	    UNVResetObject(opm[opm_num]);
	    opm_num++;
	}


	/* Set default values on the new OPM */
	opm_ptr = opm[opm_num_rtn];

	sprintf(name, "OPM #%i", sizeof(name));
	strncpy(opm_ptr->name, name, sizeof(opm_ptr->name));
	opm_ptr->name[sizeof(opm_ptr->name) - 1] = '\0';

	opm_ptr->type = type;


	return(opm_num_rtn);
}


/*
 *	Create an OPM explicitly.
 */
int OPMCreateExplicit(int opm_num, int type)
{
	int opm_count, opm_rtn;
	char name[SW_OBJ_NAME_MAX + 80];
	xsw_object_struct *opm_ptr;

	if(type <= SW_OBJ_TYPE_GARBAGE)
	    return(-2);

	/* Make sure opm_num is valid */
	if(opm_num < 0)
	{
	    fprintf(
		stderr,
"OPMCreateExplicit(): Error: Invalid OPM index %i.\n",
		opm_num
	    );
	    return(-1);
	}

	/* If object_num already exists then recycle it first */
	if(!OPMIsGarbage(opm_num))
	    OPMRecycle(opm_num);

	/* Allocate memory if opm_num is greater than total_opms */
	if(opm_num >= total_opms)
	{
	    /* Can we allocate more? */
	    if((opm_num + SWS_OBJECT_ALLOCATE_AHEAD) >= SWS_MAX_OPMS)
	    {
		fprintf(stderr, "OPM maximum %i reached.\n", SWS_MAX_OPMS);
		return(-3);
	    }

	    /* Allocate pointer array */
	    opm = DBAllocObjectPointers(
		opm,
		opm_num + SWS_OBJECT_ALLOCATE_AHEAD + 1
	    );
	    if(opm == NULL)
	    {
		total_opms = 0;
		return(-3);
	    }

	    /* Allocate each new OPM */
	    for(opm_count = total_opms;
		opm_count <= (opm_num + SWS_OBJECT_ALLOCATE_AHEAD);
		opm_count++
	    )
	    {
		/* Allocate memory for a new OPM */
		opm[opm_count] = DBAllocObject();
		if(opm[opm_count] == NULL)
		{
		    total_opms = opm_count;
		    return(-3);
		}
	    }
		
	    /* Adjust global variable total_opms */
	    total_opms = opm_num + SWS_OBJECT_ALLOCATE_AHEAD + 1;
	}


	opm_rtn = opm_num;

	/* Set default values on the new OPM */
	opm_ptr = opm[opm_rtn];
	UNVResetObject(opm_ptr);

	sprintf(name, "OPM #%i", opm_rtn);

	strncpy(opm_ptr->name, name, sizeof(opm_ptr->name));
	opm_ptr->name[sizeof(opm_ptr->name) - 1] = '\0';

	opm_ptr->type = type;

	/* Return 0 on no error, explicit OPM create does not want an
	 * OPM index return
	 */
	return(0);
}

/*
 *	Recycles OPM.
 */
void OPMRecycle(int opm_num)
{
	xsw_object_struct *opm_ptr;

	if(OPMIsGarbage(opm_num))
	    return;

	opm_ptr = opm[opm_num];

	/* Reset values and delete resources */
	UNVResetObject(opm_ptr);

	/* Setting OPM type to garbage marks it as recycled */
	opm_ptr->type = SW_OBJ_TYPE_GARBAGE;
}

/*
 *	Get highest allocated and non-garbage OPM index number.
 *
 *	Can return -1 if no allocated non-garbage OPMs are found.
 */
int OPMGetTop(void)
{
	int i, h;
	xsw_object_struct **ptr, *opm_ptr;

	for(i = 0, h = -1, ptr = opm;
	    i < total_opms;
	    i++, ptr++
	)
	{
	    opm_ptr = *ptr;

	    if(opm_ptr == NULL)
		continue;
	    if(opm_ptr->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    h = i;
	}

	return(h);
}

/*
 *	Get an OPM index number by name, returns the OPM's index
 *	number (can be assumed valid) or -1 on no match/error.
 *	Can also return -2 on ambiguous (name was not specific
 *	enough).
 *
 *	If the name contains a token string (ie "$OPM:SomeObject"),
 *	then the string tailing (ie "SomeObject") will be used
 *	in the search.
 *
 *	The type cannot be set to SW_OBJ_TYPE_GARBAGE.
 *
 *	If type is -1, then all OPM types will be matched except
 *	SW_OBJ_TYPE_GARBAGE.
 */
int OPMGetByName(const char *name, int type)
{
	int opm_num, opm_found, found_count;
	xsw_object_struct *opm_ptr;

	if(name == NULL)
	    return(-1);

	/* The specified type cannot be garbage */
	if(type == SW_OBJ_TYPE_GARBAGE)
	    return(-1);

	/* Skip leading spaces */
	while(*name == ' ')
	    name++;

	/* Token string check */
	if(*name == '$')
	{
	    /* Seek to ':' delimiter */
	    const char *s = strchr(name, ':');
	    if(s == NULL)
		return(-1);

	    name = s + 1;
	}

	/* Specified name cannot be empty */
	if(*name == '\0')
	    return(-1);


	/* Begin searching for an OPM */
	opm_found = -1;
	found_count = 0;

	/* Match regardless of type? */
	if(type < 0)
	{
	    for(opm_num = 0; opm_num < total_opms; opm_num++)
	    {
		opm_ptr = opm[opm_num];
		if(opm_ptr == NULL)
		    continue;

	        if(opm_ptr->type <= SW_OBJ_TYPE_GARBAGE)
		    continue;

	        if(!strcmp(opm_ptr->name, name))
	        {
		    opm_found = opm_num;
		    found_count++;
		}
	    }
	}
	/* Match by type */
	else
	{
	    for(opm_num = 0; opm_num < total_opms; opm_num++)
	    {
		opm_ptr = opm[opm_num];
		if(opm_ptr == NULL)
		    continue;

		if(opm_ptr->type <= SW_OBJ_TYPE_GARBAGE)
		    continue;

		/* Types must match */
		if(opm_ptr->type != type)
		    continue;

		if(!strcmp(opm_ptr->name, name))
		{
		    opm_found = opm_num;
		    found_count++;
		}
	    }
	}

	if(found_count > 1)
	    return(-2);		/* Ambiguous */
	else
	    return(opm_found);
}

/*
 *	Coppies opm values on to object, any values currently
 *	on object will be reset and set to the opm's values.
 *
 *	Allocated substructures on object will be deallocated and
 *	new ones allocated as needed depending on what the opm
 * 	has.
 *
 *	Returns 0 on success or -1 on error.
 */
int OPMModelObjectPtr(
	xsw_object_struct *obj, xsw_object_struct *opm_ptr
)
{
	int i;

	if((obj == NULL) || (opm_ptr == NULL))
	    return(-1);

	/* Free any allocated substructures on object */
	UNVResetObject(obj);

	/* Copy the OPM values to the object, and then allocate
	 * and COPY any dynamically allocated OPMs
	 */

	/* Copy the OPM values to the object */
	memcpy(obj, opm_ptr, sizeof(xsw_object_struct));

	/* Copy all allocated substructures */
	obj->password = STRDUP(opm_ptr->password);
	obj->description = STRDUP(opm_ptr->description);
	obj->elink = STRDUP(opm_ptr->elink);

	/* Weapons */  
	if((opm_ptr->weapons != NULL) && (opm_ptr->total_weapons > 0))
	{
	    const int n = opm_ptr->total_weapons;
	    obj->weapons = (xsw_weapon_struct **)calloc(
		n, sizeof(xsw_weapon_struct *)
	    );
	    if(obj->weapons == NULL)
		obj->total_weapons = 0;

	    for(i = 0; i < n; i++)
	    {
		if(opm_ptr->weapons[i] == NULL)
		    continue;

		obj->weapons[i] = (xsw_weapon_struct *)malloc(
		    sizeof(xsw_weapon_struct)
		);
		memcpy(
		    obj->weapons[i],
		    opm_ptr->weapons[i],
		    sizeof(xsw_weapon_struct)
		);
	    }
	}
	else
	{
	    obj->weapons = NULL;
	    obj->total_weapons = 0;
	}

	/* Tractored objects */
	if((opm_ptr->tractored_object != NULL) &&
	   (opm_ptr->total_tractored_objects > 0)
	)
	{
	    const int n = opm_ptr->total_tractored_objects;
	    obj->tractored_object = (int *)malloc(
		n * sizeof(int)
	    );
	    if(obj->tractored_object != NULL)
		memcpy(
		    obj->tractored_object,
		    opm_ptr->tractored_object,
		    n * sizeof(int)
		);
	}
	else
	{
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	}

	/* Scores */
	if(opm_ptr->score != NULL)
	{
	    obj->score = (xsw_score_struct *)malloc(
		sizeof(xsw_score_struct)
	    );
	    memcpy(
		obj->score,
		opm_ptr->score,
		sizeof(xsw_score_struct)
	    );
	}
	else
	{
	    obj->score = NULL;
	}

	/* Economy */
	if(opm_ptr->eco != NULL)
	{
	    obj->eco = (xsw_ecodata_struct *)malloc(
		sizeof(xsw_ecodata_struct)
	    );
	    memcpy(
		obj->eco,
		opm_ptr->eco,
		sizeof(xsw_ecodata_struct)
	    );
	    if(obj->eco != NULL)
	    {
		/* Allocate product pointers */
		obj->eco->product = (xsw_ecoproduct_struct **)calloc(
		    1,
		    opm_ptr->eco->total_products * sizeof(xsw_ecoproduct_struct *)
		);
		if(obj->eco->product == NULL)
		{
		    obj->eco->total_products = 0;
		}
		else
		{
		    obj->eco->total_products = opm_ptr->eco->total_products;

		    /* Copy each product */
		    for(i = 0; i < opm_ptr->eco->total_products; i++)
		    {
		        if(opm_ptr->eco->product[i] == NULL)
			    continue;

			obj->eco->product[i] = (xsw_ecoproduct_struct *)malloc(
			    sizeof(xsw_ecoproduct_struct)
			);
			memcpy(
			    obj->eco->product[i],
			    opm_ptr->eco->product[i],
			    sizeof(xsw_ecoproduct_struct)
			);
		    }
		}

	    }
	}
	else
	{
	    obj->eco = NULL;
	}

	return(0);
}

int OPMModelObject(int obj_num, int opm_num)
{
	xsw_object_struct *obj;
	xsw_object_struct *opm_ptr;

	if(DBIsObjectGarbage(obj_num)) 
	    return(-1);

	obj = xsw_object[obj_num];

	if(OPMIsGarbage(opm_num))
	    return(-1);

	opm_ptr = opm[opm_num];

	return(OPMModelObjectPtr(obj, opm_ptr));
}
