#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "../include/swnetcodes.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#ifdef PLUGIN_SUPPORT
#include "plugins.h"
#endif
#include "swserv.h"
#include "weapons.h"

#include "../include/swexporteventscodes.h"

extern void NetFireWeaponSyncObjectValues(
        xsw_object_struct *obj, int obj_num,
        int wep_num
);

static int REngFreqCmp(float f1, float f2);
static double REngTurnAngle(
	double theta_cur, double theta_to, double theta_limit
);

static void REngMove(
	const int obj_num,
	const char allow_sect_change, const char is_tractored
);

static void REngHeading(const int obj_num);
static void REngThrottle(const int obj_num);
static void REngThrust(const int obj_num);
static void REngLocation(const int obj_num);
static void REngVisibility(const int obj_num);
static void REngShieldVisibility(const int obj_num);
static int REngMortality(const int obj_num);
static void REngAnimated(const int obj_num);
static int REngAutoDestruct(const int obj_num);
static void REngWeaponsLock(const int obj_num);
static void REngTractorBeam(const int obj_num);
static void REngIntercept(const int obj_num);
static void REngPowerCore(const int obj_num);
static void REngDamageControl(const int obj_num);
static void REngDoRespawn(const int obj_num);
static int REngDoHit(
	const int wobj_num, const int tobj_num,
	const double power		/* Of weapon object */
);
static int REngContactCheck(const int obj_num);
static int REngAI(const int obj_num);

static int REngUpdateObject(const int obj_num);

int REngInit(void);
void REngManage(void);
void REngShutdown(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the frequencies are the same up to 2 decimal
 *	accuracy.
 */
static int REngFreqCmp(float f1, float f2)
{
	float df;

	/* Sanitize f1 and f2 */
	if(f1 < SW_FREQ_MIN)
	    f1 = SW_FREQ_MIN;
	if(f1 > SW_FREQ_MAX)
	    f1 = SW_FREQ_MAX;

	if(f2 < SW_FREQ_MIN)
	    f2 = SW_FREQ_MIN;
	if(f2 > SW_FREQ_MAX)
	    f2 = SW_FREQ_MAX;

	/* Get absolute value difference */
	df = f2 - f1;
	if(df < 0)
	    df *= -1;

	/* Compare up to 2 decimal places */
	if(df < 0.01)
	    return(1);
	else
	    return(0);
}


/*
 *	Turns the angle theta_cur towards the angle theta_to but
 *	no more than theta_limit (all units in radians).
 */
static double REngTurnAngle(
	double theta_cur, double theta_to, double theta_limit
)
{
	const double limit_tc = theta_limit * time_compensation;
	double delta = MuAngleDelta(theta_to, theta_cur);

	/* Limit delta angle */
	if(delta > limit_tc)
	    delta = limit_tc;
	else if(delta < -limit_tc) 
	    delta = -limit_tc;

	/* Turn current angle */
	theta_cur += delta;

	return(SANITIZERADIANS(theta_cur));
}


/*
 *      Moves the object by calculating one quantom shift
 *      of its 3D coordinates.
 * 
 *      If allow_sect_change is true then if the object moves out
 *      if the coordinate bounds, it will change sectors and warp to
 *      the other side.
 */
static void REngMove(
	const int obj_num,
	const char allow_sect_change, const char is_tractored
)
{
	char ch_sect_px, ch_sect_nx, ch_sect_py, ch_sect_ny;
	double thrust_dir, thrust_mag, thrust_magc;
	double velocity_mag, velocity_magc;
	double x_vel, y_vel, z_vel;      /* Velocity delta */
	double x_velc, y_velc, z_velc;   /* Velocity delta /w time compensation */
	double x_thr, y_thr, z_thr;      /* Thrust */
	double x_thrc, y_thrc, z_thrc;   /* Thrust delta /w time compensation */
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip non-moving types of object if this object is not being
	 * tractored
	 */
	if(!is_tractored)
	{
	    if((obj->type == SW_OBJ_TYPE_STATIC) ||
	       (obj->type == SW_OBJ_TYPE_PLAYER) ||
	       (obj->type == SW_OBJ_TYPE_HOME) ||
	       (obj->type == SW_OBJ_TYPE_AREA) ||
	       (obj->type == SW_OBJ_TYPE_ANIMATED) ||
	       (obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	       (obj->type == SW_OBJ_TYPE_ELINK)
	    )
		return;
	}

	/* Stream weapons have the location of their owner */
	if((obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	{
	    const int owner_object = obj->owner;
	    if(!DBIsObjectGarbage(owner_object))
	    {
		xsw_object_struct *owner_obj = xsw_object[owner_object];
		obj->sect_x = owner_obj->sect_x;
		obj->sect_y = owner_obj->sect_y;
		obj->sect_z = owner_obj->sect_z;
		obj->x = owner_obj->x;
		obj->y = owner_obj->y;
		obj->z = owner_obj->z;
	    }
	    return;
	}


	/* Begin standard object movement */

	/* Sanitize velocity heading */
	obj->velocity_heading = SANITIZERADIANS(obj->velocity_heading);

	/* Calculate movement from velocity (momentum).
	 * Fetch velocity_mag, with compensation for lapsed time.
	 */
	velocity_magc = obj->velocity * time_compensation;
	velocity_mag = obj->velocity;

	/* Get momentum (velocity) vector compoents (with and
	 * without time compensation) from momentum unit vector
	 * compoents.
	 */
	x_velc = obj->momentum_vector_compoent.i * velocity_magc;
	y_velc = obj->momentum_vector_compoent.j * velocity_magc;
	z_velc = obj->momentum_vector_compoent.k * velocity_magc;

	x_vel = obj->momentum_vector_compoent.i * velocity_mag;
	y_vel = obj->momentum_vector_compoent.j * velocity_mag;
	z_vel = obj->momentum_vector_compoent.k * velocity_mag;


	/* Calculate (not object relative) thrust direction */
	thrust_dir = SANITIZERADIANS(
	    obj->heading +
	    obj->thrust_dir +
	    (1.0 * PI)
	);


	/* Calculate thrust magnitude, taking into account the
	 * object's thrust power, current power, and power
	 * purity
	 *
	 * The power coeff is in the range of 0.5 to 1.0, this is
	 * due to the fact that if there is no power the vessel cannot
	 * move and many players complain that this is unfair
	 */
	if(obj->power_max > 0.0)
	{
	    const double power_coeff = (obj->power /
		obj->power_max / 2.0) + 0.5;
	    thrust_mag = obj->thrust * obj->power_purity *
		power_coeff;
	}
	else
	{
	    thrust_mag = obj->thrust * obj->power_purity;
	}

	/* Calculate thrust magnitude with time_compensation */
	thrust_magc = thrust_mag * time_compensation;

	/* Get thrust vector compoents with and without time
	 * compensation
	 */
	x_thrc = thrust_magc * sin(thrust_dir);
	y_thrc = thrust_magc * cos(thrust_dir);
	z_thrc = 0.0;

	x_thr = thrust_mag * sin(thrust_dir);
	y_thr = thrust_mag * cos(thrust_dir);
	z_thr = 0.0;


	/* Move object by adding the calculated velocity and thrust
	 * vector compoents.
	 */
	obj->x += x_velc + x_thrc;
	obj->y += y_velc + y_thrc;
/*      obj->z += z_velc + z_thrc; */
/* Force set z to 0 */
	obj->z = 0.0;


	/* Do sector change as needed */
	if(allow_sect_change)
	{
	    ch_sect_px = 0;
	    ch_sect_nx = 0;
	    ch_sect_py = 0;
	    ch_sect_ny = 0;

	    if(obj->x < sector_legend.x_min)
	    {
		obj->x += sector_legend.x_len;
		obj->sect_x -= 1;
		ch_sect_nx = 1;         /* Negative sect x change */
	        NetSendObjectSect(-1, obj_num);
	    }
	    else if(obj->x > sector_legend.x_max)
	    {
		obj->x -= sector_legend.x_len;
		obj->sect_x += 1;
		ch_sect_px = 1;         /* Positive sect x change */
		NetSendObjectSect(-1, obj_num);
	    }

	    if(obj->y < sector_legend.y_min)
	    {
		obj->y += sector_legend.y_len;
		obj->sect_y -= 1;
		ch_sect_ny = 1;         /* Negative sect y change */
		NetSendObjectSect(-1, obj_num);
	    }
	    else if(obj->y > sector_legend.y_max)
	    {
		obj->y -= sector_legend.y_len;
		obj->sect_y += 1;
		ch_sect_py = 1;         /* Positive sect y change */
		NetSendObjectSect(-1, obj_num);
	    }

#if 0
	    if(obj->z < sector_legend.z_min)
	    {
		obj->z += sector_legend.z_len;
		obj->sect_z -= 1;
		ch_sect_nz = 1;
		NetSendObjectSect(-1, obj_num);
	    }
	    else if(obj->z > sector_legend.z_max)
	    {
		obj->z -= sector_legend.z_len;
		obj->sect_z += 1;
		ch_sect_pz = 1;
		NetSendObjectSect(-1, obj_num);
	    }
#endif

	    /* Change sector for tractored objects */ 
	    if(ch_sect_nx || ch_sect_px ||
	       ch_sect_ny || ch_sect_py
/*
	       ch_sect_nz || ch_sect_pz
 */
	    )   
	    {
		int i, trac_obj_num;
		xsw_object_struct *trac_obj;

		for(i = 0; i < obj->total_tractored_objects; i++)
		{
		    trac_obj_num = obj->tractored_object[i];
		    if(DBIsObjectGarbage(trac_obj_num))
			continue;

		    trac_obj = xsw_object[trac_obj_num];

		    trac_obj->sect_x = obj->sect_x;
		    trac_obj->sect_y = obj->sect_y;
/*                  trac_obj->sect_z = obj->sect_z; */

		    /* Move tractored object coordinates to exactly
		     * where the source object is
		     */
		    if(ch_sect_nx)
			trac_obj->x = obj->x;
		    if(ch_sect_px)
			trac_obj->x = obj->x;

		    if(ch_sect_ny)
			trac_obj->y = obj->y;
		    if(ch_sect_py)
			trac_obj->y = obj->y;


#if 0
		    if(ch_sect_nx)
			trac_obj->x += sector_legend.x_len;
		    if(ch_sect_px)
			trac_obj->x -= sector_legend.x_len;

		    if(ch_sect_ny)
			trac_obj->y += sector_legend.y_len;
		    if(ch_sect_py)
			trac_obj->y -= sector_legend.y_len;

		    if(ch_sect_nz)
			trac_obj->z += sector_legend.z_len;
		    if(ch_sect_pz)
			trac_obj->z -= sector_legend.z_len;
#endif
		    NetSendObjectSect(-1, trac_obj_num);
		}
	    }
	}


	/* Record new velocity and velocity_heading on object
	 *
	 * This value must not have the time compensation modifications
	 * to it
	 */
	obj->velocity_heading = MuCoordinateDeltaVector(
	    x_vel + x_thr,
	    y_vel + y_thr
	);
	MuSetUnitVector2D(
	    &obj->momentum_vector_compoent,
	    obj->velocity_heading
	);

	obj->velocity = Mu3DDistance(
	    x_vel + x_thr,
	    y_vel + y_thr,
	    0.0
	);
}

/*
 *      Updates the heading of the object.
 *
 *	For intercepts.
 */
static void REngHeading(const int obj_num)
{
	char oversight = 1;
	int tar_obj_num;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip objects that do not turn */
	if((obj->type == SW_OBJ_TYPE_STATIC) ||
	   (obj->type == SW_OBJ_TYPE_HOME) ||
	   (obj->type == SW_OBJ_TYPE_AREA) ||
	   (obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (obj->type == SW_OBJ_TYPE_ELINK)
	)
	   return;

	/* Skip player objects that are not intercepting anything,
	 * when they are not intercepting, their heading is updated by
	 * the client
	 */
	if((obj->type == SW_OBJ_TYPE_PLAYER) &&
	   (obj->intercepting_object < 0)
	)
	   return;

	/* Disable oversight calculations for stream weapons */
	if(obj->type == SW_OBJ_TYPE_STREAMWEAPON)
	    oversight = 0;

	/* Is object intercepting something? */
	tar_obj_num = obj->intercepting_object;
	if(!DBIsObjectGarbage(tar_obj_num))
	{
	    /* Get rel_heading to the object we're intercepting */
	    double bearing;
	    xsw_object_struct *tar_obj = xsw_object[tar_obj_num];

	    if(Mu3DInSameSectorPtr(obj, tar_obj))
	    {
		/* In same sector */
		double src_mx, src_my, tar_mx, tar_my, delta;

		/* Calculate the distance between the two objects */
		const double d = Mu3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		);

		/* Calculate delta time (in cycles) for the source
		 * object to reach the target object. A bunch
		 * of errors may be introduced if this is the
		 * start of an intercept, but that shouldn't
		 * be too much as the intercept path neatens up.
		 */
		const double dt = (obj->velocity > 0.0) ?
		    (d / obj->velocity) : 0.0;

		/* Get current position of source object */
		src_mx = obj->x;
		src_my = obj->y;

		if(oversight)
		{
		    /* Get calculated position of target object using
		     * the target object's current momentum and multiply
		     * that by the time (in cycles) it will take for
		     * the source object to reach the target object
		     * at both object's current positions and
		     * velocities
		     */
		    tar_mx = tar_obj->x +
		        (tar_obj->momentum_vector_compoent.i *
		        tar_obj->velocity * dt);
		    tar_my = tar_obj->y +
		        (tar_obj->momentum_vector_compoent.j *
		        tar_obj->velocity * dt);
		}
		else
		{
		    tar_mx = tar_obj->x;
		    tar_my = tar_obj->y;
		}

		/* Calculate bearing from the source object's current
		 * position to target object's predicted position
		 */
		bearing = MuCoordinateDeltaVector(
		    tar_mx - src_mx,
		    tar_my - src_my
		);

		if(oversight)
		{
		    /* Calculate the shorter delta between the bearing
		     * and the source object's velocity then apply it
		     * to the bearing so it will compensate for the
		     * source object's velocity
		     */
		    delta = MuAngleDelta(bearing, obj->velocity_heading);
		    if((delta <= (0.40 * PI)) && (delta >= (-0.40 * PI)))
			bearing = SANITIZERADIANS(bearing + delta);
		}
#if 0
printf("theta1=%.0f theta2=%.0f bearing=%.0f\n",
 RADTODEG(theta1), RADTODEG(theta2), RADTODEG(bearing)
);
#endif
	    }
	    else
	    {
		/* In different sectors */
		bearing = MuCoordinateDeltaVector(
		    tar_obj->sect_x - obj->sect_x,
		    tar_obj->sect_y - obj->sect_y
	        );
	    }

	    /* Turn heading */
	    obj->heading = REngTurnAngle(
		obj->heading, bearing,
		obj->turnrate * 100.0
	    );
	}

	/* Sanitize heading */
	obj->heading = SANITIZERADIANS(obj->heading);
	MuSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
}


/*
 *	Nothing is done here, this function is just for client side.
 */
static void REngThrottle(const int obj_num)
{
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip static objects */
	if(obj->type == SW_OBJ_TYPE_STATIC)
	   return;

	/* Skip player objects, their throttle is updated on client */
	if(obj->type == SW_OBJ_TYPE_PLAYER)
	   return;

}

/*
 *	Thrust, antimatter consumption, drag, and other calculations
 *	of velocity are performed here.
 */
static void REngThrust(const int obj_num)
{
	double throttle, thrust_power;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* We want to update all types of objects, since there is
	 * no telling if an object should have power decrementation
	 *
	 * A civilization on a planet for instance, would use up
	 * antimatter, so throttle applied would simulate antimatter
	 * usage
	 */

	/* Get throttle and thrust_power */
	throttle = obj->throttle;
	if(throttle > 0.0)
	{
	    /* Get thrust power */
	    thrust_power = obj->thrust_power;

	    /* Consume antimatter based on throttle */
	    obj->antimatter -= throttle * thrust_power * time_compensation;
	    if(obj->antimatter < 0.0)
		obj->antimatter = 0.0;

	    /* Thrust output should be 0 if there is no antimatter left */
	    if(obj->antimatter <= 0.0)
		thrust_power = 0.0;
	}
	else
	{
	    thrust_power = 0.0;
	    throttle = 0.0;
	}


	/* Calculate the actual thrust output based on the
	 * throttle and thrust power
	 */
	obj->thrust = MAX(
	    throttle * thrust_power * time_compensation,
	    0.0
	);

	/* Add drag to the velocity based on the object's current
	 * velocity, velocity max, and thrust power
	 */   
	if((obj->velocity_max > 0.0) && (obj->thrust_power > 0.0))
	    obj->velocity -= (obj->velocity /
		obj->velocity_max) * obj->thrust_power *
		time_compensation;
	else 
	    obj->velocity = 0.0;

	/* If velocity is such a small number then set it to 0 */
	if(obj->velocity < 0.00003)
	    obj->velocity = 0.0;

	/* Consume antimatter if cloak is up */
	if(obj->cloak_state == CLOAK_STATE_UP)
	{
	    const double size_ru = (double)obj->size / 1000.0;

	    /* Consume antimatter based on the object's size */
	    obj->antimatter -= size_ru * sysparm.cloak_am_rate *
		time_compensation;
	    if(obj->antimatter < 0.0)
		obj->antimatter = 0.0;

	    /* Lower cloak if out of antimatter */
	    if(obj->antimatter == 0.0)
		obj->cloak_state = CLOAK_STATE_DOWN;
	}
}


/*
 *	Updates the object's location type.
 */
static void REngLocation(const int obj_num)
{
	int i;
	double d = 0.0;			/* In screen units */
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj,
				*area_obj = NULL;

	/* Skip area objects */
	if(obj->type == SW_OBJ_TYPE_AREA)
	    return;

	/* Iterate through the objects list backwards */
	for(i = total_objects - 1; i >= 0; i--)
	{
	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    /* Check only objects of type SW_OBJ_TYPE_AREA */
	    if(tar_obj->type != SW_OBJ_TYPE_AREA)
		continue;

	    /* Skip itself */
	    if(tar_obj == obj)
		continue;

	    /* Not in the same sector? */
	    if((obj->sect_x != tar_obj->sect_x) ||
	       (obj->sect_y != tar_obj->sect_y) ||
	       (obj->sect_z != tar_obj->sect_z)
	    )
		continue;

	    /* Calculate the distance in screen units */
	    d = Mu3DDistance(
		tar_obj->x - obj->x,
		tar_obj->y - obj->y,
		tar_obj->z - obj->z
	    ) * 1000.0;
	    if(d > tar_obj->size)
		continue;

	    area_obj = tar_obj;
	    break;
	}

	/* Is this object in an area? */
	if(area_obj != NULL)
	{
	    /* Was the object previously not in this area type? */
	    if(obj->loc_type != area_obj->loc_type)
	    {
		switch(area_obj->loc_type)
		{
		  case SW_LOC_TYPE_SPACE_TERRITORY:
/* TODO */
		    break;

		  case SW_LOC_TYPE_NEBULA:
		    break;
		}

		/* Set new location type */
	        obj->loc_type = area_obj->loc_type;
	    }
	}
	else
	{
	    /* Object is now in normal space, check if it wasn't */
	    if(obj->loc_type != SW_LOC_TYPE_SPACE)
	    {
		/* Set new location type */
		obj->loc_type = SW_LOC_TYPE_SPACE;
	    }
	}
}


/*
 *	Updates the object's current visiblity by checking its
 *	shield, cloak and location.
 */
static void REngVisibility(const int obj_num)
{
	double v;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Are shields up? */
	if(obj->shield_state == SHIELD_STATE_UP)
	{
	    if(obj->cur_visibility < 1)
	    {
		obj->cur_visibility = MIN(
		    obj->cur_visibility + (0.01 * time_compensation),
		    1
		);
	    }
	}            
	/* Is cloak up? */
	else if(obj->cloak_state == CLOAK_STATE_UP)
	{
	    if(obj->power_max > 0)
		v = obj->visibility - (obj->cloak_strength *
		    obj->power_purity * obj->power /
		    obj->power_max);
	    else
		v = obj->visibility;
		    
	    if(obj->cur_visibility > v)
	    {
		obj->cur_visibility = MAX(
		    obj->cur_visibility - (0.01 * time_compensation),
		    v
		);
	    }
	}
	/* Adjust visibility to normal */
	else
	{
	    if(obj->cur_visibility > obj->visibility)
	    {
		obj->cur_visibility = MAX(   
		    obj->cur_visibility - (0.01 * time_compensation),
		    obj->visibility
		);
	    }
	    else if(obj->cur_visibility < obj->visibility)
	    {   
		obj->cur_visibility = MIN(
		    obj->cur_visibility + (0.01 * time_compensation),
		    obj->visibility
		);
	    }
	}

	/* Sanitize */
	if(obj->cur_visibility < 0)
	    obj->cur_visibility = 0;
 	else if(obj->cur_visibility > 1)
	     obj->cur_visibility = 1;
}

/*
 *      Updates the object's shield visibility.
 */
static void REngShieldVisibility(const int obj_num)
{
	double x;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip if shields not visable */
	if(obj->shield_visibility <= 0)
	    return;

	/* Calculate deduction */
	x = (double)((double)SW_CYCLE_LAPSE_MS /
	    (double)SHIELD_VISIBILITY_INTERVAL) * time_compensation; 

	/* Reduce shield visibility */
	obj->shield_visibility -= x;

	/* Sanitize shield_visibility */
	if(obj->shield_visibility < 0)
	    obj->shield_visibility = 0;
}

/*
 *	Checks if the object has lived past its life span and
 *	recycles it as needed.
 *
 *	Returns true if the object has been recycled.
 */
static int REngMortality(const int obj_num)
{
	char object_must_die = 0;
	xsw_object_struct *obj = xsw_object[obj_num];
	xsw_animation_struct *anim;

	/* Handle mortality by object type */
	switch(obj->type)
	{
	  case SW_OBJ_TYPE_ANIMATED:
	    /* Animated objects die when they have a limited (non zero)
	     * cycle_times and that the number of times they have
	     * cycled exceeds that
	     */
	    anim = &obj->animation;
	    if(anim->cycle_times > 0)
	    {
		if(anim->cycle_count >= anim->cycle_times)
		    object_must_die = 1;
	    }
	    break;

	  case SW_OBJ_TYPE_PLAYER:
	    /* Player objects do not die regardless */
	    break;

	  default:
	    /* All other object types, handle by life span
	     *
	     * If the object has a defined (non-zero and positive) life
	     * span then their birth time is checked with current time
	     * to see if they must die
	     */
	    if((obj->lifespan > 0l) && (obj->birth_time > 0l))
	    {
		/* Calculate age */
		const time_t dt = cur_millitime - obj->birth_time;

		/* Age positive and exceeded life span? */
		if((dt < 0l) || (dt > obj->lifespan))
		    object_must_die = 1;
	    }
	    break;
	}


	/* Has this object died and we must recycle it? */
	if(object_must_die)
	{
	    /* Recycle this object */
	    DBRecycleObject(obj_num);

	    /* Notify all connections about this recycle */
	    NetSendRecycleObject(-1, obj_num);

	    return(1);
	}
	else
	{
	    return(0);
	}
}

/*
 *	Updates animated object's animation frame position.
 */
static void REngAnimated(const int obj_num)
{
	time_t dt;
	xsw_object_struct *obj = xsw_object[obj_num];
	xsw_animation_struct *anim = &obj->animation;

	/* Time to increment frame? */
	dt = anim->last_interval + anim->interval;
	if(dt <= cur_millitime)
	{
	    anim->current_frame++;

	    /* Looped? */
	    if(anim->current_frame >= anim->total_frames)
	    {
		anim->current_frame = 0;

		/* Increment cycled times only if there is a cycle limit */
		if(anim->cycle_times > 0)
		    anim->cycle_count++;
	    }

	    anim->last_interval = cur_millitime;
	}
}

/*
 *	Auto-destruct.
 *
 *	Returns true if an object has been recycled.
 */
static int REngAutoDestruct(const int obj_num)
{
	int objects_recycled = 0;
	xsw_object_struct *obj = xsw_object[obj_num];

	if(obj->auto_destruct_next == 0l)
	    return(objects_recycled);

	/* Time to auto-destruct? */
	if(obj->auto_destruct_next <= cur_systime)
	{
	    int opm_num, debris_num, explosion_num;
	    char	obj_name[SW_OBJ_NAME_MAX + 80],
			debris_name[SW_OBJ_NAME_MAX + 80];
	    xsw_object_struct *opm_ptr;

	    /* Reset auto-destruct marker */
	    obj->auto_destruct_next = 0l;

	    /* If the object has power left then apply damage to
	     * nearby objects
	     */
	    if(obj->power > 0.0)
	    {
		const double hit_radius = 2500.0;
		int i;
		double	distance,
			power_coeff,
			power = obj->power;
		xsw_object_struct *tar_obj;

		for(i = 0; i < total_objects; i++)
		{
		    tar_obj = xsw_object[i];
		    if(tar_obj == NULL)
			continue;

		    if(i == obj_num)
			continue;

		    if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
			continue;

		    /* Only hit players and controlled objects */
		    if((tar_obj->type != SW_OBJ_TYPE_CONTROLLED) &&
		       (tar_obj->type != SW_OBJ_TYPE_PLAYER)
		    )
			continue;

		    /* Not in the same sector? */
		    if(!Mu3DInSameSectorPtr(obj, tar_obj))
			continue;

		    /* Calculate distance apart in screen units */
		    distance = Mu3DDistance(
			tar_obj->x - obj->x,
			tar_obj->y - obj->y,
			tar_obj->z - obj->z
		    ) * 1000.0;
		    if(distance > hit_radius)
			continue;

		    /* Calculate power coefficient to determine how
		     * powerful is the self-destruct explosion based
		     * on the distance apart
		     */
		    power_coeff = MIN(
			((1.0 - (distance / hit_radius)) * 2.0),
			1.0
		    );

		    /* Perform hit on this object */
		    objects_recycled += REngDoHit(
			obj_num, i,
			power * power_coeff
		    );
		    if(DBIsObjectGarbage(obj_num))
			return(objects_recycled);
		}
	    }

	    /* Get the object's name */
	    strncpy(
		obj_name, DBGetFormalNameStr(obj_num),
		sizeof(obj_name)
	    );
	    obj_name[sizeof(obj_name) - 1] = '\0';

	    /* Create debris */
	    if(*(obj->name) != '\0')
		sprintf(
		    debris_name, "Debris of %s",
		    obj->name
		);
	    else
		strncpy(
		    debris_name, "Debris",
		    sizeof(debris_name)
		);
	    opm_num = sws_std_opm.debris_num;
	    opm_ptr = OPMIsGarbage(opm_num) ?
		NULL : opm[opm_num];
	    debris_num = DBCreateObjectByOPMPtr(
		opm_ptr,
		debris_name,
		SW_OBJ_TYPE_STATIC,
		unv_head.lost_found_owner,
		obj->sect_x, obj->sect_y, obj->sect_z,
		obj->x, obj->y, obj->z,
		obj->heading, obj->pitch, 0.0 * PI
	    );
	    if(!DBIsObjectGarbage(debris_num))
	    {
		xsw_object_struct *debris = xsw_object[debris_num];

		strncpy(
		    debris->empire, obj->empire,
		    sizeof(debris->empire)
		);
		debris->empire[sizeof(debris->empire) - 1] = '\0';

		NetSendCreateObject(-1, debris_num);
		NetSendObjectName(-1, debris_num);
	    }

	    /* Create big explosion */
	    opm_num = sws_std_opm.explosion_large_num;
	    opm_ptr = OPMIsGarbage(opm_num) ?
		NULL : opm[opm_num];
	    explosion_num = DBCreateObjectByOPMPtr(
		opm_ptr,
		NULL,
		SW_OBJ_TYPE_ANIMATED,
		unv_head.lost_found_owner,
		obj->sect_x, obj->sect_y, obj->sect_z,
		obj->x, obj->y, obj->z,
		obj->heading, obj->pitch, 0.0 * PI
	    );
	    if(!DBIsObjectGarbage(explosion_num))
		NetSendCreateObject(-1, explosion_num);

	    /* Is the object is not immune from being destroyed? */
	    if(obj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE)
	    {
		/* Object can be destroyed */

		if(1)
		{
		    int i;
		    char *s, loc_str[256];
		    sws_connection_struct *con;

#ifdef PLUGIN_SUPPORT
		    /* Notify all plugins */
		    for(i = 0; i < total_plugins; i++)
			PluginDestroyNotify(
			    i,
			    SW_DESTROY_REASON_SELF_DESTRUCT,
			    obj_num,	/* Destroyed object */
			    obj_num,	/* Object that caused destruction */
			    obj_num	/* Owner of weapon object */
			);
#endif

		    /* Format location string */
		    UNVLocationFormatString(
			loc_str,
			&obj->sect_x, &obj->sect_y, &obj->sect_z,
			&obj->x, &obj->y, &obj->z,
			sizeof(loc_str)
		    );

		    switch(obj->type)
		    {
		      case SW_OBJ_TYPE_CONTROLLED:
		      case SW_OBJ_TYPE_PLAYER:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(obj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed due to power-core overload (auto-destruct) at %s.",
				obj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(obj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed due to power-core overload (auto-destruct) at %s.",
			    obj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDPLAYER,
			    s
			);
			free(s);
		        break;

		      default:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(obj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed at %s.",
				obj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(obj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed at %s.",
			    obj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDOBJECT,
			    s
			);
			free(s);
			break;
		    }

		    /* Send destroy notify to all connections in
		     * the same sector
		     */
		    for(i = 0; i < total_connections; i++)
		    {
			con = connection[i];
			if(con == NULL)
			    continue;

			if((con->socket < 0) || (con->object_num < 0))
			    continue;

			/* Not in the same sector? */
			if(!Mu3DInSameSector(
			    xsw_object, total_objects,
			    con->object_num, obj_num
			))
			    continue;

			/* Send destroy notify to this connection */
/*			NetSendObjectName(i, obj_num); */
		        NetSendNotifyDestroy(
			    i,
			    SW_DESTROY_REASON_SELF_DESTRUCT,	/* Reason is self destruct */
			    obj_num,	/* Destroyed object */
			    obj_num,	/* Object that caused destruction */
			    obj_num	/* Owner of weapon object */
			);
		    }
		}

		/* Check if the destroyed object is a player and if
		 * players should be respawned, otherwise recycle
		 * the destroyed object
		 *
		 * Note that after this point, obj should be considered
		 * invalid. It can still be used as cross reference but
		 * it must never be referenced implicitily.
		 */
		if((obj->type == SW_OBJ_TYPE_PLAYER) &&
		   sysparm.respawn_players
		)
		{
		    REngDoRespawn(obj_num);
		}
		else
		{
		    DBRecycleObject(obj_num);
		    NetSendRecycleObject(-1, obj_num);
		    objects_recycled++;
		}
	    }
	    else
	    {
		/* Sanitize hp on obj (since we didn't recycle it */
		if(obj->hp < 0.0)
		    obj->hp = 0.0;
	    }
	}

	return(objects_recycled);
}

/*
 *	Updates weapons lock on an object.
 */
static void REngWeaponsLock(const int obj_num)
{
	int tar_obj_num;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;

	/* Get the locked on object and check if it is valid */
	tar_obj_num = obj->locked_on;
	if(DBIsObjectGarbage(tar_obj_num))
	{
	    /* Locked on object is now garbage, unlock */
	    obj->locked_on = -1;
	    return;
	}

	tar_obj = xsw_object[tar_obj_num];

	/* Weapons objects maintain their lock on the target regardless
	 * of the target's visibility and distance
	 */
	if((obj->type == SW_OBJ_TYPE_WEAPON) ||
	   (obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	    return;

	/* Check distance */
	if(Mu3DInRange(
	    xsw_object, total_objects,
	    obj_num, tar_obj_num,
	    obj->scanner_range *
	    DBGetObjectVisibilityPtr(tar_obj)
	))
	{
	    /* Object still in range */

	}
	else
	{
	    /* Object out of scanner range, unlock */
	    obj->locked_on = -1;
	}
}


/*
 *	Updates an object's tractor beam.
 *
 *	Checks if the object is tractoring any objects, consumes
 *	antimatter, and moves each tractored object.
 */
static void REngTractorBeam(const int obj_num)
{
	int i;
	int trac_obj_num;
	double size_ru, tractor_beam_range = sysparm.tractor_beam_range;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;  

	/* Skip if not tractoring anything */
	if(obj->total_tractored_objects <= 0)
	    return;

	/* Unlock tractor beam if object is out of antimatter */
	if(obj->antimatter <= 0.0)
	{
	    for(i = 0; i < obj->total_tractored_objects; i++)
	    {
		DBObjectUntractor(obj_num, obj->tractored_object[i]);
		NetSendTractorBeamLock(-1, obj_num, -1);
	    }
	    return;
	}

	/* Move each tractored object */
	for(i = 0; i < obj->total_tractored_objects; i++)
	{
	    trac_obj_num = obj->tractored_object[i];
	    if(trac_obj_num < 0)
		continue;

	    /* Check if trac_obj_num is not -1 but garbage */
	    if(DBIsObjectGarbage(trac_obj_num))
	    {
		DBObjectUntractor(obj_num, trac_obj_num);
		NetSendTractorBeamLock(-1, obj_num, -1);
	        continue;
	    }

	    tar_obj = xsw_object[trac_obj_num];

	    /* Check if the target object is tractorable */
	    if(!DBIsObjectTractorablePtr(obj, tar_obj))
	    {
		DBObjectUntractor(obj_num, trac_obj_num);
	        NetSendTractorBeamLock(-1, obj_num, -1);
		continue;
	    }

	    /* Consume antimatter */
	    size_ru = (double)tar_obj->size / 1000.0;
	    obj->antimatter -= size_ru * sysparm.tractor_beam_am_rate *
		time_compensation;
	    if(obj->antimatter < 0.0)
	        obj->antimatter = 0.0;

	    /* Move the tractored object? */
	    if(1)
	    {
		/* Calculate the delta position behind the tractoring
		 * object as dx, dy, and dz
		 *
		 * Record the tractored object's original maximum
		 * velocity since it will have to be modified before
		 * it is moved and then restored afterwards
		 */
		double	theta1, theta2, theta3,
			mag1, mag2, mag3,
			old_velocity_max = tar_obj->velocity_max,
			d = (double)obj->size * 2.0 / 1000.0,
			tx = -MuPolarRotX(obj->heading, d),
			ty = -MuPolarRotY(obj->heading, d),
			dx = obj->x - tar_obj->x + tx,
			dy = obj->y - tar_obj->y + ty,
			dz = obj->z - tar_obj->z;

	        d = Mu3DDistance(dx, dy, dz);

		/* theta1 and mag1 is the tractored object's heading
		 * and velocity
		 */
	        theta1 = tar_obj->velocity_heading;

		/* theta2 and mag2 is the heading and velocity from
		 * the tractored object to the tractoring object
		 * (or the direction of the tractor beam's pull)
		 */
	        theta2 = MuCoordinateDeltaVector(dx, dy);

		/* Calculate tractor beam strength mag2 by taking
		 * the coefficient of the distance away the objects
		 * are from each other
		 *
		 * The farther away they are, the greater the tractor
		 * beam strength
		 */
	        if(d < ((double)tar_obj->size * 2.0 / 1000.0))
		{
		    /* Slow down the tractored object's velocity */
		    if(obj->thrust_power > 0.0)
			mag1 = MAX(
			    tar_obj->velocity -
			    (obj->thrust_power * time_compensation),
			    0.0
			);
		    else
			mag1 = 0.0;
		    mag2 = 0.0;
		}
	        else
		{
		    mag1 = tar_obj->velocity;
	            mag2 = obj->velocity_max *
			MIN((d * 4.0 / tractor_beam_range), 1.0);
 		}

/* Note that objects that do not move will be more difficult to
 * tractor at higher velocities since their velocity will get set to
 * 0 after this call
 */

		/* Add target object's current velocity vector with
		 * the new velocity vector caused by the tractor beam
		 * pull
		 */
	        MuVectorAdd(
	            theta1, mag1,
	            theta2, mag2,
	            &theta3, &mag3
	        );

	        tar_obj->velocity_heading = theta3;
		MuSetUnitVector2D(
		    &tar_obj->momentum_vector_compoent,
		    tar_obj->velocity_heading
		);

	        tar_obj->velocity = mag3;

		/* Turn the tractored object to face the direction
		 * it is being pulled
		 */
		tar_obj->heading = REngTurnAngle(
		    tar_obj->heading, theta3,
		    tar_obj->turnrate * 100.0
		);

		/* Adjust the tractored object's maximum velocity
		 * as needed to accomidate the current velocity
		 */
	        if(tar_obj->velocity > tar_obj->velocity_max)
		    tar_obj->velocity_max = tar_obj->velocity;

	        /* Move the object */
	        REngMove(trac_obj_num, 0, 1);

	        /* Restore the tractored object's original maximum
		 * velocity
		 */
	        tar_obj->velocity_max = old_velocity_max;
	    }
	}
}


/*
 *	Only checks if object is in intercept range, does NOT
 *	adjust heading.
 */
static void REngIntercept(const int obj_num)
{
	int tar_obj_num;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip worm hole objects */
	if(obj->type == SW_OBJ_TYPE_WORMHOLE)
	    return;

	/* Get the target object */
	tar_obj_num = obj->intercepting_object;
	if(DBIsObjectGarbage(tar_obj_num))
	    return;

	/* Weapons objects maintain their lock on the target regardless
	 * of the target's visibility and distance
	 */
	if((obj->type == SW_OBJ_TYPE_WEAPON) ||
	   (obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)                                    
	    return;

	/* Use check if the target object is in range */
	if(UNVMatchInterceptByNumber(
	    xsw_object, total_objects,
	    obj_num, tar_obj_num
	) < 0)
	    obj->intercepting_object = -1;
}


/*
 *	Regernerates an object's power core.
 */
static void REngPowerCore(const int obj_num)
{
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Increase power in the object's power core */
	if(obj->power < obj->power_max)
	{
	    obj->power += (obj->core_efficency *
		obj->power_purity * time_compensation
	    );

	    if(obj->power > obj->power_max)
		obj->power = obj->power_max;
	}
}


/*
 *	Repairs object if the object has damage control on.
 */
static void REngDamageControl(const int obj_num)
{
	double power, power_purity, hp_inc;
	xsw_object_struct *obj = xsw_object[obj_num];

	/* Skip if damage control is not on */
	if(obj->damage_control != DMGCTL_STATE_ON)
	    return;

	/* Skip if hit points are already at max */
	if(obj->hp >= obj->hp_max)
	    return;

	/* Fetch and sanitize power */
	power = MAX(obj->power, 0);
	power_purity = MIN(obj->power_purity, 1);
	if(power_purity < 0)
	    power_purity = 0;

	/* Calculate hit points increment */
	hp_inc = MAX(
	    sysparm.dmg_ctl_rate * power * power_purity * time_compensation,
	    0.0
	);

	/* "Repair" hit points */
	obj->hp += hp_inc;
	if(obj->hp > obj->hp_max)
	{
	    hp_inc -= obj->hp - obj->hp_max;
	    if(hp_inc < 0.0)
		hp_inc = 0.0;

	    obj->hp = obj->hp_max;
	}

	/* Consume antimatter based on how much repairs were made */
/* Need better calculation */
	obj->antimatter -= hp_inc;
	if(obj->antimatter < 0.0)
	    obj->antimatter = 0.0;
}


/*
 *	Respawns the player.
 *
 *	The object will be moved back to the player starting position
 *	and its values will be restored (ie hit points and weapons).
 *
 *	Network updates, including a force pose and object values will
 *	be sent.
 */
static void REngDoRespawn(const int obj_num)
{
	int i;
	xsw_weapon_struct *wep;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Move to starting position */
	obj->heading = unv_head.player_start_heading;
	obj->pitch = unv_head.player_start_pitch;
	obj->bank = unv_head.player_start_bank;

	obj->sect_x = unv_head.player_start_sect_x;
	obj->sect_y = unv_head.player_start_sect_y;
	obj->sect_z = unv_head.player_start_sect_z;

	obj->x = unv_head.player_start_x;
	obj->y = unv_head.player_start_y;
	obj->z = unv_head.player_start_z;

	/* Restore hit points and power */
	obj->hp = obj->hp_max;
	obj->power = obj->power_max;
	if(obj->shield_state != SHIELD_STATE_NONE)
	    obj->shield_state = SHIELD_STATE_DOWN;

	obj->throttle = 0.0;
	obj->damage_control = DMGCTL_STATE_OFF;

	obj->locked_on = -1;
	obj->intercepting_object = -1;

	free(obj->tractored_object);
	obj->tractored_object = NULL;
	obj->total_tractored_objects = 0;

	/* Replenish weapons */
	obj->selected_weapon = 0;
	for(i = 0; i < obj->total_weapons; i++)
	{
	    wep = obj->weapons[i];
	    if(wep == NULL)
		continue;

/* Should wep->amount be reset to 0? */
	    wep->amount = wep->max;
	    wep->last_used = 0l;
	}

	/* Remove weapons lock of all other objects on this object */
	for(i = 0; i < total_objects; i++)
	{
	    if(DBIsObjectGarbage(i))
		continue;

	    if(xsw_object[i]->locked_on == obj_num)
		xsw_object[i]->locked_on = -1;
	}

	/* Notify all connections about respawn */
	NetSendObjectForcePose(-1, obj_num);
	NetSendObjectValues(-1, obj_num);
	for(i = 0; i < obj->total_weapons; i++)
	    NetSendWeaponValues(-1, obj_num, i);
	NetSendScore(-1, obj_num);
}

/*
 *	Hits the target object with the specified weapon object,
 *	updating the scores on the target object and the weapon
 *	object's owner, reducing the power and/or hit points on the
 *	target object, recycling the target object if it has been
 *	destroyed, create explosions and debris.
 *
 *	This function does not check if the weapon object actually
 *	hit the target object or not.
 *
 *	If the weapon object's type is SW_OBJ_TYPE_WEAPON
 *	(projectile or pulse weapon) then it will be recycled by this
 *	function.
 *
 *	The target object will be either respawned or recycled if its
 *	hit points goes to 0.
 *
 *	Returns the number of objects recycled.
 */
static int REngDoHit(
	const int wobj_num, const int tobj_num,
	const double power		/* Of weapon object */
)
{
	char	report_this = 1,	/* Log and export event */
		can_count_kill = 1;	/* Can count as a kill */
	int objects_recycled = 0, wobj_owner_num;
	double dx, theta;
	char	tobj_name[SW_OBJ_NAME_MAX + 80],
		wobj_name[SW_OBJ_NAME_MAX + 80],
		wobj_owner_name[SW_OBJ_NAME_MAX + 80];
	double	damage_to_structure = 0.0,
		damage_to_shield = 0.0;
	xsw_object_struct *wobj, *tobj;


	/* Check if weapon and target objects are valid */
	if(DBIsObjectGarbage(wobj_num))
	    return(0);
	if(DBIsObjectGarbage(tobj_num))
	    return(0);

	wobj = xsw_object[wobj_num];
	tobj = xsw_object[tobj_num];

	/* If no power is specified then do not hit anything */
	if(power <= 0.0)
	{
	    /* Recycle the weapon object if it is a projectile or
	     * pulse
	     */
	    if(wobj->type == SW_OBJ_TYPE_WEAPON)
	    {
		DBRecycleObject(wobj_num);
		NetSendRecycleObject(-1, wobj_num);
		objects_recycled++;
	    }
	    return(objects_recycled);
	}

	/* If report_destroyed_weapons is set to false, then
	 * check if target object is a weapon object.
	 * If it is, then set report_this to false so that
	 * no reporting of this hit will be made.
	 */
	if(!sysparm.report_destroyed_weapons)
	{
	    if((tobj->type == SW_OBJ_TYPE_WEAPON) ||
	       (tobj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	       (tobj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	    )
	    {
		report_this = 0;
		can_count_kill = 0;
	    }
	}


	/* Get owner of weapon object or the weapon object itself
	 * if it is a player or controlled
	 */
	if((wobj->type == SW_OBJ_TYPE_CONTROLLED) ||
	   (wobj->type == SW_OBJ_TYPE_PLAYER)
	)
	    wobj_owner_num = wobj_num;
	else
	    wobj_owner_num = wobj->owner;


	/* Get formal names of the weapon and target objects */
	strncpy(tobj_name, DBGetFormalNameStr(tobj_num), sizeof(tobj_name));
	tobj_name[sizeof(tobj_name) - 1] = '\0';

	strncpy(wobj_name, DBGetFormalNameStr(wobj_num), sizeof(wobj_name));
	wobj_name[sizeof(wobj_name) - 1] = '\0';

	strncpy(
	    wobj_owner_name,
	    DBGetFormalNameStr(wobj_owner_num),
	    sizeof(wobj_owner_name)
	);
	wobj_owner_name[sizeof(wobj_owner_name) - 1] = '\0';


	/* Log weapon object hitting target object (as needed) */
	if(report_this)
	{
	    char *s, loc_str[256];

	    /* Format location string */
	    UNVLocationFormatString(
		loc_str,
		&tobj->sect_x,
		&tobj->sect_y,
		&tobj->sect_z,
		&tobj->x,
		&tobj->y,
		&tobj->z,
		sizeof(loc_str)
	    );  

	    /* Log hit? */
	    if(sysparm.log_events)
	    {
		s = (char *)malloc(
		    STRLEN(wobj_name) +
		    STRLEN(tobj_name) +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
		    "%s hit %s at %s.",
		    wobj_name, tobj_name, loc_str
		);
		LogAppend(fname.log, s);
		free(s);
	    }

	    /* Export hit event */
	    switch(wobj->type)
	    {
	      case SW_OBJ_TYPE_WEAPON:
		s = (char *)malloc(
		    STRLEN(tobj_name) +
		    STRLEN(wobj_name) +
		    STRLEN(wobj_owner_name) +
		    STRLEN(loc_str) +
		    80
		);
	        sprintf(
		    s,
"%s hit by %s fired by %s at %s.",
		    tobj_name, wobj_name, wobj_owner_name, loc_str
	        );
	        ExportEvents(
		    fname.events_export,
		    EVST_TYPE_PROJECTILEWEAPON_HIT,
		    s
	        );
		free(s);
	        break;

	      case SW_OBJ_TYPE_STREAMWEAPON:
	      case SW_OBJ_TYPE_SPHEREWEAPON:
		s = (char *)malloc(
		    STRLEN(tobj_name) +
		    STRLEN(wobj_name) +
		    STRLEN(wobj_owner_name) +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
"%s hit by %s fired by %s at %s.",
		    tobj_name, wobj_name, wobj_owner_name, loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_STREAMWEAPON_HIT,
		    s
		);
		free(s);
		break;

	      default:
		s = (char *)malloc(
		    STRLEN(tobj_name) +
		    STRLEN(wobj_owner_name) +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
"%s hit by %s at %s.",
		    tobj_name, wobj_owner_name, loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_STREAMWEAPON_HIT,
		    s
		);
		free(s);
		break;
	    }
	}


	/* Make shields visiable on target object if the target
	 * object's shields are up and the shield frequencies on the
	 * the target object and weapon object do not match
	 *
	 * Also check if the weapon object can penetrate shields
	 */
	if((tobj->shield_state == SHIELD_STATE_UP) &&
	   (tobj->power_max > 0.0) &&
	   !REngFreqCmp(tobj->shield_frequency, wobj->shield_frequency) &&
	   !(wobj->ext_flags & SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD)
	)
	{
	    tobj->shield_visibility = tobj->power / tobj->power_max;

	    /* Send shield visibility to all connections */
	    NetSendShieldVis(-1, tobj_num);
	}

	/* Should the target object receive damage? */
	if((sysparm.homes_destroyable) ? 1 :
	   ((tobj->type != SW_OBJ_TYPE_HOME) &&
	    (tobj->type != SW_OBJ_TYPE_WORMHOLE) &&
	    (tobj->type != SW_OBJ_TYPE_ELINK)
	   )
	)
	{
	    /* Does the target object have its shields up, do the
	     * shield frequencies not match, and is the weapon object
	     * not able to penetrate shields?
	     */
	    if((tobj->shield_state == SHIELD_STATE_UP) &&
	       !REngFreqCmp(tobj->shield_frequency, wobj->shield_frequency) &&
	       !(wobj->ext_flags & SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD)
	    )
	    {
		/* Calculate extra damage to shields due to the target
		 * object's lack of power purity
		 */
		dx = MAX(
		    ((power * (2.0 - tobj->power_purity)) - power),
		    0.0
		);

		/* Add damage to shields from the weapon's power and the
		 * target object's lack of power purity
		 */
		damage_to_shield += power + dx;

	        tobj->power -= damage_to_shield;


		/* Shields went down, apply remaining damage to hull */
	        if(tobj->power < 0.0)
	        {
		    /* Target object power is now negative and represents
		     * damage to its structure
		     */
		    dx = MAX(
			(tobj->power * -1.0) - tobj->hp,
			0
		    );

		    damage_to_structure = MAX(
			(tobj->power * -1.0) - dx,
			0
		    );

		    tobj->hp -= damage_to_structure;

		    /* Undo too much damage to shields */
		    damage_to_shield += tobj->power;
	        }
	    }
	    else
	    {
		/* Target object's shields were down, no shields at all, or
		 * frequency of weapon matched frequency of target object's
		 * shields
		 */

		dx = MAX(power - tobj->hp, 0.0);

		damage_to_structure = MAX(power - dx, 0.0);

	        tobj->hp -= damage_to_structure;
	    }
	}
	/* Sanitize target object's power */
	if(tobj->power < 0.0)
	    tobj->power = 0.0;


	/* Damage to structure and shields have been calculated
	 * and processed, now add/subtract scores on each object
	 */

	/* Damage given by weapon object's owner */
	if(!DBIsObjectGarbage(wobj_owner_num))
	{
	    xsw_object_struct *wobj_owner = xsw_object[wobj_owner_num];

	    if((sysparm.homes_destroyable) ? 1 :
	       ((tobj->type != SW_OBJ_TYPE_HOME) &&
		(tobj->type != SW_OBJ_TYPE_WORMHOLE) &&
		(tobj->type != SW_OBJ_TYPE_ELINK)
	       )
	    )
	    {
		if(!UNVAllocScores(wobj_owner))
	        {
		    xsw_score_struct *score = wobj_owner->score;

		    /* Add damage_given to weapon object owner */
		    score->damage_given += damage_to_shield +
			damage_to_structure;

		    /* Add bonus credits to owner */
		    if(tobj->type == SW_OBJ_TYPE_PLAYER)
		    {
		        score->credits += (
			    (damage_to_shield + damage_to_structure) *
			    sysparm.hit_player_bonus
		        );
		    }

		    /* Add kills if target object hp's are less than 0 */
		    if((tobj->hp <= 0.0) && can_count_kill)
		    {
			score->kills++;

			/* Transfer credits on tobj to wobj_owner? */
			if(sysparm.killer_gets_credits &&
			   (tobj->score != NULL)
			)
			{
			    score->credits += tobj->score->credits;
			    tobj->score->credits = 0.0;
			}
		    }

		    /* Update scores to all connections */
		    NetSendScore(-1, wobj_owner_num);
	        }
	    }
	}

	/* Damage recieved by target object */
	if((sysparm.homes_destroyable) ? 1 :
	   ((tobj->type != SW_OBJ_TYPE_HOME) &&
	    (tobj->type != SW_OBJ_TYPE_WORMHOLE) &&
	    (tobj->type != SW_OBJ_TYPE_ELINK)
	   )
	)
	{
	    if(!UNVAllocScores(tobj))
	    {
		xsw_score_struct *score = tobj->score;

		/* Add damage_recieved to target object */
		score->damage_recieved += damage_to_shield +
		    damage_to_structure;

		NetSendScore(-1, tobj_num);
	    }

	    if(report_this)
	    {
		int i;
		const double total_damage = damage_to_shield +
		    damage_to_structure;
		sws_connection_struct *con;

		/* Log target object recieving damage */
		if(sysparm.log_events)
		{
		    char *s = (char *)malloc(
			STRLEN(tobj_name) +
			40 +
			STRLEN(wobj_owner_name) +
			80
		    );
		    if(wobj_owner_num != wobj_num)
			sprintf(
			    s,
"%s: Recieved %.2f damage by weapon fired by %s",
			    tobj_name,
			    damage_to_shield + damage_to_structure,
			    wobj_owner_name
			);
		    else
			sprintf(
			    s,
"%s: Recieved %.2f damage by %s",
			    tobj_name,
			    damage_to_shield + damage_to_structure,
			    wobj_owner_name
			);
		    LogAppend(fname.log, s);
		    free(s);
		}

	        /* Send hit notify to target object and weapon
		 * object's owner's connections
		 */

		/* Get bearing from target object to weapon object */
		theta = MuCoordinateDeltaVector(
		    wobj->x - tobj->x,
		    wobj->y - tobj->y
		);
		/* Change bearing to be relative to target object's
		 * orientation
		 */
		theta = SANITIZERADIANS(theta - tobj->heading);

#ifdef PLUGIN_SUPPORT
		/* Notify all plugins */
		for(i = 0; i < total_plugins; i++)
		    PluginHitNotify(
			i,
			wobj_num, tobj_num,
			total_damage,
			theta,		/* Target to weapon, in radians */
			damage_to_structure,
			damage_to_shield
		    );
#endif

		/* Send hit notify to all connections in sector */
		for(i = 0; i < total_connections; i++)
		{
		    con = connection[i];
		    if(con == NULL)
			continue;

		    if((con->socket < 0) || (con->object_num < 0))
			continue;

		    /* In same sector? */
		    if(!Mu3DInSameSector(
			xsw_object, total_objects,
			con->object_num, tobj_num
		    ))
			continue;

		    /* Send hit notify to this connection */
		    NetSendNotifyHit(
			i,
			wobj_num, tobj_num,
			total_damage,
			theta,			/* Target to weapon, in radians */
			damage_to_structure,
			damage_to_shield
		    );
		}
	    }	/* if(report_this) */
	}


	/* Target object destroyed? */
	if(tobj->hp <= 0.0)
	{
	    int opm_num, debris_num, explosion_num;
	    char debris_name[SW_OBJ_NAME_MAX + 80];
	    xsw_object_struct *opm_ptr;

	    /* Create debris */
	    if(*(tobj->name) != '\0')
		sprintf(
		    debris_name, "Debris of %s",
		    tobj->name
		);
	    else
		strncpy(
		    debris_name, "Debris",
		    sizeof(debris_name)
		);
	    opm_num = sws_std_opm.debris_num;
	    opm_ptr = OPMIsGarbage(opm_num) ?
		NULL : opm[opm_num];
	    debris_num = DBCreateObjectByOPMPtr(
		opm_ptr,
		debris_name,
		SW_OBJ_TYPE_STATIC,
		unv_head.lost_found_owner,
		tobj->sect_x, tobj->sect_y, tobj->sect_z,
		tobj->x, tobj->y, tobj->z,
		tobj->heading, tobj->pitch, 0.0 * PI
	    );
	    if(!DBIsObjectGarbage(debris_num))
	    {
		xsw_object_struct *debris = xsw_object[debris_num];

		strncpy(
		    debris->empire, tobj->empire,
		    sizeof(debris->empire)
		);
		debris->empire[sizeof(debris->empire) - 1] = '\0';

		NetSendCreateObject(-1, debris_num);
		NetSendObjectName(-1, debris_num);
	    }

	    /* Create big explosion */
	    opm_num = sws_std_opm.explosion_large_num;
	    opm_ptr = OPMIsGarbage(opm_num) ?
		NULL : opm[opm_num];
	    explosion_num = DBCreateObjectByOPMPtr(
		opm_ptr,
		NULL,
		SW_OBJ_TYPE_ANIMATED,
		unv_head.lost_found_owner,  
		tobj->sect_x, tobj->sect_y, tobj->sect_z,
		tobj->x, tobj->y, tobj->z,  
		tobj->heading, tobj->pitch, 0.0 * PI
	    );
	    if(!DBIsObjectGarbage(explosion_num))
		NetSendCreateObject(-1, explosion_num);

	    /* Is the target object is not immune to be destroyed? */
	    if(tobj->permission.uid > SWS_ACCESS_UID_DESTROY_IMMUNE)
	    {
		/* Target object can be destroyed */

		if(report_this)
		{
		    int i;
		    char *s, loc_str[256];
		    sws_connection_struct *con;

#ifdef PLUGIN_SUPPORT
		    /* Notify all plugins */
		    for(i = 0; i < total_plugins; i++)
			PluginDestroyNotify(
			    i,
			    SW_DESTROY_REASON_UNSPECIFIC,
			    tobj_num,		/* Destroyed object */
			    wobj_num,		/* Object that caused destruction */
			    wobj_owner_num	/* Owner of wobj */
			);
#endif

		    /* Format location string */
		    UNVLocationFormatString(
			loc_str,
			&tobj->sect_x,
			&tobj->sect_y,
			&tobj->sect_z,
			&tobj->x,
			&tobj->y,
			&tobj->z,
			sizeof(loc_str)
		    );

		    switch(tobj->type)
		    {
		      case SW_OBJ_TYPE_PLAYER:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(tobj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed due to hull structure colapse at %s.",
				tobj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(tobj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed due to hull structure colapse at %s.",
			    tobj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDPLAYER,
			    s
			);
			free(s);
		        break;

		      case SW_OBJ_TYPE_WEAPON:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(tobj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed at %s.",
				tobj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(tobj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed at %s.",
			    tobj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDWEAPON,
			    s
			);
			free(s);
			break;

		      default:
			/* Log */
			if(sysparm.log_events)
			{
			    s = (char *)malloc(
				STRLEN(tobj_name) +
				STRLEN(loc_str) +
				80
			    );
			    sprintf(
				s,
"%s: Destroyed at %s.",
				tobj_name, loc_str
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Export event */
			s = (char *)malloc(
			    STRLEN(tobj_name) +
			    STRLEN(loc_str) +
			    80
			);
			sprintf(
			    s,
"%s destroyed at %s.",
			    tobj_name, loc_str
			);
			ExportEvents(
			    fname.events_export,
			    EVST_TYPE_DESTROYEDOBJECT,
			    s
			);
			free(s);
			break;
		    }

		    /* Send destroy notify to all connections */
		    for(i = 0; i < total_connections; i++)
		    {
			con = connection[i];
			if(con == NULL)
			    continue;

			if((con->socket < 0) || (con->object_num < 0))
			    continue;

			/* Not in the same sector? */
			if(!Mu3DInSameSector(
			    xsw_object, total_objects,
			    con->object_num, tobj_num
			))
			    continue;

			/* Send destroy notify to this connection */
			NetSendObjectName(i, tobj_num);
		        NetSendNotifyDestroy(
			    i,
			    SW_DESTROY_REASON_UNSPECIFIC,	/* Reason is hit by weapon */
			    tobj_num,		/* Destroyed object */
			    wobj_num,		/* Object that caused destruction */
			    wobj_owner_num	/* Owner of wobj */
			);
		    }
		}	/* if(report_this) */

		/* Check if the destroyed object is a player and if
		 * players should be respawned, otherwise recycle
		 * the destroyed object
		 *
		 * Note that after this point, tobj should be considered
		 * invalid. It can still be used as cross reference but
		 * it must never be referenced implicitily.
		 */
		if((tobj->type == SW_OBJ_TYPE_PLAYER) &&
		    sysparm.respawn_players
		)
		{
		    REngDoRespawn(tobj_num);
		}
		else
		{
		    DBRecycleObject(tobj_num);
		    NetSendRecycleObject(-1, tobj_num);
		    objects_recycled++;
		}
	    }
	    else
	    {
		/* Sanitize hp on tobj (since we didn't recycle it */
		if(tobj->hp < 0.0)
		    tobj->hp = 0.0;
	    }
	}
	else
	{
	    /* Create small explosion */
	    int opm_num, explosion_num = -1;
	    double distance, bearing, size;
	    xsw_object_struct *opm_ptr;

	    switch(wobj->type)
	    {
	      /* Stream weapon */
	      case SW_OBJ_TYPE_STREAMWEAPON:
		opm_num = sws_std_opm.explosion_small_num;
		opm_ptr = OPMIsGarbage(opm_num) ?
		    NULL : opm[opm_num];
		distance = Mu3DDistance(
		    wobj->x - tobj->x,
		    wobj->y - tobj->y,
		    wobj->z - tobj->z
		) - (double)tobj->size / 1000.0;
		explosion_num = DBCreateObjectByOPMPtr(
		    opm_ptr,
		    NULL,
		    SW_OBJ_TYPE_ANIMATED,
		    unv_head.lost_found_owner,
		    wobj->sect_x, wobj->sect_y, wobj->sect_z,
		    wobj->x + (sin(wobj->heading) * distance),
		    wobj->y + (cos(wobj->heading) * distance),
		    wobj->z,
		    wobj->heading, wobj->pitch, 0.0 * PI
		);
		break;

	      /* Sphere weapon */
	      case SW_OBJ_TYPE_SPHEREWEAPON:
		opm_num = sws_std_opm.explosion_small_num;
		opm_ptr = OPMIsGarbage(opm_num) ?    
		    NULL : opm[opm_num];
		size = (double)tobj->size / 1000.0;
		bearing = MuCoordinateDeltaVector(
		    wobj->x - tobj->x,
		    wobj->y - tobj->y
		);
		explosion_num = DBCreateObjectByOPMPtr(
		    opm_ptr,
		    NULL,
		    SW_OBJ_TYPE_ANIMATED,
		    unv_head.lost_found_owner,
		    wobj->sect_x, wobj->sect_y, wobj->sect_z,
		    tobj->x + (sin(bearing) * size),
		    tobj->y + (cos(bearing) * size),
		    tobj->z,
		    wobj->heading, wobj->pitch, 0.0 * PI
		);
		break;

	      /* Projectile/Pulse weapon */
	      case SW_OBJ_TYPE_WEAPON:
		opm_num = sws_std_opm.explosion_small_num;
		opm_ptr = OPMIsGarbage(opm_num) ?
		    NULL : opm[opm_num];
		explosion_num = DBCreateObjectByOPMPtr(
		    opm_ptr,
		    NULL,
		    SW_OBJ_TYPE_ANIMATED,
		    unv_head.lost_found_owner,
		    wobj->sect_x, wobj->sect_y, wobj->sect_z,
		    wobj->x, wobj->y, wobj->z,
		    wobj->heading, wobj->pitch, 0.0 * PI
		);
		break;

	      /* All else create explosion over the target object */
	      default:
		opm_num = sws_std_opm.explosion_small_num;
		opm_ptr = OPMIsGarbage(opm_num) ?
		    NULL : opm[opm_num];
		explosion_num = DBCreateObjectByOPMPtr(
		    opm_ptr,
		    NULL,
		    SW_OBJ_TYPE_ANIMATED,
		    unv_head.lost_found_owner,
		    tobj->sect_x, tobj->sect_y, tobj->sect_z,
		    tobj->x, tobj->y, tobj->z,
		    tobj->heading, tobj->pitch, 0.0 * PI
		);
		break;

	    }
	    if(!DBIsObjectGarbage(explosion_num))
	    {
		NetSendCreateObject(-1, explosion_num);                
	    }
	}

	/* Target object may be invalid if it has been destroyed and
	 * recycled above
	 */

	/* Recycle the weapon object if it is a projectile or pulse */
	if(wobj->type == SW_OBJ_TYPE_WEAPON)
	{
	    DBRecycleObject(wobj_num);
	    NetSendRecycleObject(-1, wobj_num);
	    objects_recycled++;
	}

	return(objects_recycled);
}


/*
 *	Checks if the object has come in contact with another object
 *	and then performs the appropriate collision procedure.
 *
 *	Returns true if an object has been recycled.
 */
static int REngContactCheck(const int obj_num)
{
	int i;
	double heading_variance, distance;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;

	/* Projectile weapon */
	if(obj->type == SW_OBJ_TYPE_WEAPON)
	{
	    /* Begin checking if object has come in contact */

	    /* Iterate through all objects */
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip the following types of objects */
		if((tar_obj->type <= SW_OBJ_TYPE_GARBAGE) ||
		   (tar_obj->type == SW_OBJ_TYPE_WEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
		   (tar_obj->type == SW_OBJ_TYPE_ANIMATED)
		)
		    continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Check if weapon object has any antimatter left */
		if(obj->antimatter <= 0.0)
		{
		    /* No antimatter left so the weapon can be picked
		     * up, so check for contact
		     */ 
		    if(Mu3DInContactPtr(obj, tar_obj))
		    {
			if(WeaponsReclaim(
			    i,			/* Object */
			    obj_num		/* Weapon object */
			))
			    return(1);
		    }
		}
		else
		{
		    /* Weapon object has antimatter, check for hit
		     * contact
		     */

		    /* Do not hit its owner */
		    if(obj->owner == i)
			continue;

		    /* Does target's permission level make it immune to
		     * being hit?
		     */
		    if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
			continue;

		    /* Check for contact */
		    if(Mu3DInContactPtr(obj, tar_obj))
		    {
			/* Do hit */
		        REngDoHit(obj_num, i, obj->power);
			return(1);
		    }
		}
	    }
	}
	/* Stream weapon? */
	else if(obj->type == SW_OBJ_TYPE_STREAMWEAPON)
	{
	    /* Stream weapon object must have power */
	    if(obj->power <= 0.0)
		return(0);

	    /* Iterate through all objects */ 
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip the following types of objects */
		if((tar_obj->type <= SW_OBJ_TYPE_GARBAGE) ||
		   (tar_obj->type == SW_OBJ_TYPE_WEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
		   (tar_obj->type == SW_OBJ_TYPE_ANIMATED) ||
		   (obj->owner == i)	/* Do not hit its owner */
		)
			continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Does target's permission level make it immune to
		 * being hit?
		 */
		if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
		    continue;

		/* Not in the same sector? */
	        if(!Mu3DInSameSectorPtr(obj, tar_obj))
		    continue;

		/* Did it hit something? Get distance away in screen
		 * units
		 */
		distance = Mu3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		) * 1000.0;
		if(distance > 0.0)
		{
		    heading_variance = atan2(
			tar_obj->size,	/* dy */
			distance		/* dx */
		    );
		}
		else
		{
		    /* No distance away, so partical hits partical
		     * regardless of their bearings to each other
		     */
		    heading_variance = (2.0 * PI);
		}

		/* Check if in vector contact */
		if(Mu3DInVectorContact(
		    xsw_object, total_objects,
		    obj_num,			/* Source */
		    i,				/* Target */
		    obj->heading,		/* Source to target */
		    heading_variance,
		    obj->size		/* Size of source */
		))
		{
		    /* Do hit */
		    REngDoHit(obj_num, i, obj->power);

		    /* Stream weapons are not recycled in REngDoHit(),
		     * so we need to reset their power to 0 so that
		     * they do not hit a second object
		     */
		    if(!DBIsObjectGarbage(obj_num))
			xsw_object[obj_num]->power = 0.0;

		    return(1);
		}
	    }
	}
	/* Sphere weapons */
	else if(obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	{
	    int got_hit = 0;

	    /* Sphere weapon object must have power */
	    if(obj->power <= 0.0)
		return(0);

	    /* Iterate through each object
	     *
	     * Each object within the sphere weapon's radius gets
	     * equal amount of damage
	     */
	    for(i = 0; i < total_objects; i++)
	    {
		tar_obj = xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Skip the following types of objects */
		if((tar_obj->type <= SW_OBJ_TYPE_GARBAGE) ||
		   (tar_obj->type == SW_OBJ_TYPE_WEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
		   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
		   (tar_obj->type == SW_OBJ_TYPE_ANIMATED) ||
		   (obj->owner == i)	/* Do not hit its owner */
		)
			continue;

		/* Skip hidden objects */
		if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    continue;

		/* Does target's permission level make it immune to
		 * being hit?
		 */
		if(tar_obj->permission.uid <= SWS_ACCESS_UID_WEAPON_IMMUNE)
		    continue;

		/* Did it hit something? */
		if(Mu3DInContactPtr(obj, tar_obj))
		{
		    /* Do hit */
		    REngDoHit(obj_num, i, obj->power);
		    got_hit = 1;

		    if(DBIsObjectGarbage(obj_num))
			break;
		}
	    }

	    /* If the sphere weapon hit one or more objects then set
	     * its power to 0 so that it does not hit anything on
	     * the next call to this function
	     */
	    if(got_hit)
	    {
		if(!DBIsObjectGarbage(obj_num))
		    xsw_object[obj_num]->power = 0.0;
		return(1);
	    }
	}

	return(0);
}

/*
 *	AI simulation for controlled objects.
 *
 *	If the sysparm use_server_ai is set to false then this function
 *	does nothing.
 *
 *	Returns true if an object has been created or recycled.
 */
static int REngAI(const int obj_num)
{
	int	status = 0,
		i, closest_fobject, closest_uobject,
		type, sel_wep, weapons_fired, iff_code;
	char *s;
	sw_flags_t ai_flags;
	xsw_object_struct	*obj = xsw_object[obj_num],
				*tar_obj;
	xsw_weapon_struct *wep;

	long closest_wep_range;		/* In pixel units */
	double	scanner_range,		/* In real units */
		distance,		/* In real units */
		closest_fdistance,	/* In real units */
		closest_udistance;	/* In real units */
	double ovisibility;

	char obj_name[SW_OBJ_NAME_MAX + 80];

	const int loc_str_len = 256;
	char loc_str[loc_str_len];


	/* Server AI disabled or is the object not of the type
	 * controlled?
	 */
	if(!sysparm.use_server_ai ||
	   (obj->type != SW_OBJ_TYPE_CONTROLLED)
	)
	    return(0);


	/* Reset closest objects */
	closest_fobject = -1;
	closest_uobject = -1;

	/* Get this object's AI flags */
	ai_flags = obj->ai_flags;

	/* Get this object's scanner range */
	scanner_range = obj->scanner_range;
	closest_fdistance = scanner_range;
	closest_udistance = scanner_range;

	/* Iterate through the objects list */
	for(i = 0; i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];

	    /* Skip ourselves */
	    if(obj == tar_obj)
		continue;

	    /* Checks if objects are valid and in the same sector */
/* What about objects that need to follow objects across sectors? */
	    if(!Mu3DInSameSectorPtr(obj, tar_obj))
		continue;

	    /* React to only controlled and player objects */
	    type = tar_obj->type;
	    if((type != SW_OBJ_TYPE_CONTROLLED) &&
	       (type != SW_OBJ_TYPE_PLAYER)
	    )
		continue;

	    /* Skip objects that are marked hidden, AI's should
	     * ignore these
	     */
	    if(tar_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		continue;


	    /* Get the target object's current visibility */
	    ovisibility = DBGetObjectVisibilityPtr(tar_obj);

	    /* Calculate the distance to the target object in real
	     * units
	     */
	    distance = Mu3DDistance(
		tar_obj->x - obj->x,
		tar_obj->y - obj->y,
		tar_obj->z - obj->z
	    );

	    /* Check IFF */
	    iff_code = MatchIFFPtr(obj, tar_obj);
	    if(iff_code == SW_IFF_FRIENDLY)
	    {
		/* Friendly */
		/* Range check */
	        if((distance > (scanner_range * ovisibility)) ||
		   (distance > closest_fdistance) ||
		   (type == SW_OBJ_TYPE_CONTROLLED)
	        )
		    continue;

		closest_fdistance = distance;
		closest_fobject = i;
	    }
	    else
	    {
		/* Hostile or unknown */
		/* Range check */
		if((distance > (scanner_range * ovisibility)) ||
		   (distance > closest_udistance)
		)
		    continue;

		closest_udistance = distance;
		closest_uobject = i;
	    }
	}


	/* ******************************************************* */
	/* Follow: Adjust intercepting object */
	while((ai_flags & SW_AI_FIRE_ANY) ||
	      (ai_flags & SW_AI_FOLLOW_ANY)
	)
	{
	    /* Hostile or unknown (has precidence over friendly) */
	    if(((ai_flags & SW_AI_FIRE_UNKNOWN) ||
		(ai_flags & SW_AI_FIRE_HOSTILE) ||
		(ai_flags & SW_AI_FOLLOW_UNKNOWN) ||
		(ai_flags & SW_AI_FOLLOW_HOSTILE)
	       ) && (closest_uobject > -1)
	    )
	    {
		if(obj->intercepting_object != closest_uobject)
		{
		    obj->intercepting_object = closest_uobject;
		    obj->locked_on = closest_uobject;
		}   
	    }
	    /* Friendly */
	    else if(((ai_flags & SW_AI_FIRE_FRIEND) ||
		     (ai_flags & SW_AI_FOLLOW_FRIEND)
		     ) && (closest_fobject > -1)
	    )
	    {
		if(obj->intercepting_object != closest_fobject)
		{
		    obj->intercepting_object = closest_fobject;
		    obj->locked_on = closest_fobject;
		}
	    }
	    else
	    {
		if(obj->intercepting_object > -1)
		{
		    obj->intercepting_object = -1;
		    obj->locked_on = -1;
		}
	    }

	    break;      /* Break out of this while() loop */
	}
	/* ******************************************************* */
	/* Follow: Adjust throttle */
	while(ai_flags & SW_AI_FOLLOW_ANY)
	{
	    /* Actions in this section need to corrilate with
	     * those in the above intercepting object check
	     * since `following' implies concensus in throttle
	     * and intercept.
	     */

	    /* Hostile or unknown (has precidence over friendly) */
	    if(((ai_flags & SW_AI_FOLLOW_UNKNOWN) ||
		(ai_flags & SW_AI_FOLLOW_HOSTILE)
		) &&
		(closest_uobject > -1)
	    )
	    {
		/* Adjust throttle */
		if(closest_udistance > 0.6)
		{
		    obj->throttle += (0.02 * time_compensation);
		}
		else
		{
		    obj->throttle -= (0.02 * time_compensation);

		    /* Apply external dampers (brakes), this may
		     * be a bit stronger than it is equated on the
		     * client
		     */
#if 0
		    obj->velocity -= (obj->thrust_power *
			obj->power_purity * time_compensation
		    );
#endif
		    obj->velocity -= (0.04 * time_compensation);
		    if(obj->velocity < 0.0)
			obj->velocity = 0.0;
		}

		/* Sanitize throttle value */
		if(obj->throttle > 1.0)
		    obj->throttle = 1.0;
		else if(obj->throttle < 0.0)
		    obj->throttle = 0.0;
	    }
	    /* Friendly */
	    else if((ai_flags & SW_AI_FOLLOW_FRIEND) &&
		    (closest_fobject > -1)
	    )
	    {
		/* Adjust throttle */
		if(closest_fdistance > 0.6)
		{
		    obj->throttle += (0.02 * time_compensation);
		}
		else
		{
		    obj->throttle -= (0.02 * time_compensation);

		    /* Apply external dampers (brakes), this may
		     * be a bit stronger than it is equated on the
		     * client
		     */
#if 0
		    obj->velocity -= (obj->thrust_power *
			obj->power_purity * time_compensation
		    );
#endif
		    obj->velocity -= (0.04 * time_compensation); 
		    if(obj->velocity < 0.0)
			obj->velocity = 0.0; 
		}

		/* Sanitize throttle value */
		if(obj->throttle > 1.0)
		    obj->throttle = 1.0;
		else if(obj->throttle < 0.0)
		    obj->throttle = 0.0;
	    }
	    else
	    {    
		obj->throttle = 0.0;
	    }

	    break;	/* Break out of this while() loop */
	}
	/* ******************************************************* */
	/* Fire */
	while(ai_flags & SW_AI_FIRE_ANY)
	{
	    /* Same some cpu power if object has no weapons */
	    if(obj->total_weapons <= 0)
		break;

	    /* Iterate through list of weapons and set
	     * closest_wep_range to be the distance of the weapon
	     * with the longest range on the object
	     */
	    for(i = 0, closest_wep_range = 0;
		i < obj->total_weapons;
		i++
	    )
	    {
		/* Get weapon pointer and check if it's valid */
		wep = obj->weapons[i];
		if(wep == NULL)
		    continue;

		/* In pixel units */
		if(wep->range > closest_wep_range)
		    closest_wep_range = wep->range;
	    }


	    /* Check if there is a target object in range */

	    /* Hostile or unknown (precidence over friendly) */
	    if(((ai_flags & SW_AI_FIRE_UNKNOWN) ||
		(ai_flags & SW_AI_FIRE_HOSTILE)
		) && (closest_uobject > -1)
	    )
	    {
		/* Check if unknown object is out of range */
		if(closest_udistance > ((double)closest_wep_range / 1000))
		    break;

		/* Go through weapons list and select weapon with
		 * the closest range to fire on the unknown object
		 */
		for(i = 0, obj->selected_weapon = -1;
		    i < obj->total_weapons;
		    i++
		)
		{
		    /* Get weapon pointer and check if its valid */
		    wep = obj->weapons[i];
		    if(wep == NULL)
			continue;

		    /* Is the target inside this weapons range
		     * (in XSW Real units)?
		     */ 
		    if(closest_udistance > ((double)wep->range / 1000))
			continue;

		    /* Is this weapons range closer? */
		    if(wep->range <= closest_wep_range)
		    {
			/* Select this weapon */
			obj->selected_weapon = i;

			/* Remember closest weapon range */
			closest_wep_range = wep->range;
		    }
		}
	    }
	    /* Friendly */
	    else if((ai_flags & SW_AI_FIRE_FRIEND) &&
		    (closest_fobject > -1)
	    )
	    {
		/* Check if friendly object is out of range */
		if(closest_fdistance > ((double)closest_wep_range / 1000.0))
		    break;

		/* Go through weapons list and select weapon with
		 * the closest range to fire on the friendly object
		 */
		for(i = 0, obj->selected_weapon = -1;
		    i < obj->total_weapons;
		    i++
		)
		{
		    /* Get weapon pointer and check if its valid */
		    wep = obj->weapons[i];
		    if(wep == NULL)
			continue;

		    /* Is the target inside this weapons range
		     * (in real units)?
		     */
		    if(closest_fdistance > ((double)wep->range / 1000.0))
			continue;

		    /* Is this weapons range closer? */
		    if(wep->range <= closest_wep_range)
		    {
			/* Select this weapon */
			obj->selected_weapon = i;

			/* Remember closest weapon range */
			closest_wep_range = wep->range;
		    }
		}
	    }
	    else
	    {
		break;
	    }

	    /* Note: Procedure here is similar to the procedure
	     * in function NetHandleFireWeapon()
	     */ 

	    /* Get selected weapon pointer and check if its valid
	     * (if object has no weapons, then execution never
	     * gets past this point)
	     */
	    sel_wep = obj->selected_weapon;
	    if((sel_wep < 0) || (sel_wep >= obj->total_weapons))
		break;

	    wep = obj->weapons[sel_wep];
	    if(wep == NULL)
		break;

	    /* Too early to fire again? */
	    if((wep->delay + wep->last_used) >= cur_millitime)
		break;

	    /* Is the object cloaked and the weapon may not be
	     * used while cloaked?
	     */
	    if((obj->cloak_state == CLOAK_STATE_UP) &&
	       !(wep->flags & SW_WEP_FLAG_FIRE_CLOAKED)
	    )
	    {
		wep->last_used = cur_millitime;
		break;
	    }

	    /* If weapon is expendable, see if object has any left */
	    if(wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	    {
		if(wep->amount <= 0)
		    break;
	    }

	    /* Does object have enough power to fire weapon? */
	    if(obj->power < wep->create_power)
		break; 

	    /* Create (fire) the weapon object */
	    /* Stream weapons */
	    if(wep->emission_type == SW_WEP_EMISSION_STREAM)
	    {
		weapons_fired = WeaponsFire(
		    wep->ocs_code,		/* OCS */
		    obj_num,			/* Owner */
		    wep->emission_type,		/* Emission type */
		    SWS_MAX_OCS_COPPIES,	/* Create max */
		    obj->x,
		    obj->y,
		    obj->heading,
		    obj->pitch,
		    wep->power,
		    wep->range,
		    SW_FREQ_MIN,	/* Minimal frequency for now */
		    wep->flags
		);
	    }
	    /* Projectile */
	    else if(wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	    {
		weapons_fired = WeaponsFire(
		    wep->ocs_code,		/* OCS */
		    obj_num,			/* Owner */
		    wep->emission_type,		/* Emission type */
		    wep->amount,		/* Create max */
		    obj->x,
		    obj->y,
		    obj->heading,
		    obj->pitch,
		    wep->power,
		    wep->range,
		    SW_FREQ_MIN,	/* Minimal frequency for now */
		    wep->flags
		);
	    }
	    /* All else assume pulse */
	    else
	    {
		weapons_fired = WeaponsFire(
		    wep->ocs_code,		/* OCS */
		    obj_num,			/* Owner */
		    wep->emission_type,		/* Emission type */
		    SWS_MAX_OCS_COPPIES,	/* Create max */
		    obj->x,
		    obj->y,
		    obj->heading,
		    obj->pitch,
		    wep->power,
		    wep->range,
		    SW_FREQ_MIN,	/* Minimal frequency for now */
		    wep->flags
		);
	    }

	    /* Reget object pointer, creating object(s) above may
	     * change the pointer value
	     */
	    obj = xsw_object[obj_num];

	    /* Update status to report memory change */
	    status = 1;

	    /* No objects created? */
	    if(weapons_fired <= 0)
		break;


	    /* Mark the last time it was fired */
	    wep->last_used = cur_millitime;


	    /* Get object name for logging purposes */
	    strncpy(
		obj_name, DBGetFormalNameStr(obj_num),
		sizeof(obj_name)
	    );
	    obj_name[sizeof(obj_name) - 1] = '\0';

	    /* Format location string */  
	    UNVLocationFormatString(
		loc_str,
		&obj->sect_x, &obj->sect_y, &obj->sect_z,
		&obj->x, &obj->y, &obj->z,
		loc_str_len
	    );

	    /* Consume power on object for the power that was
	     * needed to create this weapon and export event
	     */
	    switch(wep->emission_type)
	    {
	      case SW_WEP_EMISSION_PROJECTILE:
		obj->power -= (wep->create_power * weapons_fired);
		/* Decrease stock for projectiles */
		wep->amount -= weapons_fired;
		if(wep->amount < 0)
		    wep->amount = 0;

		/* Export projectile weapons fire */
		s = (char *)malloc(
		    STRLEN(obj_name) +
		    SW_OBJ_NAME_MAX +
		    STRLEN(loc_str) +
		    80
		);
		sprintf(
		    s,
		    "%s fired %s at %s.",
		    obj_name,
		    DBGetOCSOPMName(wep->ocs_code),
		    loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_PROJECTILEWEAPON_FIRED,
		    s
		);
		free(s);
		break;

	      case SW_WEP_EMISSION_PULSE:
		obj->power -= (wep->create_power * weapons_fired);
		break;
	    
	      /* Default to stream weapon */
	      default:
		obj->power -= wep->create_power;
 	        break;
	    }
	    /* Sanitize power on object */
	    if(obj->power < 0.0)
		obj->power = 0.0;

            /* Report new object and weapons values to objects nearby */
            NetFireWeaponSyncObjectValues(
                obj, obj_num,
                sel_wep
            );

#if 0
	    /* Set next object values update to now so it gets updated now */
	    next.object_values = cur_millitime;
	    next.need_weapon_values = 1;
#endif

	    break;	/* Break out of this while() loop */
	}


	return(status);
}


/*
 *	Updates the object.
 *
 *	Returns true if the object was recycled during this call.
 */
static int REngUpdateObject(const int obj_num)
{
	REngVisibility(obj_num);
	REngShieldVisibility(obj_num);

	if(REngMortality(obj_num))
	    return(1);

	REngAnimated(obj_num);

	if(REngAutoDestruct(obj_num))
	    return(1);

	REngPowerCore(obj_num);
	REngDamageControl(obj_num);

	REngWeaponsLock(obj_num);
	REngTractorBeam(obj_num);
	REngIntercept(obj_num);

	REngThrottle(obj_num);
	REngThrust(obj_num);

	REngHeading(obj_num);
	REngLocation(obj_num);
	REngMove(obj_num, 1, 0);

	if(REngContactCheck(obj_num))
	    return(1);

	if(REngAI(obj_num))
	    return(1);

	return(0);
}


/*
 *      Initializes the reality engine.
 */
int REngInit(void)
{
	return(0);
}

/*
 *	Manages the reality engine.
 */
void REngManage(void)
{
	int obj_num;
	xsw_object_struct *obj;

	/* Iterate through the objects list */
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Update this object */
	    REngUpdateObject(obj_num);
	}
}

/*
 *	Shuts down the reality engine.
 */
void REngShutdown(void)
{
	return;
}
