#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"
#include "weapons.h"


static void WepSyncCreateObject(
	xsw_object_struct *obj, int obj_num
);

int WeaponsFire(
	int ocs_code,		/* OCS used to create the weapons fire */
	int owner_obj_num,	/* Who fired the weapon (must be valid) */
	int emission_type,	/* Emission type */
	int create_max,		/* Do not create more than this many weapons */
	double birth_x,		/* Location of fire (not offset) */
	double birth_y,
	double birth_heading,	/* Direction to be fired */
	double birth_pitch,
	double power,
	long range,		/* In screen units */
	double freq,
	sw_flags_t flags	/* Weapon flags */
);

int WeaponsReclaim(
	int obj_num,
	int wep_obj_num		/* Weapon object to pick up */
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sends create object and object values of the specified
 *	weapon object to all connections in scanner range.
 */
static void WepSyncCreateObject(
	xsw_object_struct *obj, int obj_num
)
{
	int i, con_obj_num;
	sws_connection_struct *con;
	xsw_object_struct *con_obj;

	/* Check if marked hidden from connection */ 
	if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
	    return;

	/* Iterate through all connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    con_obj_num = con->object_num;
	    if(DBIsObjectGarbage(con_obj_num))
		continue;

	    con_obj = xsw_object[con_obj_num];

	    /* Check if objects are valid and in range */
	    if(!Mu3DInRangePtr(
		con_obj, obj,
		con_obj->scanner_range
	    ))
		continue;

	    /* Not owned objects, check for visibility */
	    if(obj->owner != con_obj_num)
	    {
		if(DBGetObjectVisibilityPtr(obj) <= 0.0)
		    continue;
	    }

	    /* Send object values to connection */
	    NetSendCreateObject(i, obj_num);
	    NetSendObjectMaximums(i, obj_num);
	    NetSendObjectValues(i, obj_num);
	}
}

/*
 *	Sends set weapon values of the specified weapon number on
 *	the specified object to all connections with objects in
 *	scanner range.
 */
void WepSyncObjectWeaponValues(
	xsw_object_struct *obj, int obj_num,
	int wep_num
)
{
	int i, con_obj_num;
	sws_connection_struct *con;
	xsw_object_struct *con_obj;

	/* Check if marked hidden from connection */
	if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
	    return;

	/* Iterate through all connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    con_obj_num = con->object_num;
	    if(DBIsObjectGarbage(con_obj_num))
		continue;

	    con_obj = xsw_object[con_obj_num];

	    /* Check if objects are valid and in range */
	    if(!Mu3DInRangePtr(
		con_obj, obj,
		con_obj->scanner_range
	    ))
		continue;

	    /* Not owned objects, check for visibility */
	    if(obj->owner != con_obj_num)
	    {
		if(DBGetObjectVisibilityPtr(obj) <= 0.00)
		    continue;
	    }

	    /* Send weapon values to connection */
	    NetSendWeaponValues(
		i,
		obj_num,
		wep_num
	    );
	}
}


/*
 *	Creates a new weapons fire object relative to the specified
 *	owner object.
 *
 *	The ocs_code specifies the OCS to be used to create the weapons
 *	fire.
 *
 *	The owner_obj_num specifies the object that fired the weapon,
 *	it must be a valid object.
 *
 *	The create_max specifies the limit of how many weapons fire
 *	objects may be created.
 *
 *	The power specifies the power & yield of the weapon.
 *
 *	The range specifies the weapons range in screen units (for
 *	stream and sphere weapons only)
 *
 *	The freq specifies the weapons frequency.
 *
 *	The flags specify the weapon's flags.
 */
int WeaponsFire(
	int ocs_code,		/* OCS used to create the weapons fire */
	int owner_obj_num,	/* Who fired the weapon (must be valid) */
 	int emission_type,	/* Emission type */
	int create_max,		/* Do not create more than this many weapons */
	double birth_x,		/* Location of fire (not offset) */
	double birth_y,
	double birth_heading,	/* Direction to be fired */
	double birth_pitch,
	double power,
	long range,		/* In screen units */
	double freq,
	sw_flags_t flags	/* Weapon flags */
)
{
	int objects_created = 0;
	int obj_num, ocs_num, opm_num;
	double work_angle;
	double tmp_x, tmp_y, tmp_z;
	double tmp_h, tmp_p, tmp_b;
	xsw_object_struct *obj, *owner_obj, *opm_ptr;
	sws_ocs_struct *ocs_ptr;

	/* Cease fire impose? */
	if(sysparm.cease_fire)
	    return(0);

	/* OCS code must be valid */
	if(ocs_code <= OCS_TYPE_GARBAGE)
	    return(-2);

	/* Emission type must be valid */
	if(emission_type < 0)
	{
	    fprintf(
		stderr,
		"WeaponsFire(): Error: Emission type %i is invalid.\n",
		emission_type
	    );
	    return(-2);
	}

	/* Owner must be specified and valid */
	if(DBIsObjectGarbage(owner_obj_num))
	    return(-2);

	owner_obj = xsw_object[owner_obj_num];

	/* Sanitize birth angle */
	birth_heading = SANITIZERADIANS(birth_heading);
	birth_pitch = SANITIZERADIANS(birth_pitch);

	/* Sanitize power */
	if(power < 0.0)
	    power = 0.0;

	/* Sanitize range */
	if(range < 0l)
	    range = 0l;

	/* Sanitize frequency */
	if(freq < SW_FREQ_MIN)
	    freq = SW_FREQ_MIN;
	if(freq > SW_FREQ_MAX)
	    freq = SW_FREQ_MAX;


	/* Get the OCS number from the specified OCS code */
	ocs_num = OCSGetByCode(ocs_code);
	if(OCSIsGarbage(ocs_num))
	{
	    fprintf(
		stderr,
		"WeaponsFire(): Error: %s requested invalid OCS type %i.\n",
		DBGetFormalNameStr(owner_obj_num),
		ocs_code
	    );
	    return(-2);
	}

	ocs_ptr = ocs[ocs_num];


	/* Number of coppies must be positive, otherwise there is
	 * nothing to do
	 */
	if(ocs_ptr->coppies < 1)
	    return(0);

	/* Get the initial creation relativity angle */
	work_angle = SANITIZERADIANS(ocs_ptr->heading);

	/* Get the OPM specified by the OCS */
	opm_num = OPMGetByName(ocs_ptr->opm_name, -1);
	if(OPMIsGarbage(opm_num))
	{
	    fprintf(
		stderr,
		"WeaponsFire(): Error: %s: No such OPM.\n",
		ocs_ptr->opm_name
	    );
	    return(-2);
	}

	opm_ptr = opm[opm_num];


	/* Sanitize number of coppies */
	if(ocs_ptr->coppies > SWS_MAX_OCS_COPPIES)
	    ocs_ptr->coppies = SWS_MAX_OCS_COPPIES;


	/* Begin creating the new fired weapon object(s) */
	while((objects_created < ocs_ptr->coppies) &&
	      (objects_created < create_max)
	)
	{
	    /* Create a new fired weapon object */
	    obj_num = DBCreateObject(
		opm_ptr->imageset,
		opm_ptr->type,
		owner_obj_num,
		owner_obj->sect_x, owner_obj->sect_y, owner_obj->sect_z,
		birth_x, birth_y, 0.0,
		birth_heading, birth_pitch, 0.0 * PI
	    );
	    if(DBIsObjectGarbage(obj_num))
	    {
		/* Failed to create object, stop creating */
		fprintf(
		    stderr,
		    "WeaponsFire(): DBCreateObject(): Could not create object.\n"
		);
		break;
	    }

	    obj = xsw_object[obj_num];

	    /* Increment number of objects created */
	    objects_created++;

	    /* Record tempory positions and attitudes */
	    tmp_x = birth_x;
	    tmp_y = birth_y;
	    tmp_z = 0;

	    tmp_h = birth_heading;
	    tmp_p = birth_pitch;
	    tmp_b = 0;

	    /* Adjust relative position */
	    tmp_x +=
		MuPolarRotX(
		    birth_heading + work_angle,
		    ocs_ptr->radius
		);
	    tmp_y +=
		MuPolarRotY(
		    birth_heading + work_angle,
		    ocs_ptr->radius
		);


	    /* **************************************************** */

	    /* Set parameters on new object from OPM */
	    if(OPMModelObject(obj_num, opm_num))
	    {
/* What do we do if we can't model if successfully? */
		fprintf(
		    stderr,
"WeaponsFire(): Error using OPM #%i for new object #%i.\n",
		    opm_num, obj_num
		);
	    }

	    /* Set position */
	    obj->x = tmp_x;
	    obj->y = tmp_y;
	    obj->z = tmp_z;

	    obj->heading = tmp_h;
	    obj->pitch = tmp_p;
	    obj->bank = tmp_b;
	    MuSetUnitVector2D(
		&obj->attitude_vector_compoent,
		obj->heading
	    );

	    obj->sect_x = owner_obj->sect_x;
	    obj->sect_y = owner_obj->sect_y;
	    obj->sect_z = owner_obj->sect_z;


	    /* Set owner */
	    obj->owner = owner_obj_num;

	    /* Set the empire as the owner's */
	    strncpy(
		obj->empire, owner_obj->empire,
		sizeof(obj->empire)
	    );
	    obj->empire[sizeof(obj->empire) - 1] = '\0';

#if 0
	    /* Set scanner range to match the owner's */
	    obj->scanner_range = owner_obj->scanner_range;
#endif

	    /* Set emission type dependant values */
	    switch(emission_type)
	    {
	      case SW_WEP_EMISSION_STREAM:
		/* Stream weapons have size specified as the range
		 *
		 * They will intercept the owner's locked object
		 */
		obj->size = range;

		if((flags & SW_WEP_FLAG_NO_LOCK) ||
		   DBIsObjectGarbage(owner_obj->locked_on)
		)
		{
		    obj->intercepting_object = -1;
		    obj->locked_on = -1;
		}
		else
		{
		    obj->intercepting_object = owner_obj->locked_on;
		    obj->locked_on = owner_obj->locked_on;
		}
		break;

	       case SW_WEP_EMISSION_PROJECTILE:
		/* Projectiles have the size of the OPM and intercept
		 * the owner's locked object
		 */
		if((flags & SW_WEP_FLAG_NO_LOCK) ||
		   DBIsObjectGarbage(owner_obj->locked_on)
		)
		{
		    obj->intercepting_object = -1;
		    obj->locked_on = -1;
		}
		else
		{
		    obj->intercepting_object = owner_obj->locked_on;
		    obj->locked_on = owner_obj->locked_on;
		}
		break;

	      default:
		/* Default is pulse weapon, they have the size of the
		 * OPM and do not incercept the owner's locked object
		 */
		obj->intercepting_object = -1;
		break;
	    }

	    /* Set velocity to match that of the owner */
	    obj->velocity = MIN(owner_obj->velocity, obj->velocity_max);
	    obj->velocity_heading = owner_obj->velocity_heading;
	    MuSetUnitVector2D(
		&obj->momentum_vector_compoent,
		obj->velocity_heading
	    );

	    /* Set thrust as owner's thrust */
	    obj->thrust = owner_obj->thrust;

	    /* Set power from given function input */
	    obj->power = power;
	    obj->power_max = power;

	    /* Set frequency */
	    obj->shield_frequency = freq;

	    /* Penetrate shield? */
	    if(flags & SW_WEP_FLAG_PENETRATE_SHIELD)
		obj->ext_flags |= SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD;

	    /* Mark the time this object was created */
	    obj->birth_time = cur_millitime;

	    /* Record the ocs code used to create this weapon object */
	    obj->creation_ocs = ocs_code;


	    /* ***************************************************** */

	    /* Send create object and object values to all connections
	     * with nearby objects for this weapon object
	     */
	    WepSyncCreateObject(obj, obj_num);


	    /* Modify work_angle depending on emission_type */
	    if((emission_type == SW_WEP_EMISSION_PROJECTILE) ||
	       (emission_type == SW_WEP_EMISSION_PULSE)
	    )
	    {
	        /* Flip work_angle to negative? */
	        if(work_angle > 0)
	        {
		    work_angle *= -1;
	        }
	        /* Flip angle to positive, then decrease angle */
	        else
	        {
		    work_angle *= -1;
		    work_angle = work_angle *
		        (double)(1 / (double)(ocs_ptr->coppies));
		}
	    }
	}

	return(objects_created);
}


/*
 *	Places the weapon object specified by wep_obj_num back into the
 *	object specified by obj_num (if the object can carry that
 *	weapon).
 *
 *	The weapon object is assumed to be within range of the
 *	specified object and that the weapon object has no antimatter
 *	(suggesting that it is disarmed).
 *
 *	Returns 0 if the weapon object could not be picked up or 1 if
 *	the weapon object was successfully picked up and recycled.
 */
int WeaponsReclaim(
	int obj_num,
	int wep_obj_num		/* Weapon object to pick up */
)
{
	int i, ocs_code;
	xsw_weapon_struct *wep;
	xsw_object_struct *obj, *wep_obj;

	if(DBIsObjectGarbage(obj_num) ||
	   DBIsObjectGarbage(wep_obj_num)
	)
	    return(0);

	obj = xsw_object[obj_num];
	wep_obj = xsw_object[wep_obj_num];

	/* Weapon object must be a projectile weapon */
	if(wep_obj->type != SW_OBJ_TYPE_WEAPON)
	    return(0);

#if 0
	/* Check if source object is tractoring weapon object */
	for(i = 0; i < src_obj->total_tractored_objects; i++)
	{
	    if(wep_obj_num == src_obj->tractored_object[i])
		break;
	}
	if(i >= src_obj->total_tractored_objects)
	    return(0);
#endif

	/* Get the OCS of what the weapon was created with */
	ocs_code = wep_obj->creation_ocs;


	/* Search for a weapon on the object matching the OCS */
	for(i = 0; i < obj->total_weapons; i++)
	{
	    wep = obj->weapons[i];
	    if(wep == NULL)
		continue;

	    if(wep->ocs_code == ocs_code)
		break;
	}
	if(i < obj->total_weapons)
	{
	    /* Got match, the object can carry the weapon  */
	    wep = obj->weapons[i];

	    /* Matched weapon must be a projectile type emission */
	    if(wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	    {
		/* Room to add one more unit back into inventory? */
		if(wep->amount < wep->max)
		{
		    int con_num = ConGetByObject(obj_num);
		    char *wep_name = STRDUP(wep_obj->name);

		    /* Increment units */
		    wep->amount++;

		    /* Untractor the weapon object */
		    DBObjectUntractor(obj_num, wep_obj_num);
                    NetSendTractorBeamLock(-1, obj_num, -1);

		    /* Recycle the weapon object */
		    DBRecycleObject(wep_obj_num);
		    NetSendRecycleObject(-1, wep_obj_num);

		    /* Update weapon values on the object to all
		     * connections
		     */
		    WepSyncObjectWeaponValues(
			obj,
			obj_num,
			i
		    );

		    /* Notify connection */
		    if(con_num > -1)
		    {
			char *s = (char *)malloc(STRLEN(wep_name) + 80);
			sprintf(
			    s,
"Obtained a `%s'.",
			    wep_name
			);
			NetSendLiveMessage(con_num, s);
			free(s);
		    }
		    free(wep_name);

		    return(1);
		}
	    }
	}

	return(0);
}
